/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: BaseActionClass.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef BASEACTIONCLASS_H_
#define BASEACTIONCLASS_H_

#include <iostream>

class BaseActionClass;

#include "FastSet.h"
#include "BaseStateVariableClass.h"
#include "MobiusEnums.h"
#include "../Simulator/UserDistributions.h"
#include "debug.h"
#include "List.h"
#include "../Simulator/PAddress.h"

/**
 * This is a link to the distribution archive or possibly to a user
 * defined probability distribution function.
 */
extern UserDistributions* TheDistribution;

class ActionAffectsElement;
class BaseStateVariableClass;

/**
 * This class describes the basic structure of an action. All
 * formalism-specific action should be derived from this class and
 * should overload all the virtual methods.
 *
 *  @author Jay M. Doyle
 *  @version Mobius version 1.0
 */

class BaseActionClass {
public:

	/* public methods  */

	/**
	 * Base Constructor
	 */
	BaseActionClass();

	/**
	 * Destructor
	 */
	virtual ~BaseActionClass();

	/**
	 * Overloaded constructor
	 *
	 *  @param TheActionName The name of the action
	 *  @param TheGroupIDNumber The unique number identifying the action group
	 *  @param TheDistributionType The action probability distribution function
	 *  @param TheExecutionPolicy  The action's execution policy
	 *  @param TheNumberOfAffectedStateVariables The number of state variables
	 *     whose state is affected by the action
	 *  @param TheNumberOfEnablingStateVariables The number of state variables
	 *      whose state enables the action
	 *  @param The initial value of reactivation data element
	 *
	 */
	BaseActionClass(char* TheActionName, int TheGroupIDNumber, Distribution TheDistributionType, ExecutionPolicyType TheExecutionPolicy,
			int TheNumberOfAffectedStateVariables, int TheNumberOfEnablingStateVariables, bool TheReactivation);

	/**
	 * This method initializes all the internal data structures for
	 * the BaseActionClass
	 *  @param TheGroupName The name of the action group
	 *  @param TheGroupIDNumber The unique number identifying the action group
	 *  @param TheDistributionType The action probability distribution function
	 *  @param TheExecutionPolicy  The action's execution policy
	 *  @param TheNumberOfAffectedStateVariables The number of state variables
	 *    whose state is affected by the action
	 *  @param TheNumberOfEnablingStateVariables The number of state variables
	 *    whose state enables the action
	 *  @param The initial value of reactivation data element
	 */
	void BaseActionClassInitialize(char* TheGroupName, int TheGroupIDNumber, Distribution TheDistributionType, ExecutionPolicyType TheExecutionPolicy,
			int TheNumberOfAffectedStateVariables, int TheNumberOfEnablingStateVariables, bool TheReactivation);

	/**
	 * This method was added so that synchronized actions can easily be
	 * identified.
	 */
	virtual bool isSync() {
		return false;
	}

	/**
	 * This method determines whether the action is enabled in the current state.
	 * This method should be overloaded for all real actions.  The enabling
	 * conditions determine the set of states in which the action may fire.
	 */
	virtual bool Enabled() {
		return false;
	}

	/**
	 * This method return the action's weight.  Weights are used to determine the
	 * probability of selecting an action from the set of enabled actions in the
	 * current state.  The greater the weight value, the the more likely it is to
	 * be selected to fire in the enabling state.
	 */
	virtual double Weight() {
		return 0.0;
	}

	/**
	 * This method returns the rate with which an exponentially timed
	 * action fires.  If the action's distribution is not exponential
	 * then this method can be ignored. Rate is defined on the
	 * BaseActionClass since many of the analytical solver assume
	 * exponential actions and having defined on BaseActionClass
	 * helps to optimize the solvers.
	 *
	 * @param The rate of an exponential action
	 */
	virtual double Rate() {
		return 0.0;
	}

	/**
	 * This method determines whether an action is reactivable, i.e.
	 * whether it can restart on a state change.  If this is true, then
	 * the ReactivationFunction is called at state changes to see if the
	 * action does restate in the new state.
	 *
	 *  @return True if the action can possibly restart in state changes
	 */
	virtual bool ReactivationPredicate() {
		return false;
	}

	/**
	 * This method determines if an action, whose ReactivationPredicate
	 * is true, should restart after a state change in which the action is
	 * still enabled.
	 *
	 *  @return True if the action should restart in the new state
	 */
	virtual bool ReactivationFunction() {
		return false;
	}

	/**
	 * This method samples the action's distribution and returns the
	 * action's time to completion
	 *
	 * @return The time until the action completes (sampled from the
	 *  action distribution function).
	 */
	virtual double SampleDistribution() {
		return 0.0;
	}

	/**
	 * This method returns the set of distribution parameters.  The number
	 * of parameters is determined by
	 * the action distribution function.
	 *
	 * @return An array of distribution parameters.
	 */
	virtual double* ReturnDistributionParameters() {
		return nullptr;
	}

	/**
	 * This method sets the Fired data member on an action to remember
	 * the fact that the action fired.  This is used by the performance
	 * variable code.
	 *
	 *  @param value The value to which Fire should be set
	 */
	virtual void SetFired(int value);

	/**
	 *  This method defines how the action changes the state
	 *  of the model.  This method should be overloaded
	 *  for all real actions.
	 *
	 *  @return A pointer to the action that fired.
	 */
	virtual BaseActionClass* Fire() {
		/* At some point in the state space generator in order to avoid adding
		 redundant code, we have to indirectly call the BaseActionClass::Fire()
		 (i.e. called during a direct call to another method) which actually
		 does nothing and it's harmless but it was printing the following error
		 message which we decided to comment out. For more comments go to
		 Cpp/gen/StateSpaceGen.cpp and look at the comments above this statement :
		 The_Model->FireAction((void *) new_state_holder, ..... )
		 (Salem)
		 */
		//cerr<<"Error: Firing BaseActionClass::Fire()"<<endl;
		ASSERT(true);
		return nullptr;
	} // end BaseActionClass* Fire()

	/**
	 * This method returns value used to determine with action from,
	 * among the set of enabled actions, will fire.  Actions with higher
	 * ranks are guarantee to fire before any actions with lower ranks.  This
	 * means that only highest ranked actions will fire in an enabling
	 * state.  Since ranks can be state depended the highest rank value
	 * may differ from state to state. Actions are selected from among the
	 * set of enabled action in the highest ranking group based on their
	 * Weight values.   Weight determined the probability of being selected
	 * from among a "rank group."
	 *
	 *  @return The value of the action's rank
	 */
	virtual int Rank() {
		return 0;
	}

	/**
	 * This method determines if there has been a change in the enabling
	 * condition since the last time the Enabled method was called.
	 *
	 *  @return True if the enabling condition had changed since the
	 *  last time Enabled was called
	 */
	virtual bool EnablingChange() {
		return false;
	}

	/**
	 * This method is really defined for BaseGroupClass which is derived
	 * from BaseActionClass.
	 *  @see BaseGroupClass
	 *
	 *  @param TheAction An action defined within the model
	 *  @return True if action belongs to the action group
	 */
	virtual bool IsAMember(BaseActionClass* TheAction);

	/**
	 * This method is really defined for BaseGroupClass which is
	 * derived from BaseActionClass.
	 *  @see BaseGroupClass
	 *
	 *  @param TheAction An action defined within the model
	 *  @return The probability of selecting the action from the group
	 */
	virtual double Probability(BaseActionClass* TheAction);

	/**
	 * This is a debugging method that print useful information about the action
	 */
	void PrintInfo();

	/**
	 *  print name,state info about state variables affected by this action
	 */
	virtual void printAffectedState(ostream *strm);

	/**
	 *  print name,state info about state variables that enable this action
	 */
	virtual void printEnablingState(ostream *strm);

	// Needed to find the action in the group to print
	virtual BaseActionClass* getPrintedAction();

	/**
	 *  This method sets the action's name
	 *
	 *  @param TheActionName  The new name for the
	 *         action
	 */
	void defineName(char* TheNewActionName);

	/**
	 *  This returns the number of enabling state variables
	 *  defined on an action
	 */
	inline int getNumEnablingSVs() {
		return EnablingStateVariables.getNumItems();
	} // end getNumEnablingSVs()

	/**
	 *  This method allows you to replace an element of the
	 *  enabling state variable list with a new state variable
	 *  location.  This is useful for update actions with the
	 *  new location a shared state variable.
	 *
	 *  @param TheOldSV The old state variabls location
	 *  @param TheNewSV The new location of the state variable
	 *    formerly pointed to by TheOldSV
	 */
	inline void replaceEnablingSV(BaseStateVariableClass* TheOldSV, BaseStateVariableClass* TheNewSV) {
		EnablingStateVariables.replaceItem(TheOldSV, TheNewSV);
	} // end replaceAffectedSV(BaseStateVariableClass*, BaseStateVariableClass*)

	/**
	 *  This method returns the i'th method in the
	 *  action's enabling state variable list.
	 */
	inline BaseStateVariableClass* getEnablingSV(int i) {
		return EnablingStateVariables.getItem(i);
	} // end getEnablingSV(int i)

	/**
	 * This add an enabling state variable to the
	 * action's list.
	 *
	 * @param TheSV A state variable whose state affects
	 *  the action's enabling condition
	 */
	void addEnablingSV(BaseStateVariableClass* TheSV);
	void removeEnablingSV(BaseStateVariableClass* TheSV);

	/**
	 *  This method returns the number of state variables
	 *  affected by the action's firing
	 */
	inline int getNumAffectedSVs() {
		return AffectedStateVariables.getNumItems();
	} // end int getNumAffedSVs()

	/**
	 *  This method returns the i'th method in the
	 *  action's affected state variable list.
	 */
	inline BaseStateVariableClass* getAffectedSV(int i) {
		return AffectedStateVariables.getItem(i);
	} // end getAffectedSV(int i)
	/**
	 *  This method allows you to replace an element of the
	 *  affected state variable list with a new state variable
	 *  location.  This is useful for update actions with the
	 *  new location a shared state variable.
	 *
	 *  @param TheOldSV The old state variables location
	 *  @param TheNewSV The new location of the state variable
	 *    formerly pointed to by TheOldSV
	 */
	inline void replaceAffectedSV(BaseStateVariableClass* TheOldSV, BaseStateVariableClass* TheNewSV) {
		AffectedStateVariables.replaceItem(TheOldSV, TheNewSV);
	} // end replaceAffectedSV(BaseStateVariableClass*, BaseStateVariableClass*)

	/**
	 * This add an affected state variable to the
	 * action's list.
	 *
	 * @param TheSV A state variable whose state is affected
	 *  by the action's firing
	 */
	void addAffectedSV(BaseStateVariableClass* TheSV);
	void removeAffectedSV(BaseStateVariableClass* TheSV);

	//Allocation and deallocation methods for data structures dealing
	//with dependency of PVWorkers on the action (DS stands for Data Structure)
	bool IsPVRateDSAllocated();
	bool IsPVImpulseDSAllocated();
	void AllocatePVRateDS(int NoOfPVs);
	void DeallocatePVRateDS();
	void AllocatePVImpulseDS(int NoOfPVs);
	void DeallocatePVImpulseDS();

	/**
	 *  This method sets the actions's GroupID data
	 *  member.
	 *
	 *  @param TheNewID The new value of the action's
	 *   ID
	 */
	inline void setGroupId(int TheNewID) {
		GroupID = TheNewID;
	}

	/**
	 *  give each activity a unique id number (used by the simulator to
	 *  speed up model initialization; see
	 *  Simulator::GenerateConnectivityList
	 */
	inline int getID() {
		return ID;
	}
	inline int getMaxID() {
		return MaxID;
	}

	/**
	 *  This method returns the action's group ID
	 *
	 * @return GroupID
	 */
	inline int getGroupId() {
		return GroupID;
	}
	/**
	 *  This method resets the action's group ID
	 *
	 * @param i, the new GroupID
	 */

	inline void setGroupID(int i) {
		GroupID = i;
	}

	/* public variables */

	/**
	 * This is unique integer that identifies the action group to
	 * which action belongs
	 */

	virtual double ErlangParamM() {
		return 1.0;
	}
	;
	void ResetPhase();
	void Age();
	bool ReadyToFire();


	inline const char * getActionName() const
	{
		return ActionName;
	}


	inline int **const  getNumAffectedImpulseWorkers() const
	{
			return NumAffectedImpulseWorkers;
	}

	inline int ***const  getAffectedImpulseWorkers() const
	{
				return AffectedImpulseWorkers;
	}

	inline FastSet* const  getAffectedRateWorkers() const
	{
		return AffectedRateWorkers;
	}


	inline FastSet * const getAffectedImpulseRewards()const
	{
		return AffectedImpulseRewards;
	}




	/**
	 * This boolean flag is set when evaluating the enabling status of
	 * an action in a group to remember its  enabling state such that
	 * the EnablingChange method can determined whether there
	 * had been a change in the action enabling state. (??)
	 */
	bool Disabled;

	/**
	 * Initially this value is set to true if the a non-null reactivation
	 * predicate is defined. After initialization the simulator sets the
	 * value to true if the reactivation predicate is true at enabling time.
	 */
	bool Reactivation;



	/**
	 * linked list of a state variables affected by action firing
	 */
	ActionAffectsElement* Affects;

	/**
	 * This was originally used to keep track of which action in a
	 * group was fired, but since the  the Fire method returns the
	 * action that fired there is no longer a need for this data member.
	 * It being maintained for consistency sake.
	 */
	bool Fired;

	/**
	 * The action probability distribution function
	 */
	Distribution DistributionType;

	/**
	 * A pointer the action's event on the future event list
	 */
	void* EventPointer;

	/**
	 * The fraction of the action time was complete before it became
	 * disabled before firing.  This is  only important for Race-Age
	 * execution policy in which actions restart from where they left of when
	 * they became disabled.
	 */
	double FractionComplete;

	/**
	 * pointer to model that contains the activity
	 */
	void* TheModel;

	/**
	 * the set of PV nodes whose impulse reward is affected by this action
	 */
	FastSet AffectedNodesByImpulseReward;

	/**
	 * The total number of performance variables collected to date.  As
	 * each performance variable is collected over an action it increments
	 * the actions TotalNumCollected variable until it reaches it NumAffects
	 * value at which point AffectedNodes is set to NULL so that the simulator,
	 * who looks at AffectedNodes, does nothing.
	 */
	int TotalNumCollected;

	/**
	 * The set of indices of PV nodes affected by this action
	 */
	FastSet AffectedNodes;

	/**
	 * The set of indices of PV nodes whose rate reward is affected by the action
	 */
	FastSet AffectedNodesByRateReward;

	/**
	 * Moved from SAN folder where we had a separate distribution for
	 * erlang activity to this location. This is the first step to fix
	 * the Model level AFI. All this methods and variable are specific
	 * to erlang activity.
	 */
	short phase;

	/**
	* This data member desribes the action execution policy.
	*/
	ExecutionPolicyType ExecutionPolicy;

	int GroupID;

protected:

	/**
	 *  give each activity a unique id number (used by the simulator to
	 *  speed up model initialization; see
	 *  Simulator::GenerateConnectivityList
	 */
	static int MaxID;
	int ID;


	/**
	 * This is the action's name
	 */
	char* ActionName;

	/**
	 * The length of the impulse workers array in AffectedImpulseWorkers
	 */
	int **NumAffectedImpulseWorkers;

	/**
	 * The list of workers defined on the impulse affecting impulses
	 * (from AffectedImpulseRewards)
	 */
	int ***AffectedImpulseWorkers;


	/**
	 * An array of sets of rate workers. One set of rate workers is defined for
	 * each node. AffectedRateWorkers[i] is the set of rate workers
	 * that this action affects. The size of the array is the
	 * number of performance variables defined on the model
	 * AffectedRateWorkers==NULL means that this action does not affect any
	 * workers (and subsequently, nodes).
	 */
	FastSet* AffectedRateWorkers;

	/**
	 * Array of sets of impulse reward indices. AffectedImpulseRewards[i]
	 * ( 0 <= i < # of PV nodes) is the set of impulse rewards of
	 * PV node i that this action affects.
	 */
	FastSet *AffectedImpulseRewards;


	/**
	 * This data structure holds an array of pointers to all state
	 * variables affected by this action's
	 *  firing and the length of that array
	 */
	List<BaseStateVariableClass> AffectedStateVariables;

	/**
	 * This data structure holds an array of pointers to all state
	 * variables that enable this action and the length of that array.
	 */
	List<BaseStateVariableClass> EnablingStateVariables;

}; // end BaseActionClass

/**
 * This is a class describing an element in a linked list which describes
 * a set of actions that affect a state variable
 *
 *  @author Jay M. Doyle
 *  @version Mobius version 1.0
 *
 */
class ActionAffectsElement {
public:

	/**
	 * A pointer to an action that affects a particular state variable
	 */
	BaseActionClass* Action;

	/**
	 * A pointer to the next element in the linked list
	 */
	ActionAffectsElement* next;



}; // end class ActionAffectsElement

#endif // BASEACTIONCLASS_H_
