/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: BaseGroupClass.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef BASEGROUPCLASS_H
#define BASEGROUPCLASS_H

#include "BaseActionClass.h"
#include "StandardErrors.h"
#include "debug.h"
#include <string>
#include <cstddef>
#include <cstdio>
#include <iostream>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

extern UserDistributions* TheDistribution;

// Forward declaration
class BaseGroupClass;

/**
 *  This structure is used in action groups to keep track weight and their actions
 */
struct GroupWeight {

	/**
	 *  The value of an action's weight a particular enbabling state
	 */
	double Weight;

	/**
	 *  A pointer to the group whose weight is stored in the Weight data member
	 */
	BaseGroupClass* AssociatedAction;

}; // end GroupWeight

/**
 *  This class describes the basic requirements of an action
 *  group in the Mobius framwork.  An action group is a collection
 *  of  one or more actions for which there is only one representative
 *  that can be "fireable" at any given time.  The group representative is
 *  choosen from among the set of enabled actions base upon their rank and weight.
 *  The exact select algorithm is determined by the specific action group.
 *  Action group allow for a more robust execution policies including
 *  preselection.
 *
 *  @author Jay M. Doyle
 *  @version Mobius Version 1.0
 */
class BaseGroupClass: public BaseActionClass {

public:



	/**
	 *  Pointer to action selected as the group's representative
	 */
	BaseActionClass* SelectedAction;

	/**
	 *  An array of groups weights, one per every action enabled in the curent state
	 */
	GroupWeight* ActionWeights;

	/**
	 *  The highest rank of any of the enabled actions in the current enabling state
	 */
	int HighestRank;

	/**
	 *  The number of action contained in the action group
	 */
	int NumActions;

	/**
	 *   The number of action that are selectable in the current
	 *  state (i.e. the number of enabled of
	 *  action that belong to the highest ranking group).
	 */
	int NumSelectableActions;

	/**
	 *  Base constructor
	 */
	BaseGroupClass();

	/**
	 *  Overloaded constructor that take all necessary information for construction
	 *  an action group
	 *
	 *  @param TheGroupName The name of the action group
	 *  @param TheGroupIDNumber The unique number identifying the action group
	 *  @param TheDistributionType The type of distribution used to
	 *         initialize the DistributionType on
	 *          the BaseActionClass from which BaseGroupClass is derived.
	 *  @param TheExecutionPolicy  The type of execution policy used to
	 *         initialize the DistributionType on
	 *         the BaseActionClass from which BaseGroupClass is derived.
	 *  @param TheNumberOfAffectedStateVariables The number of state
	 *         variables whose state is affected by
	 *         at least one member in the action group
	 *  @param TheNumberOfEnablingStateVariables The number of state
	 *         variables whose state enables at least
	 *         one member in the action group.
	 *  @param The initial value of reactivation data element
	 *
	 */
	BaseGroupClass(char* TheGroupName, int TheGroupIDNumber, Distribution TheDistributionType, ExecutionPolicyType TheExecutionPolicyType,
			int TheNumberOfAffectedStateVariables, int TheNumberOfEnablingStateVariables, bool TheReactivation);

	/**
	 *  This method initialize all the initializes all the internal data
	 *  structures for the action group.
	 *  @param TheGroupName The name of the action group
	 *  @param TheGroupIDNumber The unique number identifying the action group
	 *  @param TheDistributionType The type of distribution used to
	 *         initialize the DistributionType on
	 *         the BaseActionClass from which BaseGroupClass is derived.
	 *  @param TheExecutionPolicy  The type of execution policy used
	 *         to initialize the DistributionType on
	 *         the BaseActionClass from which BaseGroupClass is derived.
	 *
	 *  @param TheNumberOfAffectedStateVariables The number of state
	 *         variables whose state is affected by
	 *         at least one member in the action group
	 *  @param TheNumberOfEnablingStateVariables The number of state
	 *         variables whose state enables at least
	 *         one member in the action group.
	 *
	 *  @param The initial value of reactivation data element
	 */
	void BaseGroupClassInitialize(char* TheGroupName, int TheGroupIDNumber, Distribution TheDistributionType, ExecutionPolicyType TheExecutionPolicy,
			int TheNumberOfAffectedStateVariables, int TheNumberOfEnablingStateVariables, bool TheReactivation);

	/**
	 *  This method initializes all the internal data structures
	 *
	 * @param TheNumberOfMembers The number of action that
	 *        will be part of the postselection group.
	 * @param TheGroupName The name of the group
	 *
	 */
	void initialize(int TheNumberOfMembers, char* TheGroupName);

	/**
	 *  This method initialize all the initializes all the internal data
	 *  structures for the action group.
	 *  @param TheGroupName The name of the action group
	 *  @param TheGroupIDNumber The unique number identifying the action group
	 *  @param TheDistributionType The type of distribution used to
	 *         initialize the DistributionType on
	 *         the BaseActionClass from which BaseGroupClass is derived.
	 *  @param TheExecutionPolicy  The type of execution policy used
	 *         to initialize the DistributionType on
	 *         the BaseActionClass from which BaseGroupClass is derived.
	 *
	 *  @param TheNumberOfAffectedStateVariables The number of state
	 *         variables whose state is affected by
	 *         at least one member in the action group
	 *  @param TheNumberOfEnablingStateVariables The number of state
	 *         variables whose state enables at least
	 *         one member in the action group.
	 *  @param The initial value of reactivation data element
	 *  @param NumberOfMemberGroups The number of groups contained
	 *         in this group
	 */
	BaseGroupClass(char* TheGroupName, int TheGroupIDNumber, Distribution TheDistributionType, ExecutionPolicyType TheExecutionPolicy,
			int TheNumberOfAffectedStateVariables, int TheNumberOfEnablingStateVariables, bool TheReactivation, int NumberOfMemberGroups);
	/**
	 *  Base destructor
	 */
	virtual ~BaseGroupClass();

	/**
	 *  This method set the group member at specific index
	 *
	 *   @param TheMemberGroup A group to be contained in this group
	 *   @param TheIndex The index at which the group pointer should
	 *          be stored in GroupArray
	 */
	void setMember(BaseGroupClass* TheMemberGroup, int TheIndex);

	/**
	 *  This method appends the specified group
	 *  to this group's GroupArray, forming a
	 *  multi-level group.
	 *
	 *  @param TheNewMember A new member of the group
	 */
	void appendGroup(BaseGroupClass* TheNewMember);

	/**
	 *  This method copies the group members from the group
	 *  passed in as a parameter.  If this group does not have
	 *  room in the GroupArray, new space will be allocated, the
	 *  old members will be copied to the new array, the other group
	 *  members will be appended, and the old array will be deallocated.
	 *  This method differs from appendGroup in the sense that this
	 *  method takes TheMemberGroup's members and appends them to
	 *  this group's GroupArray instead of just appending TheMemberGroup
	 *  to the GroupArray.
	 *
	 *  @param TheMemberGroup A pointer to a group to be added to
	 *         the array of contained groups
	 */
	void appendMembers(BaseGroupClass* TheMemberGroup);

	/**
	 *  This method determines if there is at least one member
	 * action enabled in the current state
	 *
	 *  @return True if there is at least one enabled
	 *          member action in the current state
	 */
	bool Enabled();

	/** This method overrides the Weight method on BaseActionClass.
	 * Essentially, it returns the group's weight.
	 *
	 *  @return The value of the group's weight in the current marking
	 */
	double Weight();

	/** This method overrides the Rate method on BaseActionClass.
	 * Essentially, it returns the the group's rate.
	 *
	 * @param The value of the group's rate in the current marking
	 */
	double Rate();

	/** This method returns the probability of selection the
	 * specified member action from amoung the set
	 * of enabled memeber action in the current state.
	 *
	 *  @param TheMemberAction A pointer to an action in the action group
	 *  @return The probability of selection the specified action in the current state
	 */
	double Probability(BaseActionClass* TheMemberAction);

	/**  This method checks to see if the specified action is a member of the action group
	 *
	 * @param TheSpecifiedAction The action in question
	 * @return True if the specified action is a member of the group
	 */
	bool IsAMember(BaseActionClass* TheSpecifiedAction);

	/**  This method replaces a member by a new one, if the action
	 *  to be replaced, is not a member nothing is done.
	 *
	 * @param theOld, the action to be replaced
	 * @param theNew, the substitute
	 */

	void replaceMember(BaseActionClass*, BaseActionClass*);

	/** This method overrides the ReactivationPredicate method defined on BaseActionClass
	 *
	 *  @return True if the group is reactivable, false otherwise
	 */
	virtual bool ReactivationPredicate();

	/** This method override sthe ReactivationFunction method defined on BaseActionClass
	 *
	 * @return True if the group reactivation function is true for the current state
	 */
	virtual bool ReactivationFunction();

	/**  This method returns a sample from the selected action's distribution
	 *  (overrides BaseActionClass::SampleDistribution())
	 *
	 *  @return A value selected from the group's selected action distribution
	 */
	virtual double SampleDistribution();

	/**  This method fires the group's selected action (overrides
	 *  BaseActionClass::SampleDistribution())
	 *
	 *  @return A pointer to the member action that fired
	 */
	virtual BaseActionClass* Fire();

	/** This method returns the groups's rank -- not necessarily
	 * the same as the selected action rank
	 */
	int Rank();

	/**  This method performs the selection algorithm on the group
	 *  and defines which one of the group's
	 *  action is selected.
	 *
	 *  @param This method returns true is there is at least
	 *  on enabled action in the current state and returns false
	 *  otherwise.  A false return code signifies that an action was
	 *  not selected
	 */
	virtual bool SelectAction();

	/** This method is used in calculating the probability
	 * of selecting each member action in the current
	 * state.
	 *
	 *  @return The sum all the enabled, highest
	 *   ranked actions in the current state.  A negative
	 *   return value indicates that there are no enabled
	 *   members in the current state
	 */
	double CalculateWeightDistribution();

	/** This method checks to see if there has been a change in
	 * the enabling state since the last time
	 * enable was called.
	 *
	 * @return True if there has been a change in the enabling state
	 */
	bool EnablingChange();

	/** This method prints out debugging information
	 */
	void Check();

	/** This is  another method that prints out debugging information
	 */
	void PrintInfo();

	/** this method prints the state variables and values necessary for
	 * enabling each action in this group
	 */
	virtual void printEnablingState(ostream *strm);

	// Needed to find the action in the group to print
	virtual BaseActionClass* getPrintedAction();

	/**
	 *  This returns the number of members
	 *  of the group.
	 *
	 *  @return The number of members in the
	 *    group.
	 */
	inline int getNumMembers() {
		return NumActions;
	}



	/*Function to return the index of the group array*/


	inline BaseGroupClass** getGroupArray()
	{
		return GroupArray;
	}

	/**
	 *  This method returns the specified element from
	 *  the member array.
	 *
	 *  @param TheIndex A valid index for the GroupArray
	 *  @return The group member at the specified index.
	 */
	inline BaseGroupClass* getMember(int TheIndex) {
		ASSERT(TheIndex>=0);
		DS3("Index into GroupArray", TheIndex);
		DS3("Number of members", NumActions);
		ASSERT(TheIndex<NumActions);
		ASSERT(GroupArray[TheIndex]!=NULL);
		return GroupArray[TheIndex];
	} // end getMember(int)

	/**
	 *  This method recursively prints out
	 *  the group members
	 */
	void printGroup();

	/**
	 *  This method changes the group ID of all the
	 *  group members to the specified value
	 *
	 *  @param TheNewID The new value ID for the group
	 */
	virtual void setGroupId(int TheNewID);

protected:



	/**
	 *  An array of pointers to member groups
	 *  contained in this group
	 */
	BaseGroupClass** GroupArray;

	/**
	 * This is the size of the member array
	 */
	int ArraySize;

	/**
	 *  This method increases the size of the
	 *  GroupArray and GroupWeight arrays to
	 *  some new size specified by the input
	 *  parameter.  This method is used in the
	 *  append methods.
	 *
	 *  @param TheNewSize The new size of the
	 *   GroupArray and GroupWeight arrays.
	 */
	void enlargeArrays(int TheNewSize);

	/**
	 *  This method is used to recursively print
	 *  out group members.  ThePrefix is the number
	 *  used before the local index number to denote
	 *  the members group parents.  The indent space
	 *  in the number of spaces to indent from the left
	 *  side for all group members.
	 */
	void printGroup(char* ThePrefix, int IndentSpace);

private:

}; // end BaseGroupClass class

#endif
