/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: BaseModelClass.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef BASEMODELCLASS_H_
#define BASEMODELCLASS_H_

#include "BaseActionClass.h"
#include "BaseGroupClass.h"
#include "List.h"
#include "MobiusStructs.h"
#include "State.h"

// Determine if the C++ compiler support the C++11 forward_list
#if defined(__has_include)
#if __has_include(<forward_list>)
#define HAS_FORWARD_LIST 1
#endif
#elif (__cplusplus >= 201103L) || (_MSC_VER >= 1800) 
#define HAS_FORWARD_LIST 1
#endif

#if (HAS_FORWARD_LIST)

#include <forward_list>
typedef std::forward_list<BaseActionClass*> base_action_list_t;

#else

#include <ext/slist>
typedef __gnu_cxx::slist<BaseActionClass*> base_action_list_t;

#endif

// this extern added so models can use trout functionality.
extern std::ostream *trout;

// we cant just include the header, because its header
// includes the header for BaseModelClass

class SharedStateVarLink;

/**
 *  This is the base class for all models defined in the
 *  Mobius modeling framework.  This class outlines all the
 *  necessary methods that needed to be overloaded to compleletly
 *  describe a Mobius model.
 *
 *  @author Jay M. Doyle
 *  @version Mobius Version 1.0
 *
 */
class BaseModelClass {
public:
	/**
	 * Default constructor and destructor
	 */
	BaseModelClass();
	virtual ~BaseModelClass();
	/**
	 * This method is used to define the name of the model
	 *
	 * @param TheModelName The new name of the model
	 */
	void DefineName(const char* const TheModelName);

	/**
	 *  This method returns the number of actions
	 *  defined in the model
	 */
	inline int getNumActions() const {
		return NumActions;
	}

	/**
	 *  This method returns the number of state variables
	 *  in the model.
	 *
	 *  @return The number of state variables defined in
	 *          this model
	 */
	virtual inline int getNumStateVariables() {
		return NumStateVariables;
	}

	/**
	 *  This method returns the number of equivalence shared
	 *  state variables in this model.
	 *
	 *  @return The numbero equivalence shared state variables
	 *          defined in this model
	 */
	inline int getNumSharedStateVariables() const {
		return NumSharedStateVariables;
	}

	/**
	 *  This method returns the array of action members
	 */
	inline BaseActionClass** getActionList() const {
		return ActionList;
	}

	/**
	 *  This method returns the array of group pointers
	 */
	inline BaseGroupClass** getGroupList() const {
		return GroupList;
	}

	/**
	 *  This method returns the number of groups
	 *  defined in the model
	 */
	inline int getNumGroups() const {
		return NumGroups;
	}

	/**
	 *  This method returns the name of the model.
	 *
	 *  @return The name given to the model by the user
	 */
	inline const char* const getName() const {
		return Name;
	}

	/**
	 * Returns the Id.
	 */
	inline int getId() const {
		return Id;
	}

	/**
		 * Returns the LastActionTime.
	*/
	inline int getLastActionTime() const {
			return LastActionTime;
	}

	/**  (added 01/28/08 by Ruth Lamprecht)
	 * This method returns the ParentId
	 */
	inline int getParentId() const {
		return ParentId;
	}

	/* The following must be defined by the formalism specific model*/

	/**
	 *  This method returns the cardinality of Mobius models with
	 *  the specified name
	 *
	 *  @param TheSpecifiedModelName The name of a Mobius model
	 *  @param TheContainer An array of pointers to models with the given name
	 */
	virtual void listModels(char* TheSpecifiedModelName, List<BaseModelClass>* TheContainer)=0;

	/**
	 *  This method retrieve the current state of the model and
	 *  puts it into the specified memory location
	 *
	 *  @param TheSpecifiedMemoryLocation The location into which
	 *    the current model state will be stored
	 */
	virtual void CurrentState(void* TheSpecifiedMemoryLocation)=0;

	/**
	 *  This method returns a list of all the actions defined
	 *   within the model
	 *
	 *  @return An array of pointers to all the actions defined
	 *     in the model
	 */
	virtual void listActions(List<BaseActionClass>* TheContainer)=0;

	/**
	 *  This method retuns a lis of all the action groups defined
	 *   within the model
	 *
	 * @return An array of pointer to all action groups defined
	 *    within the model
	 */
	virtual void listGroups(List<BaseGroupClass>* TheContainer)=0;

	/**
	 *  This method sets the model to the state value located at the
	 *    specified memory location
	 *
	 *  @param TheSpecifiedMemoryLocation The location in memory
	 *    holding the new model state
	 */
	virtual void SetState(void* TheSpecifiedMemoryLocation)=0;

	/**
	 *  This method returns the size of the model state (in bytes)
	 *
	 *  @return The size of the model state in bytes
	 */
	virtual int StateSize()=0;

	/**
	 *  This method compares to model states and determines whether
	 *  they are the same state.   Since different modeling formalism may
	 *  have different representation of state, a simple memory compare
	 *  may not realize that one state may be represented in more than
	 *  one way.
	 *
	 *  @param TheFirstState The location of the first model state
	 *  @param TheSecondState The location of the second model state
	 *  @return True if both values represent the same state, false otherwise
	 *
	 */
	virtual bool CompareState(void* TheFirstState, void* TheSecondState)=0;

	/**
	 *  This method generates an array of pointers to all the actions with
	 *  the specified name.  Different atomic models are allowed to have an
	 *  action by the same name, hence this list can have more than one
	 *  element.
	 *
	 * @param TheSpecifiedActionName The name of the specified action
	 * @param TheContainer An object containing a collection of
	 *        pointers to actions which have the specified action name
	 */
	virtual void listActions(char* TheSpecifiedActionName, List<BaseActionClass>* TheContainer)=0;

	/**
	 *  This method creates a list of pointers to state variables who
	 *  belong to the specified model and whose name is the same as the
	 *  specified place name.
	 *
	 *  @param TheSpecifiedModelName The name of the model in which to
	 *         search for the state variable
	 *  @param TheSpecifiedStateVariableName The name of the state variable
	 *  @param TheSVContainer A data structure containing an
	 *         array of pointers to state variables
	 *  @param TheModelContainer A data structure containing an
	 *         array of pointer to models in which the
	 *         state variables listed in TheSVContainer are contained.
	 */
	virtual void listSVs(char* TheSpecifiedModelName, char* TheSpecifiedStateVariableName, List<BaseStateVariableClass>* TheSVContainer,
			List<BaseModelClass>* TheModelContiner)=0;

	/**
	 * This method prints the model state in a readable format to strm
	 */
	virtual void printState(ostream *strm);
	virtual void printState();

	/**
	 * These methods return thed local and shared SVs
	 */
	virtual BaseStateVariableClass** GetLocalStateVariables(ostream *strm);
	virtual BaseStateVariableClass** GetSharedStateVariables();

	/**
	 *  This method hierarchically determines the highest-level state
	 *  variable that the state variable has been shared with through
	 *  the composer tree. The default implementation will return null
	 */
	virtual BaseStateVariableClass*
	getMainSharedVariable(BaseStateVariableClass*);

	/**
	 *  This method is to used to hierarchically build groups of equivalent
	 *  state variables shared at each level in the composer tree.
	 *  The default implementation will return null.
	 */
	virtual SharedStateVarLink*
	getListOfSharedVariables(BaseStateVariableClass*);

	/**
	 *  a convenience method that calls the affected methods for both
	 *  the action and the state variable, to reduce the amount of
	 *  code written in model definition classes.  Should be called
	 *  for each affected arc between actions and state variables.
	 */
	virtual void AddAffectArc(BaseStateVariableClass*, BaseActionClass*);

	/**
	 *  a convenience method that calls the enabled methods for both
	 *  the action and the state variable, to reduce the amount of
	 *  code written in model definition classes.  Should be called
	 *  for each enabled arc between actions and state variables.
	 */
	virtual void AddEnableArc(BaseStateVariableClass*, BaseActionClass*);

	/** (added 01/28/08 by Ruth Lamprecht)
	 * This method assigns ParentId to the value of NewValue.
	 */
	void SetParentId(int NewValue);

	/* public variables */



public:
	/**
	 * The following member declarations are used by the sybmolic SSG
	 * algorithm
	 */
	// MGMGRAPH - Need to add 'GraphJoin'
	typedef enum {
		InvalidModel, AtomicModel, JoinModel, RepModel, GraphJoinModel, GraphWrapperModel
	} ModelType;
	virtual ModelType GetModelType();
	virtual int GetNumberOfChildren();
	virtual BaseModelClass* GetChild(int i);
	virtual void GetLocalActions(base_action_list_t& actions);
	virtual void GetGlobalActions(base_action_list_t& actions);
	virtual bool HasLocalState();
	virtual void GetLocalState(State* _LS);
	virtual void SetLocalState(const State* _LS);
	virtual void ConstructLocalState(State*& _LS);
	virtual void DestructLocalState(State* _LS);
	/**
	 * The number of state variables defined in the model.
	 * This includes state variables that are shared with
	 * other state variables.  The sharing can either be
	 * equivalence or uni-directional functional sharing.
	 */
	int NumStateVariables;

	/**
	 * The number of equivalence shared state variables
	 */
	int NumSharedStateVariables;

	/**
	 * The number of actions defined in the model
	 */
	int NumActions;

	/**
	 * The number of action groups defined in the model
	 */
	int NumGroups;


protected:

	/**
	 * The name of the model
	 */
	char* Name;

	/**
	 *  An array of pointers to all actions in the model
	 */
	BaseActionClass** ActionList;

	/**
	 *  A array of pointers to all the action groups defined
	 *  within the model
	 */
	BaseGroupClass** GroupList;

	/**
	 *  Id: assigned by mobius as a way to give each model a unique identity
	 */
	int Id;

	/**  (added 01/28/08 by Ruth Lamprecht)
	 * Id of the parent node, own Id if not shared
	 */
	int ParentId;

	/**
	 * The number of performance variables defined on the model
	 */
	int NumPVs;

	/**
	 * The last time that an action was fired in the underlying model.
	 */
	static double LastActionTime;

}; // end class BaseModelClass

#endif // BASEMODELCLASS_H_
