/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: BaseStateVariableClass.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef BASESTATEVARIABLECLASS_H_
#define BASESTATEVARIABLECLASS_H_

#include "List.h"
#include "Utility.hpp"
#include "BaseActionClass.h"

/**
 *  This class describes the base functionality of all
 *  state variables in the Mobius framework.
 *  All state variable for all formalism must fulfill
 *  the pure virtual methods defined in this class.
 */
class BaseStateVariableClass {

public:

	/* public variables */
	char* StateVariableName;
	inline char* getName() {
		return (StateVariableName);
	}
	;

	/**
	 * Returns the fully qualified name for the state variable.
	 */
	inline const char* getFullyQualifiedName() {
		if(VariableNamePrefix[0]) {	//not empty string
			return (std::string(VariableNamePrefix) + "." + StateVariableName).c_str();
		} else {	//is empty; we don't want the "."
			return StateVariableName;
		}
	}

	/* public methods */

	/**
	 * Base Constructor
	 */
	BaseStateVariableClass();

	/**
	 *  Overloaded constructor
	 *
	 * @param TheName The name of the state variable
	 */
	BaseStateVariableClass(const char* TheName, const char* prefix);

	BaseStateVariableClass(const BaseStateVariableClass& rhs) = delete;
	BaseStateVariableClass& operator=(const BaseStateVariableClass&) = delete;

	/**
	 * Destructor
	 */
	virtual ~BaseStateVariableClass();

	/**
	 * Sets the StateVariableName date member
	 *
	 * @param TheNewName The new name of the
	 *     model
	 */
	void SetName(const char* TheNewName);

	/**
	 * Sets the StateVariablePrefix member
	 *
	 * @param prefix The new prefix
	 */

	void SetPrefix(const char* prefix);

	/**
	 *  This method returns number of bytes of compact
	 *  state variable representation
	 *
	 *  @return The number of bytes needed to represent
	 *      the state variable's state
	 */

	virtual int StateSize()=0;

	/**
	 *  This method sets the state of the StateVariable
	 *
	 *  @param TheNewState The location of the new value
	 *   for the state variable
	 */
	virtual void SetState(void* TheNewState)=0;

	/**
	 *  This method writes the state variables current state
	 *  to the specified memory location
	 *
	 *  @param TheMemoryLocation The location in memory to which
	 *     the state variable state  should be stored
	 */
	virtual void CurrentState(void* TheMemoryLocation)=0;

	/**
	 * This method prints state of state variable to cout
	 */
	virtual void printState(std::ostream *strm);
	virtual void printState();

	/**
	 * This methods prints just the value of the state
	 */
	virtual void printStateVal(std::ostream *strm);
	/**
	 *  This method returns the value of the Shared data member
	 *
	 * @return The value of the Shared data member
	 */
	bool getShared();

	/**
	 * This method returns the value of the Stored data member
	 *
	 * @return The value of the Stored data member
	 */
	bool getStored();

	/**
	 *  This method returns the value of the FunctionallyShared
	 *  data member
	 *
	 * @return The value of the FunctionallyShared data member
	 */
	bool getFunctionallyShared();

	/**
	 * This method returns the number of state variables that are
	 * shared with this state variable.
	 *
	 * @return The number of state variables shared with this one
	 */
	int getSharingCount();

	/**
	 *	This method tells the variable to share the same data
	 *	as the other variable
	 *
	 *	@param other: the other basestatevariableclass variable
	 */
	virtual void share(BaseStateVariableClass* other) = 0;

	/**
	 * This method set the value of the FunctionallyShared data member
	 *
	 * @param TheNewFunctionallySharedValue The new value for
	 *        functionally shared
	 */
	void setFunctionallyShared(bool TheNewFunctionallySharedValue);

	/**
	 *  This method returns a data structure with the list
	 *  of actions whose firing affects this state variable.
	 *
	 *  @return The SVAffectingActions
	 */
	List<BaseActionClass>* getAffectingActions();

	/**
	 *  This method returns the ith affecting action
	 *  from the state variable AffectingActions
	 *  List data member.  If the i is greater
	 *  than or equal to the size of List then
	 *  a null pointer in returned.
	 *
	 *  @param TheIndex The index of the affecting
	 *         action
	 *  @return The ith affecting action if i is
	 *          within bounds, else null
	 */
	BaseActionClass* getAffectingAction(int TheIndex);

	/**
	 *  This method returns the number of actions that
	 *  affect this state variable.
	 *
	 *  @return The number of actions in the state variable's
	 *    AffectingActions list.
	 */
	int getNumAffectingActions();

	/**
	 * This method appends the specified action to the state variable's
	 * AffectingActions object.   The state variable will
	 * use the append method defined on the AffectingActions class to add
	 * the action to the AffectingAction object.
	 */
	virtual void appendAffectingAction(BaseActionClass* TheAction);

	/**
	 * This method removes the specified action of the state variable's
	 * AffectingActions object.   The state variable will
	 * use the remove method defined on the AffectingActions class to add
	 * the action to the AffectingAction object.
	 */
	virtual void removeAffectingAction(BaseActionClass* TheAction);

	/**
	 * This method replaces the TheAction of the state variable's
	 * AffectingActions object by theNewAction object.   The state variable will
	 * use the replace method defined on the AffectingActions class to add
	 * the action to the AffectingAction object.
	 */
	virtual void replaceAffectingAction(BaseActionClass* TheAction, BaseActionClass* theNeNewAction);

	/**
	 *  This method copies the data structure passed in and
	 *  uses it as its list of affecting actions.
	 *
	 *  @param TheActions The new list of affecting actions
	 */
	void copyAffectingActions(List<BaseActionClass>* TheActions);

	/**
	 *  these oddly-named method allows actions to depend on all
	 *  leaf nodes of array and structure SV, without having to expose
	 *  all leaf node pointers directly to the action.  Called by
	 *  BaseActionClass addEnablingSV() and addAffectedSV()
	 */
	virtual void addSelfAsEnablingSVToAction(BaseActionClass *action);
	virtual void addSelfAsAffectedSVToAction(BaseActionClass *action);
	virtual void removeSelfFromEnablingSVOfAction(BaseActionClass *action);
	virtual void removeSelfFromAffectedSVOfAction(BaseActionClass *action);

	/**
	 * This is a debugging method that will print the name of the
	 * state variable along with the names of all the actions that
	 * affect the state variable (as specified by the AffectingActions
	 * data member).
	 */
	void printAffectingActions();

	/**
	 * This function gets the list of base state variable class leaf nodes
	 * for all the elements present in the StateVariable. For simple places,
	 * this would return only the list with single entry of itself, else it
	 * would return the leaf nodes
	 * CARE MUST BE TAKEN TO FREE THE MEMORY ALLOCATED TO THE VARIABLES AND
	 * DELETE THEM.
	 */

	virtual List<BaseStateVariableClass> * getLeafNodes();
	virtual int getNumLeafNodes();
	virtual bool isLeafNode();

	/**
	 *  This method returns a data structure with the list
	 *  of actions whose firing affects this state variable.
	 *
	 *  @return The SVEnabledActions
	 */
	List<BaseActionClass>* getEnabledActions();

	/**
	 *  This returns the ith member of the actions's
	 *  EnabledAction class data element.  If the index
	 *  is out of bounds then a NULL pointer will be
	 *  returned.
	 *
	 *  @param TheIndex The index into the EnabledActions
	 *     data structure.
	 *  @return The specificed enabled action or a null
	 *          pointer if the index is out of bounds.
	 */
	BaseActionClass* getEnabledAction(int TheIndex);

	/**
	 *  This method returns the number of actions that
	 *  affect this state variable.
	 *
	 *  @return The number of actions in the state variable's
	 *    EnabledActions list.
	 */
	int getNumEnabledActions();

	/**
	 * This method appends the specified action to the state variable's
	 * EnabledActions object.   The state variable will
	 * use the append method defined on the EnabledActions class to add
	 * the action to the EnabledAction object.
	 */
	virtual void appendEnabledAction(BaseActionClass* TheAction);

	/**
	 * This method removes the specified action from the state variable's
	 * EnabledActions object.   The state variable will
	 * use the remove method defined on the EnabledActions class to add
	 * the action to the EnabledAction object.
	 */
	virtual void removeEnabledAction(BaseActionClass* TheAction);

	/**
	 * This method replaces the TheAction action object from the state variable's
	 * EnabledActions object by the theNewAction action object.
	 * The state variable will
	 * use the replace method defined on the EnabledActions class to add
	 * the action to the EnabledAction object.
	 */
	virtual void replaceEnabledAction(BaseActionClass* TheAction, BaseActionClass* theNewAction);

	/**
	 *  This method copies the data structure passed in and
	 *  uses it as its list of affecting actions.
	 *
	 *  @param TheActions The new list of affecting actions
	 */
	void copyEnabledActions(List<BaseActionClass>* TheActions);

	/**
	 * This is a debugging method that will print the name of the
	 * state variable along with the names of all the actions that
	 * affect the state variable (as specified by the EnabledActions
	 * data member).
	 */
	void printEnabledActions();

	/**
	 *  This method set the SVAffectingActions data member
	 *  to the reference passed in a parameter.  This method
	 *  is used when two state variables are shared through
	 *  equivalence sharing.
	 *
	 *  @param TheAction The list of affecting actions for
	 *         the equivalence group to which this state
	 *         variable belongs.
	 */
	void setAffectingActions(List<BaseActionClass>* TheAction);

	/**
	 *  This method set the SVEnabledActions data member
	 *  to the reference passed in a parameter.  This method
	 *  is used when two state variables are shared through
	 *  equivalence sharing.
	 *
	 *  @param TheAction The list of enabled actions for
	 *         the equivalence group to which this state
	 *         variable belongs.
	 */
	void setEnabledActions(List<BaseActionClass>* TheAction);

protected:
	/**
	 *  This method sets the value of Shared data member
	 *
	 *  @param TheNewSharedValue The new value of the shared value
	 */
	void setShared(bool TheNewSharedValue);

	/**
	 *  This method sets the value of Stored data member
	 *
	 *  @param TheNewStoredValue The new value of the stored value
	 */
	void setStored(bool);


	/**
	 * Stores a string containing the fully qualified name of the state variable minus the local name (which is stored in StateVariableName)
	 */
	char* VariableNamePrefix;

	/**
	 * false if the state variable data
	 * structure is stored in another model
	 */
	bool Stored;

	/**
	 * true if value of state  is defined as a
	 * function on some other state-holding entity
	 */
	bool FunctionallyShared;

	/**
	 * Number of other state variable that
	 * share this member location include yourself
	 */
	int SharingCount;

	/**
	 *  This is a list of actions whose firing affects
	 *  this state variable's state
	 */
	List<BaseActionClass>* SVAffectingActions;

	/**
	 *  This is a list of actions whose Enabled function
	 *  is a function of this state variable's state.
	 */
	List<BaseActionClass>* SVEnabledActions;

private:
	/**
	 * true if state variable is shared with another model
	 */
	bool mShared = false;

}; // end BaseStateVariableClass

#endif // BASESTATEVARIABLECLASS_H_
