/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: CPUTime
 *
 *      Authors: Tod Courtney
 *      Creation Date: 2001
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/
/////////////////////////////////////////////////////////////
//
//  CPUTime - this class is used to measure the CPU time computed by a
//  process.  Specifically, it measures the time between the last call
//  to Start and the current call to GetTime.  Time is always measured
//  relative to the last call to Start(), independent of the number of
//  calls to GetTime.

#ifndef _CPU_TIME_H_
#define _CPU_TIME_H_
#include <stdlib.h>
#include <time.h>

#ifndef MINGW
// for UNIX getrusage
#include <sys/resource.h>
#endif

class CPUTime {
private:
	double _initialCPUClock;
	double _initialSYSClock;
#ifndef MINGW
	struct rusage _rusageStruct;
#endif

public:
	inline CPUTime();
	inline ~CPUTime();

	inline void Start();
	inline double GetTime();
	inline void GetTime(double &cpu, double &sys);
};

////////////////////////////////////////////////////////////////
//
//  Constructor/Destructor
//
inline CPUTime::CPUTime() {
}
inline CPUTime::~CPUTime() {
}

////////////////////////////////////////////////////////////////
//
//  Start() - This starts the timer.  Subsequent calls to GetTime
//  measure the amount of CPU used by the process between this call to
//  Start and the call to GetTime
//
inline void CPUTime::Start() {
#ifdef MINGW
	// THIS IS FOR WINDOWS
	_initialCPUClock=clock();
	_initialSYSClock=0;
#else
	// THIS IS FOR UNIX
	getrusage(RUSAGE_SELF, &_rusageStruct);
	// the initial time is the number of seconds plus the fraction
	// representing the number of milliseconds
	_initialCPUClock = _rusageStruct.ru_utime.tv_sec + _rusageStruct.ru_utime.tv_usec / 1E6;
	_initialSYSClock = _rusageStruct.ru_stime.tv_sec + _rusageStruct.ru_stime.tv_usec / 1E6;
#endif

}

inline double CPUTime::GetTime() {
	double ret;
#ifdef MINGW
	// THIS IS FOR WINDOWS
	double now = clock();
	ret=(now-_initialCPUClock)/(double)CLOCKS_PER_SEC;
#else
	// THIS IS FOR UNIX
	getrusage(RUSAGE_SELF, &_rusageStruct);
	// the current time is the number of seconds plus the fraction
	// representing the number of milliseconds
	ret = _rusageStruct.ru_utime.tv_sec + _rusageStruct.ru_utime.tv_usec / 1E6;
	// subtract the initial time
	ret -= _initialCPUClock;
#endif

	return ret;
}

inline void CPUTime::GetTime(double &cputime, double &systime) {
#ifdef MINGW
	// THIS IS FOR WINDOWS
	double now = clock();
	cputime=(now-_initialCPUClock)/(double)CLOCKS_PER_SEC;
	systime=0;
#else
	// THIS IS FOR UNIX
	getrusage(RUSAGE_SELF, &_rusageStruct);
	// the current time is the number of seconds plus the fraction
	// representing the number of milliseconds
	cputime = _rusageStruct.ru_utime.tv_sec + _rusageStruct.ru_utime.tv_usec / 1E6;
	// subtract the initial time
	cputime -= _initialCPUClock;
	systime = _rusageStruct.ru_stime.tv_sec + _rusageStruct.ru_stime.tv_usec / 1E6;
	// subtract the initial time
	systime -= _initialSYSClock;
#endif
}

#endif
