// -*- C++ -*-
/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 2014 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

/**
 * @file GlobalVariables.h
 * This file contains a set of classes that represent the primitive data type,
 * but with restricted access.  The idea is to stop people from changing the
 * values of global variables inside their models by mistake.
 */

#ifndef GLOBAL_VARIABLES_HPP_
#define GLOBAL_VARIABLES_HPP_

/**
 * \brief A "primitive type" template
 *
 * The PrimitiveType template represents a primitive data type (int, float, ...)
 * but with NO way to modify the underlining data.  This template is used as a
 * base class for the various data types supported by the global variable editor.
 */
template<typename T>
class PrimitiveType
{
 public:
	/**
	 * Default Constructor
	 *
	 * Required so we can write:  		Int i;
	 * However, this is not possible:	Int i = 3;
	 */
    PrimitiveType() : mVar(0)
    {}

    /**
     * Casting operator
     *
     * If we have a global variable, Int i, the casting operator
     * allows us to write:
     *
     * int j = i;
     * int k = j % i + 3;
     */
    operator T() const
    {
        return mVar;
    }
    
protected:
    T mVar;
 
private:
   
    /**
     * Remove all copy, move or assignment operators to restrict access!
     */
#if (__cplusplus >= 201103L) || (_MSC_VER >= 1800)
    PrimitiveType(const PrimitiveType<T>&) = delete;
    PrimitiveType<T>& operator=(const PrimitiveType<T>&) = delete;
    
    PrimitiveType(PrimitiveType<T>&&) = delete;
    PrimitiveType<T>& operator=(PrimitiveType<T>&&) = delete;
#else
    PrimitiveType(const PrimitiveType<T>&);
    PrimitiveType<T>& operator=(const PrimitiveType<T>&);
#endif
};

/**
 * \class Short
 * \brief Represents the "short" data type with restricted access.
 */

class Short : public PrimitiveType<short>
{
protected:
    void Set(short s)	{ mVar = s; }
    
    friend class BaseStudyClass;
};

/**
 * \class Int
 * \brief Represents the "int" data type with restricted access.
 */
class Int : public PrimitiveType<int>
{
protected:
    void Set(int i)		{ mVar = i; }
    
    friend class BaseStudyClass;
};

/**
 * \class Long
 * \brief Represents the "long" data type with restricted access.
 */
class Long : public PrimitiveType<long>
{
protected:
    void Set(long l)	{ mVar = l; }
    
    friend class BaseStudyClass;
};

/**
 * \class Float
 * \brief Represents the "float" data type with restricted access.
 */
class Float : public PrimitiveType<float>
{
protected:
    void Set(float f)	{ mVar = f; }
    
    friend class BaseStudyClass;
};

/**
 * \class Double
 * \brief Represents the "double" data type with restricted access.
 */
class Double : public PrimitiveType<double>
{
protected:
    void Set(double d)	{ mVar = d; }
    
    friend class BaseStudyClass;
};

/**
 * \class Bool
 * \brief Represents the "bool" data type with restricted access.
 */
class Bool : public PrimitiveType<bool>
{
protected:
    void Set(bool b)	{ mVar = b; }
    
    friend class BaseStudyClass;
};

/**
 * \class Char
 * \brief Represents the "char" data type with restricted access.
 */
class Char : public PrimitiveType<char>
{
protected:
    void Set(char c)	{ mVar = c; }
    
    friend class BaseStudyClass;
};


#endif // end GLOBAL_VARIABLES_HPP_

