/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: List.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef __MOBIUS_LIST_H_
#define __MOBIUS_LIST_H_

#include <iostream>
//#include "BaseStateVariableClass.h"

/**
 *  This defines the default size of the TheList
 *  array when the default constructor is used.
 */
#define  __MOBIUS_DEFAULT_ARRAY_SIZE 10

/**
 *  This defines the scaling factor by which
 *  the TheList array is increased if it is
 *  not large enough to append all the items
 *  passed into an append method.  The new size
 *  of the array will be __ARRAY_SCALING_FACTOR
 *  times the sum of the two affecting models.
 */
#define  __ARRAY_SCALING_FACTOR 2

#include "debug.h"
#include "StandardErrors.h"

/**
 *  This is a template class used to
 *  maintain a list of mobius entities pointer
 *  (e.g., action pointers, state variable pointers,
 *  group pointers, and model pointers).
 *  The class handles the process of appending
 *  new item pointers to an array of
 *  pointers. By default, it does not allow
 *  the same pointer on the list of items.
 *  This template class is used extensively to
 *  define methods in BaseModelClass.
 *
 *   @author Jay M. Doyle
 *   @version Mobius version 1.0
 */
template<class C> class List {

public:

	/**
	 *  This default constructor creates an array with a default size.
	 *  The default size will be defined by the symbol
	 *  __MOBIUS_DEFAULT_ARRAY_SIZE.
	 */
	List();

	/**
	 *  This overloaded constructor makes an array of size InitialSize.
	 *
	 *  @param TheInitialSize The initial length of the TheList array.
	 */
	List(int TheInitialSize);
	List(int TheInitialSize, C** TheInitialList);

	/**
	 *  The destructor deallocates the TheList array.
	 */
	~List();

	/**
	 *  This method returns the number of items in
	 *   the list.
	 *
	 *  @return The number of items in the TheList array.
	 */
	int getNumItems();

	/**
	 *  This method returns the list of items.
	 *  It returns it as a const so that the
	 *  caller cannot change the value of the
	 *  internal data member.
	 *
	 *  @return The TheList array
	 */
	C** getList();

	/**
	 *  If the sum of this.NumItems and TheOther.getNumItems
	 *  is > ArraySize then a new TheList array is created with
	 *  size equal to N times the sum of the two List
	 *  (allocating more memory reduces the likelihood of having
	 *  to do a new allocation in the future). One set of items
	 *  is copied to new array, and then the other set of items
	 *  is added to the new array (checking for duplicates). In the
	 *  case of duplicate items, the array will not be fully filled.
	 *  If the sum of this.NumItems and TheOther.getNumItems <=
	 *  ArraySize then TheOther's items can be directly added to
	 *  this's TheList array.
	 *
	 *  @param TheOtherItems  Another List object that contains
	 *     other items.
	 */
	void append(List* TheOtherList);

	/**
	 *  This method works the same way as the other append method
	 *  but with a single model.
	 *
	 *  @param TheOtherModel Another model to add to the list.
	 *
	 */
	void append(C* TheOtherThing);

	/**
	 *  This method appends an array of items.
	 *
	 *  @param TheOtherThings an array of pointers to
	 *   items that should be added to the list.
	 *  @param TheArraySize The number of items in TheOtherThings
	 *   array.
	 */
	void append(C** TheOtherThings, int TheArraySize);

	/**
	 *  This method returns a specific item from the
	 *  list. If i is >= NumItems then error message is printed
	 *  and a NULL pointer is returned.
	 *
	 *  @param i The index into the TheList array.
	 */
	C* getItem(int i);
	/**
	 *  This method works as the [i]-Operator by calling
	 *  getItem , not eff., thus shortcut see below
	 *  @param i The index into the TheList array.
	 */

	inline C* operator [](int i) {
		//return(getItem(i));
		return (TheList[i]);
	};

	/**
	 *  This method removes all the members from the list.
	 */
	void clearList();

	/**
	 *  This method replaces the instance of TheOldItem with
	 *  TheNewItem.  If TheOldItem does not exists  in the list
	 *  nothing happens.  If the TheNewItem is already a member
	 *  then the TheOldItem is removed from the list and there
	 *  will be one less member of the list.
	 *
	 *  @param TheOldItem The item to be replaced
	 *  @param TheNewItem The item to replace TheOldItem
	 */
	void replaceItem(C* TheOldItem, C* TheNewItem);

	/**
	 *  This method determines is the specified item
	 *  is an element of the list.
	 */
	bool contains(C* AnItem);

	/**
	 *   This method removes the specified item from
	 *   the list of items.  If the item is not
	 *   present in the list no error or exception is
	 *   thrown.
	 */
	void remove(C* AnItem);
	/**
	 *   This method removes a set of items from
	 *   the list of items.  If the item is not
	 *   present in the list no error or exception is
	 *   thrown.
	 */

	void removeSetOfItems(List<C> theItemSet);

	/**
	 *  This method prints the value all pointers
	 *  in the TheList array.
	 */
	void printList();

protected:

	/**
	 *  This is THE list of items
	 */
	C** TheList;

	/**
	 * The number of items in the TheList array (not necessarily
	 * the size of the array.
	 */
	int NumItems;

	/**
	 * The size of the TheList array.
	 */
	int ArraySize;

	/**
	 *  This method inserts an item into the Model array
	 *  ASSUMING that there's enough room.  The method checks
	 *  to see if the item already exists in the array, if
	 *  so it will not be added.
	 */
	void insertItem(C* TheNewModel);

	/**
	 *  This method increase the size of the TheList array
	 *  the specificed amount.  It prints an error message
	 *  if TheNewSize <= ArraySize
	 */
	void increaseArraySize(int TheNewSize);

private:

}; // end List

#endif // __MOBIUS_LIST
