/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File:  MARCAModel.h
 *
 *      Authors:  Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1999 The Board of Trustees of the University of Illinois
 *                         All rights reserved.
 *
 \*****************************************************************************/

#ifndef MARCAMODEL_H_
#define MARCAMODEL_H_

#include "../BaseModelClass.h"
#include "../StandardErrors.h"
#include "Transition.h"
#include "../SAN/Place.h"
#include "../SAN/ReadOnlyPlace.h"
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include "../MobiusTypeDefs.h"
#include "../primes.h"

/**
 * This is the base class for all B&B models.
 * This is the header file for the MARCAModel. MARCAModel is a C++
 * class that describes the function of a Bucket and Ball Model for our
 * new tool.  Actual B&B models must be expressed as derived subclasses of
 * MARCAModelClass because there are specific methods that must be defined for
 * each instance of a MARCAModel.
 *
 *  @author Jay M. Doyle, Salem Derisavi
 *  @version Mobius version 1.0
 */
class MARCAModel: public BaseModelClass {

public:

	/**
	 * Base Constructor for B&B Models
	 */
	MARCAModel();

	/**
	 * Overloaded constructor that take an assortment of parameters
	 *
	 * @param TheMARCAModelName The name of the B&B Model
	 * @param TheNumberOfBuckets The number of buckets defined within the model
	 * @param TheListOfBuckets A list of pointers to the buckets defined
	 *        within the model
	 * @param TheNumberReadOnlyBuckets The number of pointers in the list
	 *        of read only buckets
	 * @param TheListOfReadOnlyBuckets An array of pointer to all the
	 *        read only buckets in the model
	 * @param TheNumberOfTransition The number of transitions defined
	 *         within the model
	 * @param TheListOfTrasitions A list of pointers to transitions defined
	 *         within the model
	 * @param TheNumberOfGroups The number of action groups defined
	 *         within the model
	 * @param TheListOfGroups A list of pointers to action groups
	 */
	MARCAModel(char* TheMARCAModelName, int TheNumberOfBuckets, Place** TheListOfBuckets, int TheNumberOfReadOnlyBuckets,
			ReadOnlyPlace** TheListOfReadOnlyBuckets, int TheNumberOfTransitions, Transition** TheListOfTransitions, int TheNumberOfGroups,
			BaseGroupClass** TheListOfGroups);

	/**
	 *  MARCAModel destructor -- Deallocates memory taken up by B&B model
	 *  if compiled with INCLUDE_DELETE preprocessor command.
	 */
	virtual ~MARCAModel();

	/**
	 * This method preforms the initialization of all internal data structures.
	 * The method name is long to avoid the possiblity of someone naming a bucket
	 * by the same name. This method used to be called Initialize until someone
	 * created a model with a bucket called Initialize.
	 *
	 * @param TheMARCAModelName The name of the B&B Model
	 * @param TheNumberOfBuckets The number of buckets defined within the model
	 * @param TheListOfBuckets A list of pointers to the buckets defined
	 *         within the
	 *        model
	 * @param TheNumberReadOnlyBuckets The number of pointers in the list
	 *        of read only buckets
	 * @param TheListOfReadOnlyBuckets An array of pointer to all the
	 *        read only buckets in the model
	 * @param TheNumberOfTransitions The number of transitions defined within
	 *        the model
	 * @param TheListOfTransitions A list of pointers to transitions defined
	 *        within the model
	 * @param TheNumberOfGroups The number of action groups defined
	 *        within the model
	 * @param TheListOfGroups A list of pointers to action groups
	 */
	void initializeMARCAModelNow(char* TheMARCAModelName, int TheNumberOfBuckets, Place** TheListOfBuckets, int TheNumberOfReadOnlyBuckets,
			ReadOnlyPlace** TheListOfReadOnlyBuckets, int TheNumberOfTransitions, Transition** TheListOfTransitions, int TheNumberOfGroups,
			BaseGroupClass** TheListOfGroups);

	/**
	 *  This method buckets the state of the B&B into the memory
	 *  location specified
	 *
	 * @param TheStateStorageLocation The location where the current state of
	 *         the  model should be saved.
	 */
	void CurrentState(void* TheStateStorageLocation);

	/**
	 * This method returns an array of pointers to all the transitions in
	 *  the B&B Model
	 *
	 *  @return An array of pointers to transitions (casted as BaseActionClass
	 *           pointers).
	 */
	void listActions(List<BaseActionClass>* TheContainer) override;

	/**
	 *  This method returns an array of pointers to transitions with the
	 *  specified name (not that the array only contains one or zero elements
	 *  if called upon an atomic model).
	 *
	 *  @param TheSpecificActionName The name of the specified transition
	 *  @return An array of pointers to transitions with the specified name
	 */
	void listActions(char* TheSpecificActionName, List<BaseActionClass>* TheContainer) override;

	/**
	 *  This method returns a list of action groups defined within the B&B Model
	 *
	 * @return An array of pointers to action groups defined within the model
	 */
	void listGroups(List<BaseGroupClass>* TheContainer);

	/**
	 * This method returns the size of the B&B model's state (in bytes).
	 *
	 *  @return The size of the model state (in bytes)
	 */
	int StateSize();

	/**
	 * This method set the model state to the one located at the specified
	 *  memory location.
	 *
	 * @param TheModelStateLocation The location in memory holding the
	 *    state to which the model should be changed.
	 */
	void SetState(void* TheModelStateLocation);

	/**
	 * This method compares two B&B Model states and determines whether
	 * they are the same.
	 *
	 *  @param StateLocation1 The location of the first B&B model state
	 *  @param StateLocation2 The location of the second B&B model state
	 *  @return True if the states are the same, else returns false
	 */
	bool CompareState(void* StateLocation1, void* StateLocation2);

	/**
	 *  This method returns a list of models with the given name (note this
	 *  return trivial for simple atomic models).
	 *
	 * @param TheSpecifiedModelName The name of the model
	 * @return An array of models with the specified name
	 */
	void listModels(char* TheSpecifiedModelName, List<BaseModelClass>* TheContainer);

	/**
	 *  This method returns a list of pointers to B&B buckets who belong to
	 *  the specified model and whose name is the same as the specified
	 *  bucket name.
	 *
	 *  @param TheSpecifiedModelName The name of the model in which to search
	 *          for the bucket
	 *  @param TheSpecifiedBucketName The name of the B&B bucket
	 *  @param TheContainer A data structure containing an array of
	 *         pointers to state variables and the number of state
	 *         variables in the arrary
	 */
	void listSVs(char* TheSpecifiedModelName, char* TheSpecifiedBucketName, List<BaseStateVariableClass>* TheSVContainer,
			List<BaseModelClass>* TheModelContainer);

	/**
	 *  This method (defined as pure virtual in BaseModelClass) prints the
	 *  state of the B&B model in a readable format to cout.
	 */
	void printState(ostream *strm);

	/**
	 *  This method returns the number of
	 *  read only buckets.
	 *
	 *  @return The number of read only buckets
	 */
	inline int getNumReadOnlyBuckets() {
		return NumReadOnlyBuckets;
	} // end getNumReadOnlyBuckets

	/**
	 *  This method (used in RegressionTest)
	 *  tests to see if the LocalStateVariables array
	 *  has been initialized correctly.
	 *
	 *  @param TheObject The MARCAModel on which the friend
	 *         function operates
	 *  @param TheCheckArray The array to be checked
	 *         against the LocalStateVariables array
	 *  @return True if all the members match, else
	 *          false
	 */
	friend bool checkLocalSVs(MARCAModel* TheObject, Place** TheCheckArray);

	/**
	 *  This method (used in RegressionTest)
	 *  tests to see if the ReadOnlyBuckets array
	 *  has been initialized correctly.
	 *
	 *  @param TheObject The MARCAModel on which the friend
	 *         function operates
	 *  @param TheCheckArray The array to be checked
	 *         against the ReadOnlyBuckets array
	 *  @return True if all the members match, else
	 *          false
	 */
	friend bool checkReadOnlyBuckets(MARCAModel* TheObject, ReadOnlyPlace** TheCheckArray);

	/**
	 *  This method (used in RegressionTest)
	 *  tests to see if the LocalActions array
	 *  has been initialized correctly.
	 *
	 *  @param TheObject The MARCAModel on which the friend
	 *         function operates
	 *  @param TheCheckArray The array to be checked
	 *         against the LocalActionss array
	 *  @return True if all the members match, else
	 *          false
	 */
	friend bool checkLocalActions(MARCAModel* TheObject, Transition** TheCheckArray);

protected:

	/**
	 * An array of pointers to buckets defined within the B&B model
	 */
	Place** LocalStateVariables;

	/**
	 *  An array of pointers to all the read only buckets
	 *  in the B&B model.  The value of these buckets
	 *  are not considered in the SetState, CurrentState,
	 *  or StateSize method since their value is a function
	 *  of some other part of state.
	 */
	ReadOnlyPlace** ReadOnlyBuckets;

	/**
	 *  This is the number of read only buckets in the
	 *  B&B Model.  The number of non-read only buckets
	 *  is NumStateVariables-NumReadOnlyBuckets since
	 *  read only state variables are counted as their
	 *  own state variable.
	 */
	int NumReadOnlyBuckets;

	/**
	 *   This method is used in the Initialize method to create all the
	 *   appropriate data structures for the B&B buckets.
	 *
	 *   @param TheListOfBuckets An array of pointers to B&B Buckets
	 */
	void SetBuckets(Place** TheListOfBuckets);

	/**
	 *   This method is used in the Initialize method to create all the
	 *   appropriate data structures for the B&B read only buckets.
	 *
	 *   @param TheListOfBuckets An array of pointers to B&B
	 *          read only Buckets
	 */
	void SetReadOnlyBuckets(ReadOnlyPlace** TheListOfBuckets);

	/**
	 *  This method is used in the Initialize method to create all the
	 *  appropriate data structures  for the B&B transitions.
	 *
	 *   @param TheListOfTransitions An array of pointers to B&B Buckets
	 */
	void SetActions(Transition** TheListOfTransitions);

	/**
	 *  This method is used in the Initialize method to create all the
	 *  appropriate data structures for the B&B transition groups.
	 *
	 *   @param TheListOfGroups An array of pointers to B&B Buckets
	 */
	void SetGroups(BaseGroupClass** TheListOfGroups);

public:
	/**
	 * The following member declarations are used by the symbolic SSG algorithm
	 * implemented by Salem
	 */
	class LocalState: public State {
	public:
		LocalState(MARCAModel* model);
		~LocalState();
		friend class MARCAModel;
		void print();
		int hash();
		int compare(State const & S) const;
		State const& operator=(State const& S);
		int fwriteState(FILE* f);
		int freadState(FILE* f);
		void* GetStatePtr();
	protected:
		void SetState(const BaseModelClass* model);
		void GetState(const BaseModelClass* model);
	private:
		char* state;
		int size;
	};

	virtual ModelType GetModelType();
	virtual int GetNumberOfChildren();
	virtual BaseModelClass* GetChild(int i);
    virtual void GetLocalActions(base_action_list_t& actions);
    virtual void GetGlobalActions(base_action_list_t& actions);
	virtual bool HasLocalState();
	virtual void GetLocalState(State* _LS);
	virtual void SetLocalState(const State* _LS);
	virtual void ConstructLocalState(State*& _LS);
	virtual void DestructLocalState(State* _LS);

}; // end MARCAModel class

#endif // end MARCAMODEL_H_
