/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      MobiusQueue: used to store basic types or pointers.  However,
 *      if T is a pointer, note that only shallow copies of T are
 *      done, and T is never deleted -- so an outside class needs to
 *      keep track of the T and delete them on its own.
 *
 *      Authors: Tod Courtney
 *      Creation Date: 2003
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef _MOBIUS_QUEUE_H_
#define _MOBIUS_QUEUE_H_
#include "MobiusThread.h"

template<class T> class MobiusQueue {
protected:
	T *_data;
	int _queueMaxSize;
	int _queueFront;
	int _queueEnd;
	bool _endWrap;
	MobiusMutex *_mutex;

public:
	MobiusQueue(int maxsize);
	~MobiusQueue();
	int GetCount();
	T Pop();
	void Push(T d);
};

/**
 *  create a circular queue of the given size
 */
template<class Type>
MobiusQueue<Type>::MobiusQueue(int maxsize) {
	_data = new Type[maxsize];
	_queueFront = 0;
	_queueEnd = 0;
	_queueMaxSize = maxsize;
	_endWrap = false;
	_mutex = new MobiusMutex();
}

/**
 *  destructor (note it doesn't call delete on the individual _data
 *  objects -- if they are pointers, it is the responsibility of the
 *  user of this class to pop them all first and delete them).  Maybe
 *  this isn't cleanest, but it is simpler and faster and allows this
 *  class to be used by pointers and standard types without a lot of
 *  overhead.
 */
template<class Type>
MobiusQueue<Type>::~MobiusQueue() {
	delete _mutex;
	delete[] _data;
}

/**
 *  get the number of objects in the queue
 */
template<class Type>
int MobiusQueue<Type>::GetCount() {
	int ret;
	_mutex->Get();

	if (_queueEnd >= _queueFront && _endWrap == false)
		ret = _queueEnd - _queueFront;
	else
		ret = _queueEnd - _queueFront + _queueMaxSize;

	_mutex->Release();
	return ret;
}

/**
 *  remove data from the front queue
 */
template<class Type>
Type MobiusQueue<Type>::Pop() {
	Type ret = NULL;
	_mutex->Get();

	if (_queueFront == _queueEnd && _endWrap == false) {
		//printf("POP: QUEUE IS EMPTY\n");
	} else {
		ret = _data[_queueFront];
		//printf("POP: qf=%d qe=%d\n",_queueFront, _queueEnd);
		_queueFront++;
		if (_queueFront >= _queueMaxSize) {
			_endWrap = false;
			_queueFront = 0;
		}
	}

	_mutex->Release();
	return ret;
}

/**
 *  put data into the end of the queue
 */
template<class Type>
void MobiusQueue<Type>::Push(Type d) {
	_mutex->Get();
	if (_queueFront == _queueEnd && _endWrap == true) {
		//printf("PUSH: QUEUE IS FULL\n");
	} else {
		//printf("push: qf=%d qe=%d\n",_queueFront, _queueEnd);
		_data[_queueEnd] = d;
		_queueEnd++;
		if (_queueEnd >= _queueMaxSize) {
			_endWrap = true;
			_queueEnd = 0;
		}
	}
	_mutex->Release();
}

#endif
