/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: MobiusSocket
 *
 *      Authors: Tod Courtney
 *      Creation Date: 2003
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

/**
 * MobiusSocket.h and MobiusSocket.cpp contain the MobiusSocket class definition and its implementation.
 */

#ifndef _MOBIUSSOCKET_H_
#define _MOBIUSSOCKET_H_

#if MINGW || _WIN32
	#include <winsock2.h>
#else
	typedef int SOCKET;
#endif

struct NetworkInitializer
{
	NetworkInitializer();
	~NetworkInitializer();
};


/**
 * The MobiusSocket class provides an interface for socket communications. The class
 * manages the library differences between those provided by the POSIX standard in Unix
 * and the MinGW libraries used in Windows.
 */

class MobiusSocket {
public:
	/// Calls InitializeSocketLibrary() and attempts to create a new socket file descriptor if openFD is true.
	/// @param[in] openFD If true, the constructor will open a new socket file descriptor.
	MobiusSocket(bool openFD = true);

	/// Calls Close().
	~MobiusSocket();

	/// Make the socket connection blocking/non blocking.  By default the socket is blocking
	void NonBlocking(bool non_blocking);

	/// Accepts a connection from a client
	/// @param[out] retSocket Must contain a MobiusSocket instance with an unopened socket file descriptor (this function won't close the socket FD if it is open, but can't reuse either).
	bool Accept(MobiusSocket &retSocket);

	/// Binds the MobiusSocket to the provided port
	/// @param[in] port The port number to bind to.
	/// @return Returns true if successful.
	bool Bind(int port);

	/// Closes the socket file descriptor of the instance
	/// @return Returns true if successful.
	void Close();

	/// Connects to another computer's port as a client.
	/// @param[in] hostname The DNS resolvable name of the destination.
	/// @param[in] port The port number to connect to.
	bool Connect(char *hostname, int port);

	/// Listens on its port for incoming connection requests.
	/// @param[in] queueSize The number of incoming connection requests it can queue.
	bool Listen(int queueSize);

	/// Receives data on the socket.
	/// @param[out] buffer The buffer to place the received data in.
	/// @param[in] desired_len The max size to receive. This length is not guaranteed to be received, this is only a limit.
	/// @returns Returns the number of bytes read from the socket.
	int Recv(char *buffer, int desired_len);

	/// Receives a specific amount of data from the socket.
	/// @param[out] buffer The buffer to place the received data in.
	/// @param[in] desired_len The amount of data that should be (live) waited for. Will wait until this much data arrives or a read error occurs.
	/// @returns Returns the number of bytes read from the socket.
	int RecvBlock(char *buffer, int desired_len);

	/// Receives data on the socket. This function waits until the buffer has been filled or a char array matching endstr has been received.
	/// @param[out] buffer The buffer to place the received data in.
	/// @param[in] buflen The length of the passed buffer.
	/// @param[in] endstr The string to look for in the stream of data that will end this receive function.
	/// @returns Returns the number of bytes read from the socket.
	int RecvBlock(char *buffer, int buflen, char *endstr);

	/// Sends data on the socket.
	/// @param[in] str The buffer containing the data to be sent.
	/// @param[in] strlen The length of the buffer from the first argument.
	/// @returns Returns the number of bytes written to the socket. Returns -1 if an error occurred.
	int Send(char *str, int strlen);

	/// Blocks until something is ready to be read from the socket. Or the timeout has expired.
	/// @param[in] millisec_timeout The length of time to wait for activity on the socket (in milliseconds).
	/// returns Returns true if there is something waiting to be read from the socket. Returns false if the timeout expired.
	bool Select(int millisec_timeout);

	/// Gets the error string for the MobiusSocket
	/// returns Returns the error string.
	char *GetErrorString();

private:

	/// Stores the socket file descriptor for the MobiusSocket instance.
	SOCKET mSocket;

	/// This stores error messages when they are generated during the execution of the MobiusSocket.
	char * mpErrorString;

};

#endif
