/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: Activity.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef ACTIVITY_H_
#define ACTIVITY_H_

#include "../BaseGroupClass.h"
#include "../debug.h"

/**
 *  This class describes the operation of SAN Activities -- the
 *  formalism specific action for stochastic activity networks
 *
 *  @author Jay M. Doyle
 *  @version Mobius Version 1.0
 */
class Activity: public BaseGroupClass {

public:

	/**
	 * This value saves the  enabling state in the new marking
	 */
	bool NewEnabled;

	/**
	 *  This value saves the enabling state but for previous marking.
	 *  This is key to operation of the EnablingChange() method
	 */
	bool OldEnabled;

	/**
	 *  Base Constructor
	 */
	Activity();

	/**
	 *  Destructor
	 */
	~Activity();

	/**
	 *   Overloaded costructor
	 *   @param TheActionName The name of the action group
	 *   @param TheGroupIDNumber The unique number identifying the action group
	 *   @param TheDistributionType The action probability distribution function
	 *   @param TheExecutionPolicy  The action's execution policy
	 *   @param TheNumberOfAffectedStateVariables The number of state
	 *          variables whose state is affected by the action
	 *   @param TheNumberOfEnablingStateVariables The number of state
	 *          variables whose state enables the action
	 *   @param The initial value of reactivation data element
	 */
	Activity(char* TheActionName, int TheGroupIDNumber, Distribution TheDistributionType, ExecutionPolicyType TheExecutionPolicy,
			int TheNumberOfAffectedStateVariables, int TheNumberOfEnablingStateVariables, bool TheReactivation);

	/**
	 *   This method initializes all the internal data structure
	 *   @param TheActionName The name of the action group
	 *   @param TheGroupIDNumber The unique number identifying the action group
	 *   @param TheDistributionType The action probability distribution function
	 *   @param TheExecutionPolicy  The action's execution policy
	 *   @param TheNumberOfAffectedStateVariables The number of state variables
	 *         whose state is affected by the action
	 *   @param TheNumberOfEnablingStateVariables The number of state
	 *         variables whose state enables the action
	 *   @param The initial value of reactivation data element
	 */
	void ActivityInitialize(char* TheActionName, int TheGroupIDNumber, Distribution TheDistributionType, ExecutionPolicyType TheExecutionPolicy,
			int TheNumberOfAffectedStateVariables, int TheNumberOfEnablingStateVariables, bool TheReactivation);

	/**
	 *  This method initialize the affected/enabling lists
	 */
	virtual void LinkVariables()=0;

	/**
	 * The method performs the activity's state changes.  Note
	 * that this method is pure virtual because the way that a
	 * specific acitivity changes state is specific to the activity.
	 */
	virtual BaseActionClass* Fire()=0;

	/* void Fire(int);   Unsupported method */

	/**
	 *  This method is defined on BaseGroupClass.  Since an activity is
	 *  always in a group by itself the selection is trivial.
	 *  @see BaseGroupClass
	 */
	bool SelectAction();

	/**
	 *  This is another method defined on BaseGroupClass that has a trivial
	 */
	void CalculateWeightDistribution();

	/**
	 *  This method determines if there has been a change in the
	 *  enabling status of an action since enabled was last called
	 *
	 * @return True if there was a change in the enabling
	 *  status of a action, otherwise false
	 */
	bool EnablingChange();

	/**
	 *  This is defined on BaseGroupClass. Since an activity is
	 *  always in a group by itself the probability of selection
	 *  is always 1.  However the method is defined such
	 *  that if the action that is passed in is not the "this" then
	 *  Prabability returns 0;
	 *  @see BaseGroupClass
	 *
	 *  @param TheAction The specified action
	 *  @return The probability of selecting the specified action
	 *   from the action group
	 */
	double Probability(BaseActionClass* TheAction);

	/**
	 *   This method determines whether the specified state
	 *   variable is affected by this activity.
	 *
	 *   @param TheStateVariable The specified state variable
	 *   @return True if the state variable is affected by this activity
	 */
	bool IsAnAffectingStateVariable(BaseStateVariableClass* TheStateVariable);

	/**
	 *  This overloads the BaseGroupClass implementation
	 *  and sets the GroupID for the underlying action
	 *
	 *  @param TheNewID The new group ID for the action
	 */
	inline void setGroupId(int TheNewID) {
		GroupID = TheNewID;
	} // end setGroupID


protected:

	/**
	 *  This is an array used to store values of the
	 *  the distribution parameters when the ReturnDistributionParameters
	 *  method is called.  The size of the array depends upon the number
	 *  of distribution paramters
	 */
	double* TheDistributionParameters;

}; // end class Activity

#endif // end ACTIVITY_H_
