/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: PEPAModel.h
 *
 *      Authors: Graham Clark, Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 2000
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef PEPAMODEL_H_
#define PEPAMODEL_H_

#include "../BaseModelClass.h"
#include "../StandardErrors.h"
#include "PEPAActivity.h"
#include "PEPAVariable.h"
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include "../MobiusTypeDefs.h"

class PEPAActivity;

/**
 * This is the base class for all PEPA models with value passing.
 * This is the header file for the PEPAModel. PEPAModel is a C++
 * class that describes the function of a PEPA model within the UltraSAN/Mobius
 * tool.  Actual PEPA models must be expressed as derived subclasses of
 * PEPAModel because there are specific methods that must be defined for
 * each instance of a PEPA Model.
 *
 *  @author Graham Clark
 *  @version Mobius version 1.0 Development
 */
class PEPAModel: public BaseModelClass {

public:

	PEPAModel();

	PEPAModel(char* ThePEPAModelName, int TheNumberOfVariables, PEPAVariable** TheListOfVariables, int TheNumberOfActivities,
			PEPAActivity** TheListOfActivities, int NGroups, BaseGroupClass** Groups);

	virtual ~PEPAModel();

	void initializePEPAModelNow(char* TheSANModelName, int TheNumberOfVariables, PEPAVariable** TheListOfVariables, int TheNumberOfActivities,
			PEPAActivity** TheListOfActivities, int NGroups, BaseGroupClass** Groups);

	void CurrentState(void* TheStateStorageLocation);

	void listActions(List<BaseActionClass>* TheContainer) override;

	void listActions(char* TheSpecificActionName, List<BaseActionClass>* TheContainer) override;

	void listGroups(List<BaseGroupClass>* TheContainer);

	int StateSize();

	void SetState(void* TheModelStateLocation);

	bool CompareState(void* StateLocation1, void* StateLocation2);

	void PrintPEPAModelInfo();

	void listModels(char* TheSpecifiedModelName, List<BaseModelClass>* TheContainer);

	void listSVs(char* TheSpecifiedModelName, char* TheSpecifiedPlaceName, List<BaseStateVariableClass>* TheSVContainer,
			List<BaseModelClass>* TheModelContainer);

	void printState(ostream *strm);

	inline int getNumReadOnlyVariables() {
		return NumReadOnlyVariables;
	}

	void SetNumberOfComponents(int n);
	//void InitializeComponents();


	friend bool checkLocalSVs(PEPAModel* TheObject, PEPAVariable** TheCheckArray);

	friend bool checkLocalActions(PEPAModel* TheObject, PEPAActivity** TheCheckArray);

	// gcla - CompList is an array of pointers, each one pointing to the
	// location of the state item for that particular subcomponent
	// program counter
	int NumComps;
	short **CompList;

	void CurrentTermsState(void* MemoryLocation);
	void ReadCurrentTermsState(void* MemoryLocation);
	inline int CurrentTermsLength();
	void CanonicaliseCompList();

protected:

	// these variables are for the canonical representation algorithm
	int NumCanonGroups;
	int *CanonGroupLengths;
	int **CanonGroups;
	// used in CompareState - keep the space allocated
	short *tempState1;
	short *tempState2;

	PEPAVariable** LocalStateVariables;

	int NumReadOnlyVariables;

	void SetVariables(PEPAVariable** TheListOfVariables);

	void SetActions(PEPAActivity** TheListOfActivities);

	void SetGroups(BaseGroupClass** TheListOfGroups);

	inline void CanonicaliseState(short *fromState, short *toState);

public:
	/**
	 * The following member declarations are used by the symbolic SSG algorithm
	 * implemented by Salem
	 */
	class LocalState: public State {
	public:
		LocalState(PEPAModel* model);
		~LocalState();
		friend class PEPAModel;
		void print();
		int hash();
		int compare(State const & S) const;
		State const& operator=(State const& S);
		int fwriteState(FILE* f);
		int freadState(FILE* f);
		void* GetStatePtr();
	protected:
		void SetState(const BaseModelClass* model);
		void GetState(const BaseModelClass* model);
	private:
		char* state;
		int size;
	};

	virtual ModelType GetModelType();
	virtual int GetNumberOfChildren();
	virtual BaseModelClass* GetChild(int i);
	virtual void GetLocalActions(base_action_list_t& actions);
	virtual void GetGlobalActions(base_action_list_t& actions);
	virtual bool HasLocalState();
	virtual void GetLocalState(State* _LS);
	virtual void SetLocalState(const State* _LS);
	virtual void ConstructLocalState(State*& _LS);
	virtual void DestructLocalState(State* _LS);

};

#endif
