/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: PostselectGroup.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef POSTSELECTGROUP_H
#define POSTSELECTGROUP_H

#include <stdlib.h>
#include "./StandardErrors.h"
#include "./BaseGroupClass.h"
#include "./debug.h"

/**
 * This class describes a particular type of action
 * group in which the selected action is choosen at
 * firing time.  In order for this selection policy
 * to be valid, all member actions must have the
 * same probability distribution function,
 * reactivation predicate, and reactivation
 * function.
 *  This type of action grouping is used in stocha stic
 * activity networks to mimic the functionality of
 * multiple case activities in which activity cases
 * are choosen at firing time.
 *
 *  @author Jay M. Doyle
 *  @version Mobius Version 1.0
 */
class PostselectGroup: public BaseGroupClass {

public:

	/**
	 * Base constructor
	 */
	PostselectGroup();

	/**
	 * Overloaded constructor
	 *
	 *  @param TheNumberOfAction The number of actions
	 *         that will be part of the post selection
	 *         group.
	 */
	PostselectGroup(int TheNumberOfActions);

	/**
	 * Destructor
	 */
	~PostselectGroup();

	/** This method initializes all the internal data structures
	 *
	 * @param TheNumberOfActions The number of action that
	 *        will be part of the postselection group.
	 */
	//void Initialize(int TheNumberOfActions);

	/** This method allows you to specify each one of the
	 * group member individually
	 *
	 *  @param TheActionPointer A pointer to a group member action
	 *  @param TheArrayIndex The index at which the action will be added
	 */
	//void SetAction(BaseActionClass* TheActionPointer,
	//		 int TheArrayIndex);

	/** This method calls the reactivation predicate
	 * on the first action in the array of member
	 * actions.  This valid since ALL MEMBER ACTIONS
	 * MUST HAVE THE SAME REACTIVATION PREDICATE!!
	 */
	virtual bool ReactivationPredicate();

	/** This method calls the reactivation function
	 * on the first action in the array of member
	 * actions.  This valid since ALL MEMBER ACTIONS
	 * MUST HAVE THE SAME REACTIVATION FUNCTION!!
	 */
	virtual bool ReactivationFunction();

	/** This method
	 */
	virtual double SampleDistribution();

	/**
	 *  The default Rank value for the
	 *  group.  We don't want to use the
	 *  default implemenation of Rank
	 *  since it involves the Rank of the
	 *  selected action.  The selected action
	 *  is choosen at firing time and Rank() is
	 *  called before firing by CalculateWeightDistribution
	 *  is the postselect group is a member of another
	 *  group.  Thus the group itself should have a
	 *  Rank value is separate from the Rank values
	 *  associated with its members.  This method
	 *  is declared virtual so that specific instances
	 *  of postselect groups can redefined the rank value.
	 *
	 *  @return The postselect group's Rank
	 */
	inline virtual int Rank() {
		CF("PostselectGroup::Rank()");
		return 1;
	} // end PostselectGroup::Rank()

	/**
	 *  The same reasoning as with PostselectGroup::Rank()
	 *
	 *  @return The weight value of the postselect group.
	 */
	inline virtual double Weight() {
		CF("PostselectGroup::Weight()");
		return 1.0;
	} // end PostselectGroup::Weight()

	/**
	 *  This method fires the postselect's selected
	 *  action
	 */
	BaseActionClass* Fire();

private:

}; // end PostselectGroup class

#endif // end POSTSELECTGROUP_H
