/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: PreselectGroup.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef PRESELECTGROUP_H
#define PRESELECTGROUP_H

#include <stdlib.h>
#include "./StandardErrors.h"
#include "./BaseGroupClass.h"

/**
 *  This action group class is used to implement a
 *  preselection execution policy among a group of
 *  actions
 */
class PreselectGroup: public BaseGroupClass {

public:

	/**
	 *  Base Constructor
	 */
	PreselectGroup();

	/**
	 *  Overloaded constructor that takes
	 *   the number of member groups
	 *
	 *  @param NumberOfMembers The number of member
	 *    groups contained in the preselect group
	 */
	PreselectGroup(int NumberOfMembers);

	/**
	 *  Preselect destructor
	 */
	~PreselectGroup();

	/**
	 *  This methods returns the
	 *  reactivation function of the
	 *  selected action
	 */
	bool ReactivationPredicate();

	/**
	 *  This methods returns the
	 *  reactivation function of the
	 *  selected action
	 */
	bool ReactivationFunction();

	/**
	 *  In preselect groups the selected
	 *  action is choosen when the distribution
	 *  is sampled.  Rank can be called before the
	 *  distribution is sampled, therefore the
	 *  implementation cannot be defined in terms
	 *  of the SelectedAction.  Since the selected
	 *  action will always have the highest rank among
	 *  the set of enabled action, we can determined the
	 *  highest rank value and return that value without
	 *  actually selecting an action.
	 */
	int Rank();

	/**
	 *  This method selects an action
	 *  from the set of enabled actions
	 *  and samples for the selected action's
	 *  distribution
	 */
	double SampleDistribution();

	/**
	 *  This method calls BaseGroupClass's implementation
	 *  and then resets the NeedToReselect data member
	 *  to false.
	 */
	bool SelectAction();

	/**
	 *  This method fires the selected action's
	 *  fire method
	 */
	BaseActionClass* Fire();

	/**
	 *  This method first checks to see if a
	 *  representative needs to be selected
	 *  then uses the representative Enabled
	 *  method until the next reselection time.
	 */
	bool Enabled();

protected:

	/**
	 *  This data member is set true when
	 *  the next method call on the group
	 *  should first reselect a representative
	 *  action.  This is set after an action firing,
	 *  when the selected action becomes disabled
	 *  before firing, or the selected action's reactivation
	 *  predicate and function are true in an enabling state.
	 */
	bool NeedToReselect;

private:

};

#endif // end PRESELECTGROUP_H
