/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: Place.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef EXTENDEDPLACE_H_
#define EXTENDEDPLACE_H_

#include "../state/SharableSV.h"
#include "../StandardErrors.h"
#include "../debug.h"
#include <iostream>

/** This class describes the base state variable class for stochastic activity networks
 *
 *   @author Jay M. Doyle
 *   @version Mobius version 1.0
 */
template<typename T>
class ExtendedPlace: public SharableSV<T> {

public:
	/**
	 *  This contructor takes both the place's initial name and marking.
	 *
	 *  @param ThePlaceName The initial name of the place
	 *  @param TheInitialMarking The initial place marking
	 */

	ExtendedPlace() :
		SharableSV<T> () {
		CF("Entering Place::Place");
		this->SetName("default");
		this->Register(&this->TheState);
		ASSERT(this->TheState != NULL);
		DS2("Address of Mark After Assignment",this->TheState);
		*this->TheState = this->zero();
	}

	explicit ExtendedPlace(char* TheName) :
		SharableSV<T> () {

		CF("Entering Place(char*, short)");
		this->SetName(TheName);
		this->Register(&this->TheState);
		DS2("Mark Address After Assignment", this->TheState);
		*this->TheState = this->zero();

#ifdef DEBUG
		if (GlobalDebugLevel > 1) {
			cout << "Initial Marking for " << this->StateVariableName << " is " << *(this->TheState) << endl;
		} // end if
#endif
	}
	
	ExtendedPlace(char* TheName, char* ThePrefix) :
		SharableSV<T> () {

		CF("Entering Place(char*, short)");
		this->SetName(TheName);
		this->SetPrefix(ThePrefix);
		this->Register(&this->TheState);
		DS2("Mark Address After Assignment", this->TheState);
		*this->TheState = this->zero();

#ifdef DEBUG
		if (GlobalDebugLevel > 1) {
			cout << "Initial Marking for " << this->StateVariableName << " is " << *(this->TheState) << endl;
		} // end if
#endif
	}
	
	ExtendedPlace(char* TheName, T InitialMarking) :
		SharableSV<T> () {
		CF("Entering Place(char*, short)");
		this->SetName(TheName);
		this->Register(&this->TheState);
		DS2("Mark Address After Assignment", this->TheState);
		*this->TheState = InitialMarking;

#ifdef DEBUG
		if (GlobalDebugLevel > 1) {
			cout << "Initial Marking for " << this->StateVariableName << " is " << *(this->TheState) << endl;
		} // end if
#endif
	}
	
	ExtendedPlace(char* TheName, char* ThePrefix, T InitialMarking) : SharableSV<T>() {
		this->SetName(TheName);
		this->SetPrefix(ThePrefix);
		this->Register(&this->TheState);
		*this->TheState = InitialMarking;
		
#ifdef DEBUG
		if (GlobalDebugLevel > 1) {
			cout << "Initial Marking for " << this->StateVariableName << " is " << *(this->TheState) << endl;
		} // end if
#endif
	}

	/**
	 * This method returns the value of the SAN place as an Lvalue
	 *
	 * @return The value of the mark (as an Lvalue)
	 */
	inline T& Mark() {
		return *(this->TheState);
	}

	/** This method returns the place's marking (by value)
	 */
	T getMark() {
		CF("Entering Place::getMark()");
		return *(this->TheState);
	}

	/** This method sets the value of the place's mark
	 *
	 *  @param TheNewMarking The new value of the SAN place
	 */
	void setMark(T TheNewMarking) {
		CF("Entering Place::setMark(short)");
		*(this->TheState) = TheNewMarking;
	}

	/**
	 *  A method used for debugging
	 */
	void test();

protected:

	/**
	 *	Return the initial value of a type T
	 */
	T zero() {
		const int byteCount = sizeof(T) / sizeof(char);
		T returnValue;
		char* returnBytes = reinterpret_cast<char*> (&returnValue);
		for (int i = 0; i < byteCount; i++)
			returnBytes[i] = '\0';
		return returnValue;
	}
	;

	/**
	 *  This method returns number of bytes of compact
	 *  state variable representation
	 *
	 *  @return The number of bytes needed to represent
	 *      the state variable's state
	 */

	virtual int StateSize() {
		return sizeof(T);
	}

	/**
	 *  This method sets the state of the StateVariable
	 *
	 *  @param TheNewState The location of the new value
	 *   for the state variable
	 */
	virtual void SetState(void* NewState) {
		readMemory2<T> (NewState, this->TheState);
	}
	;

	/**
	 *  This method writes the state variables current state
	 *  to the specified memory location
	 *
	 *  @param TheMemoryLocation The location in memory to which
	 *     the state variable state  should be stored
	 */
	virtual void CurrentState(void* MemoryLocation) {
		writeMemory<T> (*(this->TheState), MemoryLocation);
	}
	;

}; // end class Place

#endif

