/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: SANModel.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef SANMODEL_H_
#define SANMODEL_H_

#include "../BaseModelClass.h"
#include "../StandardErrors.h"
#include "Activity.h"
#include "Place.h"
#include "ReadOnlyPlace.h"
#include <stdlib.h>
#include <stdio.h>
#include <iostream>
#include "../MobiusTypeDefs.h"
#include "../BaseStateVariableClass.h"
#include "../primes.h"


/**
 * This is the base class for all stochastic activity networks.
 * This is the header file for the SANModel. SANModel is a C++
 * class that describes the function of a Stochastic Activity Model for our
 * new tool.  Actual SAN models must be expressed as derived subclasses of
 * SANModelClass because there are specific methods that must be defined for
 * each instance of a SANModel.
 *
 *  @author Jay M. Doyle
 *  @version Mobius version 1.0
 */
class SANModel: public BaseModelClass {

public:

	/**
	 * Base Constructor for SAN Models
	 */
	SANModel();

	/**
	 * Overloaded constructor that take an assortment of parameters
	 *
	 * @param TheSANModelName The name of the SAN Model
	 * @param TheNumberOfPlaces The number of places defined within the model
	 * @param TheListOfPlaces A list of pointers to the places defined
	 *        within the model
	 * @param TheNumberReadOnlyPlaces The number of pointers in the list
	 *        of read only places
	 * @param TheListOfReadOnlyPlaces An array of pointer to all the
	 *        read only places in the model
	 * @param TheNumberOfActivities The number of activities defined
	 *         within the model
	 * @param TheListOfActivities A list of pointers to activities defined
	 *         within the model
	 * @param TheNumberOfGroups The number of action groups defined
	 *         within the model
	 * @param TheListOfGroups A list of pointers to action groups
	 */
	SANModel(char* TheSANModelName, int TheNumberOfPlaces, BaseStateVariableClass** TheListOfPlaces, int TheNumberOfReadOnlyPlaces,
			BaseStateVariableClass** TheListOfReadOnlyPlaces, int TheNumberOfActivities, Activity** TheListOfActivities, int TheNumberOfGroups,
			BaseGroupClass** TheListOfGroups);

	/**
	 *  SANModel destructor -- Deallocates memory taken up by SAN model
	 *  if compiled with INCLUDE_DELETE preprocessor command.
	 */
	virtual ~SANModel();

	/**
	 * This method preforms the initialization of all internal data structures.
	 * The method name is long to avoid the possiblity of someone naming a place
	 * by the same name. This method used to be called Initialize until someone
	 * created a model with a place called Initialize.
	 *
	 * @param TheSANModelName The name of the SAN Model
	 * @param TheNumberOfPlaces The number of places defined within the model
	 * @param TheListOfPlaces A list of pointers to the places defined
	 *         within the
	 *        model
	 * @param TheNumberReadOnlyPlaces The number of pointers in the list
	 *        of read only places
	 * @param TheListOfReadOnlyPlaces An array of pointer to all the
	 *        read only places in the model
	 * @param TheNumberOfActivities The number of activities defined within
	 *        the model
	 * @param TheListOfActivities A list of pointers to activities defined
	 *        within the model
	 * @param TheNumberOfGroups The number of action groups defined
	 *        within the model
	 * @param TheListOfGroups A list of pointers to action groups
	 */
	void initializeSANModelNow(char* TheSANModelName, int TheNumberOfPlaces, BaseStateVariableClass** TheListOfPlaces, int TheNumberOfReadOnlyPlaces,
			BaseStateVariableClass** TheListOfReadOnlyPlaces, int TheNumberOfActivities, Activity** TheListOfActivities, int TheNumberOfGroups,
			BaseGroupClass** TheListOfGroups);

	/**
	 *  This method places the state of the SAN into the memory
	 *  location specified
	 *
	 * @param TheStateStorageLocation The location where the current state of
	 *         the  model should be saved.
	 */
	void CurrentState(void* TheStateStorageLocation);

	/**
	 * This method returns an array of pointers to all the activities in
	 *  the SAN Model
	 *
	 *  @return An array of pointers to activities (casted as BaseActionClass
	 *           pointers).
	 */
	void listActions(List<BaseActionClass>* TheContainer) override;

	/**
	 *  This method returns an array of pointers to activities with the
	 *  specified name (not that the array only contains one or zero elements
	 *  if called upon an atomic model).
	 *
	 *  @param TheSpecificActionName The name of the specified activity
	 *  @return An array of pointers to activities with the specified name
	 */
	void listActions(char* TheSpecificActionName, List<BaseActionClass>* TheContainer) override;

	/**
	 *  This method returns a list of action groups defined within the SAN Model
	 *
	 * @return An array of pointers to action groups defined within the model
	 */
	void listGroups(List<BaseGroupClass>* TheContainer);

	/**
	 * This method returns the size of the SAN model's state (in bytes).
	 *
	 *  @return The size of the model state (in bytes)
	 */
	int StateSize();

	/**
	 * This method set the model state to the one located at the specified
	 *  memory location.
	 *
	 * @param TheModelStateLocation The location in memory holding the
	 *    state to which the model should be changed.
	 */
	void SetState(void* TheModelStateLocation);

	/**
	 * This method compares two SAN Model states and determines whether
	 * they are the same.
	 *
	 *  @param StateLocation1 The location of the first SAN model state
	 *  @param StateLocation2 The location of the second SAN model state
	 *  @return True if the states are the same, else returns false
	 */
	bool CompareState(void* StateLocation1, void* StateLocation2);

	/**
	 * This method prints infomation about the SAN to cout
	 */
	void PrintSANModelInfo();

	/**
	 *  This method returns a list of models with the given name (note this
	 *  return trivial for simple atomic models).
	 *
	 * @param TheSpecifiedModelName The name of the model
	 * @return An array of models with the specified name
	 */
	void listModels(char* TheSpecifiedModelName, List<BaseModelClass>* TheContainer);

	/**
	 *  This method returns a list of pointers to SAN places who belong to
	 *  the specified model and whose name is the same as the specified
	 *  place name.
	 *
	 *  @param TheSpecifiedModelName The name of the model in which to search
	 *          for the place
	 *  @param TheSpecifiedPlaceName The name of the SAN place
	 *  @param TheContainer A data structure containing an array of
	 *         pointers to state variables and the number of state
	 *         variables in the arrary
	 */
	void listSVs(char* TheSpecifiedModelName, char* TheSpecifiedPlaceName, List<BaseStateVariableClass>* TheSVContainer,
			List<BaseModelClass>* TheModelContainer);

	/**
	 *  This method (defined as pure virtual in BaseModelClass) prints the
	 *  state of the SAN model in a readable format to strm.
	 */
	void printState(ostream *strm);

	/**
	 * The method returns an array containing the local state variables
	 */
	BaseStateVariableClass** GetLocalStateVariables(ostream *strm);

	/**
	 *  This method returns the number of
	 *  read only places.
	 *
	 *  @return The number of read only places
	 */
	inline int getNumReadOnlyPlaces() {
		return NumReadOnlyPlaces;
	} // end getNumReadOnlyPlaces

	/**
	 *  This method (used in RegressionTest)
	 *  tests to see if the LocalStateVariables array
	 *  has been initialized correctly.
	 *
	 *  @param TheObject The SANModel on which the friend
	 *         function operates
	 *  @param TheCheckArray The array to be checked
	 *         against the LocalStateVariables array
	 *  @return True if all the members match, else
	 *          false
	 */
	friend bool checkLocalSVs(SANModel* TheObject, Place** TheCheckArray);

	/**
	 *  This method (used in RegressionTest)
	 *  tests to see if the ReadOnlyPlaces array
	 *  has been initialized correctly.
	 *
	 *  @param TheObject The SANModel on which the friend
	 *         function operates
	 *  @param TheCheckArray The array to be checked
	 *         against the ReadOnlyPlaces array
	 *  @return True if all the members match, else
	 *          false
	 */
	friend bool checkReadOnlyPlaces(SANModel* TheObject, BaseStateVariableClass** TheCheckArray);

	/**
	 *  This method (used in RegressionTest)
	 *  tests to see if the LocalActions array
	 *  has been initialized correctly.
	 *
	 *  @param TheObject The SANModel on which the friend
	 *         function operates
	 *  @param TheCheckArray The array to be checked
	 *         against the LocalActionss array
	 *  @return True if all the members match, else
	 *          false
	 */
	friend bool checkLocalActions(SANModel* TheObject, Activity** TheCheckArray);

protected:

	/**
	 * An array of pointers to places defined within the SAN model
	 */
	BaseStateVariableClass** LocalStateVariables;
	/**
	 *  An array of pointers to all the read only places
	 *  in the SAN model.  The value of these places
	 *  are not considered in the SetState, CurrentState,
	 *  or StateSize method since their value is a function
	 *  of some other part of state.
	 */
	BaseStateVariableClass** ReadOnlyPlaces;

	/**
	 *  This is the number of read only places in the
	 *  SAN Model.  The number of non-read only places
	 *  is NumStateVariables-NumReadOnlyPlaces since
	 *  read only state variables are counted as their
	 *  own state variable.
	 */
	int NumReadOnlyPlaces;

	/**
	 *   This method is used in the Initialize method to create all the
	 *   appropriate data structures for the SAN places.
	 *
	 *   @param TheListOfPlaces An array of pointers to SAN Places
	 */
	void SetPlaces(BaseStateVariableClass** TheListOfPlaces);

	/**
	 *   This method is used in the Initialize method to create all the
	 *   appropriate data structures for the SAN read only places.
	 *
	 *   @param TheListOfPlaces An array of pointers to SAN
	 *          read only Places
	 */
	void SetReadOnlyPlaces(BaseStateVariableClass** TheListOfPlaces);

	/**
	 *  This method is used in the Initialize method to create all the
	 *  appropriate data structures  for the SAN activities.
	 *
	 *   @param TheListOfActivities An array of pointers to SAN Places
	 */
	void SetActions(Activity** TheListOfActivities);

	/**
	 *  This method is used in the Initialize method to create all the
	 *  appropriate data structures for the SAN activity groups.
	 *
	 *   @param TheListOfGroups An array of pointers to SAN Places
	 */
	void SetGroups(BaseGroupClass** TheListOfGroups);

public:
	/**
	 * The following member declarations are used by the symbolic SSG algorithm
	 * implemented by Salem
	 */
	class LocalState: public State {
	public:
		LocalState(SANModel* model);
		~LocalState();
		friend class SANModel;
		void print();
		int hash();
		int compare(State const & S) const;
		State const& operator=(State const& S);
		int fwriteState(FILE* f);
		int freadState(FILE* f);
		void* GetStatePtr();
	protected:
		void SetState(const BaseModelClass* model);
		void GetState(const BaseModelClass* model);
	private:
		char* state;
		int size;
	};

	virtual ModelType GetModelType();
	virtual int GetNumberOfChildren();
	virtual BaseModelClass* GetChild(int i);
	virtual void GetLocalActions(base_action_list_t& actions);
	virtual void GetGlobalActions(base_action_list_t& actions);
	virtual bool HasLocalState();
	virtual void GetLocalState(State* _LS);
	virtual void SetLocalState(const State* _LS);
	virtual void ConstructLocalState(State*& _LS);
	virtual void DestructLocalState(State* _LS);

}; // end SANModel class

#endif // end SANModel
