/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: SingleGroup.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef SINGLEGROUP_H
#define SINGLEGROUP_H

#include <stdlib.h>
#include "StandardErrors.h"
#include "BaseGroupClass.h"

/**
 * This class represents a specific type of action
 * group with only one member.
 *
 * @author Jay M. Doyle
 * @version Mobius Version 1.0
 */
class SingleGroup: public BaseGroupClass {

public:

	/**
	 * Base constructor
	 */
	SingleGroup();

	/**
	 * Overloaded constructor that takes the one member
	 *
	 * @param TheMemberAction The only member of the action group
	 */
	SingleGroup(BaseGroupClass* TheMemberAction);

	/**
	 *  Base destructor
	 */
	~SingleGroup();

	/**
	 * This method set the member action to the one specified as a parameter
	 *
	 * @param TheSpecifiedMember The new member of the action group
	 */
	void SetAction(BaseActionClass* TheSpecifiedMember);

	/**
	 * Overloaded method from BaseActionClass
	 * @see BaseActionClass
	 */
	bool ReactivationPredicate();

	/**
	 * Overloaded method from BaseActionClass
	 *
	 * @see BaseActionClass
	 */
	bool ReactivationFunction();

	/**
	 * Since there is only one member in the group, this method set
	 * the selected action to that member
	 */
	bool SelectAction();

	/**
	 * This method samples the selected action's distribution
	 *
	 * @return A sample from the selected action's probability distribution function
	 */
	double SampleDistribution();

	/**
	 * This method fires the selected action
	 *
	 * @return A pointer to the action that fired
	 */
	BaseActionClass* Fire();

	/**
	 *  This method returns the probability of selecting the
	 *  specified method from the group. For single action groups
	 *  this means this method always returns 0 or 1
	 *
	 *  @param TheSpecifiedAction The action in question
	 *  @return 0 or 1 depending if the specified action is the member action
	 */
	double Probability(BaseActionClass* TheSpecifiedAction);

	/**
	 * Overloaded method from BaseActionClass
	 * @see BaseActionClass
	 */
	bool EnablingChange(void);

	/**
	 * Overloaded method from BaseActionClass
	 * @see BaseActionClass
	 */
	bool Enabled();

	/**
	 * verloaded method from BaseActionClass
	 * @see BaseActionClass
	 */
	double Weight();

	/**
	 * Overloaded method from BaseActionClass
	 * @see BaseActionClass
	 */
	double Rate();

	/**
	 * Overloaded method from BaseActionClass
	 * @see BaseActionClass
	 */
	int Rank();

private:

	/**
	 * A pointer to the group's only member
	 */
	BaseActionClass *SingleAction;

	/**
	 * The current and previous enabling states--they are used
	 * to determined it there has been an enabling change
	 */
	int NewEnabled, OldEnabled;

}; // end SingleGroup class

#endif // end SINGLEGROUP_H
