/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2011 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

/**
 * @file ADVISEModel.h
 * This file contains the ADVISEModel class which is a Mobius atomic model base
 * class. This file also contains the ExploreNode class, which is used for a
 * state space exploration by the adversary when deciding on an appropriate
 * next attack step to attempt.
 */

#ifndef ADVISE_MODEL_H_
#define ADVISE_MODEL_H_

#include "Step.h"
#include "Goal.h"
#include "BeginAdversaryDecision.h"
#include "BaseADVISEAction.h"
#include "ADVISEStateVariable.h"
#include "../../BaseStateVariableClass.h"
#include "../../BaseGroupClass.h"
#include "../../BaseModelClass.h"
#include "../../BaseActionClass.h"
#include "../../jagged_array.hpp"
#include "StateLookAheadTree.hpp"
#include <cstddef>
#include <iostream>
#include <string>
#include <map>
#include <set>
#include <vector>


class Step;
class ExploreNode;
class StateLookAheadNode;

/**
 * This is the base class for all ADVISE atomic models.
 */
class ADVISEModel : public BaseModelClass {
public:
	/**
	 * Constructor
	 */
	ADVISEModel();
	/**
	 * Constructor
	 * @param[in] theModelName The name of the ADVISE model
	 * @param[in] numSVs The number of state variables in the model.
	 * @param[in] theSVs The set of state variables in the model.
	 * @param[in] numActions The number of actions in the model.
	 * @param[in] theActions The set of actions in the model.
	 * @param[in] numGroups The number of groups in the model.
	 * @param[in] groups The set of groups in the model.
	 * @param[in] numSteps The number of sets in the model.
	 * @param[in] numOutcomes An array of the number of outcomes for each step.
	 * @param[in] steps The two dimensional array of step/outcome pairs.
	 * @param[in] numGoals The number of goals in the model.
	 * @param[in] goals The set of goals in the model.
	 */
	ADVISEModel(char *theModelName, std::size_t numSVs, BaseStateVariableClass** theSVs,
			std::size_t numActions, BaseADVISEAction** theActions,
			std::size_t numGroups, BaseGroupClass** groups,
			std::size_t numSteps, int* numOutcomes, Step** steps,
			std::size_t numGoals, Goal** goals);
	/**
	 * Destructor
	 */
	virtual ~ADVISEModel();
	/**
	 * This function chooses the next attack step for the adversary to attempt.
	 */
	void makeAttackDecision();
	/**
	 * This method initializes the model with the given data.
	 * @param[in] theModelName The name of the ADVISE model
	 * @param[in] numSVs The number of state variables in the model.
	 * @param[in] theSVs The set of state variables in the model.
	 * @param[in] numActions The number of actions in the model.
	 * @param[in] theActions The set of actions in the model.
	 * @param[in] numGroups The number of groups in the model.
	 * @param[in] groups The set of groups in the model.
	 * @param[in] numSteps The number of sets in the model.
	 * @param[in] numOutcomes An array of the number of outcomes for each step.
	 * @param[in] steps The two dimensional array of step/outcome pairs.
	 * @param[in] numGoals The number of goals in the model.
	 * @param[in] goals The set of goals in the model.
	 */
	void commonInit(char *theModelName, std::size_t numSVs, BaseStateVariableClass** theSVs,
			std::size_t numActions, BaseADVISEAction** theActions,
			std::size_t numGroups, BaseGroupClass** groups,
			std::size_t numSteps, int* numOutcomes, Step** steps,
			std::size_t numGoals, Goal** goals);
	/**
	 * This function sets the set of step/outcome pairs in the model
	 * @param[in] steps The set of step/outcome pairs
	 */
	void setSteps(Step** steps, std::size_t numSteps);
	/**
	 * This function sets the set of goals in the model
	 * @param[in] goals The set of goals
	 */
	void setGoals(Goal** goals, std::size_t numGoals);

	/**
	 * This function explores the model state space.
	 * @param[in] the ID value of the step
	 * @return A pointer to the Step or null in the ID does not exists
	 */
	Step* getStep(int stepID);

	/**
	 * This function explores the model state space.
	 * @param[in] the ID value of the step
	 * @param[in] the ID value of the outcome within the step
	 * @return A pointer to the Step or null in the ID does not exists
	 */
	Step* getOutcome(int stepID, int outcomeID);

	/**
	 * This function recalculates the adversary decision weight values.
	 */
	////dummy function to use until the calling function parameters are changed elsewhere
	virtual void recalculateAdversaryDecisionWeights();

	/**
	 * This function recalculates the adversary decision weight values according to the supplied algorithm.
	 * @param[in] algorithmID The id of the algorithm to use.
	 */
	static const int algorithmSTANDARD_NO_STORED_STATE_TREE = 0;
	static const int algorithmSTANDARD_WITH_STORED_STATE_TREE = 1;
	static const int algorithmPATH_ATTRACTIVENESS_BOUNDS = 2;
	static const int algorithmCACHE = 3;

	virtual void recalculateAdversaryDecisionWeights(int algorithmID);

	/**
	 * This function computes the path attractiveness bounds
	 * @param[in] horizon The planning horizon
	 */
	void computePathAttractivenessBounds(int horizon);
	/**
	 * This function computes the attractiveness bounds starting at a specific node
	 * @param[in] node Starting node.
	 * @param[in] horizon The planning horizon
	 * @param[in] branchID The branch id
	 */
	void computeBounds(StateLookAheadNode* const node, int horizon, int branchID);

	/**
	 * This function uses the attractiveness bounds to decide the next attack step
	 * @param[in] horizon The planning horizon
	 */
	void useBoundsToDecide(int horizon);
	/**
	 * This function stores the current state of the model in the parameter
	 * @param[out] theState The memory space to store the model state.
	 */
	void CurrentState(void* theState);
	/**
	 * This function stores a list of pointers to all the actions in the model.
	 * @param[out] theContainer The memory space to store the list of actions
	 */
	void listActions(List<BaseActionClass>* theContainer) override;
	/**
	 * This function stores the action pointers of the provided name in the second parameter
	 * @param[in] actionName The name of the action to find
	 * @param[out] theContainer The pointer to the action if found
	 */
	void listActions(char *actionName, List<BaseActionClass>* theContainer) override;
	/**
	 * This function stores a list of pointers to all the groups in the model.
	 * @param[out] theContainer The memory space to store the list of groups.
	 */
	void listGroups(List<BaseGroupClass>* theContainer);
	/**
	 * This function returns the current size of the model state
	 */
	int StateSize();
	/**
	 * This function sets the model's state to the data in the parameter
	 * @param[in] theState The state to set the model to.
	 */
	void SetState(void* theState);
	/**
	 * This function compares two model states to see if they are equal.
	 * @return True if the states are equal
	 */
	bool CompareState(void* state1, void* state2);
	/**
	 * This function prints various information about the model
	 */
	void PrintModelInfo();
	/**
	 * This function returns a list of models with the provided name.
	 * @param[in] theModelName The model name to find
	 * @param[out] theContainer The list to put the models in.
	 */
	void listModels(char* theModelName, List<BaseModelClass>* theContainer);
	/**
	 * The function returns a list of state variable pointers who match a given name and are in a model with the given name.
	 * @param[in] theModelName The name of the model to search inside
	 * @param[in] theSVName The name of the state variable to find
	 * @param[out] The list of pointers
	 */
	void listSVs(char* theModelName, char* theSVName, List<BaseStateVariableClass>* theSVContainer, List<BaseModelClass>* theModelContainer);
	/**
	 * This function prints the current model state to the output stream
	 * @param[in] strm The output stream.
	 */
	void printState(std::ostream *strm);
	/**
	 * This function prints the local state variables to the output stream
	 * @param[in] strm The output stream.
	 */
//BAF	BaseStateVariableClass** GetLocalStateVariables(std::ostream *strm);
	/**
	 * This function converts the provided cost to utility
	 * @param[in] cost The cost
	 * @returns The utility
	 */
	virtual double convertCostToUtility(double cost) = 0;
	/**
	 * This function converts the provided payoff to utility
	 * @param[in] payoff The payoff
	 * @returns The utility
	 */
	virtual double convertPayoffToUtility(double payoff) = 0;
	/**
	 * This function converts the provided detection risk to utility
	 * @param[in] detection The detection risk
	 * @returns The utility
	 */
	virtual double convertDetectionToUtility(double detection) = 0;

	inline BaseModelClass * getRootModel() {
		return this;
	}

	/**
	 * The state decision cache.
	 */
	static std::map<std::string,int> cache;

	/**
	 * The following is required for the symbolic SSG although even under simulation an assert fails if these don't exist...
	 */

	class LocalState: public State {
		public:
			LocalState(ADVISEModel* model);
			~LocalState();
			friend class ADVISEModel;
			void print();
			int hash();
			int compare(State const & S) const;
			State const& operator=(State const& S);
			int fwriteState(FILE* f);
			int freadState(FILE* f);
			void* GetStatePtr();
		protected:
			void SetState(const BaseModelClass* model);
			void GetState(const BaseModelClass* model);
		private:
			char* state;
			int size;
	};

	ModelType GetModelType() override;
	int GetNumberOfChildren() override;
	BaseModelClass* GetChild(int i) override;
	void GetLocalActions(base_action_list_t& actions) override;
	void GetGlobalActions(base_action_list_t& actions) override;
	bool HasLocalState() override;
	void GetLocalState(State* _LS) override;
	void SetLocalState(const State* _LS) override;
	void ConstructLocalState(State*& _LS) override;
	void DestructLocalState(State* _LS) override;

protected:

	void exploreStateSpace(int horizon,
			Step** bestStep,
			double* expectedPayoff,
			double* expectedCost,
			double* expectedDetection,
			StateLookAheadNode* const parent,
			int level,
			int* latestFirstDNLevel);

	/**
	 * This function builds the state tree from the starting node until the given depth is reached.
	 * @param[in] parent The starting node
	 * @param[in] horizon The planning horizon
	 * @param[out] latestFirstDNHeight The height of the subtree of the tree rooted at parent corresponding to the latest do-nothing step
	 */
	void buildStateTree(StateLookAheadNode* const parent, int horizon, int *latestFirstDNHeight);

	/**
	 * This function builds the state tree but doesn't calculate attractiveness
	 * @param[in] parent The starting node
	 * @param[in] horizon The planning horizon
	 */
	void buildStateTreeNoAttractCalc(StateLookAheadNode* const parent, int horizon);

	/**
	 * This function traverses the state tree and calculates the attractiveness at each node
	 * @param[in] parent The starting node
	 * @param[in] horizon The planning horizon
	 * @param[out] latestFirstDNHeight The height of the subtree of the tree rooted at parent corresponding to the latest do-nothing step
	 */
	void traverseStateTree(StateLookAheadNode* const parent, int horizon, int *latestFirstDNHeight);

	/**
	 * This is the two dimensional array of step/outcome pair pointers. The first index is the step, the second is the outcome of that step.
	 */
	mobius::jagged_array<Step*> mSteps;

	/**
	 * This is a one dimensional array of goal pointers.
	 */
	std::vector<Goal*> theGoals;

	/**
	 * The state look-ahead tree for this model.
	 */
	StateLookAheadTree fullSLAT;
	StateLookAheadTree activeSLAT;

	/**
	 * The adversary's planning horizon
	 */
	int planningHorizon = 1;
	
	/**
	 * The adversary's cost of detection
	 */
	int costOfDetection = 0;

	/**
	 * The list of state variables in the model
	 */
	std::vector<BaseStateVariableClass*> mStateVariables;

	/**
	 * The state variable that indicates the adversary should choose another attack step
	 */
	BeginAdversaryDecision* MakeDecision;

private:
	/**
	 * This function sets the set of state variables
	 */
	void setSVs(BaseStateVariableClass** svs, std::size_t numSVs);
	/**
	 * This function sets the set of actions
	 */
	void setActions(BaseADVISEAction** actions, std::size_t numActions);
	/**
	 * This function sets the set of groups.
	 */
	void setGroups(BaseGroupClass** groups, std::size_t numGroups);

	/**
	 * This function calculates the payoff of the current state.
	 */
	double calculatePayoff();
};
#endif /* ADVISEMODEL_H_ */
