/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2011 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

/**
 * @file ADVISEStateVariable.h
 * This file contains the ADVISEStateVariable class.
 */
#ifndef ADVISESTATEVARIABLE_H_
#define ADVISESTATEVARIABLE_H_

#include "../../state/SharableSV.h"
#include "../../StandardErrors.h"
#include "../../debug.h"
#include <iostream>

/**
 * The ADVISEStateVariable class is the parent class for all ADVISE state variables.
 */
template <class C>
class ADVISEStateVariable: public SharableSV<C> {

public:
	/**
	 * Constructor
	 */
	ADVISEStateVariable() {
		CF("Entering ADVISEStateVariable::ADVISEStateVariable");
		BaseStateVariableClass::SetName("default");
		SharableSV<C>::Register(&theMark);
		ASSERT(theMark != nullptr);
		DS3("Address of Mark after assignment", theMark);
		*theMark = 0;
	}
	/**
	 * Constructor
	 * @param[in] theName The state variable name
	 */
	explicit ADVISEStateVariable(char *theName) {
		CF("Entering ADVISEStateVariable::ADVISEStateVariable(char*)");
		BaseStateVariableClass::SetName(theName);
		SharableSV<C>::Register(&theMark);
		DS3("Address of Mark after assignment", theMark);
		ASSERT(theMark != nullptr);
		*theMark = 0;
	}
	/**
	 * Constructor
	 * @param[in] theName The state variable name
	 * @param[in] initialValue The state variable's initial value
	 */
	ADVISEStateVariable(char *theName, const C& initialValue) {
		CF("Entering ADVISEStateVariable::ADVISEStateVariable(char*, const C&)");
		BaseStateVariableClass::SetName(theName);
		SharableSV<C>::Register(&theMark);
		DS3("Address of Mark after assignment", theMark);
		*theMark = initialValue;

#ifdef DEBUG
		if(GlobalDebugLevel > 1) {
			std::cout << "Initial marking for "
					<< BaseStateVariableClass::StateVariableName
					<< " is " << *theMark << std::endl;
		}
#endif
	}

	ADVISEStateVariable(const ADVISEStateVariable&) = delete;
	ADVISEStateVariable& operator =(const ADVISEStateVariable&) = delete;

	/**
	 * Destructor
	 */
	virtual ~ADVISEStateVariable() = default;

	/**
	 * This method returns the value of the state variable as an Lvalue
	 *
	 * @return The value of the state variable (as an Lvalue)
	 */
	inline C& Mark() {
		return *theMark;
	}
	/**
	 * This method returns the value of the state variable by value
	 */
	C getMark() const {
		CF("Entering ADVISEStateVariable::getMark()");
		return *theMark;
	}
	/**
	 * This method sets the value of the state variable by value
	 * @param[in] newVal The new value.
	 */
	void setMark(C newVal) {
		CF("Entering ADVISEStateVariable::setMark(C)");
		*theMark = newVal;
	}

protected:
	/**
	 * The value of the state variable.
	 */
	C *theMark;
};

#endif /* ADVISESTATEVARIABLE_H_ */
