/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2011 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

/**
 * @file BaseADVISEAction.h
 * This file contains the BaseADVISEAction class.
 */

#ifndef BASEADVISEACTION_H_
#define BASEADVISEACTION_H_

#include "../../BaseGroupClass.h"

class ADVISEModel;
/**
 * This class provides a parent class for all ADVISE actions.
 */
class BaseADVISEAction: public BaseGroupClass {
public:
	/**
	 * This variable stores the enabling state in the new marking
	 */
	bool NewEnabled;
	/**
	 * This variable stores the enabling state in the previous marking
	 */
	bool OldEnabled;
	/**
	 * This is the ADVISEModel that this state variable belongs to.
	 */
	ADVISEModel* parent;
	/**
	 * Constructor
	 */
	BaseADVISEAction();
	/**
	 * Destructor
	 */
	virtual ~BaseADVISEAction();
	/**
	 * Constructor
	 * @param[in] theActionName The name of the action
	 * @param[in] theGroupId The group id
	 * @param[in] theDistType The action probability distribution function
	 * @param[in] theExecPolicy The execution policy
	 * @param[in] numAffectedSVs The number of affecting state variables
	 * @param[in] numEnablingSVs The number of enabling state variables
	 * @param[in] theReactivation Initial value of reactivation
	 */
	BaseADVISEAction(char *theActionName, int theGroupId,
			Distribution theDistType, ExecutionPolicyType theExecPolicy,
			int numAffectedSVs, int numEnablingSVs, bool theReactivation);
	/**
	 * This method initializes the object data.
	 * @param[in] theActionName The name of the action
	 * @param[in] theGroupId The group id
	 * @param[in] theDistType The action probability distribution function
	 * @param[in] theExecPolicy The execution policy
	 * @param[in] numAffectedSVs The number of affecting state variables
	 * @param[in] numEnablingSVs The number of enabling state variables
	 * @param[in] theReactivation Initial value of reactivation
	 */
	void commonInit(char *theActionName, int theGroupId,
			Distribution theDistType, ExecutionPolicyType theExecPolicy,
			int numAffectedSVs, int numEnablingSVs, bool theReactivation);
	/**
	 * This function initializes the affected and enabling lists
	 */
	virtual void LinkVariables() = 0;
	/**
	 * This function changes the model state.
	 */
	virtual BaseActionClass *Fire() = 0;
	/**
	 * This function selects the action from the group to fire.
	 */
	bool SelectAction();
	/**
	 * This function calculates the weights which are used by SelectAction
	 */
	void CalculateWeightDistribution();
	/**
	 * This function determines if the enabling state has changed.
	 */
	bool EnablingChange();
	/**
	 * This function returns the probability of the action in the group.
	 * @param[in] theAction The action whose probability should be returned.
	 */
	double Probability(BaseActionClass *theAction);
	/**
	 * This function determines if the given state variable is an affecting state variable
	 * @param[in] theSV The state variable to check
	 */
	bool IsAnAffectingStateVariable(BaseStateVariableClass *theSV);
	/**
	 * This function sets the group id of the action
	 * @param[in] theId The ID
	 */
	inline void setGroupId(int theId) {
		GroupID = theId;
	}
	/**
	 * This function sets the parent ADVISE model
	 * @param[in] theParent The parent ADVISE model
	 */
	inline void setParent(ADVISEModel* theParent) {
		ASSERT(theParent != nullptr);
		parent = theParent;
	}

protected:
	/**
	 * The parameters for the action distribution.
	 */
//	double *TheDistributionParameters;
};

#endif /* BASEADVISEACTION_H_ */
