/*
 * StateLookAheadTree.hpp
 *
 *  Created on: Oct 29, 2015
 *      Author: bfeddersen
 */

#ifndef STATELOOKAHEADTREE_HPP_
#define STATELOOKAHEADTREE_HPP_

#include <string>
#include <vector>
#include <map>

class ADVISEModel;
class StateLookAheadNode;
class Step;

/**
 * @file StateLookAheadTree.hpp
 * This file contains an implementation of the state look-ahead tree (SLAT)
 * as described in Elizabeth LeMay's Thesis, page 16;
 */
class StateLookAheadTree
{
public:
	typedef std::map<int, double> attractivenessBounds;

	StateLookAheadTree();
	~StateLookAheadTree();

	bool isEmpty() const;
	void reset(ADVISEModel* model);

	StateLookAheadNode* const appendTo(StateLookAheadNode* parentNode,
				int stepIDValue, int outcomeIDValue,
				double outcomeProbabilityValue,
				double outcomeDetectionValue,
				double attackStepCostValue,
				double payoffValue,
				ADVISEModel* model) ;

	const unsigned char* getState() const;
	Step* bestNextStep() const;
	void bestNextStep(Step* newNextBestStep);

	const StateLookAheadNode& currentNode() const
	{
		return *mpCurrentNode;
	}

	StateLookAheadNode* const currentNode()
	{
		return mpCurrentNode;
	}

	const StateLookAheadNode* const currentNodePtr() const
	{
		return mpCurrentNode;
	}

	void currentNode(StateLookAheadNode* node)
	{
		mpCurrentNode = node;
	}

	bool hasChildren() const;
	std::multimap<int, StateLookAheadNode*>& getChildren() const;

	void updateCurrentState(ADVISEModel* model);

	attractivenessBounds& lowerBoundAttractiveness()
	{
		return mLowerBoundAttractiveness;
	}

	attractivenessBounds& upperBoundAttractiveness()
	{
		return mUpperBoundAttractiveness;
	}

	void clearAttractivenessFactors();

private:
	// make this node the root node of the state tree, deleting all nodes
	// above this node
	void makeRoot(StateLookAheadNode* pNode);

	StateLookAheadNode* mpRoot = nullptr;
	StateLookAheadNode* mpCurrentNode = nullptr;

	// most attractive next attack step based on analysis looking
	// ahead to the planning horizon states
	Step* mpBestNextStep = nullptr;

	// (only for bounds calculation)
	attractivenessBounds mLowerBoundAttractiveness;  // key = attackStepID; value = bound
	attractivenessBounds mUpperBoundAttractiveness;  // key = attackStepID; value = bound
};


class StateLookAheadNode
{
public:

	// these data members need to be accessed by functions in the ADVISEModel class
	// store this set of values on backwards traversal of tree (for exact calculation)

	// most attractive next attack step based on analysis looking
	// ahead to the planning horizon states
	Step* mpBestNextStep = nullptr;

	// expected sum of costs of attack steps between this state and the
	// planning horizon states
	double expectedCostSum = 0.0;

	// expected �sum� of detection of attack steps between this state and
	// the planning horizon states
	double expectedDetectionSum = 0.0;

	// expected payoff in this state derived from the planning horizon states
	double expectedPayoffSum = 0.0;

	// store these two path values (only for bounds calculation)

	// sum of costs of attack steps between root node state and this horizon state
	double pathCost = 0.0;

	// �sum� of detection of attack steps between root node state and this horizon state
	double pathDetection = 0.0;

	double computedCost = 0.0;
	double computedDetection = 0.0;
	double computedPayoff = 0.0;
	double costUtility = 0.0;
	double detectionUtility = 0.0;
	double payoffUtility = 0.0;
	double attractiveness = 0.0;

	const char* getName() const;

	bool isEmpty() const {
		return (mStepID == -1) && (mOutcomeID == -1);
	}

	bool hasChildren() const {
		return ! children.empty();
	}

	inline const std::multimap<int, StateLookAheadNode*>& getChildren() const {
		return children;
	}

	inline std::multimap<int, StateLookAheadNode*>& getChildren() {
		return children;
	}

	inline const unsigned char* getState() const {
		return modelState.data();
	}
	inline int getStepID() const {
		return mStepID;
	}
	inline int getOutcomeID() const {
		return mOutcomeID;
	}
	inline double getOutcomeProbability() const {
		return outcomeProbability;
	}
	inline double getOutcomeDetection() const {
		return outcomeDetection;
	}
	inline double getAttackStepCost() const {
		return attackStepCost;
	}
	inline double getPayoff() const {
		return payoff;
	}

	StateLookAheadNode* addChild(int stepID, int outcomeID,
				double outcomeProbabilityValue,
				double outcomeDetectionValue,
				double attackStepCostValue,
				double payoffValue);

private:

	StateLookAheadNode(StateLookAheadNode*& rootNode,
				StateLookAheadNode* parentNode,
				int stepID, int outcomeID,
				double outcomeProbabilityValue,
				double outcomeDetectionValue,
				double attackStepCostValue,
				double payoffValue,
				ADVISEModel* model);

	~StateLookAheadNode();

	inline void addChild(int attackStepID, StateLookAheadNode *child) {
		// in this (parent) node, add link to child node
		children.insert(std::make_pair(attackStepID, child));
	}

	double computePathEscapeDetection();

	StateLookAheadNode* mpRoot = nullptr; // pointer to root node
	StateLookAheadNode* mpParent = nullptr; // pointer to parent node
	std::multimap<int, StateLookAheadNode*> children; // set of pointers to children nodes
	std::vector<unsigned char> modelState; // model state represented by this state node

	// ID of attack step to execute to get from parent state to this state
	int mStepID = -1;

	// ID of attack step outcome to execute to get from parent state to this state
	int mOutcomeID = -1;

	// probability of the attack step outcome to get from parent state to
	// this state, given that the attack step is attempted
	double outcomeProbability = 0.0;

	// probability of detection for the attack step outcome to get from parent
	// state to this state
	double outcomeDetection = 0.0;

	// cost of the attack step to get from parent state to this state
	double attackStepCost = 0.0;

	// total payoff of all goals achieved in this state
	double payoff = 0.0;

	ADVISEModel* mpModel = nullptr;

friend class StateLookAheadTree;
};


std::string to_string(const StateLookAheadTree& slat);

#endif /* STATELOOKAHEADTREE_HPP_ */
