/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2011 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

/**
 * @file Step.h
 * This file contains the Step class which is a Mobius action.
 */

#ifndef STEP_H_
#define STEP_H_

#include "BaseADVISEAction.h"
#include "ADVISEModel.h"
#include "StepWeight.h"
#include "StepChosen.h"

/**
 * This class represents the ADVISE attack step/outcome pairs that are found in attack execution graphs. An attack step is a Mobius action and therefore changes state.
 */
class Step : public BaseADVISEAction {
public:
	/**
	 * Constructor
	 */
	Step();
	/**
	 * Destructor
	 */
	virtual ~Step();
	/**
	 * This function executes the attack steps effects and changes the model state.
	 */
	virtual BaseActionClass *Fire();
	/**
	 * This function determines if the attack step's preconditions have been met in the current state.
	 */
	virtual bool preconditionsMet() = 0;
	/**
	 * This function executes the attack step's effects.
	 */
	virtual void executeEffects() = 0;
	/**
	 * This function returns the cost of the attack step
	 */
	virtual double getCost() = 0;
	/**
	 * This function returns the outcome probability
	 */
	virtual double getOutcomeProbability() = 0;
	/**
	 * This function returns the detection of the attack step outcome pair
	 */
	virtual double getDetection() = 0;
	/**
	 * This function sets the values of the StepChosen and StepWeight associated with this attack step/outcome pair.
	 */
	inline void setSVs(StepChosen* theChosen, StepWeight* theWeight) { theWeightSV = theWeight; theChosenSV = theChosen; }
	/**
	 * This function gets the value of the StepWeight associated with this attack step
	 */
	inline StepWeight* getWeightSV() { return theWeightSV; }
	/**
	 * This function gets the value of the StepChosen associated with this attack step
	 */
	inline StepChosen* getChosenSV() { return theChosenSV; }
	/**
	 * This function sets the step index
	 */
	inline void setStepIndex(int index) { stepIndex = index; }
	/**
	 * This function gets the step index.
	 */
	inline int getStepIndex() { return stepIndex; }

	const char* getStepName() const { return mpStepName; }

protected:
	void setStepName(const char * pName) { mpStepName = pName; }
	/**
	 * The step weight associated with this step/outcome
	 */
	StepWeight* theWeightSV;
	/**
	 * The step chosen associated with this step/outcome
	 */
	StepChosen* theChosenSV;
	/**
	 * The step index
	 */
	int stepIndex;

	const char* mpStepName;
};

#endif /* STEP_H_ */
