/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: debug.h
 *
 *      Authors: Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef DEBUG_H
#include <cstdlib>
#include <iostream>

using namespace std;


/* Turn on/off debugging */

//Off for performance testing.
// this is turned on by the debug target in the Makefile.  shouldn't
// need to turn it on by modifying it here anymore
//#define DEBUG

/* Indicates debug level:
 0 : No debugging
 1 : Important events printed
 2 : Important data structures printed
 3 : Detailed data structures printed
 4 : Detailed control flow information printed
 */

extern int GlobalDebugLevel;

/* Assertions:
 ASSERT(TheAssertion)	: Assert the expression TheAssertion.  Report
 an error including file name and line number
 if TheAssertion is FALSE
 ASSERTV(TheAssertion,TheVariable,TheValue)
 : Assert the expression TheAssertion.  Report
 an error including file name, line number,
 a variable name (TheVariable), and it's
 value (TheValue)
 ASSERTE(TheAssertion,TheVariable,TheValue,TheExplanation)
 : Do the same as ASSERTV except also give the
 explanation in the string TheExplanation
 */

#ifdef DEBUG
// RK: undef first to get rid of warning
# undef ASSERT
# define ASSERT(TheAssertion) \
do { \
  if (!(TheAssertion)) \
  { \
    std::cerr << "!! Assertion failed in file " << __FILE__ << ", line " \
         << __LINE__ << std::endl; \
    abort(); \
  } /* end TheAssertion */ \
} while (0)

#undef ASSERTV
#define ASSERTV(TheAssertion,TheVariable,TheValue) \
do { \
  if (!(TheAssertion)) \
  { \
    std::cerr << "!! Assertion failed in file " << __FILE__ << ", line " \
         << __LINE__ << endl; \
    std::cerr << "!! Variable " << TheVariable << " = " << TheValue << std::endl; \
    abort(); \
  } /* end TheAssertion */ \
} while (0)

#undef ASSERTE
#define ASSERTE(TheAssertion,TheVariable,TheValue,TheExplanation) \
do { \
  if (!(TheAssertion)) \
  { \
    std::cerr << "!! Assertion failed in file " << __FILE__ << ", line " \
         << __LINE__ << std::endl; \
    std::cerr << "!! Variable " << TheVariable << " = " << TheValue << std::endl; \
    std::cerr << "!! " << TheExplanation << std::endl; \
    abort(); \
  } /* end TheAssertion */ \
} while (0)

/* Trace information
 IE(TheMessage)	 : Print the message TheMessage if GlobalDebugLevel >= 1
 DS1(TheName,TheValue) : Print out the name and value of the data
 structure if GlobalDebugLevel >= 2
 DS2(TheName,TheValue) : Print out the name and value of the data
 structure if GlobalDebugLevel >= 3
 CF(TheMessage)	 : Control flow debugging; message TheMessage is
 printed if GlobalDebugLevel >= 4
 */

// Important Event
#undef IE
#define IE(TheMessage) \
do { \
  if (GlobalDebugLevel >= 1) \
  { \
    std::cout << "** " << TheMessage <<  "\n"; \
  } /* end GlobalDebugLevel */ \
} while (0)

// Basic Data Structures
#undef DS1
#define DS1(TheName,TheValue) \
do { \
  if (GlobalDebugLevel >= 1) \
  { \
    std::cout << "** " << TheName << " = " << TheValue <<  "\n"; \
  } /* end GlobalDebugLevel */ \
} while (0)

// Important Data Structures
#undef DS2
#define DS2(TheName,TheValue) \
do { \
  if (GlobalDebugLevel >= 2) \
  { \
    std::cout << "** " << TheName << " = " << TheValue <<  "\n"; \
  } /* end GlobalDebugLevel */ \
} while (0)

// Detailed Data Structures
#undef DS3
#define DS3(TheName,TheValue) \
do { \
  if (GlobalDebugLevel >= 3) \
  { \
    std::cout << "** " << TheName << " = " << TheValue <<  "\n"; \
  } /* end GlobalDebugLevel */ \
} while (0)

// Most Detailed Data Structures
#undef DS4
#define DS4(TheName,TheValue) \
do { \
  if (GlobalDebugLevel >= 4) \
  { \
    std::cout << "** " << TheName << " = " << TheValue <<  "\n"; \
  } /* end GlobalDebugLevel */ \
} while (0)

// Control Flow - Level 1 (minimal)
#undef CF1
#define CF1(TheMessage) \
do { \
  if (GlobalDebugLevel >= 1) \
  { \
    std::cout << "** " << TheMessage << "\n"; \
  } /* end GlobalDebugLevel */ \
} while(0)

// Control Flow - Level 2 (basic)
#undef CF2
#define CF2(TheMessage) \
do { \
  if (GlobalDebugLevel >= 2) \
  { \
    std::cout << "** " << TheMessage << "\n"; \
  } /* end GlobalDebugLevel */ \
} while(0)

// Control Flow - Level 3 (intermediate)
#undef CF3
#define CF3(TheMessage) \
do { \
  if (GlobalDebugLevel >= 3) \
  { \
    std::cout << "** " << TheMessage << "\n"; \
  } /* end GlobalDebugLevel */ \
} while(0)

 // Control Flow - Level 4 (most detailed)
 #undef CF4
 #define CF4(TheMessage) \
 do { \
   if (GlobalDebugLevel >= 4) \
   { \
     std::cout << "** " << TheMessage << "\n"; \
   } /* end GlobalDebugLevel */ \
 } while(0)

// Detailed Control Flow
#undef CF
#define CF(TheMessage) \
do { \
  if (GlobalDebugLevel >= 4) \
  { \
    std::cout << "** " << TheMessage <<  "\n"; \
  } /* end GlobalDebugLevel */ \
} while(0)

#else /* DEBUG */
#define ASSERT(TheAssertion)
#define ASSERTV(TheAssertion,TheVariable,TheValue)
#define ASSERTE(TheAssertion,TheVarible,TheValue,TheExplanation)
#define IE(TheMessage)
#define DS1(TheName,TheValue)
#define DS2(TheName,TheValue)
#define DS3(TheName,TheValue)
#define DS4(TheName,TheValue)
#define CF1(TheMessage)
#define CF2(TheMessage)
#define CF3(TheMessage)
#define CF4(TheMessage)
#define CF(TheMessage)
#endif /* DEBUG */

#endif /* DEBUG_H */
