
/*
 * jagged_array.hpp
 *
 *  Created on: Oct 15, 2015
 *      Author: bfeddersen
 */

#ifndef JAGGED_ARRAY_HPP_
#define JAGGED_ARRAY_HPP_

#include <cstddef>
#include <algorithm>
#include <vector>
#include <type_traits>
#include <stdexcept>

/***
 * A jagged array from Wikipedia:
 * In computer science, a jagged array, also known as a ragged array, is an
 * array of arrays of which the member arrays can be of different sizes,
 * producing rows of jagged edges when visualized as output.  In contrast,
 * C-styled arrays are always rectangular.
 */
namespace mobius
{

template<typename T>
class jagged_array {
public:
	// publish our template parameter and variations thereof
	typedef T 						value_type;
	typedef T* 						pointer_type;
	typedef const T*				const_pointer_type;
	typedef T& 						reference_type;
	typedef const T& 				const_reference_type;
    typedef std::size_t				size_type;

    typedef std::vector<T>								row_type;
    typedef row_type&									row_reference_type;
    typedef const row_type&								const_row_reference_type;
    typedef typename row_type::iterator					iterator;
    typedef typename row_type::const_iterator			const_iterator;
    typedef typename row_type::reverse_iterator			reverse_iterator;
    typedef typename row_type::const_iterator			const_reverse_iterator;

    typedef std::vector<row_type>								row_container_type;
    typedef typename row_container_type::iterator				row_iterator;
    typedef typename row_container_type::const_iterator			const_row_iterator;
    typedef typename row_container_type::reverse_iterator		row_reverse_iterator;
    typedef typename row_container_type::const_reverse_iterator	const_row_reverse_iterator;


	jagged_array() = default;
    explicit jagged_array(size_type numRows);
    ~jagged_array() = default;

    jagged_array(const jagged_array& rhs) = delete;
    jagged_array(jagged_array&& rhs) = delete;
    jagged_array& operator=(const jagged_array& x) = delete;
    jagged_array& operator=(jagged_array&& x) = delete;


// Accessors

    row_reference_type		 at(size_type row);
    const_row_reference_type at(size_type row) const;

    row_reference_type		 operator[](size_type row);
    const_row_reference_type operator[](size_type row) const;

    pointer_type 		data(size_type row);
    const_pointer_type 	data(size_type row) const;

// Iterators

    row_iterator 		begin();
    iterator 			begin(size_type row);
    const_row_iterator	cbegin() const;
    const_iterator		cbegin(size_type row) const;

    row_iterator 		end();
    iterator 			end(size_type row);
    const_row_iterator	cend() const;
    const_iterator		cend(size_type row) const;

    row_reverse_iterator 		rbegin();
    reverse_iterator 			rbegin(size_type row);
    const_row_reverse_iterator	crbegin() const;
    const_reverse_iterator		crbegin(size_type row) const;

    row_reverse_iterator 		rend();
    reverse_iterator 			rend(size_type row);
    const_row_reverse_iterator	crend() const;
    const_reverse_iterator		crend(size_type row) const;

// Capacity

    bool empty() const;
    bool empty(size_type row) const;

    size_type size() const;
    size_type size(size_type row) const;

//    void reserve(size_type new_cap);
//    void reserve(size_type row, size_type new_cap);

//    size_type capacity() const;
//    size_type capacity(size_type row) const;

// Modifiers

    void resize(size_type numRows);
    void resize(size_type row, size_type count);

private:

    row_container_type mData;
};


/******************************************************************************
 * Implementation Details
******************************************************************************/

template<typename T>
jagged_array<T>::jagged_array(size_type numRows)
	:
		mData(numRows)
{}


////////////////////////
// Accessors
////////////////////////

template<typename T>
typename jagged_array<T>::row_reference_type
inline jagged_array<T>::at(size_type row)
{
	return mData.at(row);
}

template<typename T>
inline typename jagged_array<T>::const_row_reference_type
jagged_array<T>::at(size_type row) const
{
	return mData.at(row);
}

template<typename T>
inline typename jagged_array<T>::row_reference_type
jagged_array<T>::operator[](size_type row)
{
	return mData[row];
}

template<typename T>
inline typename jagged_array<T>::const_row_reference_type
jagged_array<T>::operator[](size_type row) const
{
	return mData[row];
}

template<typename T>
typename jagged_array<T>::pointer_type
jagged_array<T>::data(size_type row)
{
	return mData.at(row).data();
}

template<typename T>
typename jagged_array<T>::const_pointer_type
jagged_array<T>::data(size_type row) const
{
	return mData.at(row).data();
}

////////////////////////
// Iterators
////////////////////////

template<typename T>
typename jagged_array<T>::row_iterator jagged_array<T>::begin()
{
	return mData.begin();
}

template<typename T>
typename jagged_array<T>::iterator jagged_array<T>::begin(size_type row)
{
	return mData.at(row).begin();
}

template<typename T>
typename jagged_array<T>::const_row_iterator jagged_array<T>::cbegin() const
{
	return mData.cbegin();
}

template<typename T>
typename jagged_array<T>::const_iterator jagged_array<T>::cbegin(size_type row) const
{
	return mData.at(row).cbegin();
}

template<typename T>
typename jagged_array<T>::row_iterator jagged_array<T>::end()
{
	return mData.end();
}

template<typename T>
typename jagged_array<T>::iterator jagged_array<T>::end(size_type row)
{
	return mData.at(row).end();
}

template<typename T>
typename jagged_array<T>::const_row_iterator jagged_array<T>::cend() const
{
	return mData.cend();
}

template<typename T>
typename jagged_array<T>::const_iterator jagged_array<T>::cend(size_type row) const
{
	return mData.at(row).cend();
}

template<typename T>
typename jagged_array<T>::row_reverse_iterator
jagged_array<T>::rbegin()
{
	return mData.rbegin();
}

template<typename T>
typename jagged_array<T>::reverse_iterator
jagged_array<T>::rbegin(size_type row)
{
	return mData.at(row).rbegin();
}

template<typename T>
typename jagged_array<T>::const_row_reverse_iterator
jagged_array<T>::crbegin() const
{
	return mData.crbegin();
}

template<typename T>
typename jagged_array<T>::const_reverse_iterator
jagged_array<T>::crbegin(size_type row) const
{
	return mData.at(row).crbegin();
}

template<typename T>
typename jagged_array<T>::row_reverse_iterator
jagged_array<T>::rend()
{
	return mData.rend();
}

template<typename T>
typename jagged_array<T>::reverse_iterator
jagged_array<T>::rend(size_type row)
{
	return mData.at(row).rend();
}

template<typename T>
typename jagged_array<T>::const_row_reverse_iterator
jagged_array<T>::crend() const
{
	return mData.crend();
}

template<typename T>
typename jagged_array<T>::const_reverse_iterator
jagged_array<T>::crend(size_type row) const
{
	return mData.at(row).crend();
}

////////////////////////
// Capacity
////////////////////////

template<typename T>
bool jagged_array<T>::empty() const
{
	return mData.empty();
}

template<typename T>
bool jagged_array<T>::empty(size_type row) const
{
	return mData.at(row).empty();
}

template<typename T>
typename jagged_array<T>::size_type jagged_array<T>::size() const
{
	return mData.size();
}

template<typename T>
typename jagged_array<T>::size_type jagged_array<T>::size(size_type row) const
{
	return mData.at(row).size();
}

//template<typename T>
//void jagged_array<T>::reserve(size_type new_cap )
//{
//	return mData.reserve(new_cap);
//}
//
//template<typename T>
//void jagged_array<T>::reserve(size_type row, size_type new_cap )
//{
//	return mData.at(row).reserve(new_cap);
//}
//
//template<typename T>
//typename jagged_array<T>::size_type jagged_array<T>::capacity() const
//{
//	return mData.capacity();
//}
//
//template<typename T>
//typename jagged_array<T>::size_type jagged_array<T>::capacity(size_type row) const
//{
//	return mData.at(row).capacity();
//}

////////////////////////
// Modifiers
////////////////////////

template<typename T>
void jagged_array<T>::resize(size_type numRows)
{
	mData.resize(numRows);
}

template<typename T>
void jagged_array<T>::resize(size_type row, size_type count)
{
	mData.at(row).resize(count);
}

/******************************************************************************
 * Private Helper Methods
 *****************************************************************************/


} // end namespace mobius

#endif /* JAGGED_ARRAY_HPP_ */
