/*
/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: observer_ptr.hpp
 *
 *      Copyright (c) 2015 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef OBSERVER_PTR_HPP_
#define OBSERVER_PTR_HPP_

#include <memory>
#include <utility>
#include <algorithm>

namespace mobius
{

template <class P>
class observer_ptr {
public:
	// publish our template parameter and variations thereof
	typedef P value_type;
	typedef P* pointer_type;
	typedef const P* const_pointer_type;
	typedef P& reference_type;
	typedef const P& const_reference_type;

	// default c’tor
	constexpr observer_ptr() noexcept : mPtr(nullptr) {}

	// pointer-accepting c’tors
	explicit observer_ptr(pointer_type ptr) noexcept : mPtr(ptr) {}

	// copying c’tors (in addition to compiler-generated copy c’tor)
	observer_ptr(const observer_ptr& o) noexcept : mPtr(o.mPtr) {}
	observer_ptr(observer_ptr&& o) noexcept : mPtr(o.mPtr)
	{
		o.mPtr = nullptr;
	}

	// copying assignments (in addition to compiler-generated copy assignment)
	observer_ptr& operator=(const observer_ptr& o) noexcept
	{
		mPtr = o.mPtr;
		return *this;
	}

	~observer_ptr() = default;

	// observers
	pointer_type get() const noexcept { return mPtr; }
	reference_type operator*() const noexcept {return *mPtr; }
	pointer_type operator->() const noexcept {return mPtr; }
	explicit operator bool() const noexcept {return mPtr != nullptr; }

	// modifiers
	pointer_type release() noexcept
	{
		pointer_type oldPtr(mPtr);
		mPtr = nullptr;
		return oldPtr;
	}
	void reset(pointer_type p = nullptr) noexcept {mPtr = p;}
	void swap(observer_ptr& o) noexcept	{ std::swap(mPtr, o.mPtr); }

private:
	pointer_type mPtr;

}; // observer_ptr<>}
} // end namespace mobius

#endif /* OBSERVER_PTR_HPP_ */
