/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: ArrayStateVariable.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef ARRAY_STATE_VARIABLE
#define ARRAY_STATE_VARIABLE

# include <vector>
# include <iostream>

# include "../BaseStateVariableClass.h"
# include "../MobiusEnums.h"
# include "../Utility.hpp"
# include "../StandardErrors.h"

/** This is the base class for all array state variable objects
 *
 *  The class is separated into a BaseArraySV and a derived Template
 *  class. All non-template methods are put in the baseclass, which
 *  can be compiled into the base library. All template type dependent
 *  methods should be put in the derived Template class, which must be
 *  defined in the header file.
 */

class BaseArrayStateVariable: public BaseStateVariableClass {
protected:
	/**  BaseConstructor
	 */
	BaseArrayStateVariable();

	/** Overloaded constructor
	 *
	 *  @param TheName The name of state variable
	 */
	BaseArrayStateVariable(char* TheName, char* prefix);

	/**
	 *	Constructor initializing the array to be used
	 *
	 *  @param name The name of state variable
	 *  @param elementType The type of the elements
	 *  @param size the size of this array
	 */
	BaseArrayStateVariable(char* name, char* prefix, ArrayType elementType, int size);

	virtual ~BaseArrayStateVariable();

public:
	/** This method builds a list of pointers pointing to shared memory location.
	 *
	 * @param TheOtherStateVariable TheOtherVariable The other state
	 * variable which which will share this state variables state
	 */
	virtual void share(BaseStateVariableClass* other);

	virtual void ShareWith(BaseArrayStateVariable* other);

	/** This method fulfills the pure virtual method defined on
	 * BaseStateVariableClass.  It returns the size of the state
	 * variable in bytes
	 *
	 * @return The size of the state variable in bytes
	 */
	int StateSize();

	/** This methods set the state of the state variable to the value at
	 * the specificed memory location.
	 *
	 * @param TheSpecifiedMemoryLocation
	 */
	void SetState(void* TheSpecifiedMemoryLocation);

	/** This method stores the value of the current state in the
	 * specified memory location
	 *
	 * @param StateStorageLocation The location in memory to which the
	 * current state should be saved.
	 */
	void CurrentState(void* StateStorageLocation);

	/** This method prints the state of the state variable to cout
	 */
	void printState(ostream *strm);

	/**  This method returns the type this array is made of
	 *
	 *  @return The type of the array
	 */
	ArrayType getType();

	/**  This method returns the size of this array
	 *
	 *  @return The size of the array
	 */
	int getSize();

	/**
	 * This method appends the specified action to the state variable's
	 * AffectingActions object.   The state variable will
	 * use the append method defined on the AffectingActions class to add
	 * the action to the AffectingAction object.
	 */
	void appendAffectingAction(BaseActionClass* TheAction);

	/**
	 * This method appends the specified action to the state variable's
	 * EnabledActions object.   The state variable will
	 * use the append method defined on the EnabledActions class to add
	 * the action to the EnabledAction object.
	 */
	void appendEnabledAction(BaseActionClass* TheAction);

	/**
	 * This function gets the list of base state variable class leaf nodes
	 * for all the elements present in the StateVariable. For simple places,
	 * this would return only the list with single entry of itself, else it
	 * would return the leaf nodes
	 * CARE MUST BE TAKEN TO FREE THE MEMORY ALLOCATED TO THE VARIABLES AND
	 * DELETE THEM.
	 */

	List<BaseStateVariableClass> * getLeafNodes();
	virtual int getNumLeafNodes();
	virtual bool isLeafNode();

	/**
	 *  these oddly-named method allows actions to depend on all
	 *  leaf nodes of array and structure SV, without having to expose
	 *  all leaf node pointers directly to the action.  Called by
	 *  BaseActionClass addEnablingSV() and addAffectedSV()
	 */
	virtual void addSelfAsEnablingSVToAction(BaseActionClass *action);
	virtual void addSelfAsAffectedSVToAction(BaseActionClass *action);
	virtual void removeSelfFromEnablingSVOfAction(BaseActionClass *action);
	virtual void removeSelfFromAffectedSVOfAction(BaseActionClass *action);

protected:
	typedef std::vector<BaseStateVariableClass*> field_list;
	typedef field_list::const_iterator fieldIterator;
	ArrayType type; //< type of the array
	field_list fields;
};

template<class Type> class ArrayStateVariable: public BaseArrayStateVariable {
public:
	/**  BaseConstructor
	 */
	ArrayStateVariable() :
		BaseArrayStateVariable() {
	}

	/** Overloaded constructor
	 *
	 *  @param TheName The name of state variable
	 */
	ArrayStateVariable(char* TheName, char* ThePrefix) :
		BaseArrayStateVariable(TheName, ThePrefix) {
	}

	/**
	 *	Constructor initializing the array to be used
	 *
	 *  @param name The name of state variable
	 *  @param elementType The type of the elements
	 *  @param size the size of this array
	 */
	ArrayStateVariable(char* name, char* ThePrefix, ArrayType elementType, int size) :
		BaseArrayStateVariable(name, ThePrefix, elementType, size) {
	}

	virtual ~ArrayStateVariable() {
	}

	/**
	 *  access operator
	 */
	inline Type* operator[](int index) {
		return Index(index);
	}

	/**
	 *  access operator
	 */
	inline Type* Index(int index) {
		if ((index < fields.size()) && (index >= 0))
			return dynamic_cast<Type*> (fields[index]);
		else {
			//run time error: index out of bounds
			cerr << "Runtime error: index out of bounds. ";
			cerr << StateVariableName << "[0.." << (fields.size() - 1);
			cerr << "] indexed with " << index << "\n";
			assert(0);
		}
	}
	/**
	 *  access operator
	 */
	inline Type* index(int index) {
		return Index(index);
	}

	// The Model-specific way to get the contents of this variable
	// In a San: get the marking of a place
	inline Type* getMark() {
		Type* temp;
		CurrentState(static_cast<void*> (&temp));
		return temp;
	}

	// The model-specific way to set the contents of this variable
	// In a SAN: set the marking of a place
	inline void setMark(Type* new_value) {
		SetState(static_cast<void*> (&new_value));
	}

};

#endif // end ARRAY_STATE_VARIABLE
