/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: SharableSV.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef __SHARABLE_SV
#define __SHARABLE_SV

#include <iostream>
#include <cassert>

#include "../BaseStateVariableClass.h"
#include "SharedPointer.h"
#include "../Utility.hpp"
#include "../StandardErrors.h"
#include "../timestuff.h"

/**
 *   This is the base class for all short state variable objects
 *
 *    @author Jay M. Doyle
 *    @version BaseClasses version 2.0
 *
 */
template<class C> class SharableSV: public BaseStateVariableClass {
public:
	/**
	 * BaseConstructor
	 */
	SharableSV() :
		BaseStateVariableClass(), TheState(nullptr), Head(nullptr), Leader(this) {
		if ((TheState = new C) == nullptr)
			MemoryAllocationError();

		*TheState = 0;

		registerMyself();
	}

	/**
	 * New Constructor that takes the name
	 */
	explicit SharableSV(char* TheName) :
		BaseStateVariableClass(TheName, ""), TheState(0), Head(0), Leader(this) {
		if ((TheState = new C) == nullptr)
			MemoryAllocationError();
		*TheState = 0;

		registerMyself();
	}

	/**
	 *  Destructor
	 */
	virtual ~SharableSV() {
		SharedPointer<C>* NextNeighbor;
		SharedPointer<C>* LinkedListElement;

		// If Stored == false, it implies
		// that this state variable was pointing
		// to somebodyelse state, therfore we do
		// not need to deallocate it.
		//
		if (Stored) {
			delete TheState;
		}

		if (Head != nullptr) {
			// Remove all the elements in the linked list
			LinkedListElement = Head;
			while (LinkedListElement != nullptr) {
				NextNeighbor = LinkedListElement->getNeighbor();
				delete LinkedListElement;
				LinkedListElement = NextNeighbor;
			}
		}
	}

	/**
	 * Implements equivalence sharing with another state variable
	 * of type SharableSV
	 */
	void ShareWith(SharableSV<C>* TheOtherStateVariable) {

		//cout << "SharableSV.h Calling ShareWith now\n";
		if (TheOtherStateVariable->getStored()) {
			if (getStored()) {
				setShared(true);
				setState(TheOtherStateVariable->getState()); // Copy initial value
				Register(TheOtherStateVariable);
			} else {
				Leader->Register(TheOtherStateVariable);
			}
		} else {
			if (getStored()) {
				TheOtherStateVariable->getLeader()->Register(this);
			} else {
				if (SharingCount > TheOtherStateVariable->getLeader()->getSharingCount())
					Register(TheOtherStateVariable->getLeader());
				else
					TheOtherStateVariable->getLeader()->Register(Leader);
			}
		}

		// Now ensure that all the state variables
		// have the same, correct AffectingActions
		// data structures.
		//     buildAffectsList();

		// Note:  Only a single Enabled and Affecting actions
		// list is necessary for the equivalence so all
		// members should point to the leader's data structures
		// which will be keep consistant by always adding
		// new members affecting and enabled actions to the
		// leader's list

		// We can also take this opportunity to update
		// the data structures of the enabled and affecting
		// actions.  Each has two data structures: EnablingStateVariables
		// and AffectedStateVariable which consist of a list
		// of pointers to state variables that either affect
		// or enable the action.  We would like this pointers to
		// point to the lead state variable in the case where
		// one of the state variable is shared.  This way
		// we can easily determine if one action's firing affects
		// another action's enabling condition by changing
		// the value of a shared state variable.  This update
		// procedure used to be done by the models in updateAffectsList
		// method.  However this implementation was unnecessarly
		// inefficient since if check each action in the model.

		int numActions = TheOtherStateVariable->getNumAffectingActions();
		BaseActionClass* tempAction;

		for (int i = 0; i < numActions; ++i) {
			tempAction = TheOtherStateVariable->getAffectingAction(i);
			tempAction->replaceAffectedSV(TheOtherStateVariable, Leader);
		}

		numActions = TheOtherStateVariable->getNumEnabledActions();
		for (int i = 0; i < numActions; ++i) {
			tempAction = TheOtherStateVariable->getEnabledAction(i);
			tempAction->replaceEnablingSV(TheOtherStateVariable, Leader);
		}
	}
	;

	/**
	 *  Builds a list of pointer pointing to shared memory location
	 */
	void Register(SharableSV<C>* TheOtherPointer) {
		SharedPointer<C>* NextElement;
		SharedPointer<C>* LastElement;

		delete TheOtherPointer->TheState;

		TheOtherPointer->setLeader(this); // make this other's leader
		TheOtherPointer->setPointer(TheState); // change all pointers pointing to
		// the TheOtherPointer

		TheOtherPointer->setShared(true);
		TheOtherPointer->setStored(false);

		// All members of an equivalence group
		// should share the same AffectingActions
		// and EnabledActions data structures.
		// The actions in these data structures
		// should be the union of all individual
		// members' data strtures.

		// Add the new equivalence group's
		// affecting actions and enabled actions
		// the master list
		SVAffectingActions->append(TheOtherPointer->SVAffectingActions);
		SVEnabledActions->append(TheOtherPointer->SVEnabledActions);

		// delete old data structure
		delete TheOtherPointer->SVAffectingActions;
		delete TheOtherPointer->SVEnabledActions;

		TheOtherPointer->SVAffectingActions = nullptr;
		TheOtherPointer->SVEnabledActions = nullptr;

		assert(SVAffectingActions);
		assert(SVEnabledActions);

		// Set new members data structures to point
		// to equivalence group's data structures.
		TheOtherPointer->SVAffectingActions = SVAffectingActions;
		TheOtherPointer->SVEnabledActions = SVEnabledActions;

		// Now set all members of TheOtherPointer's equivalence
		// group to point to the master list of affecting
		// and enabled actions by going through TheOtherPointer's
		// linked list of equivalence group members.
		assert(TheOtherPointer->getHead());
		NextElement = TheOtherPointer->getHead();
		LastElement = nullptr;
		while (NextElement != nullptr) {
			// Ignore elements in the linked list
			// that corespond to short cut pointers
			// and not state variables in the equivalence
			// group.  Elements in the linked list associated
			// with short cut pointer have Owner == nullptr.
			if (NextElement->getOwner() != nullptr) {
				NextElement->getOwner()->setAffectingActions(SVAffectingActions);
				NextElement->getOwner()->setEnabledActions(SVEnabledActions);
			}
			// Get next element of linked list
			LastElement = NextElement;
			NextElement = NextElement->getNeighbor();
		}

		// TAC 6/21/2006: add TheOtherPointer's list ot the front of
		// Head's list, to avoid scanning through Head to find the end
		if (LastElement != nullptr) {
			// set last element in other list to point to the first element
			// in this list
			LastElement->setNeighbor(Head);
			// set the head of this list to point to the first element in
			// the other list
			Head = TheOtherPointer->getHead();
			// Increment the total number of sharers
			SharingCount += TheOtherPointer->getSharingCount();
		}

		// Now that the other state variable is
		// shared, it no longer needs it pointer
		// to the linked list.
		TheOtherPointer->setHeadToNull();

#ifdef DEBUG
		if (GlobalDebugLevel >= 4) {
			std::cout << "Now " << TheOtherPointer->StateVariableName
					  << " has a head pointer = " << TheOtherPointer->getHead()
					  << std::endl;

			std::cout << StateVariableName << "  has " << SharingCount
					  << " elements." << std::endl;

			int TempCount = 1;
			NextElement = Head;
			while (NextElement->getNeighbor() != nullptr) {
				NextElement = NextElement->getNeighbor();
				TempCount++;
			}

			std::cout << StateVariableName << " really has " << TempCount << std::endl;

			printf("%s Linked List Contains :\n", StateVariableName);
			while (NextElement->getNeighbor() != nullptr) {
				printf("%p\n", NextElement->getPointer());
				NextElement = NextElement->getNeighbor();
			}
			printf("\n");
		}
#endif

	}
	;

	/**
	 *	same as Register
	 */
	virtual void share(BaseStateVariableClass* other) {
		SharableSV<C>* otherSV = dynamic_cast<SharableSV<C>*> (other);
		//printf("SharableSV.h Calling This Function\n");
		assert(otherSV);
		ShareWith(otherSV);
		//Register(otherSV);
	}

	/**
	 * Register pointer to state
	 */
	void Register(C** TheOtherPointer) {
		SharedPointer<C> * NextElement;

		// TheState should be created in the
		// constructor
		ASSERT(TheState != nullptr);

		// Set the C pointer to
		// point at the shared state
		*TheOtherPointer = TheState;

		// Increase Sharing count to reflect
		// that there is another member of
		// the equivalence group
		SharingCount++;

		// Create a new element to add the linked list
		if ((NextElement = new SharedPointer<C> ()) == nullptr) {
			MemoryAllocationError();
		}

		// add memeber to our list of sharing state variables
		NextElement->setNeighbor(Head);
		NextElement->setPointer(TheOtherPointer);
		NextElement->setOwner(nullptr);
		Head = NextElement;
	}

	/**
	 *  Fulfills the pure virtual method defined on BaseStateVariableClass
	 */
	int StateSize() {
		return sizeof(C);
	}
	;

	/**
	 * Sets value of TheState
	 */
	void setPointer(C* TheNewLocation) {
		SharedPointer<C>* TheElement;

		TheState = TheNewLocation;

		// Change the value for all C variables that also
		// share this state variable
		TheElement = Head;
		while (TheElement->getNeighbor() != nullptr) {

			TheElement->changePointer(TheNewLocation);
			TheElement = TheElement->getNeighbor();
		}

		TheElement->changePointer(TheNewLocation);
	}
	;

	/**
	 * Fullfills requirement for BaseStateVariable
	 */
	void SetState(void* NewState) {
		readMemory2<C> (NewState, TheState);
	}
	;

	/**
	 * Fullfills requirement for BaseStateVariable
	 */
	void CurrentState(void* MemoryLocation) {
		writeMemory<C> (*TheState, MemoryLocation);
	}
	;

	/**
	 * Fullfills requirement for BaseStateVariable
	 */
	void printState(ostream *strm) {
		(*strm) << StateVariableName << " = " << *TheState << std::endl;
	}
	;

	/**
	 * This returns TheState(for testing)
	 */
	void printStateVal(ostream *strm) {
		(*strm) << *TheState;
	}
	/**
	 * This returns TheState(for testing)
	 */
	C* getPointer() {
		return TheState;
	}
	;

	/**
	 * Sets the Leader pointer
	 */
	void setLeader(SharableSV<C>* TheLeader) {
		Leader = TheLeader;
	}
	;

	/**
	 * return Leader pointer;
	 */
	SharableSV<C>* getLeader() {
		return Leader;
	}
	;

	/**
	 * This method returns head of linked list of
	 * shared pointers. If the sharable state
	 * variable is not the leader it returns the
	 * leader's Head since this is the head
	 * of the equivalence's group linked list.
	 */
	SharedPointer<C>* getHead() {
		return Head;
	}
	;

	/**
	 * Return the value of the state
	 */
	C getState() {
		return *TheState;
	}
	;

	/**
	 *  This method returns the value
	 *  of the state by reference
	 */
	inline C& State() {
		return *TheState;
	}

	/**
	 * Set the value of TheState
	 */
	void setState(C TheValue) {
		*TheState = TheValue;
	}

	/**
	 *  This method overloads the method defined
	 *  in BaseStateVariableClass
	 */
	void appendAffectingAction(BaseActionClass* TheAction) {
		SharedPointer<C>* NextElement;

		// Since the linked list is shared by all members
		// of the equivalence group only the leader performs
		// the appending

		NextElement = Head;
		while ((NextElement != nullptr) && (this == getLeader())) {
			// some elements of the linked list represent C**
			// that are not associated with any particular state variable.
			// Such pointer have a representative on the linked list with
			// owner == NULL.  See method register(C**).
			if (NextElement->getOwner() != nullptr) {
				NextElement->getOwner()-> BaseStateVariableClass::appendAffectingAction(TheAction);
				DS3("Appended affecting action to",
						NextElement->getOwner()->StateVariableName);
			}
			NextElement = NextElement->getNeighbor();
		}
	}

protected:

	/**
	 *  Pointer to state
	 */
	C* TheState;

	/**
	 * Head of linked list
	 */
	SharedPointer<C> * Head;

	/**
	 * This is the "leader" of a group of shared
	 * state variables.  The leader is the one
	 * that actually stores the state and also
	 * maintains the list of all shared state variable
	 *  pointers
	 */
	SharableSV<C>* Leader;

	/**
	 *  This method is used to build a common
	 *  AffectingActions object for all the state
	 *  variables that share the same state.
	 */
	void buildAffectsList() {
		List<BaseActionClass> MasterList;
		SharedPointer<C>* NextNeighbor;

		// First create a master list of affecting
		// actions by polling each state variable
		// in the equivalence sharing group.

		NextNeighbor = getLeader()->getHead();
		while (NextNeighbor != nullptr) {
			// some elements of the linked list represent C**
			// that are not associated with any particular state variable.
			// Such pointer have a representative on the linked list with
			// owner == NULL.  See method register(C**).
			if (NextNeighbor->getOwner() != nullptr) {
				MasterList.append(NextNeighbor->getOwner()->getAffectingActions());
			}
			NextNeighbor = NextNeighbor->getNeighbor();
		}

		// Now tell each state variable in the
		// equivalence sharing group to use the
		// master list a its list of affecting actions
		NextNeighbor = getLeader()->getHead();
		while (NextNeighbor != nullptr) {
			// some elements of the linked list represent C**
			// that are not associated with any particular state variable.
			// Such pointer have a representative on the linked list with
			// owner == NULL.  See method register(C**).
			if (NextNeighbor->getOwner() != nullptr) {
				NextNeighbor->getOwner()->copyAffectingActions(&MasterList);
			}
			NextNeighbor = NextNeighbor->getNeighbor();
		}
	}

	/**
	 * This method sets the Head data member to
	 * NULL.  This is necessary as part of the
	 * registration process done when two state variables
	 * are shared.
	 */
	void setHeadToNull() {
		Head = nullptr;
	}

	/**
	 *  This method creates the initial entry into the
	 *  linked list of shares.
	 */
	void registerMyself() {
		SharedPointer<C>* NextElement;

		SharingCount++;
		// Create a new element to add the linked list
		if ((NextElement = new SharedPointer<C> ()) == nullptr) {
			MemoryAllocationError();
		}

		// add memeber to our list of sharing state variables
		NextElement->setNeighbor(Head);
		NextElement->setPointer(&TheState);
		NextElement->setOwner(this);
		assert(NextElement->getOwner());
		Head = NextElement;
	}
	;

private:

	/**
	 *  For debugging
	 */
	void printLinkedList() {
		SharedPointer<C>* NextNeighbor;

		NextNeighbor = Head;

		if (getLeader()->getHead()->getOwner() != nullptr) {
			std::cout << "** Sharer Name: ";
			std::cout << getLeader()->getHead()->getOwner()->StateVariableName << std::endl;
		}
		std::cout << "** Head = " << Head << std::endl;
		std::cout << "** ";
		while (NextNeighbor != nullptr) {
			cout << NextNeighbor << "-->";
			NextNeighbor = NextNeighbor->getNeighbor();
		}
		std::cout << "NULL" << std::endl;
	}
	;

};

#endif // end __SHARABLE_SV
