/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: StructStateVariable.h
 *
 *      Authors: Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef STRUCT_STATE_VARIABLE
#define STRUCT_STATE_VARIABLE

#include "../BaseStateVariableClass.h"
#include "../state/SharableSV.h"
#include "../MobiusEnums.h"
#include "../MobiusTypeDefs.h"
#include "../Utility.hpp"
#include "../StandardErrors.h"
#include "../debug.h"

#include <iostream>
#include <vector>

/**
 * This is the base class for all structured
 * state variable objects.
 *
 *  The class is separated into a BaseStructSV and a derived Template
 *  class. All non-template methods are put in the baseclass, which
 *  can be compiled into the base library. All template type dependent
 *  methods should be put in the derived Template class, which must be
 *  defined in the header file.
 */

class BaseStructStateVariable: public BaseStateVariableClass {
public:
	/**
	 * BaseConstructor
	 */
	BaseStructStateVariable();

	/**
	 *  Initialize the state variable with
	 *  its name.
	 *
	 *  @param TheName The name of the structured
	 *    state variable.
	 */
	BaseStructStateVariable(char* TheName, char* prefix);

	/**
	 *  Overloaded constructor
	 *
	 *  @param TheName The name of state variable
	 *  @param TheTypeArray An array of state variable type
	 *         include in the unordered list
	 *  @param TheSize The size of the structlist
	 */
	BaseStructStateVariable(char* TheName, char* prefix, const ArrayType* TheTypeArray, int TheSize);

	/**
	 *  Class destructor
	 */
	virtual ~BaseStructStateVariable();

	/**
	 * This method builds a list of pointers pointing to
	 * shared memory location.
	 *
	 * @param TheOtherStateVariable   TheOtherVariable The
	 *        other state variable which will share this
	 *        state variables state
	 */
	virtual void share(BaseStateVariableClass* other);

	virtual void ShareWith(BaseStructStateVariable* other);
	/**
	 * This method fulfills the pure virtual method defined
	 * on BaseStateVariableClass.  It returns the size of
	 * the state variable in bytes
	 *
	 * @return The size of the state variable in bytes
	 */
	int StateSize();

	/**
	 * This methods set the state of the state variable
	 * to the value at the specified memory location.
	 *
	 * @param TheSpecifiedMemoryLocation
	 */
	void SetState(void* TheSpecifiedMemoryLocation);

	/**
	 * This method stores the value of the current state
	 * in the specified memory location
	 *
	 * @param StateStorageLocation The location in memory
	 *        to which the current state should be saved.
	 */
	void CurrentState(void* StateStorageLocation);

	/**
	 *  This method prints the state of the state variable to cout
	 */
	void printState(std::ostream* strm);

	/**
	 * This method returns the number of items in the struct
	 *
	 * @return The number of items in the struct
	 */
	int getSize();

	/**
	 *  This method returns type of element at TheIndex
	 *
	 *  @param TheIndex The index into the ArrayType array
	 */
	ArrayType getType(int TheIndex);

	/**
	 * This method appends the specified action to the state variable's
	 * AffectingActions object.   The state variable will
	 * use the append method defined on the AffectingActions class to add
	 * the action to the AffectingAction object.
	 */
	void appendAffectingAction(BaseActionClass* TheAction);

	/**
	 * This method appends the specified action to the state variable's
	 * EnabledActions object.   The state variable will
	 * use the append method defined on the EnabledActions class to add
	 * the action to the EnabledAction object.
	 */
	void appendEnabledAction(BaseActionClass* TheAction);

	/**
	 *  these oddly-named method allows actions to depend on all
	 *  leaf nodes of array and structure SV, without having to expose
	 *  all leaf node pointers directly to the action.  Called by
	 *  BaseActionClass addEnablingSV() and addAffectedSV()
	 */
	virtual void addSelfAsEnablingSVToAction(BaseActionClass *action);
	virtual void addSelfAsAffectedSVToAction(BaseActionClass *action);
	virtual void removeSelfFromEnablingSVOfAction(BaseActionClass *action);
	virtual void removeSelfFromAffectedSVOfAction(BaseActionClass *action);

	/**
	 * This function gets the list of base state variable class leaf nodes
	 * for all the elements present in the StateVariable. For simple places,
	 * this would return only the list with single entry of itself, else it
	 * would return the leaf nodes
	 * CARE MUST BE TAKEN TO FREE THE MEMORY ALLOCATED TO THE VARIABLES AND
	 * DELETE THEM.
	 */

	List<BaseStateVariableClass> * getLeafNodes();
	virtual int getNumLeafNodes();
	virtual bool isLeafNode();

protected:

	int calculateStateSize(void);

	int calculatedStateSize;

	// convenience types
	typedef struct fieldType {
		BaseStateVariableClass* field;
		ArrayType type;
	};
	typedef std::vector<fieldType> field_list;
	typedef field_list::const_iterator fieldIterator;
	/**
	 *  Pointer to formalism state
	 */
	field_list fields;

	/**
	 *  Constructs a fieldType
	 */
	virtual fieldType field(const ArrayType type, BaseStateVariableClass* sv) const {
		fieldType result;
		result.type = type;
		result.field = sv;
		return result;
	}

	/**
	 *  This method initializes the TheType and
	 *  StructPointer.  Both arrays will be made of
	 *  size TheSize.
	 *
	 *  @param TheNewTypes The types state variables
	 *    in structured state variable.
	 *  @param TheSize The number of state variables
	 *    making up this structured state variable.
	 */
	void makeArrays(const ArrayType* TheNewTypes, int TheSize);
};

template<class Type>
class StructStateVariable: public BaseStructStateVariable {
protected:
	/**
	 * BaseConstructor
	 */
	StructStateVariable() :
		BaseStructStateVariable() {
	}

	/**
	 *  Initialize the state variable with
	 *  its name.
	 *
	 *  @param TheName The name of the structured
	 *    state variable.
	 */
	StructStateVariable(char* TheName, char* ThePrefix) :
		BaseStructStateVariable(TheName, ThePrefix) {
	}

	/**
	 *  Overloaded constructor
	 *
	 *  @param TheName The name of state variable
	 *  @param TheTypeArray An array of state variable type
	 *         include in the unorder list
	 *  @param TheSize The size of the structlist
	 */
	StructStateVariable(char* TheName, char* ThePrefix, const ArrayType* TheTypeArray, int TheSize) :
		BaseStructStateVariable(TheName, ThePrefix, TheTypeArray, TheSize) {
	}

	/**
	 *  Class destructor
	 */
	virtual ~StructStateVariable() {
	}

public:
	// The Model-specific way to get the contents of this variable
	// In a san: get the marking of a place
	Type getMark() {
		Type temp;
		CurrentState(static_cast<void*> (&temp));
		return temp;
	}

	// The model-specific way to set the contents of this variable
	// In a SAN: set the marking of a place
	void setMark(Type &new_value_struct) {
		SetState(static_cast<void*> (&new_value_struct));
	}
};

#endif
