/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef DynamicObjects_H_
#define DynamicObjects_H_

#include <vector>
#include <cassert>
#include <cstring>
#include "BaseClasses/TraceFileMobius.h"

#if defined(__has_include)
#if __has_include(<forward_list>)
#define HAS_FORWARD_LIST 1
#endif
#elif (__cplusplus > 199711L) || (_MSC_VER >= 1800)
#define HAS_FORWARD_LIST 1
#endif

#if (HAS_FORWARD_LIST)
#include <forward_list>
#else
#include <ext/slist>
using __gnu_cxx::slist;
#endif

using namespace std;

template<class T>
class SparseMatrix {
public:
	SparseMatrix() {
		current_row = -1;
	}
	;
	~SparseMatrix() {
		elements.clear();
		cols.clear();
		row_start_idx.clear();
	}
	void AppendElement(int row, int col, T element) {
		if (row > current_row) {
			for (int i = current_row; i < row; i++)
				row_start_idx.push_back(cols.size());
			current_row = row;
		}
		//The sparse matrix should be filled in sequential order
		assert(row == current_row);
		elements.push_back(element);
		cols.push_back(col);
	}
	//This function should be called when the matrix is completely built.
	void FinishLastRow() {
		row_start_idx.push_back(cols.size());
	}
	//This methods is supposed to compress each row of the matrix so that
	//it takes the least memory possible
	void Compress() {
	}
	int getNoRows() {
		return row_start_idx.size() - 1;
	}
	int getNoElements() {
		return elements.size();
	}
	vector<T> elements;
	vector<int> cols;
	//A vector of size n+1 where n is the number of rows of the matrix
	//row_start_idx[i] shows the index in 'elements' and 'cols' arrays where
	//the ith row starts (for i < n). row_start_idx[n] is the size of 'elements'
	//and 'cols' arrays.
	vector<int> row_start_idx;
private:
	int current_row;
};

template<class T, bool has_default_value> class DynamicArray;

template<class T>
class DynamicArray<T, true> {
public:
	DynamicArray(int _capacity, T _default_value) {
		Construct(_capacity, _default_value);
	}
	DynamicArray(T _default_value) {
		Construct(5, _default_value);
	}
	DynamicArray() :
		default_value(),
		elements(nullptr),
		capacity(0),
		size(0)

	{}

	~DynamicArray() {
		delete[] elements;
	}

	T& operator[](int i) {
		if (i >= size)
			size = i + 1;
		if (i >= capacity) {
			int new_capacity = capacity;
			while (i >= new_capacity)
				new_capacity *= 2;
			extend_capacity(new_capacity);
		}
		return elements[i];
	}

	void push_back(const T& _value) {
		if (size >= capacity)
			extend_capacity(capacity * 2);
		elements[size++] = _value;
	}

	void push_back() {
		if (size >= capacity)
			extend_capacity(capacity * 2);
		size++;
	}

	bool operator==(const DynamicArray& a) {
		if (this == &a)
			return true;
		if (size == a.size) {
			for (int i = 0; i < size; i++)
				if (elements[i] != a.elements[i])
					return false;
			return true;
		}
		return false;
	}

	bool operator!=(const DynamicArray& a) {
		return !(*this == a);
	}

	void operator=(const DynamicArray& a) {
		if (capacity < a.size) {
			int new_capacity = capacity;
			do {
				new_capacity *= 2;
			} while (a.size > new_capacity);
			extend_capacity(new_capacity);
		}
		for (int i = 0; i < a.getSize(); i++)
			elements[i] = a.elements[i];
	}

	T& getElement(int i) {
		return elements[i];
	}
	void clear() {
		reset();
	}
	int getSize() const {
		return size;
	}
	void reset() {
		for (int i = 0; i < size; i++)
			elements[i] = default_value;
		size = 0;
	}

	void Construct(int _capacity) {
		capacity = _capacity;
		elements = new T[capacity];
		size = 0;
	}

	void Construct(int _capacity, T _default_value) {
		Construct(_capacity);
		default_value = _default_value;
		for (int i = 0; i < capacity; i++)
			elements[i] = default_value;
	}

private:
	void extend_capacity(int new_capacity) {
		T* elements_larger = new T[new_capacity];
		memcpy(elements_larger, elements, capacity * sizeof(T));
		for (int j = capacity; j < new_capacity; j++)
			elements_larger[j] = default_value;
		delete[] elements;
		elements = elements_larger;
		capacity = new_capacity;
	}

private:
	T default_value;
	T* elements;
	int capacity;
	int size;
};

template<class T>
class DynamicArray<T, false> {
public:
	DynamicArray(int _capacity) {
		Initialize(_capacity);
	}

	DynamicArray() {
		Initialize(5);
	}

	~DynamicArray() {
		delete[] elements;
	}

	T& operator[](int i) {
		if (i >= size)
			size = i + 1;
		if (i >= capacity) {
			int new_capacity = capacity;
			while (i >= new_capacity)
				new_capacity *= 2;
			extend_capacity(new_capacity);
		}
		return elements[i];
	}

	void push_back(const T& _value) {
		if (size >= capacity)
			extend_capacity(capacity * 2);
		elements[size++] = _value;
	}

	bool operator==(const DynamicArray& a) {
		if (this == &a)
			return true;
		if (size == a.size) {
			for (int i = 0; i < size; i++)
				if (elements[i] != a.elements[i])
					return false;
			return true;
		}
		return false;
	}

	bool operator!=(const DynamicArray& a) {
		return !(*this == a);
	}

	void operator=(const DynamicArray& a) {
		if (capacity < a.size) {
			int new_capacity = capacity;
			do {
				new_capacity *= 2;
			} while (a.size > new_capacity);
			extend_capacity(new_capacity);
		}
		for (int i = 0; i < a.size; i++)
			elements[i] = a.elements[i];
		size = a.size;
	}

	T& getElement(int i) {
		return elements[i];
	}
	void reset() {
		size = 0;
	}
	void clear() {
		reset();
	}
	int getSize() {
		return size;
	}

private:
	void Initialize(int _capacity) {
		capacity = _capacity;
		elements = new T[capacity];
		size = 0;
	}

	void extend_capacity(int new_capacity) {
		T* elements_larger = new T[new_capacity];
		memcpy(elements_larger, elements, capacity * sizeof(T));
		delete[] elements;
		elements = elements_larger;
		capacity = new_capacity;
	}

private:
	int capacity;
	int size;
public:
	T* elements;
};

template<class T>
class SparseVectorArray2 {
public:
	SparseVectorArray2() {
	}
	;
	~SparseVectorArray2() {
	}
	typedef long long int INDEX_TYPE;
	void Transform(SparseVectorArray2& vec, const T& scale, INDEX_TYPE disp) {
		const int size = vec.getSize();
		for (int i = 0; i < size; i++)
			InsertElement(vec.getIndex(i) + disp, vec.getElement(i) * scale);
	}
	;

	void InsertElement(INDEX_TYPE i, const T& value) {
		assert ((getSize() == 0) || i> getIndex(getSize()-1));
		elements.push_back(TElement(i, value));
	}

	int getSize() {
		return elements.getSize();
	}
	;
	T& getElement(int i) {
		return elements.getElement(i).elem;
	}
	INDEX_TYPE getIndex(int i) {
		return elements.getElement(i).index;
	}
	void setElement(int i, T& elem) {
		elements.getElement(i).elem = elem;
	}
	void setIndex(int i, INDEX_TYPE index) {
		elements.getElement(i).index = index;
	}
	void Zero() {
		elements.reset();
	}
	void operator=(const SparseVectorArray2<T>& vec) {
		elements = vec.elements;
	}

	bool empty() {
		return getSize() == 0;
	}
	void print() {
		for (int i = 0; i < getSize(); i++) {
			Traceprintf(2, "[%3d] = %f\n", getIndex(i), getElement(i));
		}
	}

public:
	struct TElement {
		TElement(const INDEX_TYPE _index, const T& _elem) {
			index = _index;
			elem = _elem;
		}
		TElement() :
			index(),
			elem()
		{}
		INDEX_TYPE index;
		T elem;
	};
	DynamicArray<TElement, false> elements;
	//DynamicArray2<TElement> elements;
};

template<class T>
class SparseVector2 {
public:
	SparseVector2() {
	}

	~SparseVector2() {
	}
	typedef long long int INDEX_TYPE;

	/* This version of Transform does NOT keep track of the last element
	 of the linked list, that is, end pointer
	 */
	void Transform(SparseVector2<T> & vec, const T& scale, INDEX_TYPE disp) {
		if (vec.elements.empty())
			return;
		iterator_type it1, it2;
		if (elements.empty()) {
			it2 = vec.elements.begin();
			it1 = elements.insert(elements.begin(), make_pair(it2->first + disp, it2->second * scale));
			for (it2++; it2 != vec.elements.end(); it2++) {
				it1 = elements.insert_after(it1, make_pair(it2->first + disp, it2->second * scale));
			}
			return;
		}
		it1 = elements.begin();
		it2 = vec.elements.begin();
		iterator_type prev_it1 = it1;
		//Writes to the dest elements of source that are less than the first
		//element of the dest.
		bool first_element = true;
		while (it2 != vec.elements.end() && it2->first + disp < it1->first) {
			assert(false);
			if (first_element) {
				prev_it1 = elements.insert(prev_it1, make_pair(it2->first + disp, it2->second * scale));
				first_element = false;
			} else
				prev_it1 = elements.insert_after(prev_it1, make_pair(it2->first + disp, it2->second * scale));
			it2++;
		}

		prev_it1 = it1++;
		bool source_finish = false;
		while (it2 != vec.elements.end()) {
			if (it1 == elements.end())
				source_finish = true;
			if (prev_it1->first == it2->first + disp) {
				prev_it1->second += it2->second * scale;
				it2++;
			} else if (prev_it1->first < (it2->first + disp) && (source_finish || (it2->first + disp) < it1->first)) {
				prev_it1 = elements.insert_after(prev_it1, make_pair(it2->first + disp, it2->second * scale));
				it2++;
			} else {
				assert(it1 != vec.elements.end());
				prev_it1 = it1++;
			}
		}
	}
	;

	void InsertElement(INDEX_TYPE i, const T& value) {
		assert (elements.empty());
		elements.push_front(make_pair(i, value));
	}

	int getSize() {
		return elements.size();
	}
	;
	void Zero() {
		elements.clear();
	}
	bool empty() {
		return elements.empty();
	}
	void operator=(SparseVector2<T>& vec) {
		elements = vec.elements;
	}

	bool operator==(SparseVector2<T>& vec) {
		return elements == vec.elements;
	}

	void print() {
		for (iterator_type it = elements.begin(); it != elements.end(); it++) {
			Traceprintf(2, "[%3lld] = %f\n", it->first, it->second);
		}
		//printf("[%3d] = %f\n", end->first, end->second);
	}

#if (HAS_FORWARD_LIST)
	typedef forward_list<pair<INDEX_TYPE, T> > Type;
#else
	typedef slist<pair<INDEX_TYPE, T> > Type;
#endif
    typedef	typename Type::iterator iterator_type;
	Type elements;
};

template<class T>
class SparseVectorHash {
public:
	typedef long long int INDEX_TYPE;
	SparseVectorHash() {
		InitializeTable(5);
	}
	SparseVectorHash(short power) {
		InitializeTable(power);
	}
	void InitializeTable(short power) {
		mask = (1 << power) - 1;
		int s = mask + 1;
		t = new HashElement[s];
		memset(t, 0, sizeof(HashElement) * (mask + 1));
		loops = 0;
	}
	~SparseVectorHash() {
		delete[] t;
	}

	void Transform(SparseVectorHash<T>& vec, const T& scale, INDEX_TYPE disp) {
		for (int i = 0; i < vec.size(); i++) {
			if (vec.Occupied(i))
				InsertElement(vec.Index(i) + disp, vec.Value(i) * scale);
		}
	}

	void InsertElement(INDEX_TYPE _index, T _value) {
		short i = _index & mask, j = 0;
		while ((t[i].status == occupied) && (t[i].index != _index)) {
			assert( i < size());
			j = (j + 1) & mask;
			i = (i + j) & mask;
			assert( i < size());
			if (j == 0) {
				Traceprintf(2, "Sparse vector too small. Increase size.\n");
				exit(1);
			}
			loops++;
		}
		if (t[i].status != occupied) {
			t[i].status = occupied;
			t[i].index = _index;
			t[i].value = _value;
		} else {
			t[i].value += _value;
		}
	}
	short size() {
		return mask + 1;
	}
	bool Occupied(int i) {
		return t[i].status == occupied;
	}
	T Value(int i) {
		return t[i].value;
	}
	INDEX_TYPE Index(int i) {
		return t[i].index;
	}
	void Zero() {
		for (int i = 0; i <= mask; i++)
			t[i].status = empty;
	}

public:
	typedef enum {
		empty, occupied
	} ElementStatus;
private:
	short mask;
	typedef struct {
		INDEX_TYPE index;
		ElementStatus status;
		T value;
	} HashElement;
public:
	int loops;
	HashElement* t;
};

template<class T>
class DynamicFullMatrix {
public:
	DynamicFullMatrix(int _row_capacity, int _col_capacity, T _default_value) {
		Initialize(_row_capacity, _col_capacity, _default_value);
	}
	DynamicFullMatrix(T _default_value) {
		Initialize(5, 5, _default_value);
	}
	DynamicFullMatrix() {
		Initialize(5, 5, (T) 0);
	}
	void Initialize(int _row_capacity, int _col_capacity, T _default_value) {
		default_value = _default_value;
		row_capacity = _row_capacity;
		col_capacity = _col_capacity;
		elements = new T[row_capacity * col_capacity];
		for (int i = 0; i < row_capacity * col_capacity; i++)
			elements[i] = default_value;
		//row_size = 0;
		//col_size = 0;
	}
	~DynamicFullMatrix() {
		delete[] elements;
	}

	void ResizeNumColumns(int min_col_capacity) {
		Resize(row_capacity, min_col_capacity);
	}

	void ResizeNumRows(int min_row_capacity) {
		Resize(min_row_capacity, col_capacity);
	}

	void Resize(int min_row_capacity, int min_col_capacity) {
		int new_row_capacity = row_capacity, new_col_capacity = col_capacity;
		bool grown = false;
		while (min_row_capacity >= new_row_capacity) {
			new_row_capacity *= 2;
			grown = true;
		}
		while (min_col_capacity >= new_col_capacity) {
			new_col_capacity *= 2;
			grown = true;
		}
		if (!grown)
			return;

		T* elements_larger = new T[new_row_capacity * new_col_capacity];
		for (int i = 0; i < row_capacity; i++) {
			memcpy(elements_larger + i * new_col_capacity, elements + i * col_capacity, col_capacity * sizeof(T));
			T* p = elements_larger + i * new_col_capacity + col_capacity;
			for (int j = col_capacity; j < new_col_capacity; j++, p++)
				*p = default_value;
		}

		for (int i = row_capacity; i < new_row_capacity; i++) {
			T* p = elements_larger + i * new_col_capacity;
			for (int j = 0; j < new_col_capacity; j++, p++)
				*p = default_value;
		}

		delete[] elements;
		elements = elements_larger;
		row_capacity = new_row_capacity;
		col_capacity = new_col_capacity;
	}

	T& access(int i, int j) {
		return elements[i * col_capacity + j];
	}

	T& operator()(int i, int j) {
		//we don't need to keep the actual size of the matrix. This helps us
		//speed up this method
		/*
		 if (i >= row_size)
		 row_size = i+1;
		 if (j >= col_size)
		 col_size = j+1;
		 */
		Resize(i, j);
		return elements[i * col_capacity + j];
	}

	void reset() {
		for (int i = 0; i < row_capacity; i++)
			for (int j = 0; j < col_capacity; j++)
				elements[i * col_capacity + j] = default_value;
		//row_size = 0;
		//col_size = 0;
	}

	//int getRowSize() { return col_size;}
	//int getColSize() { return row_size;}
private:
	T default_value;
	T* elements;
	int row_capacity;
	//int row_size;
	int col_capacity;
	//int col_size;

};

#endif
