/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef __MDD_H_
#define __MDD_H_

#include "mddnode.h"
#include "cache.h"
#include <vector>
#include <stdio.h>
#include <map>
//#include "model.h"

typedef enum {
	ParentN, Parent1, Nothing
} OptType;

class CaseParam {
public:
	CaseParam(unsigned short _size) {
		size = _size;
		//assert(size >= 3); // There should be at least three parameters
		parameter = new MDDNode*[size];
		hash_value = UINT_MAX;
	}
	;
	CaseParam();
	CaseParam(MDDNode* n1, MDDNode* n2, MDDNode* n3);
	CaseParam(const CaseParam& Param);
	~CaseParam();
	//This operator is used when using CaseParam with ComputedCache
	CaseParam const& operator=(const CaseParam& Param);
	MDDNode*& operator[](unsigned short i) {
		return parameter[i];
	}
	int operator==(CaseParam& Param);
	size_t operator()(CaseParam& Param);
	void print();
	unsigned short getSize() {
		return size;
	}
	void reset();
	void incRef() {
		for (short i = 0; i < size; i++)
			parameter[i]->incRef();
	}
	;
	void decRef() {
		for (short i = 0; i < size; i++)
			parameter[i]->decRef();
	}
	;
	void resetHash() {
		hash_value = UINT_MAX;
	}
private:
	unsigned short size;
	MDDNode** parameter;
	unsigned int hash_value;
};

template<class TNode>
class MDD {
public:
	MDD(int _UniqueTableSize, int _CaseCacheSize, TNode* _DefaultValue);
	~MDD();
	MDDNode* Case(CaseParam& Params) {
		OptType Opt;
		MDDNode* R = Case(Params, Opt);
		switch (Opt) {
		case Parent1:
			R->incRefOneLevel();
		case Nothing:
			R->decRef();
			break;
		}
		return R;
	}
	MDDNode* Case(CaseParam& Params, OptType& Opt);
	void addNode(MDDNode* node);
	void deleteNode(MDDNode* node) {
		node->decRef();
	}
	MDDNode* insertNode(MDDNode* node);
	MDDNode* insertNodeUnique(MDDNode* node);
	void removeNode(MDDNode* node);
	pair<MDDNode*, bool> findOrAddNode(MDDNode* node);
	typedef std::map<int, TNode*> mapping;
	TNode* readTree(FILE *f);
	TNode* dfsReadTree(FILE* f, mapping& nodetable);
	void print();
	void GC();
	void EnableGC() {
		GCEnabled = true;
	}
	void DisableGC() {
		GCEnabled = false;
	}
	void report();
	int getNoOfNodes() {
		return UniqueTable.getNoOfNodes();
	}
	void resize(int _UniqueTableSize, int _CaseCacheSize) {
		UniqueTable.resize(_UniqueTableSize);
		CaseCache.resize(_CaseCacheSize);
	}
	void registerCache(ComputedCacheBase* Cache) {
		UniqueTable.registerCache(Cache);
	}

	NodeUniqueTable<MDDNode>::HashTableType::iterator begin() {
		return UniqueTable.begin();
	}
	NodeUniqueTable<MDDNode>::HashTableType::iterator end() {
		return UniqueTable.end();
	}
private:

	class CCaseCache: public ComputedCache<CaseParam, MDDNode> {
		void GC();
	} CaseCache;
	NodeUniqueTable<MDDNode> UniqueTable;
	//These three counters count the number of times Case operation is computed
	//using the base cases
	int BaseCase1; //When the node is a leaf
	int BaseCase2; //When all arguments are the same
	int BaseCase3; //When the ith argument equals

	//The maximum allowable load factor(number of elements divided by
	//the number of buckets) for the unique table. When this load factor
	//is exceeded after adding a new node, the hashtable size is
	//increased by 1. (The elements in the hashtable will not
	//redistributed)
	int MAXLOADFACTOR;

	//The maximum allowable ratio of number of dead nodes and number of
	//all nodes ( < 1). Garbage collection is ignited when this ratio is
	//violated.
	float MAXDEAD;
protected:
	TNode* DefaultValue;
	bool GCEnabled;
};

/******* inline definition of CaseParam's functions *******/
//This constructor is required for initializing the vector defined
//in CaseComputedCache because its elements are pair<CaseParam, MDDNode*>
inline CaseParam::CaseParam() {
	size = 0;
	parameter = NULL;
	hash_value = UINT_MAX;
}

inline int CaseParam::operator==(CaseParam& Param) {
	if (size == Param.size) {
		for (short i = 0; i < size; i++) {
			if (parameter[i] != Param.parameter[i])
				return 0;
		}
		return 1;
	} else
		return 0;
}

/*******************************************************************
 ******************** MDD class definition **************************
 ********************************************************************/

template<class TNode>
void MDD<TNode>::CCaseCache::GC() {
	bool reset;
	for (int i = 0; i < Cache.size(); i++)
		if (Cache[i].second) {
			reset = false;
			if (Cache[i].second->isDead())
				//If the return value of the operation i.e. Cache[i].second
				//(which is of type MDDNode*) or ....
				reset = true;
			else {
				for (short j = 0; j < Cache[i].first.getSize(); j++) {
					if (Cache[i].first[j]->isDead()) {
						//... any of the nodes the parameter points to are dead
						//then the cache item should be reset
						reset = true;
						break;
					}
				}
			}
			if (reset) {
				Cache[i].first.reset();
				Cache[i].second = NULL;
			}
		}
}

//Initializing the MDD structure
template<class TNode>
MDD<TNode>::MDD(int _UniqueTableSize, int _CaseCacheSize, TNode* _DefaultValue) {
	BaseCase1 = BaseCase2 = BaseCase3 = 0;
	MAXLOADFACTOR = 1;
	MAXDEAD = 0.1;
	DefaultValue = _DefaultValue;
	GCEnabled = true;
	UniqueTable.resize(_UniqueTableSize);
	CaseCache.resize(_CaseCacheSize);
	UniqueTable.registerCache(&CaseCache);
}

template<class TNode>
MDD<TNode>::~MDD() {
	GC();
}

template<class TNode>
void MDD<TNode>::GC() {
	if (GCEnabled) {
		Traceprintf(2, "MDD Garbage collection....");
		UniqueTable.GC();
		Traceprintf(2, "done.\n");
	}
}

template<class TNode>
MDDNode* MDD<TNode>::Case(CaseParam& Params, OptType& Opt) {
	if (!(Params[0]->isVariable())) {
		BaseCase1++;
		Opt = ParentN;
		return Params[Params[0]->getValue() + 1];
	}

	//Whether Params[1] through Params[size-1] are equal
	short i;
	for (i = 2; i < Params.getSize(); i++)
		if (Params[i] != Params[1])
			break;
	if (i == Params.getSize()) { //This means that all the elements are equal
		BaseCase2++;
		Opt = ParentN;
		return Params[1];
	}

	//Whether Params[i]==i-1 for all i>0
	for (i = 1; i < Params.getSize(); i++)
		if (Params[i]->isVariable() || (Params[i]->getValue() != (i - 1)))
			break;
	if (i == Params.getSize()) { //This means that all the elements are equal
		BaseCase3++;
		Opt = ParentN;
		return Params[0];
	}

	//Look up the operation in the Case cache
	bool found;
	MDDNode* result = CaseCache.lookup(Params, found);
	if (found) {
		Opt = ParentN;
		return result;
	}

	//Find the most top variable among Params[0..size-1]
	short TopVar = Params[0]->topVariable();
	short TopVarSize = Params[0]->getSize();
	for (short i = 1; i < Params.getSize(); i++)
		if (TopVar < Params[i]->topVariable()) {
			TopVar = Params[i]->topVariable();
			TopVarSize = Params[i]->getSize();
		} else if (TopVar == Params[i]->topVariable()) {
			if (TopVarSize < Params[i]->getSize())
				TopVarSize = Params[i]->getSize();
		}

	//This prevents nodes pointed to by Params from getting garbage
	//Logically also, this should be done because there is a new external
	//pointer (i.e. Params[i]) that points to that node. We do the reverse
	//thing at the end.
	//Params.incRef();

	MDDNode* H = new TNode(TopVar, TopVarSize, DefaultValue);
	OptType Opt2;
	//H is an "external" pointer to the node so we have to increment the
	//RefCount. Since there is no downward pointer we can just use the shallow increment
	//If we don't do this and GC runs, this node will be deleted.
	H->incRefOneLevel();
	MDDNode* R; //The result of the operation (return value of the function)
	CaseParam coFactors(Params.getSize());
	for (short i = 0; i < TopVarSize; i++) {
		coFactors.resetHash();
		for (short j = 0; j < coFactors.getSize(); j++) {
			coFactors[j] = Params[j]->coFactor(TopVar, i);
		}
		//Does not need coFactors.IncRef() and DecRef() because they are called in Case(...)
		R = Case(coFactors, Opt2);
		switch (Opt2) {
		case ParentN:
			H->setPointerN(i, R);
			break;
		case Parent1:
			R->incRefOneLevel();
		case Nothing:
			H->assign(i, R);
			break;
		}
	}
	if ((H->getd(0)->topVariable() < 0) && (H->getd(0)->getValue() == 0) && (H->isDownPointersEqual())) {
		R = H->getd(0);
		Opt = ParentN;
		H->decRef();
		delete H;
	} else {
		pair<MDDNode*, bool> result = findOrAddNode(H);
		if (result.second) { // If the node is already in unique table H is repetitive and must be deleted
			//This needs a big good explanation for why it's implemented this
			//way. We know that H points to some node v that H' points to, i.e.
			//H is repetitive. If H' is dead, instead of calling
			//H->decRef() and then calling H'->incRef() in the function that
			//calls Case() (caller) we keep the RefCount of the children
			//of H intact and only decrease H.RefCount and then in the caller
			//just increase H.RefCount (by setting Opt=Parent1)
			//If H' is not dead, calling H->decRef only changes H.RefCount and RefCount
			//of all its children (only two levels) since we are sure that RefCount
			//of all the children is greater than 1 (Since H' points to them too)
			//In this case also, it's enough to increase H'.RefCount in the caller
			//(by setting Opt=Parent1)
			if (result.first->isDead())
				H->decRefOneLevel();
			else
				H->decRef();
			Opt = Parent1;
			delete H;
		} else
			//In this part we know that H is unique. We also know that
			//H.RefCount = 1 Instead of calling H.DecRef() and then call
			//H.IncRef() by the caller We send a signal to the caller (by
			//setting Opt=Nothing) to ask him not to fiddle with RefCount of
			//any node (i.e. everything is already set correctly)
			Opt = Nothing;
		R = result.first;
	}
	//Put it into Case operation cache
	CaseCache.insert(Params, R);
	//Params.decRef();
	return R;
}

//This method is only used for inserting "zero" and "one" nodes in each
//level of the MDD structure
template<class TNode>
void MDD<TNode>::addNode(MDDNode* node) {
	pair<MDDNode*, bool> result = findOrAddNode(node);
	result.first->incRef();
	//assert(!result.second);
}

template<class TNode>
MDDNode* MDD<TNode>::insertNode(MDDNode* node) {
	node->resetHash();
	const pair<MDDNode*, bool> result = findOrAddNode(node);
	/*
	 if (result.second) { // If node is found it is repetitive and must be deleted
	 //node->decRef();
	 //result.first->incRef();
	 //The following code was changed because the following scenario was happening:
	 //Some nodes refer to 'node'. Then we modify 'node' in-place (more specificly,
	 //we saturate it). After modification, 'node' will not be unique anymore.
	 //The program has to take care of the pointers that points to 'node' and change
	 //all of them one by one to the unique node. However, before this changings
	 //are done, the RefCount of 'node' will not be zero. So it should not be
	 //deleted. That's why we put the 'if' statement below.
	 //delete node;
	 //if (node->isDead())
	 //  delete node;
	 }
	 */
	//node = result.first;
	return result.first;
}

template<class TNode>
MDDNode* MDD<TNode>::insertNodeUnique(MDDNode* node) {
	MDDNode* Temp = insertNode(node);
	if (Temp != node)
		delete node;
	return Temp;
}

template<class TNode>
void MDD<TNode>::removeNode(MDDNode* node) {
	UniqueTable.removeNode(node);
}

/** Finds a MDDNode if it already exists, else it adds it.
 *
 * If hash table is "overloaded," performs GC if too many
 * dead nodes, else resizes hashtable.
 *
 * \param node The MDDNode to add or find.
 * \returns A pair<MDDNode*, bool> of found node and bool
 * value if found or not
 */
template<class TNode>
pair<MDDNode*, bool> MDD<TNode>::findOrAddNode(MDDNode* node) {
	pair<MDDNode*, bool> result = UniqueTable.findOrAddNode(node);
	if (!result.second) { //node is a new element (not already in UniqueTable)
		if (getNoOfNodes() > UniqueTable.getNoOfBuckets() * MAXLOADFACTOR) {
			if (node->getNoOfDeads() > getNoOfNodes() * MAXDEAD) {
				result.first->incRef(); //This is done to avoid result to be GCed
				GC();
				result.first->decRef();
			} else
				UniqueTable.resize(UniqueTable.getNoOfBuckets() + 1);
		}
	}
	return result;
}

template<class TNode>
TNode* MDD<TNode>::readTree(FILE *f) {
	mapping nodetable;
	DisableGC();
	TNode* result = dfsReadTree(f, nodetable);
	EnableGC();
	return result;
}

template<class TNode>
TNode* MDD<TNode>::dfsReadTree(FILE* f, mapping& nodetable) {
	vector<int> children_id;
	int id;
	TNode* node = new TNode;
	if (node->readNode(f, DefaultValue, id, children_id) != SUCCESSFUL_READ)
		return NULL;
	if (node->topVariable() >= 0) {
		for (short i = 0; i < children_id.size(); i++) {
			typename mapping::iterator it = nodetable.find(children_id[i]);
			if (it != nodetable.end())
				node->setPointer(i, nodetable[children_id[i]]);
			else {
				TNode* child_tree = dfsReadTree(f, nodetable);
				if (child_tree == NULL)
					return NULL;
				node->setPointer(i, child_tree);
			}
		}
	}
	node = (TNode*) insertNodeUnique(node);
	nodetable[id] = node;
	return node;
}

template<class TNode>
void MDD<TNode>::print() {
	UniqueTable.print();
}

template<class TNode>
void MDD<TNode>::report() {
	Traceprintf(2, "Case Cache\n");
	CaseCache.report();
	Traceprintf(2, "BaseCase1 happened %d times.\n", BaseCase1);
	Traceprintf(2, "BaseCase2 happened %d times.\n", BaseCase2);
	Traceprintf(2, "BaseCase3 happened %d times.\n", BaseCase3);
}

#endif
