/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef __MDDNODE_H_
#define __MDDNODE_H_

#include "typedefs.h"
#include "BaseClasses/TraceFileMobius.h"
#include "BaseClasses/primes.h"

#include <cstdlib>
#include <iostream>
#include <limits.h>
#include <cassert>
#include <cstdio>
#include <set>
#include <vector>
#include <sstream>
#include <cstring>

#if (HAS_FORWARD_LIST)
	#include <forward_list>
#else
	#include <ext/slist>
	using __gnu_cxx::slist;
#endif


using namespace std;

//BE CAREFUL: You can not use MDDNode::setPointer on a node after it
//is inserted in the UniqueTable because its hash value will change
//and this causes problems. In the only algorithm that uses
//setPointer, the node is created only after all its downward pointers
//are set so there won't be a problem. You should take care if you are
//gonna implement other MDD-based algorithms.

template<class ValType>
class BaseMDDNode {
public:
	BaseMDDNode(short variable, short _size) {
		Construct(variable, _size);
	}
	BaseMDDNode(short variable, short _size, BaseMDDNode* initial_value) {
		Construct(variable, _size, initial_value);
	}
	BaseMDDNode(ValType val) {
		Construct(val);
	}
	BaseMDDNode(BaseMDDNode* v);
	BaseMDDNode() {
	}
	;
	virtual ~BaseMDDNode();
	virtual void Construct(short _var, short _size, BaseMDDNode* _initial_value = NULL);
	virtual void Construct(ValType _val);
	void initializeNode();
	ValType getValue() const {
		assert(var < 0);
		return value;
	}
	short getSize() const {
		return size;
	}
	/** Returns the index of the level of the MDD tree */
	short topVariable() const {
		return var;
	}
	bool isVariable() const {
		return (var >= 0);
	}
	bool isDead() const {
		return (RefCount == 0);
	}
	virtual void print(bool detailed = true);
	void writeTree(FILE* f);
	virtual void setPointer(unsigned short i, BaseMDDNode* node);
	virtual void setPointerN(unsigned short i, BaseMDDNode* node);
	virtual void assign(unsigned short i, BaseMDDNode* node) {
		d[i] = node;
	}
	virtual BaseMDDNode* getPointer(unsigned short i) const {
		return d[i];
	}
	BaseMDDNode* getd(unsigned short i) const {
		return d[i];
	}
	virtual int operator==(const BaseMDDNode& a);
	virtual BaseMDDNode* coFactor(short variable, short value);
	virtual size_t hash(void);
	void resetHash() {
		hash_value = UINT_MAX;
	}
	void incRef();
	void decRef();
	void incRefOneLevel();
	void decRefOneLevel();
	bool isDownPointersEqual();
	int getID() const {
		return id;
	}
	static int getNoOfDeads() {
		return NoOfDeads;
	}
	static int getNoOfNodes() {
		return NoOfNodes;
	}
	static int getNoOfAlives() {
		return NoOfAlives;
	}
	static int getMaxNoOfAlives() {
		return MaxNoOfAlives;
	}
	static int getCounter() {
		return counter;
	}
	static void AdjustNoOfDeads(int amount) {
		NoOfDeads += amount;
		NoOfAlives -= amount;
		MaxNoOfAlives = NoOfAlives > MaxNoOfAlives ? NoOfAlives : MaxNoOfAlives;
	}
	static void AdjustNoOfNodes(int amount) {
		NoOfNodes += amount;
		NoOfAlives += amount;
		MaxNoOfAlives = NoOfAlives > MaxNoOfAlives ? NoOfAlives : MaxNoOfAlives;
	}

	virtual int readNode(FILE* f, BaseMDDNode* _initial_value, int& id, vector<int>& children_id);
	typedef set<BaseMDDNode<ValType>*> node_set_type;

#if (HAS_FORWARD_LIST)
	typedef forward_list<BaseMDDNode<ValType>*> node_slist_type;
#else
	typedef slist<BaseMDDNode<ValType>*> node_slist_type;
#endif
    
	void ConstructNodesLists(node_slist_type* nodes_list);
protected:
	void DepthFirstSearch(node_set_type& tagged_set, node_slist_type* nodes_list);
	virtual void writeNode(FILE* f);
	void dfsWriteTree(FILE* f, set<BaseMDDNode*>& tagged_set);

	//NOTE: The following design can not be used to maintain more than one
	//type of MDD structure separately (different number of levels or sizes).
	// We have to redesign it if we want to have such a feature.

	//Number of nodes created so far (including those that has been deleted)
	//This is used for giving an ID to each node
	static int counter;
	//Number of all nodes in the MDD structure (including dead ones)
	static int NoOfNodes;
	//Number of dead nodes (those with RefCount == 0) in the MDD structure.
	static int NoOfDeads;
	static int NoOfAlives;
	static int MaxNoOfAlives;
	//The index of the variable of the MDD node. index -1 are the
	//values (leaves) of the MDD.
	//Higher numbers mean higher in the tree
	short var;
	//The number of references to the node. Only external pointers and
	//nodes that have non-zero RefCount count.
	int RefCount;

	//ID of the node. IDs are assigned to nodes incrementally.
	int id;

	union {
		unsigned short size; // The number of the downward pointers When var >= 0
		ValType value; // The value of the leaf when var < 0
	};

	//an array of size 'size' (defined just above) that holds the downward
	//pointers
	BaseMDDNode** d;
	//Hash value of the node. Since the hash value depends on the downward
	//pointers, it should be computed after all downward pointers are assigned.
	//Once compute
	unsigned int hash_value;

};

//Increments the reference count of a node without affecting the reference
//count of lower levels (non-recursive version of incRef())
template<class ValType>
void BaseMDDNode<ValType>::incRefOneLevel() {
	if (RefCount == 0)
		//NoOfDeads--;
		AdjustNoOfDeads(-1);
	RefCount++;
}

template<class ValType>
void BaseMDDNode<ValType>::decRefOneLevel() {
	assert(RefCount> 0);
	RefCount--;
	if (RefCount == 0)
		//NoOfDeads++;
		AdjustNoOfDeads(1);
}

/*****************************************************************
 ******************* MDDNode class definition *********************
 ******************************************************************/
template<class ValType>
int BaseMDDNode<ValType>::counter = 0;
template<class ValType>
int BaseMDDNode<ValType>::NoOfDeads = 0;
template<class ValType>
int BaseMDDNode<ValType>::NoOfNodes = 0;
template<class ValType>
int BaseMDDNode<ValType>::NoOfAlives = 0;
template<class ValType>
int BaseMDDNode<ValType>::MaxNoOfAlives = 0;

//default value for initial_value is NULL
template<class ValType>
void BaseMDDNode<ValType>::Construct(short _var, short _size, BaseMDDNode<ValType>* _initial_value) {
	var = _var;
	size = _size;
	d = new BaseMDDNode<ValType>*[size];
	BaseMDDNode<ValType> **p = d;
	for (short i = 0; i < size; i++, p++)
		*p = _initial_value;
	if (_initial_value) {
		_initial_value->incRef();
		_initial_value->RefCount += size - 1;
	}
	initializeNode();
}

template<class ValType>
void BaseMDDNode<ValType>::Construct(ValType _val) {
	var = -1;
	value = _val;
	d = NULL;
	initializeNode();
}

template<class ValType>
void BaseMDDNode<ValType>::initializeNode() {
	id = counter++;
	//NoOfDeads++;
	AdjustNoOfDeads(1);
	RefCount = 0;
	//NoOfNodes++;
	AdjustNoOfNodes(1);
	hash_value = UINT_MAX;
}

template<class ValType>
BaseMDDNode<ValType>::BaseMDDNode(BaseMDDNode<ValType>* v) {
	var = v->var;
	initializeNode();
	if (var < 0) {
		value = v->value;
		d = v->d; //which has to be NULL
	} else {
		size = v->size;
		d = new BaseMDDNode<ValType>*[size];
		memcpy(d, v->d, size * sizeof(BaseMDDNode<ValType>*));
	}
}

template<class ValType>
BaseMDDNode<ValType>::~BaseMDDNode() {
	//Making sure nobody refers to this node
	assert(RefCount == 0);
	//There is no need to apply decRef() to all downward pointers because
	//this has been already and indirectly done by setPointer()
	//Moreover, recall the definition of RefCount.
	delete[] d;
	//NoOfDeads--;
	//NoOfNodes--;
	AdjustNoOfDeads(-1);
	AdjustNoOfNodes(-1);
}
;

/** Sets a downward pointer and update the appropriate reference counts.
 *
 * \param i The index of the child in the MDD Node
 * \param node The downward pointer to point to
 */
template<class ValType>
void BaseMDDNode<ValType>::setPointer(unsigned short i, BaseMDDNode<ValType>* node) {
	assert(var >= 0);
	assert(i < size);
	if (d[i] != node) {
		if (RefCount > 0) {
			if (d[i])
				d[i]->decRef();
			node->incRef();
		}
		d[i] = node;
	}
}

//Sets a downward pointer and update the appropriate reference counts
//This method is used when we are sure that d[i]==NULL just before
//the function call. Otherwise is equivalent to setPointer(...).
//This way we get a slight performance improvement
template<class ValType>
void BaseMDDNode<ValType>::setPointerN(unsigned short i, BaseMDDNode<ValType>* node) {
	if (RefCount > 0) {
		node->incRef();
	}
	d[i] = node;
}

template<class ValType>
BaseMDDNode<ValType>* BaseMDDNode<ValType>::coFactor(short variable, short idx) {
	if (variable > var) //respect to a higher variable
		return this;
	else if (variable == var)
		return getPointer(idx);
	else
		assert(false);
}

//Recursive version of incRef (one that conforms with the definition
// of RefCount)
template<class ValType>
void BaseMDDNode<ValType>::incRef() {
	if (RefCount == 0) {
		if (var >= 0)
			//If the node was dead before, now it gets alive and we have to
			//update the number of references for its children
			for (short i = 0; i < size; i++)
				if (d[i] != NULL)
					d[i]->incRef();
		//NoOfDeads--;
		AdjustNoOfDeads(-1);
	}
	RefCount++;
}

//Recursive version of decRef (one that conforms with the definition
// of RefCount)
template<class ValType>
void BaseMDDNode<ValType>::decRef() {
	assert(RefCount> 0);
	RefCount--;
	if (RefCount == 0) {
		//NoOfDeads++;
		AdjustNoOfDeads(1);
		if (var >= 0)
			//If the node is going to die, we have to update the number of
			//references for its children
			for (short i = 0; i < size; i++)
				if (d[i] != NULL)
					d[i]->decRef();
	}
}

//Checks whether all the downward pointers point to the same things.
//This is used in Case operator.
template<class ValType>
bool BaseMDDNode<ValType>::isDownPointersEqual() {
	for (short i = 1; i < size; i++)
		if (d[i] != d[0])
			return false;
	return true;
}

template<class ValType>
void BaseMDDNode<ValType>::print(bool detailed) {
	Traceprintf(2, "Node ID : %d (Ref#: %d)\n", id, RefCount);
	if (var >= 0) {
		Traceprintf(2, "  var# : %d size : %d\n", var, size);
		if (detailed) {
			for (unsigned short i = 0; i < size; i++) {
				Traceprintf(2, "  d[%d] = ", i);
				if (d[i])
					Traceprintf(2, "%d\n", d[i]->id);
				else
					Traceprintf(2, "NULL\n");
			}
		}
	} else
		Traceprintf(2, "  value: %d\n", value);
}
;

/******** Beginning of the definition of the inline functions *********/

//Computes and returns the hash value of a node.
template<class ValType>
size_t BaseMDDNode<ValType>::hash(void) {
	if (hash_value == UINT_MAX) { //if the hash value of the node is not computed before
		hash_value = 0;
		if (var >= 0) {
			hash_value = var * primes[50]; //i.e. a rather big prime number
			for (unsigned short i = 0; i < size; i++) {
				if (d[i])
					hash_value += primes[(i + 1) & (PRIMES_SIZE-1)] * (unsigned int) (d[i]->id);
			}
		} else
			hash_value = (int) value * primes[50];
	}
	return hash_value;
}
;

//Checks the equality of two nodes
template<class ValType>
int BaseMDDNode<ValType>::operator==(const BaseMDDNode<ValType>& a) {
	if (this == &a)
		return 1;
	else if ((var != a.var) || (size != a.getSize()))
		return 0;
	else if (var < 0)
		return getValue() == a.getValue();
	else {
		for (unsigned short i = 0; i < size; i++) {
			if (getPointer(i) != a.getPointer(i))
				return 0;
		}
		return 1;
	}
}
;

template<class ValType>
void BaseMDDNode<ValType>::writeTree(FILE* f) {
	set<BaseMDDNode<ValType>*> tagged_set;
	dfsWriteTree(f, tagged_set);
}

template<class ValType>
void BaseMDDNode<ValType>::dfsWriteTree(FILE* f, set<BaseMDDNode<ValType>*>& tagged_set) {
	if (tagged_set.find(this) == tagged_set.end()) {
		writeNode(f);
		tagged_set.insert(this);
		if (var >= 0)
			for (short i = 0; i < size; i++)
				if (d[i])
					d[i]->dfsWriteTree(f, tagged_set);
	}
}

template<class ValType>
void BaseMDDNode<ValType>::ConstructNodesLists(node_slist_type* nodes_list) {
	node_set_type tagged_set;
	DepthFirstSearch(tagged_set, nodes_list);
}

template<class ValType>
void BaseMDDNode<ValType>::DepthFirstSearch(node_set_type& tagged_set, node_slist_type* nodes_list) {
	if (tagged_set.find(this) == tagged_set.end() && topVariable() >= 0) {
		tagged_set.insert(this);
		nodes_list[topVariable()].push_front(this);
		if (topVariable() > 0)
			for (int i = 0; i < getSize(); i++)
				if (d[i])
					d[i]->DepthFirstSearch(tagged_set, nodes_list);
	}
}

template<class ValType>
int BaseMDDNode<ValType>::readNode(FILE* f, BaseMDDNode<ValType>* _initial_value, int& id, vector<int>& children_id) {
	short var;
	int data_read; //the number of elements successfully read from the file

#ifdef DASCII
	data_read = fscanf(f, "%hd %d", &var, &id);
#else
	data_read = 0;
	data_read += fread(&var, sizeof(short), 1, f);
	data_read += fread(&id, sizeof(int), 1, f);
#endif
	if (data_read < 2)
		return INCOMPLETE_DATA;
	if (var < 0) {
		ValType _value;

#ifdef DASCII
		//TODO: error-checking should be done for file reading. I don't know how
		//to do it for the following special case
		char s[50];
		fscanf(f, "%s", s);
		istringstream sbuf(s);
		sbuf >> _value;
#else
		if (fread(&_value, sizeof(ValType), 1, f) != 1)
		return INCOMPLETE_DATA;
#endif
		Construct(_value);
	} else {
		unsigned short _size;
#ifdef DASCII
		if (fscanf(f, "%hd", &_size) != 1)
			return INCOMPLETE_DATA;
#else
		if (fread(&_size, sizeof(short), 1, f) != 1)
		return INCOMPLETE_DATA;
#endif
		Construct(var, _size, _initial_value);
		children_id.resize(_size);
		for (short i = 0; i < _size; i++) {
#ifdef DASCII
			if (fscanf(f, " %d", &children_id[i]) != 1)
				return INCOMPLETE_DATA;
#else
			if (fread(&children_id[i], sizeof(int), 1, f) != 1)
			return INCOMPLETE_DATA;
#endif
		}
	}
	return SUCCESSFUL_READ;
}

template<class ValType>
void BaseMDDNode<ValType>::writeNode(FILE* f) {
#ifndef DASCII
	fwrite(&var, sizeof(short), 1, f);
	fwrite(&id, sizeof(int), 1, f);
	if (var >= 0) {
		fwrite(&size, sizeof(unsigned short), 1, f);
		for (short i = 0; i < size; i++) {
			int child_id;
			if (d[i])
			child_id = d[i]->getID();
			else
			child_id = -1;
			fwrite(&child_id, sizeof(int), 1, f);
		}
	}
	else
	fwrite(&value, sizeof(ValType), 1, f);
#else
	fprintf(f, "%hd %d", var, id);
	if (var >= 0) {
		fprintf(f, " %hd", size);
		for (short i = 0; i < size; i++) {
			if (d[i])
				fprintf(f, " %d", d[i]->getID());
			else
				fprintf(f, " -1");
		}
	} else {
		ostringstream sbuf;
		sbuf << value;
		fprintf(f, " %s", sbuf.str().c_str());
	}
	fprintf(f, "\n");
#endif
}

typedef BaseMDDNode<short> MDDNode;

#endif
