/*
 * saucy.c
 * Automorphism groups of graphs
 *
 * by Paul T. Darga <pdarga@umich.edu>
 * and Mark Liffiton <liffiton@umich.edu>
 *
 * Copyright (C) 2004, The Regents of the University of Michigan
 * See the LICENSE file for details.
 */
/*
 LICENSE file contains:
 COPYRIGHT (C) 2004
 THE REGENTS OF THE UNIVERSITY OF MICHIGAN
 ALL RIGHTS RESERVED

 PERMISSION IS GRANTED TO USE, COPY, CREATE DERIVATIVE WORKS AND
 REDISTRIBUTE THIS SOFTWARE AND SUCH DERIVATIVE WORKS FOR ANY PURPOSE,
 SO LONG AS THE COPYRIGHT NOTICE ABOVE, THIS GRANT OF PERMISSION,
 AND THE DISCLAIMER BELOW APPEAR IN ALL COPIES MADE; AND SO LONG AS
 THE NAME OF THE UNIVERSITY OF MICHIGAN IS NOT USED IN ANY ADVERTISING
 OR PUBLICITY PERTAINING TO THE USE OR DISTRIBUTION OF THIS SOFTWARE
 WITHOUT SPECIFIC, WRITTEN PRIOR AUTHORIZATION.

 THIS SOFTWARE IS PROVIDED AS IS, WITHOUT REPRESENTATION FROM THE
 UNIVERSITY OF MICHIGAN AS TO ITS FITNESS FOR ANY PURPOSE, AND WITHOUT
 WARRANTY BY THE UNIVERSITY OF MICHIGAN OF ANY KIND, EITHER EXPRESS
 OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED WARRANTIES OF
 MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE REGENTS
 OF THE UNIVERSITY OF MICHIGAN SHALL NOT BE LIABLE FOR ANY DAMAGES,
 INCLUDING SPECIAL, INDIRECT, INCIDENTAL, OR CONSEQUENTIAL DAMAGES,
 WITH RESPECT TO ANY CLAIM ARISING OUT OF OR IN CONNECTION WITH THE
 USE OF THE SOFTWARE, EVEN IF IT HAS BEEN OR IS HEREAFTER ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGES.
 */
#include <string>
#include <iostream>
#include <sstream>
#include "Permutation.h"
using namespace std;

#ifndef SAUCY_H
#define SAUCY_H

/*
 * The alarm() call and SIGALRM are POSIX.  getopt() is
 * (apparently) an XSI extension.  Defining _XOPEN_SOURCE at 600
 * implies _POSIX_C_SOURCE at 200112L.
 */

#define _XOPEN_SOURCE 600
#include <unistd.h> /* getopt, alarm, SIGALRM */

/* We need this for strdup */
#define _XOPEN_SOURCE_EXTENDED 1

/* ANSI stuff */
#include <limits.h> /* CHAR_BIT */
#include <stdlib.h> /* malloc, calloc, free, exit, EXIT_* */
#include <string.h> /* memset */
#include <signal.h> /* signal, SIG_IGN */
#include <stdio.h>  /* FILE, fopen, fscanf, printf, fclose */
#include <time.h>/* clock, CLOCKS_PER_SEC */

#define SAUCY_MALLOC(count_,type_) ((type_ *) malloc((count_)*sizeof(type_)))
#define SAUCY_CALLOC(count_,type_) ((type_ *) calloc((unsigned) (count_),sizeof(type_)))
#define SAUCY_REALLOC(ptr_,type_,num_) ((type_ *) realloc((ptr_),sizeof(type_)))
#define SAUCY_FREE(ptr_) free((ptr_))

#define CPUTIME (((double)clock())/((double)CLOCKS_PER_SEC))

#define MULTIPLY(s1_,s2_,i_)   \
  {                            \
    if ((s1_ *= i_) >= 1e10) { \
      s1_ /= 1e10; s2_ += 10;  \
    }                          \
  }

/* Bit manipulation routines */
#define BITS            ((int)(sizeof(int)*8))
#define BSIZE(n_)       (((n_-1)/BITS)+1)
#define BALLOC(b_,n_)   ((b_)=SAUCY_CALLOC(BSIZE((n_)),int))
#define BWIPE(b_,n_)    (memset((b_),0,(BSIZE((n_))*sizeof(int))))
#define BISSET(b_,i_)   ((b_)[(i_)/BITS]&(1<<((i_)%BITS)))
#define BSET(b_,i_)     ((b_)[(i_)/BITS]|=(1<<((i_)%BITS)))
#define BCLEAR(b_,i_)   ((b_)[(i_)/BITS]&=~(1<<((i_)%BITS)))
#define BERASE(b_,i_)   ((b_)[(i_)/BITS]=0)

/* Array-of-bit-vectors manipulation routines */
#define BARR_ALLOC(b_,n_,k_)  ((b_)=SAUCY_CALLOC((k_*BSIZE((n_))),int))
#define BARR(b_,i_,k_)        ((b_)+((i_)*BSIZE(k_)))
#define FIX(fix_,i_,n_)       (BARR((fix_),(i_),(n_)))
#define MCR(mcr_,i_,n_)       (BARR((mcr_),(i_),(n_)))

/* To add stuff to alpha/beta */
#define refine_add_induce(saucy_, single_, who_) \
  {                                                    \
    if (single_) {                                     \
      saucy->beta[saucy_->beta_num++] = who_;          \
    } else {                                           \
      saucy->alpha[saucy_->alpha_num++] = who_;        \
    }                                                  \
    saucy->abmark[who_] = 1;                           \
  }

typedef struct saucy_struct saucy_t;
typedef struct saucy_cycle_struct saucy_cycle_t;
typedef struct saucy_permutation_struct saucy_permutation_t;

struct saucy_struct {
	int vertex_num; /* Number of vertices searched over */
	int edge_num; /* Number of edges */
	int cell_num; /* Number of cells in partition */
	int lev; /* Current search tree level */
	int flag; /* Was the last leaf node an automorphism? */
	int target; /* Front of cell about to be split */
	int target_min; /* Location of minimum element in target cell */
	int anc; /* Level of greatest common ancestor with zeta */
	int have_zeta; /* Have we hit the first leaf node yet? */
	int saved_auto_num;/* Number of auto's saved */
	int front; /* Position of refining cell */
	int alpha_num; /* Size of alpha stack */
	int beta_num; /* Size of beta stack */
	int cell_list_num;/* Number of cells in cell list */
	int perm_num; /* permutation count */
	int max_saved_auto_num;/* Number of saved auto's for backtracking from bads*/
	int timeout; /* Seconds before quitting after refinement */
	int stats_mode; /* Print out stats when we're done */
	int zeta_mode; /* Stop when we hit the first leaf node */
	int quiet_mode; /* Don't output automorphisms */
	int gap_mode; /* Do GAP I/O (interface with Shatter) */
	int grpsize_indmin;/* used to compute size of automorphism group */
	int grpsize_exp;/* used to compute size of automorphism group */
	double grpsize_base;/* used to compute size of automorphism group */
	int *edg; /* Connection info */
	int *adj; /* Adjacency list sizes and positions */
	int *lab; /* Labelling of objects (half of coloring) */
	int *ptn; /* Where/what level cells split (other half) */
	int *nextnonsing;/* Forward next-nonsingleton pointers */
	int *prevnonsing;/* Backward next-nonsingleton pointers */
	int *cell_front;/* Pointer to front of cells */
	int *cell_len; /* Length of cells (defined for cfront's) */
	int *gamma; /* Container for permutations at leaf nodes */
	int *bit_contents;/* Workspace bit vector */
	int *start; /* Location of target at each level */
	int *fixed; /* Label fixed at each level */
	int *zeta; /* The first leaf node */
	int *theta; /* Running approximation of orbit partition */
	int *bit_fixed; /* Bit vector version of fixed */
	int *fix; /* Labels left fixed by each saved auto */
	int *mcr; /* Min cell representatives of each saved auto */
	int *alpha; /* Nonsingletons that might induce refinement */
	int *beta; /* Singletons that might induce refinement */
	int *bucket; /* Workspace */
	int *vertex_count;/* Num vertices with same adj count to ref cell */
	int *cell_count;/* Number of connections to refining cell */
	int *cell_list; /* List of cells marked for refining */
	char *abmark; /* Those things that are marked */
	char *stuff; /* bit_contents replacement (faster) */
	char *cell_mark;/* Is this cell marked for refining? */
	saucy_permutation_t *permutations;/* saved permutation list */
	FILE *file; /* File we're reading input from */
};

struct saucy_cycle_struct {
	saucy_cycle_t *next_cycle; /* next cycle */
	int cycle_size; /* cycle size */
	int *cycle; /* cycle */
};

struct saucy_permutation_struct {
	saucy_permutation_t *next_permutation; /* next permutation */
	int cycle_num; /* number of cycles */
	saucy_cycle_t *cycles; /* cycles */
};

class saucy_obj {
public:
	saucy_obj();

	///* MGM Additions */
	vector<Permutation> PermVec;
	vector<Permutation> convert_perm(string perm_string, int num_nodes, int num_switches, int num_each);
	int num_switches, num_total_elements;

	/* symmetry computation pi */
	saucy_t *saucy_begin(int argc, char **argv);
	saucy_t *saucy_begin2(char * fname);
	void saucy_init(saucy_t *saucy);
	void saucy_end(saucy_t *saucy);
	int saucy_search(saucy_t *saucy);

	/* options pi */
	void saucy_data_init_options(saucy_t *saucy);
	void saucy_data_set_options(saucy_t *saucy, int quiet_mode, int gap_mode, int zeta_mode, int stats_mode, int max_saved, int timeout);
	void saucy_parse_options(saucy_t *saucy, int argc, char **argv);

	/* graph pi */
	void saucy_graph_create(saucy_t *saucy, int node_num, int edge_num);
	void saucy_graph_vertex_edge_count(saucy_t *saucy, int node, int edge_num);
	void saucy_graph_vertex_complete(saucy_t *saucy);
	void saucy_graph_insert_edge(saucy_t *saucy, int j, int k);
	void saucy_graph_insert_partition_separator(saucy_t *saucy, int k);
	void saucy_graph_insert_color(saucy_t *saucy, int i, int j);
	void saucy_graph_edge_complete(saucy_t *saucy);

	/* permutation pi */
	saucy_permutation_t *saucy_permutations_get_first_permutation(saucy_t *saucy);
	saucy_permutation_t *saucy_permutations_get_next_permutation(saucy_permutation_t *permutation);
	int saucy_permutations_get_permutation_count(saucy_t *saucy);
	int saucy_permutations_get_cycle_count(saucy_permutation_t *permutation);
	saucy_cycle_t *saucy_permutations_get_first_cycle(saucy_permutation_t *permutation);
	saucy_cycle_t *saucy_permutations_get_next_cycle(saucy_cycle_t *cycle);
	int saucy_permutations_get_cycle_size(saucy_cycle_t *cycle);
	int saucy_permutations_get_cycle_element(saucy_cycle_t *cycle, int index);

private:
	void saucy_search_init(saucy_t *saucy);
	void saucy_search_cleanup(saucy_t *saucy);
	void saucy_search_descend_left(saucy_t *saucy);
	void saucy_search_update_theta(saucy_t *saucy);
	void saucy_search_backtrack(saucy_t *saucy);
	void saucy_search_theta_prune(saucy_t *saucy);
	void saucy_search_orbit_prune(saucy_t *saucy);
	void saucy_search_descend(saucy_t *saucy);
	void saucy_search_min_target(saucy_t *saucy);
	void saucy_search_target_cell(saucy_t *saucy);
	void saucy_search_is_automorphism(saucy_t *saucy);
	void saucy_data_init(saucy_t *saucy);
	void saucy_data_cleanup(saucy_t *saucy);
	void saucy_data_mark(saucy_t *saucy, int x);
	void saucy_data_unmark(saucy_t *saucy, int x);
	void saucy_data_count(saucy_t *saucy, int x);
	void saucy_data_read_amorph(saucy_t *saucy); /*Read graph in amorph format*/
	void saucy_data_read_gap(saucy_t *saucy);/*Read graph in GAP format*/
	void saucy_data_vertex_complete(saucy_t *saucy);/*Fix vertices for adjacency array*/
	void saucy_data_edge_complete(saucy_t *saucy);/*Fix edges for adjacency array*/
	void saucy_refine_init(saucy_t *saucy);
	void saucy_refine_cleanup(saucy_t *saucy);
	void saucy_refine(saucy_t *saucy);
	void saucy_refine_singleton(saucy_t *saucy);/*Refine with a single cell*/
	void saucy_refine_nonsingle(saucy_t *saucy);/*Refine with cell (>=2 objs)*/
	void saucy_refine_sort_cell_list(int *cell_list, int cell_size);/*Shell sort*/
	int saucy_parse_opt_int(int min);
	void saucy_parse_usage(void);
};
#endif
