/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: BaseComposerClass.h
 *
 *      Authors: Aaron Stillman, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef BASE_COMPOSER_CLASS_H_
#define BASE_COMPOSER_CLASS_H_

#include "../BaseClasses/BaseModelClass.h"
#include "../BaseClasses/List.h"
#include "../BaseClasses/BaseStateVariableClass.h"
#include "../BaseClasses/PreselectGroup.h"
#include "../BaseClasses/timestuff.h"
#include "../BaseClasses/MobiusTypeDefs.h"
#include "SharedStateVarLink.h"
#include <list>
#include "../BaseClasses/state/SharableSV.h"
#include "../BaseClasses/state/StructStateVariable.h"
#include "../BaseClasses/state/ArrayStateVariable.h"

/**
 *  Governs the size of the hash table that will be used to hold
 *  the action names.  It will be multiplied by the number of
 *  actions to create the initial number of entries in the table.
 */
const unsigned int JOIN_HASH_TABLE_RATIO = 4;

/**  This is the base composer class, derived from the base model class.
 *  All composer node objects are derived from this class.
 *
 *  @author Aaron J. Stillman
 *  @version Mobius Version 1.0
 *
 */
class BaseComposerClass: public BaseModelClass {
public:
	/** Base composer class constructor
	 *
	 * @param TheName The name to be given to the composer node
	 * @param NumModels The number of submodels
	 * @param NumSharedVariables The number of shared state variables
	 * @param SharedVariableNames An array of strings specifying the names of the shared state variables
	 *
	 */
	BaseComposerClass(char *, int, int, char **);

	/** Base composer class destructor
	 */
	~BaseComposerClass();

	/** An array of pointers to the submodels
	 */
	BaseModelClass ** ModelArray;
	inline BaseModelClass ** getModelArray() {
		return (ModelArray);
	}
	;
	inline int getNumModels() {
		return (NumModels);
	}
	;

	template <class T> inline SharableSV<T>* getSharableSVPointer(SharableSV<T>* pointer) {
		return pointer;
	}
	template <class T> inline SharableSV<T>* getSharableSVPointer(SharableSV<T> &pointer) {
		return &pointer;
	}
	inline BaseStructStateVariable* getSharableSVPointer(BaseStructStateVariable* pointer) {
		return pointer;
	}
	inline BaseStructStateVariable* getSharableSVPointer(BaseStructStateVariable &pointer) {
		return &pointer;
	}
	inline BaseArrayStateVariable* getSharableSVPointer(BaseArrayStateVariable* pointer) {
		return pointer;
	}
	inline BaseArrayStateVariable* getSharableSVPointer(BaseArrayStateVariable &pointer) {
		return &pointer;
	}


	/** The number of submodels
	 */
	int NumModels;

	/** An array of pointers to the shared state variables
	 */
	BaseStateVariableClass ** SharedStatePtrs = nullptr;

	/** An array of pointers to the list of equivalent state variables for each shared state variable */
	SharedStateVarLink ** SharedStateVarList = nullptr;

	/** An array of linked lists each of which corresponds to one (shared)
	 state variable SV in the composer node. The linked list contains the
	 list of state variables in direct children of the composer node
	 that are shared to create SV.
	 */
	list<BaseStateVariableClass*>* ChildrenSharedSVList = nullptr;

	/** An array of strings specifying the names of the shared state variables
	 */
	char ** SharedStateNames = nullptr;

	/** The number of shared state variables that have been initialized in the composer node */
	unsigned int AddedSharedPtrs;

	/** Utility variables for use in class methods
	 */
	char * stateindex = nullptr;
	int index = 0;
	int MyStateSize = 0;
	int ObjectCounter = 0;

	/**
	 * Gets the actions defined on the composer node
	 *
	 * @return ActionList An array of pointers to actions defined on the node
	 *
	 */
	virtual void listActions(List<BaseActionClass>* TheContainer) override {
		TheContainer->append(ActionList, getNumActions());
	} // end listActions(List<BaseActionClass>*)


	/** Gets all the actions with the given name defined on the composer node
	 *
	 * @param TheActionName The desired action naem
	 * @return TheActions An array of pointers to actions with the given name
	 */
	virtual void listActions(char *, List<BaseActionClass>*) override;

	/** Gets the groups defined on the composer node
	 *
	 * @return GroupList An array of pointers to groups defined on the node
	 *
	 */
	virtual void listGroups(List<BaseGroupClass>* TheContainer) {
		TheContainer->append(GroupList, getNumGroups());
	} // end listGroup(List<BaseGroupClass>*)

	/**
	 * Gets all submodels with the given name
	 *
	 * @param ModelName The desired submodel name
	 * @param TheContainer An array of pointers to models with the given name
	 *
	 */
	virtual void listModels(char * ModelName, List<BaseModelClass>* TheContainer);

	/**
	 * Returns the total size of all the shared state variables at the node
	 *
	 * @return The total size (in bytes) of all the shared state variables
	 *
	 */
	int SharedStateSize();

	/** Returns the state size of the composer node
	 *
	 * @return The size (in bytes) of the composer node's state
	 *
	 */
	virtual int StateSize()=0;

	/** Compares two states
	 *
	 * @param State1 The first state to be compared
	 * @param State2 The second state to be compared
	 * @return A boolean denoting whether the two states were equivalent
	 *
	 */
	virtual bool CompareState(void*, void*);

	/**
	 *  Sets up the data structures of the composer node
	 */
	void GeneralSetup();

	/** Hashes a string value
	 *
	 * @param Key The string to be hashed
	 * @param HashSize The size of the hash table
	 * @return The hashed value of the string
	 */
	unsigned int Hash(const char * Key, int HashSize);

	/**
	 *  Writes the current state of the composed model to the specified
	 *  memory location.
	 *
	 *  @param location The memory location to be written to.
	 */
	virtual void CurrentState(void * location)=0;

	/**
	 *  Sets the state of the composed model to the state residing
	 *  in the specified memory location.
	 *
	 *  @param location  The location holding the state to be copied.
	 */
	virtual void SetState(void * location)=0;

	/**
	 *  Will look to see if the state variable is the main variable
	 *  shared with a set of other state variables.  If so, it will
	 *  return a linked list of the variables that it is shared with.
	 *  This is used to build lists of equivalently shared state
	 *  variables in composed models.
	 *
	 *  @param var  The base state variable whose shared variables we want
	 *  @return A linked list of state variables that are shared with it.
	 */
	virtual SharedStateVarLink * getListOfSharedVariables(BaseStateVariableClass * var);

	/**
	 *  Will find the representative state variable that contains the state var
	 *  in its equivalent class.
	 *
	 *  @param var  The base state variable whose representative state var
	 *              we want.
	 *  @return The representative state variable or NULL.
	 */
	virtual BaseStateVariableClass * getMainSharedVariable(BaseStateVariableClass * var);

	/**
	 *  Outputs the state of the composed model to standard out.
	 */
	void printState(ostream *strm);

	BaseStateVariableClass** GetSharedStateVariables();
	/**
	 *  Used for debugging.
	 */
	void printAllSharedVariables();

	/**
	 *  Adds the state variable and its name to the bookkeeping
	 *  data structures in the composed node.
	 *
	 *  @param var  The state variable to be added.
	 *  @param name  The name of the state variable to be added
	 */
	void addSharedPtr(BaseStateVariableClass * var, char * name);

	/**
	 *  Adds the shared state variable pointer from the submodel to the
	 *  linked list of state variable pointers that have been shared
	 *  with the new state variable created in the composed node.
	 *
	 *  @param oldVar  The shared state variable from the submodel
	 *  @param newVar  The newly created state var in the composed model,
	 *                 with which the submodel's state var is shared
	 */
	void addSharingInfo(BaseStateVariableClass * oldVar, BaseStateVariableClass * newVar, BaseModelClass* TheParent = NULL);

	/**
	 *  Gets the list of affected state variables in the submodels
	 *  having the given name and having the specified state variable name.
	 *
	 *  @param ModelName The name of the model in which to
	 *     search for the state variable
	 *  @param VarName The name of the state variable
	 *  @param TheContainer A data structure containing an array of pointers to state
	 *    variables and the number of state variables in the arrary
	 */
	void listSVs(char* ModelName, char* VarName, List<BaseStateVariableClass>* TheSVContainer, List<BaseModelClass>* TheModelContainer);

	/**
	 *  Initializes the data structures for actions within the composed
	 *  model node.
	 */
	void SetupActions();

protected:

	/**
	 *  This groups is used a container for all
	 *  instantaneous groups contained in submodels
	 */
	PreselectGroup InstantaneousActions;

	bool AllChildrenEmpty();

};

#endif
