/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: Join.h
 *
 *      Authors: Aaron Stillman, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef JOIN_H_
#define JOIN_H_

#include "BaseComposerClass.h"
#include "../BaseClasses/primes.h"
#include <vector>


/**  This is the join class, derived from the base composer class.
 *  It is used to create join nodes in a composed model
 *
 *  @author Aaron J. Stillman
 *  @version Mobius Version 1.0
 *
 */
class Join: public BaseComposerClass {
public:
	/** The join constructor
	 *
	 * @param TheName The name given to the join node
	 * @param NumModels The number of submodels at the join node
	 * @param NumSharedVariables The number of shared variables at the join node
	 * @param SharedVariableNames An array of strings specifying the names of the shared variables
	 *
	 */
	Join(char *, int, int, char **);

	/* The class destructor
	 */
	~Join();

	/** An array of integers giving the cardinalities of the actions
	 */
	int * ActionNameCardinalities = nullptr;

	/** An array giving the state sizes (in bytes) of each submodel
	 */
	int * SubmodelStateSize = nullptr;

	/** Utility variable used for searching the action hash table
	 */
	bool LookForward;

	/** Various utility variables
	 */
	unsigned int currentStateIndex, currentStateCounter;
	unsigned int stateSizeCounter, stateSizeTotal;
	unsigned int hashActionsCounter, hashActionsIndex;
	unsigned int numActionsCounter, numActionsIndex;
	unsigned int numModelsCounter, numModelsIndex;
	unsigned int setStateIndex, setStateCounter;

	/** Returns the state size (in bytes) of the join node
	 *
	 * @return The size, in bytes, of the join's state
	 */
	virtual int StateSize();

	/** Sets the current state of the join
	 *
	 * @param TheAddress The address in memory of the state to be assigned to the join
	 *
	 */
	virtual void SetState(void * TheAddress);

	/** Initializes submodels and sets up the data structures of the join node
	 */
	void Setup();

	/** Hashes the actions in to the action hash table
	 */
	void HashActions();

	/** Writes the current state of the join node to a location in memory
	 *
	 * @param TheLocation The location where the state will be written
	 *
	 */
	virtual void CurrentState(void * TheLocation);

private:
	void SetSharedState(void *Address);
	void CurrentSharedState(void *Address);

public:
	/**
	 * The following member declarations are used by the symbolic SSG algorithm
	 * implemented by Salem
	 */
	class LocalState: public State {
	public:
		LocalState(Join* model);
		~LocalState();
		friend class Join;
		void print();
		int hash();
		int compare(State const & S) const;
		State const& operator=(State const& S);
		int fwriteState(FILE* f);
		int freadState(FILE* f);
		void* GetStatePtr();
	protected:
		void SetState(const BaseModelClass* model);
		void GetState(const BaseModelClass* model);
	private:
		char* state;
		int size;
	};

	virtual ModelType GetModelType();
	virtual int GetNumberOfChildren();
	virtual BaseModelClass* GetChild(int i);
	virtual void GetLocalActions(base_action_list_t& actions);
	virtual void GetGlobalActions(base_action_list_t& actions);
	virtual bool HasLocalState();
	virtual void GetLocalState(State* _LS);
	virtual void SetLocalState(const State* _LS);
	virtual void ConstructLocalState(State*& _LS);
	virtual void DestructLocalState(State* _LS);


protected:
	/** The size of the hash table used to store the actions
	 */
	int ActionNameTableSize;

	/** The hash table storing the actions
	 */
	std::vector<const char*> ActionNameHashTable;
//BAF	char ** ActionNameHashTable;
};

#endif

