/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: Rep.h
 *
 *      Authors: Aaron Stillman, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef REP_H_
#define REP_H_

#include "BaseComposerClass.h"
#include "../BaseClasses/primes.h"

/** This variable denotes the array size when we start to use insertion
 sort instead of quicksort.  (Insertion sort becomes faster than
 quicksort after the array becomes small enough). */
const int CUTOFF = 10;

/**  This is the rep class, derived from the base composer class.
 *  It is used to create replicate nodes in a composed model
 *
 *  @author Aaron J. Stillman
 *  @version Mobius Version 1.0
 *
 */
class Rep: public BaseComposerClass {
public:
	/** The rep constructor
	 *
	 * @param TheName The name given to the rep node
	 * @param NumModels The number of submodels for the rep node
	 * @param NumSharedVariables The number of shared variables at the rep node
	 * @param SharedVariableNames An array of strings specifying the shared variable names
	 *
	 */
	Rep(char * TheName, int NumModels, int NumSharedVariables, char** SharedVariableNames);

	/** The rep destructor
	 */
	~Rep();

	/** This method returns the state size of the rep node
	 *
	 * @return The size of the state (in bytes)
	 */
	virtual int StateSize();

	/** This method writes the current state to the given address
	 *
	 * @param TheLocation The address that the current state will be written to
	 */
	virtual void CurrentState(void * TheLocation);

	/** This method sets the state of the model to the state stored at the given address
	 *
	 * @param TheLocation The address in memory of the state to be set
	 */
	virtual void SetState(void * TheLocation);

	/** Executes the quicksort sorting algorithm (not implemented yet)
	 *
	 */
	void QuickSort(const int, const int, const int);

	/** This method executes the insertion sort sorting algorithm
	 *
	 * @param TheLength The number of array elements (adjusted to start with 0)
	 * @param TheSize The size of the elements (states) to be lumped
	 */
	void InsertionSort(const unsigned int, const int);

	/**
	 *  Swaps the values of the two integers.
	 *
	 *  @param A  The location of the first integer.
	 *  @param B  The location of the second integer.
	 */
	inline void Swap(unsigned int & A, unsigned int & B);

	/**
	 *  Swaps the values of the two text strings.
	 *
	 *  @param text1  The location of the first string.
	 *  @param text2  The location of the second string.
	 */
	inline void Swap(char ** text1, char ** text2);

	/**
	 *  Calculates the median of three values
	 *
	 *  @param v1  The first value
	 *  @param v2  The second value
	 *  @param v3  The third value
	 *  @return The median of the three values
	 */
	int Median3(const unsigned int v1, const unsigned int v2, const int v3);

	/**
	 *  Performs various initializations to set up the rep node
	 *
	 * @param TheSubmodelName The name to be given to the rep's submodel nodes
	 */
	void Setup(char * TheSubmodelName);

	/** A utility pointer for use in currentstate
	 */
	char * tempChar = nullptr;

	/** If minbits is 4, this boolean tells us whether we will write to the left 4 bits of the byte next
	 */
	bool leftBit;

	/** Various and sundry utility variables for the different methods
	 */
	int currentStateCounter, currentStateIndex;
	int lumpStatesCounter, lumpStatesStateCounter;
	unsigned int setStateCounter;
	char * setStateIndex = nullptr;
	char * setStateCardLoc = nullptr;
	int writeStatesIndex, writeStatesCounter;
	int isIndex = 0;
	int isCounter = 0;
	int isCard = 0;
	int numCopies;

	/** The minimum number of bits needed to store a given submodel's state
	 */
	short MinBits;

	/** The minimum number of bytes needed to store all the submodels' states
	 */
	short MinBytes;

	/** A utility char array equal in size to a submodel's state size
	 */
	char * dummystate = nullptr; // dummy char pointer

	/**
	 *  A helper variable to let us know when a value has been
	 *  located in an array.
	 */
	bool FoundIt;

	/** This array of states holds the lumped states of the rep's submodels
	 */
	char ** LumpedStates;

	/**  An array of integers holding the cardinalities of the
	 *   submodel states.
	 */
	unsigned int * LumpedCardinalities;

	/** The total number of different lumped states
	 */
	unsigned int TotalDiffLumped;

	/** The unlumped state of the rep node when the currentstate was called last time
	 */
	char * UnlumpedLastTime;

	/** The unlumped state of the rep node on the present call to currentstate
	 */
	char * UnlumpedThisTime;

	/** The size, in bytes, of any of the rep node's submodels
	 */
	int SubmodelStateSize;

private:
	char* SetSharedState(char *Address);
	void CurrentSharedState(char* Address);

public:
	/**
	 * The following member declarations are used by the symbolic SSG algorithm
	 * implemented by Salem
	 */
	class LocalState: public State {
	public:
		LocalState(Rep* model);
		~LocalState();
		friend class Rep;
		void print();
		int hash();
		int compare(State const & S) const;
		State const& operator=(State const& S);
		int fwriteState(FILE* f);
		int freadState(FILE* f);
		void* GetStatePtr();
	protected:
		void SetState(const BaseModelClass* model);
		void GetState(const BaseModelClass* model);
	private:
		char* state;
		int size;
	};

	virtual ModelType GetModelType();
	virtual int GetNumberOfChildren();
	virtual BaseModelClass* GetChild(int i);
	virtual void GetLocalActions(base_action_list_t& actions);
	virtual void GetGlobalActions(base_action_list_t& actions);
	virtual bool HasLocalState();
	virtual void GetLocalState(State* _LS);
	virtual void SetLocalState(const State* _LS);
	virtual void ConstructLocalState(State*& _LS);
	virtual void DestructLocalState(State* _LS);

};

#endif
