/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: GraphJoin.h
 *
 *      Authors: Aaron Stillman, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef GRAPH_JOIN_H_
#define GRAPH_JOIN_H_

#include "BaseGraphComposerClass.h"
#include "Wrapper.h"
#include <stdlib.h>
#include <vector>

/**  This is the graph join class, derived from the base composer class.
 *  It is used to create join nodes in a composed graph model
 *
 *  @author Aaron J. Stillman
 *  @version Mobius Version 1.0
 *
 */
class GraphJoin: public BaseGraphComposerClass {
public:
	/** The GraphJoin constructor
	 *
	 * @param TheName The name given to the join node
	 * @param NumModels The number of submodels at the join node
	 * @param NumSharedVariables The number of shared variables at the join node
	 * @param SharedVariableNames An array of strings specifying the names of the shared variables
	 *
	 */
	GraphJoin(char *, int, int, char **, Wrapper *);

	/* The class destructor
	 */
	~GraphJoin();

	/** The wrapper that holds this join
	 */
	Wrapper * MyWrapper;

	/** An array of integers giving the cardinalities of the actions
	 */
	int * ActionNameCardinalities;

	/** An array giving the state sizes (in bytes) of each submodel
	 */
	int * SubmodelStateSize;

	/** Utility variable used for searching the action hash table
	 */
	bool LookForward;

	/** Various utility variables
	 */
	unsigned int currentStateIndex, currentStateCounter;
	unsigned int stateSizeCounter, stateSizeTotal;
	unsigned int hashActionsCounter, hashActionsIndex;
	unsigned int numActionsCounter, numActionsIndex;
	unsigned int numModelsCounter, numModelsIndex;
	unsigned int setStateIndex, setStateCounter;

	/** Returns the state size (in bytes) of the join node
	 *
	 * @return The size, in bytes, of the join's state
	 */
	virtual int StateSize();

	/** Sets the current state of the join
	 *
	 * @param TheAddress The address in memory of the state to be assigned to the join
	 */
	virtual void SetState(void*);

	/** Initializes submodels and sets up the data structures of the join node
	 */
	void Setup();

	/** Hashes the actions in to the action hash table
	 */
	void HashActions();

	/** Writes the current state of the join node to a location in memory
	 *
	 * @param TheLocation The location where the state will be written
	 */
	virtual void CurrentState(void *);

	/** This function is not used and will be removed later */
	void CorrectSubmodelStateSize(int) {
	} // no submodel statesize stored yet

	/**
	 *  Will look to see if the state variable is the main variable
	 *  shared with a set of other state variables.  If so, it will
	 *  return a linked list of the variables that it is shared with.
	 *  This is used to build lists of equivalently shared state
	 *  variables in composed models.
	 *
	 *  @param var  The base state variable whose shared variables we want
	 *  @return A linked list of state variables that are shared with it.
	 */
	SharedStateVarLink * getListOfSharedVariables(BaseStateVariableClass *);

	/**
	 *  Will find the representative state variable that contains the state var
	 *  in its equivalent class.
	 *
	 *  @param var  The base state variable whose representative state var
	 *              we want.
	 *  @return The representative state variable or NULL.
	 */
	BaseStateVariableClass * getMainSharedVariable(BaseStateVariableClass *);

	/**
	 *  Adds the state variable and its name to the bookkeeping
	 *  data structures in the composed node.
	 *
	 *  @param var  The state variable to be added.
	 *  @param name  The name of the state variable to be added
	 */
	void addSharedPtr(BaseStateVariableClass *, char *);

	/**
	 *  Adds the shared state variable pointer from the submodel to the
	 *  linked list of state variable pointers that have been shared
	 *  with the new state variable created in the composed node.
	 *
	 *  @param oldVar  The shared state variable from the submodel
	 *  @param newVar  The newly created state var in the composed model,
	 *                 with which the submodel's state var is shared
	 */
	void addSharingInfo(BaseStateVariableClass * oldVar, BaseStateVariableClass * newVar, BaseModelClass * parent = NULL);

	virtual void SetupActions();

	/**
	 *  Returns a list of the models have the given name
	 *  @param name  The name of the model being looked for.
	 *  @return A list of pointers to BaseModelClass objects having the given name
	 */
	virtual BaseModelClass ** ListOfModels(char * name);

	/**
	 *  Initializes the array of submodel state sizes as a speed
	 *  optimization (see calling function in Wrapper.h for more
	 *  details).
	 */
	void setupSubmodelStateSizes();

	/** Methods needed for the Symbolic SSG */

	class LocalState: public State {
	public:
		LocalState(GraphJoin* model);
		~LocalState();
		friend class GraphJoin;
		void print();
		int hash();
		int compare(State const & S) const;
		State const& operator=(State const& S);
		int fwriteState(FILE* f);
		int freadState(FILE* f);
		void* GetStatePtr();
	protected:
		void SetState(const BaseModelClass* model);
		void GetState(const BaseModelClass* model);
	private:
		char* state;
		int size;
	};

	virtual ModelType GetModelType();
	virtual int GetNumberOfChildren();
	virtual BaseModelClass* GetChild(int i);
	//virtual void GetLocalActions(slist<BaseActionClass*>& actions);
	//virtual void GetGlobalActions(slist<BaseActionClass*>& actions);
	virtual bool HasLocalState();
	virtual void GetLocalState(State* _LS);
	virtual void SetLocalState(const State* _LS);
	virtual void ConstructLocalState(State*& _LS);
	virtual void DestructLocalState(State* _LS);

protected:

	/** The size of the hash table used to store the actions
	 */
	int ActionNameTableSize;

	/** The hash table storing the actions
	 */
	std::vector<const char *> ActionNameHashTable;
};

#endif

