/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: Wrapper.h
 *
 *      Authors: Aaron Stillman, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef WRAPPER_H_
#define WRAPPER_H_

#include "BaseGraphComposerClass.h"
#include "../BaseClasses/primes.h"
#include <vector>

/**  This is the wrapper class, derived from the base composer class.
 *  It is used to create wrapper nodes in a composed model
 *
 *  @author Aaron J. Stillman
 *  @version Mobius Version 1.0
 *
 */
class Wrapper: public BaseGraphComposerClass {
public:
	/** The wrapper constructor
	 *
	 * @param TheName The name given to the wrapper node
	 * @param NumModels The number of submodels at the wrapper node
	 * @param NumSharedVariables The number of shared variables at the wrapper node
	 * @param NumJoins The number of joins in the composed model
	 * @param SharedVariableNames An array of strings specifying the names of the shared variables
	 *
	 */
	Wrapper(char *, int, int, int, char **);

	/* The class destructor
	 */
	~Wrapper();

	/** The joins in this composed model
	 */
	BaseGraphComposerClass ** JoinArray;

	/** The number of joins
	 */
	int NumJoins;

	/** An array of integers giving the cardinalities of the actions
	 */
	int * ActionNameCardinalities;

	/** An array giving the shared state size (in bytes) of each join
	 */
	int * JoinSharedStateSize;

	/** An array giving the state sizes (in bytes) of each submodel
	 */
	int * SubmodelStateSize;

	/** Utility variable used for searching the action hash table
	 */
	bool LookForward;

	/** Various utility variables
	 */
	unsigned int currentStateIndex, currentStateCounter;
	unsigned int stateSizeCounter, stateSizeTotal;
	unsigned int hashActionsCounter, hashActionsIndex;
	unsigned int numActionsCounter, numActionsIndex;
	unsigned int numModelsCounter, numModelsIndex;
	unsigned int setStateIndex, setStateCounter;

	/** Returns the state size (in bytes) of the wrapper node
	 *
	 * @return The size, in bytes, of the wrapper's state
	 */
	virtual int StateSize();

	/** Sets the current state of the wrapper
	 *
	 * @param TheAddress The address in memory of the state to be assigned to the wrapper
	 *
	 */
	virtual void SetState(void * TheAddress);

	/** Initializes submodels and sets up the data structures of the wrapper node
	 */
	void Setup();

	/** Hashes the actions in to the action hash table
	 */
	void HashActions();

	/** Writes the current state of the wrapper node to a location in memory
	 *
	 * @param TheLocation The location where the state will be written
	 *
	 */
	virtual void CurrentState(void * TheLocation);

	/**
	 *  Will look to see if the state variable is the main variable
	 *  shared with a set of other state variables.  If so, it will
	 *  return a linked list of the variables that it is shared with.
	 *  This is used to build lists of equivalently shared state
	 *  variables in composed models.
	 *
	 *  @param var  The base state variable whose shared variables we want
	 *  @return A linked list of state variables that are shared with it.
	 */
	SharedStateVarLink * getListOfSharedVariables(BaseStateVariableClass * var);

	/**
	 *  Used to find the representative state variable that contains the state var
	 *  in its equivalent class
	 *
	 *  @param TheVariable  The state variable whose main variable is desired
	 */
	BaseStateVariableClass * getMainSharedVariable(BaseStateVariableClass * TheVariable);

	/**
	 *  Adds a new pointer to the shared pointers list for the new
	 *  shared state variable.
	 *
	 *  @param TheVariable  The base state variable
	 *  @param TheName  The name of the variable
	 */
	void addSharedPtr(BaseStateVariableClass * TheVariable, char *);

	/**
	 *  Adds the shared state variable pointer from the submodel to the
	 *  linked list of state variable pointers that have been shared
	 *  with the new state variable created in the composed node.
	 *
	 *  @param oldVar  The shared state variable from the submodel
	 *  @param newVar  The newly created state var in the composed model,
	 *                 with which the submodel's state var is shared
	 */
	void addSharingInfo(BaseStateVariableClass * oldVar, BaseStateVariableClass * newVar, BaseModelClass* TheParent = NULL);

	/**
	 *  Returns a pointer to the model with the given name
	 *
	 *  @param TheName  The name of the desired base model
	 */
	BaseModelClass * getModelNamed(char * TheName);

	/**
	 * Initializes action and group data structures.
	 */
	virtual void SetupActions();

	/**
	 *  This function initializes all of the join's submodel sizes
	 *  in an integer array (a speed optimization, where we don't
	 *  need to query the submodels on the fly for their state
	 *  sizes).  This needs to be done after all of the joins and
	 *  their state sharing is done, or else we can get the wrong
	 *  state size values for submodels.  Note that when a state
	 *  variable is shared in a submodel, it no longer counts it
	 *  as part of its state size.
	 */
	void setupJoinSubmodelStateSizes();

	/**
	 *  This function should not be called on the wrapper classes.
	 */
	virtual void setupSubmodelStateSizes();

	/**
	 *  Returns a list of the models have the given name
	 *  @param name  The name of the model being looked for.
	 *  @return A list of pointers to BaseModelClass objects having the given name
	 */
	virtual void listModels(char * name, List<BaseModelClass>* TheContainer);

	/** Methods needed for the Symbolic SSG */
	class LocalState: public State {
	public:
		LocalState(Wrapper* model);
		~LocalState();
		friend class Wrapper;
		void print();
		int hash();
		int compare(State const & S) const;
		State const& operator=(State const& S);
		int fwriteState(FILE* f);
		int freadState(FILE* f);
		void* GetStatePtr();
	protected:
		void SetState(const BaseModelClass* model);
		void GetState(const BaseModelClass* model);
	private:
		char* state;
		int size;
	};

	virtual ModelType GetModelType();
	virtual int GetNumberOfChildren();
	virtual BaseModelClass* GetChild(int i);
	//virtual void GetLocalActions(slist<BaseActionClass*>& actions);
	//virtual void GetGlobalActions(slist<BaseActionClass*>& actions);
	virtual bool HasLocalState();
	virtual void GetLocalState(State* _LS);
	virtual void SetLocalState(const State* _LS);
	virtual void ConstructLocalState(State*& _LS);
	virtual void DestructLocalState(State* _LS);

protected:

	/** The size of the hash table used to store the actions
	 */
	int ActionNameTableSize;

	/** The hash table storing the actions
	 */
	std::vector<const char *> ActionNameHashTable;

};

#endif

