/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: IntervalOfTime.hpp
 *
 *      Authors: Gerard P. Kavanaugh, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef INTERVAL_OF_TIME_HPP_
#define INTERVAL_OF_TIME_HPP_

#include "PerformanceVariableWorker.hpp"
#include "../BaseClasses/BaseActionClass.h"
#include "../BaseClasses/MobiusEnums.h"

// I know putting the following two functions in here is not the good way
// of doing it. I just temporarily put them in here until we decide about
// a file that we put these kinds of functions in.

#ifndef COMMON
#define COMMON

template<class T>
inline T Max(T a, T b) {
	if (a > b)
		return a;
	else
		return b;
}

template<class T>
inline T Min(T a, T b) {
	if (a < b)
		return a;
	else
		return b;
}

template<class T>
inline T CommonInterval(T x1, T x2, T y1, T y2) {
	return Max(0.0, Min(x2, y2) - Max(x1, y1));
}
#endif

/**
 *  This class defines the operation of a performance
 *  variable worker for an interval of time and
 *  time averaged interval of time performance
 *  variables.  An interval of time performance variable
 *  in a performance variable whose value is accumulated
 *  over and time interval defined by the Start and Stop
 *  values in PerformanceVariableWorker.  A time averaged
 *  interval of time is accumulated in the same fashion
 *  as an interval of time and its value is divide by
 *  the interval length.
 *
 *  @author Gerard P. Kavanaugh
 *  @version 1.0
 */
class IntervalOfTime: public PerformanceVariableWorker {
public:

	/**
	 *  Base constructor
	 */
	IntervalOfTime();

	/**
	 *  Overload constructor.
	 *  This constructor is used to create
	 *  a time-averaged interval of time
	 *  PVWorker (by passing in a true value);
	 *
	 *  @param IsTimeAveraged True implies that
	 *         the performance variable will be
	 *         calculated as a time averaged
	 *         interval of time value.  A false
	 *         value means that the reward will
	 *         be calculated as interval of time
	 *         performanc variable.
	 *
	 */
	IntervalOfTime(bool IsTimeAveraged);

	/**
	 *  Destructor
	 */
	~IntervalOfTime();

	/**
	 *  This is the normal accumulate method that accumulate
	 *  reward based upon the current simulation time and
	 *  and the last event time.
	 *
	 *  @param new_time The current time of the simulation
	 *  @param last_event_time The time of the previous
	 *         simulation event
	 */
	void Accumulate(double new_time, double last_event_time);

	/**
	 *  This method calculates the current value
	 *  of the reward function and saves the
	 *  result in the Reward data member.
	 */
	void CurrentReward(void);

	/**
	 *  This method resets the performance
	 *  variable such that it appears as it
	 *  would before simulation starts
	 *  (with reward value = 0, last modification
	 *  time = 0, etc.).
	 */
	void Reset(void);

	//virtual double Reward_Function(){//cout<<"!! IntervalOfTime::RewardFunction()"<<endl;
	//}

protected:

	/**
	 *  This data member stores the accumulation state
	 *  of the performance variable worker.  There
	 *  are three possible states: <BR>
	 *
	 *  <UL>
	 *  <LI> later -- the simulations has not reach the
	 *                start time of the interval
	 *  <LI> now   -- the simulation is currently inside
	 *                the accumulation interval
	 *  <LI> done  -- the simulation is past the accumulation
	 *                interval
	 *  </UL>
	 *
	 */
	AccumulationState Accumulating;

	/**
	 *  This data member donotes if the performance
	 *  variable value should be calculated as
	 *  a time-averaged interval of time value
	 *  or just a plain interval of time.  True
	 *  implies that the performance variable
	 *  is a time averaged interval of time
	 *  performance variable.
	 */
	bool TimeAveraged;

};

#endif //INTERVAL_OF_TIME_HPP_
