/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: IntervalOfTimeImpulse.hpp
 *
 *      Authors: Gerard P. Kavanaugh, Jay M. Doyle,,
 *                Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef INTERVAL_OF_TIME_IMPULSE_HPP_
#define INTERVAL_OF_TIME_IMPULSE_HPP_

#include "ImpulseNodeClass.hpp"
#include "../BaseClasses/BaseActionClass.h"
#include "../BaseClasses/MobiusEnums.h"

/**
 *  This is performance variable node class used
 *  to calculate the impulse reward values for
 *  interval of time and time averaged interval
 *  of time performance variables.
 *
 *  @author Gerard P. Kavanaugh
 *  @author Jay M. Doyle
 *  @version 1.0
 */
class IntervalOfTimeImpulse: public ImpulseNodeClass {

public:

	/**
	 *  Base constructor.  By default this
	 *  creates an interval of time impulse
	 *  reward.
	 */
	IntervalOfTimeImpulse();

	/**
	 *  Overloaded constructor.  Depending on the
	 *  value of the input parameter this constructor
	 *  creates either a interval of time impulse
	 *  performance variable or a time averaged
	 *  interval of time impulse reward.
	 *
	 *  @param IsTimeAveraged If true the the impulse reward
	 *         will be calculated as a time average reward
	 */
	IntervalOfTimeImpulse(bool IsTimeAveraged);

	/**
	 *  Destructor
	 */
	~IntervalOfTimeImpulse();

	/**
	 *  This method accumulates the reward variable
	 *  impulse value.
	 *
	 *  @param new_time The current simulation time
	 *  @param last_event_time The time of the last simulation
	 *         event
	 */
	void Accumulate(double new_time, double last_event_time);

	/**
	 *  This method is used to calculate the final value
	 *  of the impulse reward.
	 *
	 *  @param finish_time The final stopping time of the
	 *         simulation
	 *  @param last_event_time The time of the previous
	 *         simulation event
	 */
	//  void EndAccumulate(double finish_time,
	//		     double last_event_time);

	/**
	 *  This method calculates the current value
	 *  of the impulse reward and saves the value
	 *  in the Impulse data member.
	 */
	void CurrentImpulse(void);

	/**
	 *  This apparently does nothing ...
	 */
	void completeAccumulation(double);

	/**
	 *  This method resets the value of the
	 *  impulse reward to zero.
	 */
	void Reset(void);

protected:

	/**
	 *  This data member stores the accumulation state
	 *  of the performance variable worker.  There
	 *  are three possible states: <BR>
	 *
	 *  <UL>
	 *  <LI> later -- the simulations has not reach the
	 *                start time of the interval
	 *  <LI> now   -- the simulation is currently inside
	 *                the accumulation interval
	 *  <LI> done  -- the simulation is past the accumulation
	 *                interval
	 *  </UL>
	 *
	 */
	AccumulationState Accumulating;

	/**
	 *  This data member donotes if the performance
	 *  variable value should be calculated as
	 *  a time-averaged interval of time value
	 *  or just a plain interval of time.  True
	 *  implies that the performance variable
	 *  is a time averaged interval of time
	 *  performance variable.
	 */
	bool TimeAveraged;

}; // end IntervalOfTimeImpulse class

#endif //INTERVAL_OF_TIME_IMPULSE_HPP_
