/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: PVModel.hpp
 *
 *      Authors: Gerard P. Kavanaugh, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef PVMODEL_HPP_
#define PVMODEL_HPP_

#include "PerformanceVariableWorker.hpp"
#include "ModelGroupClass.hpp"
#include "../BaseClasses/BaseModelClass.h"
#include "../BaseClasses/BaseActionClass.h"
#include "../BaseClasses/BaseGroupClass.h"
#include "../BaseClasses/BaseStateVariableClass.h"
#include "../BaseClasses/MobiusTypeDefs.h"
#include "PerformanceVariableNode.hpp"
#include <iosfwd>
#include "../BaseClasses/debug.h"
#include <cstdlib>
#include <vector>


/**
 * This model is the top model.  It contains all the performance
 * variables defined on the model as well as functionality for
 * determining their instant and accumulated values whether the
 * solution mechanism is through analysis or simulation.  It also
 * contains the user specified model, perhaps a composed model, and if
 * the model collector is being used it may contain other PVModels.
 *
 */
class PVModel: public BaseModelClass {

public:

	/**
	 * The list of performance variables defined on the underlying model.
	 */
	PerformanceVariableNode **PerformanceVariableList;

	/**
	 * The number of performance variables defined on
	 * the underlying model.
	 */
	int Number_Of_Performance_Variables;

	/**
	 * A flag indicating whether the variable values ar being traced or not.
	 */
	int Trace;

	/**
	 * A pointer to the underlying model.
	 */
	BaseModelClass *TheModel;

	/**
	 * A pointer which "remembers" the last action to fire so that it's
	 * fired property may be reset when the next action fires.
	 */
	BaseActionClass *LastAction;

	/**
	 *
	 */
	BaseActionClass **PVListOfActions;

	/**
	 * Latest time which influences a performance variable.
	 * It s used by the simulator to tell when simulation should end.
	 */
	double Simulation_Stop_Time;

	/**
	 *  The minimum of all the performance variable
	 *  start times? (JMD)
	 */
	double AccumulateStartTime;

	/**
	 *  The maximum of all the performance variable
	 *  end times? (JMD)
	 */
	double AccumulateStopTime;

	/**
	 *
	 */
	BaseActionClass **AffectingActions;

	/**
	 * an array of pointers to FastSets
	 */
	FastSet** ListOfAffectedLists;

	/**
	 *
	 */
	int NumAffectingActions;

	/**
	 *
	 */
	int NumActionsWithImpulseAffects;

	/**
	 *
	 */
	int TotalNumAffectingActions;

	/**
	 *
	 */
	int PVStateIndexLength;

	/**
	 *
	 */
	int ModelStateSize;

	/**
	 *
	 */
	int PVModelStateSize;

	/**
	 *
	 */
	int AdditionalStateSize;

	/**
	 *
	 */
	int StateMode;

	/**
	 *
	 */
	void **PVStateIndex;

	/**
	 *  From what I can deduce from the code
	 *  this data member is a dummy action that
	 *  is used when that action that just fired
	 *  did not have an impulsed defined on it.
	 *  (JMD)
	 */
	BaseActionClass ZeroImpulseAction;

	/**
	 *
	 */
	BaseActionClass StateLastAction;

	/************************METHODS***********************/

	/**
	 *
	 */
	void CurrentState(void*);

	/**
	 *
	 */
	BaseActionClass **ListOfActions(void);

	/**
	 *
	 */
	void SetState(void *TheState);

	/**
	 *
	 */
	int StateSize(void);

	/**
	 *
	 */
	BaseActionClass *FireAction(BaseActionClass *, double);

	/**
	 *
	 */
	void FireAction(void *, BaseActionClass *);

	/**
	 *
	 */
	void AgeAction(char *, BaseActionClass *);

	/**
	 *
	 */
	void RemakeAffectedLists(void);

	/**
	 *
	 */
	void Accumulate(double);

	/**
	 * 	The simulator calls this method if the user has
	 *   specified a time to terminate simulation which is
	 *   earlier than Simulation_Stop_Time.  It finishes the
	 *   accumulation of any variables which have not yet been
	 *   completely determined.
	 */
	void EndAccumulate(double);

	/**
	 *
	 */
	PVModel();

	/**
	 *
	 */
	virtual ~PVModel();

	/**
	 *
	 */
	void PVReset();

	/**
	 *
	 */
	bool CompareState(void*, void*);

	/**
	 *
	 */
	void SteadyState();

	/**
	 *
	 */
	int StateIndexLookup(void*);

	/**
	 *
	 */
	void listGroups(List<BaseGroupClass>* TheContainer);

	/**
	 *
	 */
	int BytesFor(int);

	/**
	 *
	 */
	void PutPVState(void*);

	/**
	 *
	 */
	void GetPVState(void*);

	/**
	 *  This method has to be called some time after initialization of
	 *  the PV model and asks the PVModel to only process those PV nodes
	 *  whose indices are given in NodeIndices array. This is to avoid
	 *  performing Node->Accumulating test against acc_ignore for each
	 *  node before we process it.
	 */
	void ConsiderOnly(int* NodeIndices, int size);

	/**
	 *
	 */
	void printState(std::ostream *strm);

	/**
	 *  This method returns a list of all the actions defined
	 *   within the model
	 *
	 *  @return An array of pointers to all the actions defined
	 *     in the model
	 */
	void listActions(List<BaseActionClass>* TheContainer) override;

	/**
	 *  This method list all the actions with
	 *  the specificed name that are part of this
	 *  model.
	 *
	 *  @param TheActionName The name of an action in the model
	 *  @param TheContainer A list data structure that maintains
	 *   the list of all the actions that have the specified name.
	 */
	void listActions(char* TheActionName, List<BaseActionClass>* TheContainer) override;

	/**
	 *  This method returns a list of models with the specificed
	 *  name.
	 *
	 *  @param TheModel The name of a contained model
	 *  @param TheContainer A list data structure that maintains
	 *    the list of models with the specified name
	 */
	void listModels(char* TheModelName, List<BaseModelClass>* TheContainer);

	/**
	 *  This method creates a list of pointers to state variables who
	 *  belong to the specified model and whose name is the same as the
	 *  specified place name.
	 *
	 *  @param TheSpecifiedModelName The name of the model in which to
	 *         search for the state variable
	 *  @param TheSpecifiedStateVariableName The name of the state variable
	 *  @param TheSVContainer A data structure containing an
	 *         array of pointers to state variables
	 *  @param TheModelContainer A data structure containing an
	 *         array of pointer to models in which the
	 *         state variables listed in TheSVContainer are contained.
	 */
	void listSVs(char* TheSpecifiedModelName, char* TheSpecifiedStateVariableName, List<BaseStateVariableClass>* TheSVContainer,
			List<BaseModelClass>* TheModelContiner);

	/**
	 *
	 */
	int CountAffectedVars(char*, char*);

	/**
	 *  this method writes the pv type file used by numerical solvers to
	 *  determine the PV type and times for PV in the state space files.
	 *  this method should be called by any ssg that desires this file to
	 *  be created.  support for reading this file is included in the
	 *  BaseStateLevel objects.
	 */
	void WritePVTypeFile(char *filename);

	//(derisavi) I've changed any locally-used member from public
	//to protected
protected:
	virtual PerformanceVariableNode* createPVNode(int pvindex, int timeindex)=0;
	/**
	 *
	 */
	void Initialize();
	void CreatePVList(int numPVs, bool expandtimearrays);

	/**
	 *  Deallocates variables allocated in Initialize()
	 */
	void Destruct();

	/**
	 *
	 */
	void CreateTotalPVAffectsList(BaseActionClass*);

	/**
	 *
	 */
	void MakeAffectedLists();

	/**
	 *
	 */
	void MakeImpulseAffectedLists();

	/**
	 *
	 */
	void CreateAffectedLists();

	/**
	 *
	 */
	void CreatePVStateIndex();

	/**
	 * This list is built in ConsiderOnly method and is a vector of
	 * pointers to all nodes that should be processed. See ConsiderOnly
	 * method for more info.
	 */
	std::vector<PerformanceVariableNode*> RelevantNodes;

	/*
	 void setupPredicates(void);
	 */

}; // end PVModel class

#endif // PVMODEL_HPP_
