/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: PerformanceVariableNode.hpp
 *
 *      Authors: Gerard P. Kavanaugh, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef PV_NODE_HPP_
#define PV_NODE_HPP_

#include "ImpulseNodeClass.hpp"
#include "ModelGroupClass.hpp"
#include "PerformanceVariableWorker.hpp"
#include "../BaseClasses/BaseModelClass.h"
#include "../BaseClasses/BaseActionClass.h"
#include "../BaseClasses/BaseSyncActionClass.h"
#include "../BaseClasses/debug.h"
#include "VarModelPairClass.hpp"
#include "ActionModelPairClass.hpp"


/**
 *
 */

#ifndef COMMON
#define COMMON

template <class T>
inline T Max(T a, T b) {
	if (a> b)
	return a;
	else
	return b;
}

template <class T>
inline T Min(T a, T b) {
	if (a < b)
	return a;
	else
	return b;
}

template <class T>
inline T CommonInterval(T x1, T x2, T y1, T y2) {
	return Max(0.0, Min(x2, y2) - Max(x1, y1));
}

#endif

class PerformanceVariableNode {

	/**********************DATA MEMBERS********************/

public:

	/**
	 *
	 */
	BaseModelClass **TheModelPtr;

	/**
	 *
	 */
	char *Name;

	/**
	 *
	 */
	double TotalReward;

	double TotalRewardSqr;

	/**
	 *
	 */
	double Start;
	/**
	 *
	 */
	double Stop;

	int NumTimes;
	double *StartList;
	double *StopList;
	/**
	 *
	 */
	double TotalImpulse;
	/**
	 *
	 */
	double TotalImpulseSqr;

	/**
	 *
	 */
	double batchSize;

	/**
	 *
	 */
	PVType Type;

	RewardType PVRewardType;

	/**
	 *
	 */
	int Accumulating;

	/**
	 *  give each pv node a unique id
	 */
	int id;
	// global id, so each node gets a unique id
	static int the_id;

	/**
	 *
	 */
	int Trace;

	//Reward Function data members

	/**
	 *
	 */
	BaseActionClass **ActionDependencies;

	/**
	 *
	 */
	int NumberOfActionDependencies;

	/**
	 *
	 */
	ModelGroupClass *ModelList;

	/**
	 *
	 */
	int NumberOfWorkers;

	/**
	 *
	 */
	PerformanceVariableWorker **WorkerList;

	/**
	 * list of (state variable name, model name) pairs that the PV node
	 * depends on
	 */
	VarModelPairClass *VarNamesList;

	/**
	 * the number of elements in VarNamesList
	 */
	int NumberOfVarNames;

	//impulse function data members

	/**
	 * The number of impulse rewards of the PV node
	 */
	int NumberOfImpulses;

	/**
	 *
	 */
	int *NumberOfImpulseModelNames;

	/**
	 * An array that holds the number of impulse workers for each
	 * impulse reward of the PV node. Its size is NumberOfImpulses.
	 */
	int *NumberOfImpulseWorkers;

	/**
	 *
	 */
	int *NumberOfImpulseActions;

	/**
	 *
	 */
	ImpulseNodeClass **ImpulseNodeList;

	/**
	 *
	 */
	ImpulseNodeClass ***ImpulseWorkersList;

	/**
	 *
	 */
	ActionModelPairClass *ActionNamesList;

	/**
	 *
	 */
	char ***ImpulseModelNamesList;

	/**
	 *
	 */
	ModelGroupClass **ImpulseModelList;

	/*
	 //predicate function data members

	 char **ModelsWithPredicates;
	 BaseModelClass ***ListOfModelsWithPredicates;
	 int *NumModelsWithPredicates;
	 BasePredicateClass **PredicateList;
	 int NumPredicates;
	 */
	/************************METHODS***********************/

	/**
	 *  initialize this pv for a given time point and pv type
	 */
	void Initialize(const char*, RewardType, int, double*, double*, int, int, int, int);

	/**
	 *  initialize an array of start and stop time points.  for instants
	 *  of time, the start and stop points should be the same.
	 */
	void InitializeTimeArrays(int NumPts, double *StartPts, double*StopPts);
	/**
	 *
	 */
	PerformanceVariableNode(void);

	/**
	 *
	 */
	virtual ~PerformanceVariableNode(void);

	/**
	 * Accumulate is called at the end of a state (possibly a stable
	 * state), before the next transition occurs. This allows us to
	 * calculate all of the appropriate rewards on the previous interval
	 * of time.
	 *
	 * @param new_time The current time in the simulation
	 * @param last_event_time The time that the last event occured
	 * @param LastAction Pointer to the last action that completed,
	 * needed for impulse rewards
	 */
	void Accumulate(double, double, BaseActionClass*);

	/**
	 *
	 */
	void EndAccumulate(double, double, BaseActionClass*);

	/**
	 *
	 */
	void UpdateAccumulate(double);

	/**
	 * Initialized the model pointer, computes
	 * the number of workers, and creates the worker list
	 *
	 */
	void FixReferences(BaseModelClass*);

	/**
	 *
	 */
	void CurrentReward(BaseActionClass*);

	/**
	 *
	 */
	void UpdateDependencies(void);

	/**
	 *
	 */
	void DefineName(char*);

	/**
	 *
	 */
	void Operation(double*, double);

	//state variable methods

	/**
	 *
	 */
	void CreateModelGroupList(void);

	/**
	 *
	 */
	virtual void CreateWorkerList(void) = 0;

	void AddPVWorkersToAction(BaseActionClass* Action, BaseModelClass* Model);
	/**
	 *
	 */
	virtual void Reset(void);

	/**
	 *
	 */
	void TotalAccumulation(void);

	/**
	 *
	 */
	void AddVariableDependency(char*, char*);

	/**
	 *
	 */
	void FixWorkerList(void);

	//impulse methods

	/**
	 *
	 */
	void AddImpulse(char*, char*, ImpulseNodeClass*);

	/**
	 *
	 */
	void AddImpulseActionDependency(char*, char*, int);

	/**
	 *
	 */
	void AddImpulseModelDependency(char*, ImpulseNodeClass*);

	/**
	 *
	 */
	void AddImpulseModelDependency(char*, int);

	/**
	 *
	 */
	void CreateImpulseWorkerLists(void);

	/**
	 *
	 */
	int GetImpulseNumber(ImpulseNodeClass*);

	/**
	 *
	 */
	void CreateImpulseModelGroupList(void);

	/**
	 *
	 */
	void FixImpulseWorkerLists(void);

	/**
	 *  get the reward type defined for this PVNode
	 *  (see RewardType enum in MobiusEnums.h)
	 */
	RewardType GetRewardType(void);

	/**
	 *  returns the number of time points that are defined for this pv node
	 */
	int GetNumTimePoints();
	/**
	 *  return the start time for the specific time point
	 */
	double GetStartTime(int index);
	/**
	 *  return the stop time for the specific time point
	 */
	double GetStopTime(int index);

	/*
	 //predicate methods

	 void addPredicate(char*,BasePredicateClass*);
	 void checkAllModelPredicates(void);
	 */

	//(derisavi) I've changed any locally-used member from public
	//to protected
protected:
	/**
	 *
	 */
	bool ImpulseAffectedBy(BaseActionClass*);

	/**
	 *
	 */
	void AddModelDependency(char*);

	/**
	 *
	 */
	bool IsDependentOnModel(char*);

	/**
	 *
	 */
	bool IsDependentOnImpulseModel(char*, int);

	/**
	 *  Computes the number of workers of the PV node and assigns it to
	 *  NumberOfWorkers
	 */
	void GetNumberOfWorkers(void);

	/**
	 * Computes the number of impulse workers for each impulse reward of
	 * the PV node and puts the results in NumberOfImpulseWorkers array
	 */
	void GetNumberOfImpulseWorkers(void);

	/**
	 * The set (i.e., no duplicates) of names of models that have
	 * state variables that the PV node depends on.
	 */
	char **ModelNamesList;

	/**
	 * The number of elements of ModelNamesList
	 */
	int NumberOfModelNames;

private:
	void CommonAccumulate(double next_tiem, double last_event_time, BaseActionClass* LastAction);

}; // end PerformanceVariableNode class


#endif //PV_NODE_HPP_
