/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: PerformanceVariableWorker.hpp
 *
 *      Authors: Gerard P. Kavanaugh, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef PV_NODE_WORKER_HPP_
#define PV_NODE_WORKER_HPP_

#include "../BaseClasses/BaseModelClass.h"
#include "../BaseClasses/MobiusEnums.h"
#include "../BaseClasses/observer_ptr.hpp"
//#include "PerformanceVariableNode.hpp"
#include <iostream>

class PerformanceVariableNode;

/**
 *   The performance variable worker class is
 *   a class that encapsulates the functionality
 *   needed to calculate the appropriate reward
 *   variable for performance variable type rewards.
 *   This is an abstract class that is used to create
 *   derived class that correspond to the four types
 *   of performance variables:  instant of time,
 *   interval of time, time averaged interval of time,
 *   and steady state.  The interval of time and the
 *   time averaged interval of time performance variables
 *   are realized through the same class (@see IntervalOfTime)
 *   since they share much of the same functionality. (JMD)
 *
 *   @author Gerard P. Kavanaugh
 *   @author Jay M. Doyle
 *   @version 1.0
 */
class PerformanceVariableWorker {

public:

	/************************METHODS***********************/

	/**
	 *  Base constructor
	 */
	PerformanceVariableWorker(void);

	/**
	 *  Destructor
	 */
	virtual ~PerformanceVariableWorker(void);

	/**
	 *  The function that defines the value
	 *  of the reward function in the current
	 *  model state. (?)
	 */
	virtual double Reward_Function(void) = 0;

	/**
	 *
	 */
	virtual void Accumulate(double, double) = 0;

	/**
	 *  I assume that this method calculates the reward
	 *  for the current state and saves the value
	 *  in the Reward data member.
	 */
	virtual void CurrentReward(void) {
		std::cerr << "!! Calling PerformanceVariableWorker::CurrentReward()" << std::endl;
	} // end CurrentReward

	/**
	 *  This method should reset the value of the
	 *  Reward to 0.
	 */
	virtual void Reset(void);

	/**
	 *  Defines the name of ther PerformanceVariableWorker
	 *  and sets the value of the Number data member
	 *
	 *  @param TheName The name of the performance variable
	 *         worker
	 *  @param TheNumber The new value for the Number data
	 *         member
	 */
	void DefineName(char* TheName, int TheNumber);


	/**
	 * 	Sets the pointer to the PerofrmanceVariableNode instance that controls this particular PerformanceVariableWorker.
	 *
	 * 	@param ptr The pointer of the PerformanceVariableNode which owns this worker.
	 */
	void SetPerformanceVariableNodePtr(PerformanceVariableNode *ptr);


	/*
	 * 	returns the TotalReward data member of the PerformanceVariableNode that owns this worker
	 */
	virtual double getTotalReward();


	/**********************DATA MEMBERS********************/

	/**
	 *
	 */
	BaseModelClass ***TheModelPtr;

	/**
	 * 	Pointer to the PerformanceVariableNode that controls this worker
	 */

	//PerformanceVariableNode *PerformanceVariableNodePtr;
	mobius::observer_ptr<PerformanceVariableNode> PerformanceVariableNodePtr;

	/**
	 *
	 */
	int NumModels;

	/**
	 *
	 */
	int lastQualification;

	/**
	 *  The name of the PVWorker
	 */
	char *Name;

	/*
	 *
	 */
	int Trace;

	/**
	 *
	 */
	int Number;

	/**
	 *
	 */
	int id;

	/**
	 *  This is the accumulated reward value up to
	 *  the current simulation time.
	 */
	double Reward;
	double RewardSqr;
	double RewardFraction; // function to track the reward fraction

	/**
	 *  The start of the interval over which
	 *  the reward is accumulated.
	 */
	double Start;

	/**
	 *  The end of the interval over which
	 *  the reward is accumulated
	 */
	double Stop;

protected:

	/*
	 int ModelsQualify(void);
	 int shouldAccumulate(double,double);
	 virtual void disqualify(double,double) = 0;
	 */

};

#endif //PV_WORKER_HPP_
