/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: Distributions.h
 *
 *      Authors: Alex Williamson, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef DISTRIBUTIONS_CLASS
#define DISTRIBUTIONS_CLASS

#include "RandomNumberGenerator.h"
#include "LaggedFibonacci.h"
#include "Tausworthe.h"
#include "MersenneTwister.h"
#include <memory>


/**
 * This class is the base class for numerical distributions.  See "Discrete Event
 * Simulation in the Mobius Modling Framework" for parameters.
 *
 * @author Alex Williamson
 * @version Mobius v 1.0
 */

class Distributions {

public:

	/**
	 *  Constructor
	 */
	Distributions();

	/**
	 *  Destructor
	 */
	~Distributions();

	/**
	 *  Samples from an exponential distribution with
	 *  the specified rate.  The rate must be greater
	 *  than zero.
	 *
	 *  @param The rate of the exponentially distributed
	 *         random variable.
	 */
	double Exponential(double rate);

	/**
	 *  Returns the deterministic value.
	 *
	 *  @param The value of the deterministic
	 *  distribution.
	 */
	double Deterministic(double value);

	/**
	 *  Returns the a sample from the geometrically
	 *  distributed random variable with interval
	 *  probability p.  This is a discrete time version
	 *  of the exponential distribution.
	 *
	 *  @param p The probability that next interval will
	 *   be selected.
	 */
	double Geometric(double p);

	/**
	 *  Samples from a Weibull distribution with
	 *  parameters alpha and beta.
	 *
	 *  @param alpha First parameter of the Weibull distribution
	 *  @param beta  Second parameter of a Weibull distribution
	 */
	double Weibull(double alpha, double beta);

	/**
	 *  Samples from a Weibull distribution given that t time
	 *  has already passed.  The sampled value from the distribution
	 *  will be the time to completion measure from the current time
	 *  (i.e., the sampled value can be < than t).
	 *
	 *  @param alpha First parameter of the Weibull distribution
	 *  @param beta  Second parameter of the Weibull distribution
	 *  @param t The amount of time that has already passed.
	 */
	double ConditionalWeibull(double alpha, double beta, double t);

	/**
	 *  Returns a sample from a Normal (Gaussian, bell-shaped)
	 *  distributed random variable with the specified mean
	 *  and variance.
	 *
	 *  @param mean The mean value of the distribution
	 *  @param variance The second centered momement of the
	 *         distributione
	 */
	double Normal(double mean, double variance);

	/**
	 *   Returns a sample from the lognormally
	 *   distributed random variable.
	 *
	 *   @param mu  A parameter for the lognormal distribution
	 *   @param alphasqr A parameter for the lognormal distribution
	 */
	double Lognormal(double mu, double alphasqr);

	/**
	 *   Returns a sample from a Erlangian distributed
	 *   random variable.  Erlang distributions are
	 *   characterized by the sum of series of exponentially
	 *   distributed random variables (all with the same rate).
	 *
	 *   @param m The number of exponential stages (should be an integer)
	 *   @param beta The rate of the independent exponentially distributed
	 *     random variable at each stage
	 */
	double Erlang(double m, double beta);

	/**
	 *  Returns a sample from a triangularly distributed
	 *  random variable.  A triangle distributed random
	 *  variable has a density function that looks like
	 *  a triangle.
	 *
	 *  @param a The lower bound of the triangle such that
	 *         F(a) == 0.
	 *  @param b The upper bound of the triangle distribution
	 *         such that F(b) == 1.
	 *  @param c The point at which the probability density function
	 *         has the highest value.
	 */
	double Triangular(double a, double b, double c);

	/**
	 *  Returns a sample from a gamma distributed random
	 *  variable.  The gamma distribution is a more general
	 *  form of the Erlang distribution in which the number
	 *  of independent exponentially distributed stages can
	 *  be any positive real number.
	 *
	 *  @param alpha The number of exponential stages (any positive
	 *         real number)
	 *  @param beta  The rate of the exponential stages
	 */
	double Gamma(double alpha, double beta);

	/**
	 *  Returns a sample from a beta distributed random
	 *  variable with parameters alpha1 and beta1.
	 *
	 *  @param alpha1 A parameter of the beta distribution
	 *  @param beta1 A parameter of the beta distribution
	 */
	double Beta(double alpha1, double beta1);

	/**
	 *  Returns a sample from a uniformally distributed
	 *  random variable.  Uniformally distributed random
	 *  variable model random variables for which all variables
	 *  with a certain range are equally probable.
	 *
	 *  @param lbound The lower bound of the uniform distribution
	 *  @param ubound The upper bound of the uniform distribution
	 */
	double Uniform(double lbound, double ubound);

	/**
	 *  Returns number of success in t independent Bernoulli
	 *  trials with probability p.
	 *
	 *  @param t The number of Bernoulli trials
	 *  @param p The probability of success for each
	 *         Bernoulli trial.
	 */
	double Binomial(double t, double p);

	/**
	 *  Returns the number of failures before the sth
	 *  success in a sequence of independent Bernoulli
	 *  trials with probability p of success on each
	 *  trial.
	 *
	 *  @param s The end of the Bernoulli trials (should
	 *         be an integer).
	 *  @param p The probability of success for each
	 *         Bernoulli trial.
	 */
	double NegativeBinomial(double s, double p);

	/**
	 *  Returns a sample from a hyperexponential random variable.  A
	 *  hyperexponential random variable is characterized by a
	 *  probabilitistic selection among 'stages' exponential random
	 *  variables.
	 *
	 *  @param stages The number of exponential random variables
	 *  @param params the parameters of the exponential random variables
	 *         given in the order rate0, p0, rate1, p1, ...
	 */
	double HyperExponential(int phases, double* params);

	/**
	 *  Returns a sample from a hypoexponential random variable.  A
	 *  hypoexponential random variable is the sum of a number of
	 *  independent exponential random variables.
	 *
	 *  @param phases the number of exponential random variables
	 *  @param params the rates of the exponential random variables
	 *         given in the order rate0, rate1, ...
	 */
	double HypoExponential(int phases, double* rates);

	/**
	 *  Returns a sample from a Pareto random variable.
	 *
	 *  @param xm the minimum value of the random variable (>0)
	 *  @param k  the shape of the pareto distribution (>0)
	 *
	 */
	double Pareto(double xm, double k);

	/**
	 * The RNG pointer may be set to any class derived
	 * from RandomNumberGenerator.
	 */
	std::unique_ptr<RandomNumberGenerator> RNG;

};

#endif
