/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: EventList.h
 *
 *      Authors: Salem Derisavi, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef EVENTLIST_CLASS
#define EVENTLIST_CLASS

#include "../BaseClasses/BaseActionClass.h"
#include "Event.h"
#include "PAddress.h"
#include <list>
#include <vector>
#include <iostream>

/**
 * @author Salem Derisavi
 * @version Mobius v 1.0
 */

/**
 * This is the data structure used for the FEL (future event list). It's implemented
 * as a heap data structure. The important operations of this data structure are :
 *  1) Insertion : Inserts a new element in the data structure
 *  2) Deletion : Remove an element in the data structure assuming we have a pointer
 *                to that element
 *  3) ModifyKey : Modifies the key of the element assuming we have a pointer
 *                 to the element. ModifyKey is faster than a deletion followed by an
 *                 insertion.
 *  4) Earliest : Returns a pointer to the highest priority element in the heap
 *                (which is the earliest event in the FEL in this case)
 *
 *  Key in FEL is the pair of the "execution time" and "rank" of the event.
 *
 *  Helper functions used to implement the above methods are :
 *     BubbleUp(i): Replaces the ith element with its parent if it has more priority
 *                  than its parent
 *     BubbleDown(i): if the ith children has any children, it replaces it with the
 *                    child has is more prior than the other (if any) also is more
 *                    prior than the ith element (its parent)
 *       BubbleDown and BubbleUp are used to keep the data structure proper after
 *       insertion, deletion and key modification.
 *     Swap(i, j): swap the contents of the ith and the jth element of the heap.
 *
 *    The elements of the heap are actually pointers to Event objects not Event
 *  themselves. This way swapping two elements will be more efficient.
 *    As you can see in Event.h each event has exec_time, start_time and action_ptr
 *  as its data members. action_ptr points to the action (of type BaseActionClass)
 *  corresponding to that event. BaseActionClass has a data member, EventPointer,
 *  of type void* that is supposed to use to keep track of the event corresponding
 *  to the action.
 *    In the current implementation it points to objects of type PAddress (declared
 *  below) which has two essential data members:
 *    1) event : which points to the Event pointer pointing to the event corresponding
 *               to the action. (So it needs double dereferencing to access the event)
 *               The reason why it's pointer to pointer to Event is that the heap
 *               itself stores pointers to Event objects as I mentioned above.
 *    2) react : An iterator (a pointer-like concept in STL) to the location of the
 *               action in the reactivation list. This way the action can be deleted
 *               from the list in constant time.
 *
 *    Having the above pointer structure we can access the event from the action and
 *  vice versa. We also can access the location of the action in the reactivation list
 *  from the event and the action.
 *
 *  If Earliest is called and there are more than one event with the same priority
 *  (i.e. with the same exec_time and rank) then FindEquals comes into play. It
 *  basically finds all the events that have the same priority and puts their
 *  indices in the heap structure in an array called Equalities. Afterward, an
 *  event is chosen from those based on their weight.
 *
 *    In order to make things even faster, Event objects are not allocated and
 *  deleted every time an event is inserted and removed. Instead there is a linked
 *  list of pointers to free Event objects (EventList::FreeList). These pointers
 *  are used and recycled everytime an element is inserted or removed. This linked
 *  list is first initialized with pointers to the maximum number of elements the
 *  heap can have. This is done in EventList constructor.

 *
 *    There are a few other methods that are described below:
 *
 *  ClearEventList() removes all the elements in the heap and recycles the Event
 *                   pointers to FreeList.
 *  PrintEventList() prints all the elements in the heap for debugging
 *                   (not implemented yet)
 *  IsValid is a debugging function that checks for the following condition to be
 *     hold:
 *      1) The execution time of every element (event) is greater than both of its
 *         children (heap property)
 *      2) For all the events e in the heap we have
 *         (**a.action_ptr->EventPointer->event) = a
 *      3) None of the pointers stored in the FreeList is really used by any
 *         element in the heap.
 *    IsValid return a specific value based on the type of error it finds in the
 *    data structure.
 **/

class EventList {
public:
	explicit EventList(int _capacity);
	~EventList();

	void Insert(Event const & event);
	void Insert(Event*);
	void Remove(void* event);
	void Remove(Event** event);
	void ModifyKey(Event** e);
	Event** Earliest();
	void ClearEventList();
	void PrintList(std::ostream& stream);

	// Used for testing the integrity of the heap data structure
	int IsValid();

private:
	void FindEquals(int);
	void BubbleDown(int i);
	int BubbleUp(int i);
	inline void Swap(int i, int j);

	EventEarlier EvEarlier;
	EventEqual EvEqual;

	// The free list is a list of pre-allocated or recycled Event objects.
	// We do this to avoid allocations during runtime.
	std::list<Event*> mFreeList;

	// The flat heap array represents a binary tree flatten into a linear array.
	// The root is the first item (index zero) in the array.  Given k-th element:
	//		parent is at index (k-1)/2
	//		left child is at index 2*k+1
	//		right child is at index 2*k+2
	std::vector<Event*> mFlatHeapArray;

	// The number of elements currently in the heap.  Note: this is not the allocated
	// size of the flat heap array.  You can find that out by using the size function
	// on the flat heap array container.
	int mNumElementInHeap;

	// The array containing indices of events with equal priorities
	std::vector<int> mEqualities;
};

inline void EventList::Swap(int i, int j) {
//	std::swap<Event*>(A[j], A[i]);
//	((PAddress*) A[j]->action_ptr->EventPointer)->event = &A[j];
//	((PAddress*) A[i]->action_ptr->EventPointer)->event = &A[i];
	std::swap(mFlatHeapArray[j], mFlatHeapArray[i]);
	((PAddress*) mFlatHeapArray[j]->action_ptr->EventPointer)->event = &mFlatHeapArray[j];
	((PAddress*) mFlatHeapArray[i]->action_ptr->EventPointer)->event = &mFlatHeapArray[i];
}

#endif
