/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef _HEAP_H_
#define _HEAP_H_

#include <utility>

using namespace std;

template<typename Key>
class Heap {
public:
	explicit Heap(int _capacity = 10);
	~Heap();
	void Insert(const Key& key, int index);
	void ModifyKey(int index, const Key& new_key);
	void Earliest(pair<Key, int>& earliest);
	const Key& getElement(int idx);

private:
	void BubbleDown(int i);
	int BubbleUp(int i);
	void Swap(int i, int j);

	int capacity; // The maximum number of elements of heap
	Key* A; // An array of pointers to Event objects
	//KeyPointers[idx] is the current position in array A where the element originally
	//stored with index idx is stored.
	int* KeyPointers;
	//KeyPointers[InvKeyPointers[i]] = i for any 0 <= i < size
	int* InvKeyPointers;
	int MaxIndex;
	int size; // The number of the events in the heap
};

template<typename Key>
Heap<Key>::Heap(int _capacity) {
	capacity = _capacity;
	MaxIndex = _capacity;
	size = 0;
	A = new Key[capacity];
	KeyPointers = new int[MaxIndex];
	InvKeyPointers = new int[capacity];
	memset(A, 0, sizeof(Key) * capacity);
	memset(KeyPointers, 0, sizeof(int) * MaxIndex);
	memset(InvKeyPointers, 0, sizeof(int) * capacity);
}

template<typename Key>
Heap<Key>::~Heap() {
	delete[] A;
	delete[] KeyPointers;
	delete[] InvKeyPointers;
}

/**
 * When the key of an element is changed in the heap data structure,
 * it should either go higher or lower in the tree. In other words it
 * does not need to change its branch in the tree. That's why
 * ModifyKey is so simple to implement.
 **/
template<typename Key>
void Heap<Key>::ModifyKey(int index, const Key& new_key) {
	int indexA = KeyPointers[index];
	A[indexA] = new_key;
	if (BubbleUp(indexA) == indexA)
		BubbleDown(indexA);
}

/**
 * If the ith element has less priority than either of its children, it swaps it
 * with the one with more priority.
 **/
template<typename Key>
void Heap<Key>::BubbleDown(int i) {
	int SwapWith;
	int temp = (i << 1) + 1;
	while (temp < size) {
		SwapWith = temp;
		if ((temp + 1 < size) && (A[temp + 1] < A[temp]))
			SwapWith = temp + 1;
		if (A[SwapWith] < A[i]) {
			Swap(SwapWith, i);
			i = SwapWith;
			temp = (i << 1) + 1;
		} else
			break;
	}
}

template<typename Key>
void Heap<Key>::Insert(const Key& elem, int index) {
	// Put the new element at the end of the heap and bubble it up
	if (size == capacity) {
		Key* new_A = new Key[capacity * 2];
		memcpy(new_A, A, sizeof(Key) * capacity);
		delete[] A;
		A = new_A;

		int* new_InvKeyPointers = new int[capacity * 2];
		memcpy(new_InvKeyPointers, InvKeyPointers, sizeof(int) * capacity);
		delete[] InvKeyPointers;
		InvKeyPointers = new_InvKeyPointers;
		capacity *= 2;
	}

	if (index >= MaxIndex) {
		int newMaxIndex = MaxIndex;
		while (index >= newMaxIndex)
			newMaxIndex *= 2;
		int* new_KeyPointers = new int[newMaxIndex];
		memcpy(new_KeyPointers, KeyPointers, sizeof(int) * MaxIndex);
		delete[] KeyPointers;
		MaxIndex = newMaxIndex;
		KeyPointers = new_KeyPointers;
	}

	A[size] = elem;
	KeyPointers[index] = size;
	InvKeyPointers[size] = index;
	BubbleUp(size++);
}

/**
 * Keeps swapping the ith element with its parent until the parent is
 * more prior or it gets to the root of the heap.
 **/
template<typename Key>
int Heap<Key>::BubbleUp(int i) {
	while (1) {
		int parent = (i - 1) >> 1;
		if ((parent >= 0) && (A[i] < A[parent])) {
			Swap(i, parent);
			i = parent;
		} else
			break;
	}
	return i;
}

template<typename Key>
void Heap<Key>::Swap(int i, int j) {
	std::swap<Key>(A[i], A[j]);
	std::swap<int>(KeyPointers[InvKeyPointers[i]], KeyPointers[InvKeyPointers[j]]);
	std::swap<int>(InvKeyPointers[i], InvKeyPointers[j]);
}

template<typename Key>
const Key& Heap<Key>::getElement(int idx) {
	return A[KeyPointers[idx]];
}

template<typename Key>
void Heap<Key>::Earliest(pair<Key, int>& earliest) {
	assert(size > 0);
	earliest.first = A[0];
	earliest.second = InvKeyPointers[0];
}

#endif
