/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: SimulatorClass.h
 *
 *      Authors: Alex Williamson, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef SIMULATOR_CLASS
#define SIMULATOR_CLASS

#include <float.h>
#include "../BaseClasses/MobiusSocket.h"
#include "../BaseClasses/VersionNum.h"
#include "../BaseClasses/CPUTime.h"
#include "../BaseClasses/EVUtil.h"
#include "../Performance_Variables/PVModel.hpp"
#include "../Study/BaseStudyClass.hpp"
#include "Heap.h"
#include "Event.h"
#include "EventList.h"
#include "../BaseClasses/TraceFile.h"
#include "../BaseClasses/TraceFileContainer.h"
#include "../BaseClasses/TraceFileMobius.h"
#include "../BaseClasses/TraceFileXML.h"
#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <setjmp.h>
#include "DataPacket.h"
#include "../BaseClasses/debug.h"
#include "../BaseClasses/MobiusQueue.h"
#include "../BaseClasses/MobiusThread.h"
#include <list>
#include <stdint.h>
#include <iosfwd>
#include <string>
#ifndef MINGW
#include <sys/time.h>
#include <netinet/in.h>
#include <netdb.h>
#include <sys/resource.h>
#else
#include <process.h>
#endif

const long DEFAULT_REPORT_SIZE	= 1000;

class MobiusSocket;
class CPUTime;

extern void GenerateGlobalValues(char *);
extern PVModel*	GenerateModel();
extern BaseStudyClass*	GenerateStudy();
extern void DestructStudy();

struct TimeData {
	double PacketType;
	double ExperimentNumber;
	double ClientID;
	double Seconds;
};

/**
 * Class used to reference an action in the "enabling" and "affected"
 * pointers of the VarListElement class.  Acts as simple linked list
 * elements.
 * @see VarListElement
 *
 * @author Alex Williamson
 * @version Mobius v 1.0
 */
class VarListData {
public:
	~VarListData();
	VarListData();

	/**
	 * Pointer to representative action.
	 */
	BaseActionClass* action;
	/**
	 * Pointer to next element in list.
	 */
	VarListData* next;
};

/**
 * This class maintains a list of affecting and enabling action for
 * a state vaiable such that the connectivity list can be created.
 * @see VarListData
 *
 * @author Alex Williamson
 * @version Mobius v 1.0
 */
class VarListElement {
public:
	~VarListElement();
	VarListElement();
	/**
	 * Construct an element with the var data element set.
	 * @param BaseStateVariableClass* the value for var.
	 */
	VarListElement(BaseStateVariableClass*);
	/**
	 * Add the specified action to the affected list.
	 * @param BaseActionClass* The Action to add.
	 */
	short AddAffected(BaseActionClass*);
	/**
	 * Add the specified action to the enabling list.
	 * @param BaseActionClass* The Action to add.
	 */
	short AddEnabling(BaseActionClass*);
	/**
	 * The state variable for this element.
	 */
	BaseStateVariableClass* var;
	/**
	 * The enabling linked list.
	 */
	VarListData* enabling;
	/**
	 * The affected linked list.
	 */
	VarListData* affected;
	/**
	 * The next VarListElement in the VarList.
	 */
	VarListElement* next;
};

/**
 * This class is the end-all be-all simulation class.  All
 * initialization methods and simulation methods are contained in this
 * class.
 *
 * @author Alex Williamson
 * @version Mobius v 1.0
 */
class SimulatorClass {
protected:
	MobiusSocket * WaitForSimMgrConnection(int timeout_ms = -1);
	void ListenToSimMgrLoop(MobiusSocket *ms);
	void ListenToSimMgr(MobiusSocket *ms);

	///Used to override double global variables
	char *OverrideString;
	char *db_entity;
	double EndianConvertDouble(double);
	void SendSSResults(int *ResultCounters, int *ResultSqrCounters);
	bool CheckReportSize(double seconds);
	DataPacket * CreateDataPacket(int numPV);
	void ResetDataPacketBuffer();
	int Output(MobiusSocket *ms, std::ostream *ostr, double *array, int sizeArray);
	int OutputInt(MobiusSocket *ms, std::ostream *ostr, int number);
	int OutputShort(MobiusSocket *ms, std::ostream *ostr, short number);
	int OutputByte(MobiusSocket *ms, std::ostream *ostr, char number);
	void OutputCPUTime(MobiusSocket *ms, double cpuSeconds);
	void OutputDataBinary(MobiusSocket *ms, std::ostream *ostr, DataPacket *dp, DataPacket *dpsqr);
	void OutputDataCSV(std::ostream &ostr, DataPacket *dp, DataPacket *dpsqr);
	inline void RemoveTrailingZeros(char *str, double d);

	std::ofstream * CreateBinFile(char *);
	void UpdateBinFileHeader(std::ofstream &);
	void CloseBinFile(std::ofstream &);

	std::ofstream * CreateAsciiFile(char *);
	void CloseAsciiFile(std::ofstream &);

	void HandleInvalidDistParams(BaseActionClass *action, std::string str);
	void CheckCompletedBatches(PerformanceVariableNode** ThePVList, int* ResultCounters, int* ResultSqrCounters);
	Heap<double>* StopTimes;
	Heap<int>* CompletedBatches;
	CPUTime *_cpuTimer;

public:
	/**
	 * Constructor which uses the command line argc and argv parameters.
	 * @param int Number of arguments (argc).
	 * @param char** Pointer to character array of arguments (argc).
	 */
	SimulatorClass(int, char**);
	~SimulatorClass();

	/**
	 *  method that calls either SteadyState or Terminating simulator
	 *  and runs the whole job
	 */
	void Run();

	/**
	 * Simulation loop step command which does not accumulate reward variables.
	 */
	bool SimStepNoAccum();
	/**
	 * Simulation loop step command which does accumulate reward variables for
	 * terminating simulation.
	 */
	bool SimStep();
	/**
	 * Simulation loop step command which does accumulate reward variables for
	 * terminating simulation.
	 */
	bool SimStepSS();

	/**
	 * This method initializes the simulator for terminating simulation and uses
	 * the proper loops.
	 */
	void TerminatingSimulation();
	/**
	 * This method initializes the simulator for steady state simulation
	 * and uses the proper loops.
	 */
	void SteadyStateSimulation();

	/**
	 * This method parses the command line options given to this class.
	 * @param int argc parameter
	 * @param char** argv parameter
	 */
	short ParseCommandLineOptions(int, char**);
	/**
	 * Generates the action-to-action connectivity list.
	 */
	short GenerateConnectivityList();
	short GenerateConnectivityListOld();

	/**
	 * Deallocated the memory allocated in GenerateConnectivityList()
	 * by clearing the Affect list of each of the actions of the model
	 */
	void DestructConnectivityList();
	/**
	 * Print out the action-to-action connectivity list.
	 */
	void PrintConnectivityList(std::ostream *strm);

	/**
	 * Generate the complete event list by checking all actions in the
	 * model for enabling.
	 */
	short GenerateCompleteEventList();
	short GenerateCompleteEventListSS();

	/**
	 * Generate a list of initially enabled actions, and then use
	 * this list for GenerateCompleteEventList() each replication
	 *
	 */
	short GenerateInitialEnabledList();

	/**
	 * AddActionAffects adds the pointer to the first action to the
	 * list of actions affected by the second action.
	 * @param BaseActionClass* add: Pointer to add.
	 * @param BaseActionClass* existing: Action to add pointer to.
	 * @return 0 = success, 1 = failure
	 */
	short AddActionAffects(BaseActionClass* add, BaseActionClass* existing);
	short AddActionAffectsOld(BaseActionClass* add, BaseActionClass* existing);

	/**
	 * Clears the Affects list of "action". This list is constructed
	 * by calls to AddActionAffects
	 */
	void ClearActionAffectsList(BaseActionClass* action);

	/**
	 * Print the command line arguments of the simulator.
	 */
	void PrintUsage();

	PerformanceVariableNode** PVList;
	BaseActionClass** ActionList;
	List<BaseGroupClass> *TheGroupContainer;
	List<BaseActionClass> *TheActionContainer;

	static THREADRETURN StaticSocketThreadFunc(void *thisptr);
	THREADRETURN SocketThreadFunc();
	static THREADRETURN StaticOutputThreadFunc(void *thisptr);
	THREADRETURN OutputThreadFunc();

	void CheckReactivationList();
	void RemoveAffectedActions();
	void InsertAffectedActions();

	Event* CurrEvent;
	Event* TempEvent;
	Event** NextEvent;
	Event* NullEvent;

	BaseActionClass* CurrAction;
	BaseActionClass* TempAction;
	BaseActionClass* OtherAction;
	BaseActionClass* NullAction;

	ActionAffectsElement* AffectsAction;
	//A pointer to the future event list. Refer to EventList.h for a full
	//description of the way FEL is implemented.
	EventList* TheEventList;

	//Refer to EventList.h for a description of this array.
	PAddress* EventPointerArray;
	BaseGroupClass** GroupList;

	//this is the list of groups enabled at the beginning of
	//each replication
	BaseActionClass** InitiallyEnabled;
	BaseActionClass** ResamplingList;
	long NumberResampling;

	//The reactivation list which is implemented as a doubly linked list
	//for fast deletion and insertion. This list stores that activities that
	//have their reactivation predicate true. After every state change, these
	//activities are checked for their reactivation function.
	std::list<BaseActionClass*> TheReactivationList;
	BaseActionClass** CheckAlwaysList;
	long NumberCheckAlways;
	int** TempPVAffects;

	//this is the number of actions initially enabled
	int NumInitiallyEnabled;
	double AccumulateStartTime;
	double AccumulateStopTime;

	int SimControl;
	double CurrentTime;
	PVModel* TheModel;
	BaseStudyClass* TheStudy;

	// thread and output data structures
	MobiusQueue<DataPacket*> *UnusedPackets;
	MobiusQueue<DataPacket*> *OutputPackets;
	DataPacket *CurPacket;
	DataPacket *CurSqrPacket;
	int MaxDataPackets;
	long NumberReports;
	MobiusSocket *SimMgrSocket;
	MobiusSocket *InSocket;
	MobiusSocket *ConnectionSocket;
	int AcceptSelectDelay;
	int ReadSelectDelay;

	// command line flag variables
	long ReportSize;
	short SocketPort;
	int MyClientID;
	unsigned long TheExperimentNumber;
	int NumPVs;
	double CPUSeconds;
	int isTerminatingSimulation;
	std::string mTraceFilename;
	int TraceLevel;
	bool EnableMobiusTrace;
	bool EnableXMLTrace;
	TraceFileContainer *trace;
	TraceFileMobius *traceMob;
	TraceFileXML *traceXML;
	unsigned long RandomNumberSeed;
	short EventListType;
	int MaxNumOfBatches;
	char *ASCIIFilename;
	char *BinFilename;
	char* LicenseFilename;
	bool WaitForSocket;
	long _fireActionCount;

	enum RandomNumberGeneratorType
	{
		LAGGED_FIBONACCI,
		TAUSWORTHE,
		MERSENNE_TWISTER,
		MAX_NUM_GENERATORS
	};
//	int RandomNumberGenerator;
	RandomNumberGeneratorType RandomNumberGenerator;

private:
	// don't call this
	SimulatorClass();

	std::ofstream * FileCSV;
	std::ofstream * FileBin;
	int BatchRuns;

	std::string CreateBinFileHeader(int fileVersion);
	std::string to_string(RandomNumberGeneratorType e);
};

// SimControl enum
enum {
	GO, STOPSOFT, STOPHARD
};

/**
 *  this method formats a double in a string so there are a max number
 *  of significant digits (currently 10), but a minimum number of
 *  trailing zeros (currently 1).  It is required the str be
 *  predefined and large enough
 */
inline void SimulatorClass::RemoveTrailingZeros(char *str, double d) {
	int n;
	sprintf(str, "%.10lf", d);
	for (n = strlen(str) - 1; n >= 1; n--)
		if (str[n] != '0' || str[n - 1] == '.')
			break;
	str[n + 1] = '\0';
}
#endif
