/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef __BASELTS_H_
#define __BASELTS_H_

#define SizeType int
#define RateType double
#define LabelType char*

using namespace std;

#include <fstream>
#include <vector>
#include <cstring>
#include <cstdlib>

typedef struct {
	int min;
	int max;
	int best;
	//Example: For three blocks (0-9, 10-19, and 20-29) it stores four
	//numbers 0, 10, 20, 30. The size of the array will be 1+size (described below)
	int* regions;
	//This is the number of blocks in the partition that the state-level object
	//agrees with
	int size;
} Partition;

class Edge {
public:
	virtual SizeType& row() = 0;
	virtual SizeType& col() = 0;
	virtual RateType& rate() = 0;
	virtual LabelType& label() = 0;
};

typedef struct {
	SizeType row;
	SizeType col;
	RateType rate;
	LabelType label;
	bool end;
} EdgeBuffer;

typedef struct {
	RateType rate;
	RateType impulse;
} PVValue;

typedef struct {
	PVValue reward;
	bool end;
} PVBuffer;

class BaseStateLevel {
public:
	// constructor
	// the constructor expects a parameter file to read information of
	// the Kronecker representation in APNN format from
	BaseStateLevel();
	virtual ~BaseStateLevel();
	static BaseStateLevel* createStateLevelObject(int argc, char**argv);

	// gives the number of states
	virtual SizeType getNumberOfStates() = 0;

	// gives the index of the initial state
	virtual SizeType getInitialState() = 0;

	virtual int getNumberOfPVs() = 0;
	virtual char* getPVName(int i) = 0;

	class Iterator {
	public:
		Iterator() {
		}
		;
		virtual ~Iterator() {
		}
		;
		virtual void operator++() = 0;
		virtual bool end() = 0;
		//virtual bool operator==(self &it) = 0;
		//virtual bool operator!=(self &it) = 0;
		virtual const Iterator& operator=(Iterator const &it) = 0;
	};

	class EdgeBaseIterator: public Iterator {
	public:
		virtual Edge* operator->() = 0;
		virtual void next(EdgeBuffer& t) = 0;
	};

	class PVBaseIterator: public Iterator {
	public:
		virtual PVValue* operator->() = 0;
		virtual PVValue& operator*() = 0;
		virtual void next(PVBuffer& t) = 0;
	};

	class Container {
	public:
		Container() {
		}
		;
		virtual ~Container() {
		}
		;
		virtual void begin(Iterator* it) = 0;
	};

	class EdgeBaseContainer: public Container {
	public:
		// I do not know why this line should be here since it is already declared
		// in Container class. Without it, the compiler made a complain about
		// not founding this method.
		virtual void begin(Iterator* it) {
		}
		virtual void begin(Iterator* it, EdgeBuffer& t) = 0;
	};

	class PVBaseContainer: public Container {
	public:
		virtual void begin(Iterator* it, PVBuffer& t) = 0;
		virtual void operator()(SizeType state_index, PVBuffer& t) = 0;
	};

	virtual void getAllEdges(EdgeBaseContainer* edges)=0;
	virtual void allocateAllEdgesContainer(EdgeBaseContainer*& edges)=0;
	virtual void deallocateAllEdgesContainer(EdgeBaseContainer* edges)=0;
	virtual void allocateAllEdgesIterator(EdgeBaseIterator*& it)=0;
	virtual void deallocateAllEdgesIterator(EdgeBaseIterator* it)=0;

	virtual void getRow(SizeType row_index, EdgeBaseContainer* r)=0;
	virtual void allocateRowContainer(EdgeBaseContainer*& r)=0;
	virtual void deallocateRowContainer(EdgeBaseContainer* r)=0;
	virtual void allocateRowIterator(EdgeBaseIterator*& it)=0;
	virtual void deallocateRowIterator(EdgeBaseIterator* it)=0;

	virtual void getColumn(SizeType row_index, EdgeBaseContainer* c)=0;
	virtual void allocateColumnContainer(EdgeBaseContainer*& c)=0;
	virtual void deallocateColumnContainer(EdgeBaseContainer* c)=0;
	virtual void allocateColumnIterator(EdgeBaseIterator*& it)=0;
	virtual void deallocateColumnIterator(EdgeBaseIterator* it)=0;

	virtual void getRowPartition(Partition& row_part) = 0;
	virtual void getColPartition(Partition& col_part) = 0;
	virtual void getSubMatrixPartition(Partition& _col_part, Partition& _row_part) = 0;
	virtual void getSubMatrix(SizeType _start_col, SizeType _end_col, SizeType _start_row, SizeType _end_row, EdgeBaseContainer* s) = 0;
	virtual void allocateSubMatrixContainer(EdgeBaseContainer*& edges) = 0;
	virtual void deallocateSubMatrixContainer(EdgeBaseContainer* edges) = 0;
	virtual void allocateSubMatrixIterator(EdgeBaseIterator*& it) = 0;
	virtual void deallocateSubMatrixIterator(EdgeBaseIterator* it) = 0;

	virtual void getPV(PVBaseContainer* pv, int PVIndex) = 0;
	virtual void allocatePVContainer(PVBaseContainer*& pv)=0;
	virtual void deallocatePVContainer(PVBaseContainer* pv)=0;
	virtual void allocatePVIterator(PVBaseIterator*& it)=0;
	virtual void deallocatePVIterator(PVBaseIterator* it)=0;

	inline int getType(int pvindex) {
		return TypeList[pvindex];
	}
	inline int getNumTimePoints(int pvindex) {
		return NumTimePtsList[pvindex];
	}
	inline double getStartTime(int pvindex, int timeindex) {
		return StartPtsList[pvindex][timeindex];
	}
	inline double getStopTime(int pvindex, int timeindex) {
		return StopPtsList[pvindex][timeindex];
	}
protected:
	/**
	 *  read the PV type file created by WritePVTypeFile in PVModel, and
	 *  executed by all SSG code
	 */
	bool readPVTypeFile(const char* filename);
private:
	vector<int> TypeList;
	vector<int> NumTimePtsList;
	vector<double*> StartPtsList;
	vector<double*> StopPtsList;
	void setType(int pvindex, int type);
	void setTimePoints(int pvindex, int numtimes, double*start, double*stop);
	bool readKeyValue(ifstream &ifs, const char *desiredkey, char *key, char *value);
};

/********************    Exception classes   ************************/

//general exception base class
class AccessException {
};
//exception class to use when row access is not available
class RowAccessNAException: public AccessException {
};
//exception class to use when column access is not available
class ColumnAccessNAException: public AccessException {
};

class PVRandomAccessNAException: public AccessException {
};

//exception class to use when block access is not available
class BlockAccessNAException: public AccessException {
};
#endif
