/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#include "../BaseStateLevel.h"
#include <iostream>
#include <vector>
#include "BaseClasses/misc.h"
#include "StateLevelSolvers/common/solvers.h"
#include <math.h>
#include <assert.h>


#define BINSEARCH_THRESHOLD 15
using namespace std;

class FlatStateLevel: public BaseStateLevel {
public:

	FlatStateLevel(const string& _filename);
	~FlatStateLevel();

	SizeType getNumberOfStates();
	SizeType getInitialState();
	int getNumberOfPVs();
	char* getPVName(int i);

	void getAllEdges(EdgeBaseContainer* edges);
	void allocateAllEdgesContainer(EdgeBaseContainer*& edges);
	void deallocateAllEdgesContainer(EdgeBaseContainer* edges);
	void allocateAllEdgesIterator(EdgeBaseIterator*& it);
	void deallocateAllEdgesIterator(EdgeBaseIterator* it);

	void getRow(SizeType row_index, EdgeBaseContainer* edges);
	void allocateRowContainer(EdgeBaseContainer*& edges);
	void deallocateRowContainer(EdgeBaseContainer* edges);
	void allocateRowIterator(EdgeBaseIterator*& it);
	void deallocateRowIterator(EdgeBaseIterator* it);

	void getColumn(SizeType column_index, EdgeBaseContainer* edges);
	void allocateColumnContainer(EdgeBaseContainer*& edges);
	void deallocateColumnContainer(EdgeBaseContainer* edges);
	void allocateColumnIterator(EdgeBaseIterator*& it);
	void deallocateColumnIterator(EdgeBaseIterator* it);

	void getSubMatrix(SizeType _start_col, SizeType _end_col, SizeType _start_row, SizeType _end_row, EdgeBaseContainer* edges);
	void getRowPartition(Partition& row_part);
	void getColPartition(Partition& col_part);
	void getSubMatrixPartition(Partition& _col_part, Partition& _row_part);
	void allocateSubMatrixContainer(EdgeBaseContainer*& edges);
	void deallocateSubMatrixContainer(EdgeBaseContainer* edges);
	void allocateSubMatrixIterator(EdgeBaseIterator*& it);
	void deallocateSubMatrixIterator(EdgeBaseIterator* it);

	void getPV(PVBaseContainer* pv, int PVIndex);
	void allocatePVContainer(PVBaseContainer*& pv);
	void deallocatePVContainer(PVBaseContainer* pv);
	void allocatePVIterator(PVBaseIterator*& it);
	void deallocatePVIterator(PVBaseIterator* it);

private:
	void getSubMatrixAux(SizeType _start_col, SizeType _end_col, SizeType _start_row, SizeType _end_row, EdgeBaseContainer* edges);
	void allocateSubMatrixAuxContainer(EdgeBaseContainer*& edges);
	void deallocateSubMatrixAuxContainer(EdgeBaseContainer* edges);
	void allocateSubMatrixAuxIterator(EdgeBaseIterator*& it);
	void deallocateSubMatrixAuxIterator(EdgeBaseIterator* it);
	void transformDSForBlockAccess();

private:
	RateType** rates;
	SizeType** minor_index;
	SizeType* major_size;
	PVValue** PVValues;

	//optimization for block access
	Partition row_part, col_part;
	RateType** rates_block;
	SizeType** column_block;
	SizeType** row_block;
	SizeType* number_of_elements;

	SizeType NumberOfStates;
	SizeType InitialState;
	enum TOrientation {
		row_oriented, col_oriented
	};
	TOrientation orientation;

	//Optimization for block access
	bool block_optimized;

	int NumberOfPVs;
	vector<char*> PVNames;
	string base_filename;

private:
	/*********************** OrderlessEdge class **************************/
	/* This class is used to store the edges of a matrix in no particular */
	/* order. It is used by the allEdgesIterator class which corresponds  */
	/* to the allEdges container class                                    */
	/**********************************************************************/
	class OrderlessEdge: public Edge {
	public:
		SizeType& row() {
			return *_row;
		}
		SizeType& col() {
			return _col;
		}
		RateType& rate() {
			return *_rate;
		}
		LabelType& label() {
			return _label;
		}

		void setRow(SizeType* row_ptr) {
			_row = row_ptr;
		}
		void setRate(RateType* rate_ptr) {
			_rate = rate_ptr;
		}
		void nextEdge();

		bool end() {
			return end_flag;
		}
		//bool operator==(OrderlessEdge &e) { return _row==e._row; }
		//bool operator!=(OrderlessEdge &e) { return _row!=e._row; }
		void init(BaseStateLevel *Model);

	private:
		void setModel(BaseStateLevel *_model);
		SizeType n;
		SizeType rowMax;
		bool end_flag;
		SizeType _col, *_row;
		RateType *_rate;
		LabelType _label;
		FlatStateLevel *model;
	};

	/*********************** RowOrientedEdge class **************************/
	/* This class is used to store the edges of a matrix in row order     */
	/* It is used by the rowIterator class which corresponds              */
	/* to the row container class                                         */
	/**********************************************************************/
	class RowOrientedEdge: public Edge {
	public:
		SizeType& row() {
			return _row;
		}
		SizeType& col() {
			return *_col;
		}
		RateType& rate() {
			return *_rate;
		}
		LabelType& label() {
			return _label;
		}
		void setCol(SizeType* col_ptr) {
			_col = col_ptr;
		}
		void setRate(RateType* rate_ptr) {
			_rate = rate_ptr;
		}
		void nextEdge() {
			_col++;
			_rate++;
			colMax--;
		}
		bool end() {
			return colMax == 0;
		}
		//bool operator==(RowOrientedEdge &e) {return _col==e._col;}
		//bool operator!=(RowOrientedEdge &e) {return _col!=e._col;}
		void init(BaseStateLevel *Model, SizeType rowIndex);

		//LTSType* getModel() {return model;}
	private:
		void setModel(BaseStateLevel *_model) {
			model = (FlatStateLevel*) _model;
		}

	private:
		SizeType colMax;
		SizeType _row, *_col;
		RateType *_rate;
		LabelType _label;
		FlatStateLevel *model;
	};

	/*********************** ColumnOrientedEdge class **************************/
	/* This class is used to store the edges of a matrix in column order    */
	/* It is used by the columnIterator class which corresponds             */
	/* to the column container class                                        */
	/************************************************************************/
	class ColumnOrientedEdge: public Edge {
	public:
		SizeType& row() {
			return *_row;
		}
		SizeType& col() {
			return _col;
		}
		RateType& rate() {
			return *_rate;
		}
		LabelType& label() {
			return _label;
		}
		void setRow(SizeType* row_ptr) {
			_row = row_ptr;
		}
		void setRate(RateType* rate_ptr) {
			_rate = rate_ptr;
		}
		void nextEdge() {
			_row++;
			_rate++;
			rowMax--;
		}
		bool end() {
			return rowMax == 0;
		}
		//bool operator==(ColumnOrientedEdge &e){ return _row==e._row;}
		//bool operator!=(ColumnOrientedEdge &e){ return _row!=e._row;}
		void init(BaseStateLevel *Model, SizeType colIndex);
	private:
		void setModel(BaseStateLevel *_model) {
			model = (FlatStateLevel*) _model;
		}
		;

	private:
		SizeType rowMax;
		SizeType _col, *_row;
		RateType *_rate;
		LabelType _label;
		FlatStateLevel *model;
	};

	/*********************** SubMatrixEdge class **********************
	 * This class is used to store the edges of a matrix in block order
	 * It is used by the SubMatrixAuxIterator class
	 ********************************************************************/
	//This class is used in construction of the optimized block access
	//data structure
	class SubMatrix;
	class SubMatrixEdge: public Edge {
	public:
		SizeType& row() {
			return *_row;
		}
		SizeType& col() {
			return *_col;
		}
		RateType& rate() {
			return *_rate;
		}
		LabelType& label() {
			return _label;
		}
		void nextEdge() {
			_row++;
			_rate++;
			_col++;
			rowMax--;
		}
		bool end() {
			return (rowMax == 0);
		}
		//bool operator==(SubMatrixEdge &e){return _row==e._row;}
		//bool operator!=(SubMatrixEdge &e){return _row!=e._row;}
		void init(BaseStateLevel *Model, SubMatrix *_sm);
	private:
		void setModel(BaseStateLevel *_model) {
			model = (FlatStateLevel*) _model;
		}
		; //(2)

	private:
		SizeType rowMax;
		SizeType *_col, *_row;
		RateType *_rate;
		LabelType _label;
		FlatStateLevel *model; //(2)
		SubMatrix *sm;
	};

	/*********************** SubMatrixAuxEdge class ***********************
	 * This class is used to store the edges of a matrix in block order
	 * while the optimized data structure for block access is being
	 * constructed. It is used in the SubMatrixAuxIterator class.
	 **********************************************************************/
	class SubMatrixAux;
	class SubMatrixAuxEdge: public Edge {
	public:
		SizeType& row() {
			return *_row;
		}
		SizeType& col() {
			return _col;
		}
		RateType& rate() {
			return *_rate;
		}
		LabelType& label() {
			return _label;
		}
		void setRow(SizeType* row_ptr) {
			_row = row_ptr;
		}
		void setRate(RateType* rate_ptr) {
			_rate = rate_ptr;
		}
		void nextEdge();
		bool end() {
			return end_flag;
		}
		//bool operator==(SubMatrixAuxEdge &e){return _row==e._row;}
		//bool operator!=(SubMatrixAuxEdge &e){return _row!=e._row;}
		void init(BaseStateLevel *Model, SubMatrixAux *_sm);

	private:
		void setModel(BaseStateLevel *_model) {
			model = (FlatStateLevel*) _model;
		}
		; //(2)
		void update_end_flag();
	private:
		bool end_flag;
		SizeType rowMax;
		SizeType _col, *_row;
		RateType *_rate;
		LabelType _label;
		FlatStateLevel *model; //(2)
		SubMatrixAux *sm;
	};

private:

	/*************** allEdgesIterator and allEdges classes *************/
	// Iterator and container classes for all edges of the LTS
	class allEdgesIterator: public EdgeBaseIterator {
	public:
		typedef allEdgesIterator self;

		allEdgesIterator() {
		}
		allEdgesIterator(BaseStateLevel *Model) {
			edge.init(Model);
		}
		allEdgesIterator(const self& it) {
			edge = it.edge;
		}
		~allEdgesIterator() {
		}
		void init(BaseStateLevel *Model) {
			edge.init(Model);
		}

		void init(BaseStateLevel *Model, EdgeBuffer& t) {
			edge.init(Model);
			getEdgeBuffer(t);
		}

		//Edge& operator*() { return edge;}
		Edge* operator->() {
			return &edge;
		}

		void next(EdgeBuffer& t) {
			edge.nextEdge();
			getEdgeBuffer(t);
		}

		void operator++() {
			edge.nextEdge();
		}
		bool end() {
			return edge.end();
		}
		//bool operator==(self &it) { return edge==it.edge;}
		//bool operator!=(self &it) { return !(edge==it.edge);}

		const Iterator& operator=(Iterator const &it) {
			//avoid doing this cause its expensive
			edge = ((self *) &it)->edge;
			return *((Iterator *) this);
		}

	private:
		void getEdgeBuffer(EdgeBuffer& t) {
			t.end = edge.end();
			if (!t.end) {
				t.row = edge.row();
				t.col = edge.col();
				t.rate = edge.rate();
				//t.label = edge.label();
			}
		}

	private:
		OrderlessEdge edge;
	};

	class allEdges: public EdgeBaseContainer {
	public:
		allEdges(BaseStateLevel* _model = 0) {
			model = _model;
		}
		;
		void init(BaseStateLevel* _model) {
			model = _model;
		}
		~allEdges() {
		}
		;

		void begin(Iterator* it) {
			((allEdgesIterator*) it)->init(model);
		}

		void begin(Iterator* it, EdgeBuffer& t) {
			((allEdgesIterator*) it)->init(model, t);
		}

	private:
		BaseStateLevel *model;
	};

	/*************** rowIterator and row classes *************/
	// Iterator and container classes for row-oriented access to the LTS
	class rowIterator: public EdgeBaseIterator {
	public:
		typedef rowIterator self;

		rowIterator() {
		}
		rowIterator(BaseStateLevel *Model, SizeType rowIndex) {
			edge.init(Model, rowIndex);
		}

		rowIterator(const self& it) {
			//avoid doing this cause its expensive
			edge = it.edge;
		}

		~rowIterator() {
		}

		void init(BaseStateLevel *Model, SizeType rowIndex) {
			edge.init(Model, rowIndex);
		}

		void init(BaseStateLevel *Model, SizeType rowIndex, EdgeBuffer& t) {
			edge.init(Model, rowIndex);
			getEdgeBuffer(t);
		}

		//Edge& operator*() { return edge;}
		Edge* operator->() {
			return &edge;
		}

		void next(EdgeBuffer &t) {
			edge.nextEdge();
			getEdgeBuffer(t);
		}

		void operator++() {
			edge.nextEdge();
		}
		bool end() {
			return edge.end();
		}
		//bool operator==(self &it) {return edge==it.edge;}
		//bool operator!=(self &it) {return edge!=it.edge;}

		const Iterator& operator=(Iterator const &it) {
			edge = ((self*) &it)->edge;
			return *((Iterator*) this);
		}

	private:
		void getEdgeBuffer(EdgeBuffer& t) {
			t.end = edge.end();
			if (!t.end) {
				t.row = edge.row();
				t.col = edge.col();
				t.rate = edge.rate();
				//t.label = edge.label();
			}
		}

	private:
		RowOrientedEdge edge;
	};

	class row: public EdgeBaseContainer {
	public:
		row(BaseStateLevel* _model = 0, SizeType _row_index = 0) {
			row_index = _row_index;
			model = _model;
		}
		;
		void init(BaseStateLevel* _model, SizeType _row_index) {
			model = _model;
			row_index = _row_index;
		}
		~row() {
		}
		;

		void begin(Iterator* it) {
			((rowIterator*) it)->init(model, row_index);
		}
		;

		void begin(Iterator* it, EdgeBuffer& t) {
			((rowIterator*) it)->init(model, row_index, t);
		}

	private:
		SizeType row_index;
		BaseStateLevel *model;
	};

	/*************** columnIterator and column classes *************/
	// Iterator and container classes for column-oriented access to the LTS
	class columnIterator: public EdgeBaseIterator {
	public:
		typedef columnIterator self;

		columnIterator() {
		}
		columnIterator(BaseStateLevel *Model, SizeType columnIndex) {
			edge.init(Model, columnIndex);
		}

		//avoid doing this cause its expensive
		columnIterator(const self& it) {
			edge = it.edge;
		}

		~columnIterator() {
		}

		void init(BaseStateLevel *Model, SizeType columnIndex) {
			edge.init(Model, columnIndex);
		}

		void init(BaseStateLevel *Model, SizeType columnIndex, EdgeBuffer& t) {
			edge.init(Model, columnIndex);
			getEdgeBuffer(t);
		}

		//Edge& operator*() { return edge;}
		Edge* operator->() {
			return &edge;
		}

		void next(EdgeBuffer &t) {
			edge.nextEdge();
			getEdgeBuffer(t);
		}

		void operator++() {
			edge.nextEdge();
		}
		bool end() {
			return edge.end();
		}
		//bool operator==(self &it) {return edge==it.edge;}
		//bool operator!=(self &it) {return edge!=it.edge;}

		const Iterator& operator=(Iterator const &it) {
			edge = ((self*) &it)->edge;
			return *((Iterator*) this);
		}

	private:
		void getEdgeBuffer(EdgeBuffer& t) {
			t.end = edge.end();
			if (!t.end) {
				t.row = edge.row();
				t.col = edge.col();
				t.rate = edge.rate();
				//t.label = edge.label();
			}
		}

	private:
		ColumnOrientedEdge edge;
	};

	class column: public EdgeBaseContainer {
	public:
		column(BaseStateLevel* _model = 0, SizeType _column_index = 0) {
			column_index = _column_index;
			model = _model;
		}
		;
		~column() {
		}
		;

		void init(BaseStateLevel* _model, SizeType _column_index) {
			model = _model;
			column_index = _column_index;
		}

		void begin(Iterator* it) {
			((columnIterator*) it)->init(model, column_index);
		}
		;

		void begin(Iterator* it, EdgeBuffer& t) {
			((columnIterator*) it)->init(model, column_index, t);
		}

	private:
		SizeType column_index;
		BaseStateLevel *model;
	};

	/*************** SubMatrixIterator and SubMatrix classes *************/
	// Iterator and container classes for block access to the state space
	class SubMatrixIterator: public EdgeBaseIterator {
	public:
		typedef SubMatrixIterator self;

		SubMatrixIterator() {
		}
		SubMatrixIterator(BaseStateLevel *Model, SubMatrix* sm) {
			edge.init(Model, sm);
		}
		//avoid doing this cause its expensive
		SubMatrixIterator(const self& it) {
			edge = it.edge;
		}
		~SubMatrixIterator() {
		}

		void init(BaseStateLevel *Model, SubMatrix* sm) {
			edge.init(Model, sm);
		}

		void init(BaseStateLevel *Model, SubMatrix* sm, EdgeBuffer& t) {
			edge.init(Model, sm);
			getEdgeBuffer(t);
		}

		//Edge& operator*() { return edge; }
		Edge* operator->() {
			return &edge;
		}

		void next(EdgeBuffer &t) {
			edge.nextEdge();
			getEdgeBuffer(t);
		}

		void operator++() {
			edge.nextEdge();
		}
		bool end() {
			return edge.end();
		}
		//bool operator==(self &it) { return edge==it.edge; }
		//bool operator!=(self &it) { return !(edge==it.edge); }

		//avoid doing this cause its expensive
		const Iterator& operator=(Iterator const &it) {
			edge = ((self*) &it)->edge;
			return *((Iterator*) this);
		}

	private:
		void getEdgeBuffer(EdgeBuffer& t) {
			t.end = edge.end();
			if (!t.end) {
				t.row = edge.row();
				t.col = edge.col();
				t.rate = edge.rate();
				//t.label = edge.label();
			}
		}
	private:
		SubMatrixEdge edge;
	};

	class SubMatrix: public EdgeBaseContainer {
	public:
		SubMatrix(BaseStateLevel* _model = 0, SizeType _start_col = 0, SizeType _end_col = 0, SizeType _start_row = 0, SizeType _end_row =
				0) {
			start_col = _start_col;
			end_col = _end_col;
			start_row = _start_row;
			end_row = _end_row;
			model = _model;
		}

		~SubMatrix() {
		}

		void init(BaseStateLevel* _model, SizeType _start_col, SizeType _end_col, SizeType _start_row, SizeType _end_row) {
			model = _model;
			start_col = _start_col;
			end_col = _end_col;
			start_row = _start_row;
			end_row = _end_row;
		}

		void begin(Iterator* it) {
			// We pass 'this' so that the iterator can access the submatrix
			// boundaries.
			((SubMatrixIterator*) it)->init(model, this);
		}

		void begin(Iterator* it, EdgeBuffer& t) {
			// We pass 'this' so that the iterator can access the submatrix
			// boundaries.
			((SubMatrixIterator*) it)->init(model, this, t);
		}

	public:
		SizeType start_col;
		SizeType end_col;
		SizeType start_row;
		SizeType end_row;
	private:
		BaseStateLevel *model;
	};

	/******************************************************************/
	class SubMatrixAuxIterator: public EdgeBaseIterator {
	public:
		typedef SubMatrixAuxIterator self;

		SubMatrixAuxIterator() {
		}
		SubMatrixAuxIterator(BaseStateLevel* Model, SubMatrixAux* sm) {
			edge.init(Model, sm);
		}

		SubMatrixAuxIterator(const self& it) {
			//avoid doing this cause its expensive
			edge = it.edge;
		}

		~SubMatrixAuxIterator() {
		}

		void init(BaseStateLevel *Model, SubMatrixAux* sm) {
			edge.init(Model, sm);
		}

		void init(BaseStateLevel *Model, SubMatrixAux* sm, EdgeBuffer& t) {
			edge.init(Model, sm);
			getEdgeBuffer(t);
		}

		// SubMatrixAuxEdge& operator*() { return edge; }
		SubMatrixAuxEdge* operator->() {
			return &edge;
		}

		void next(EdgeBuffer &t) {
			edge.nextEdge();
			getEdgeBuffer(t);
		}

		void operator++() {
			edge.nextEdge();
		}
		bool end() {
			return edge.end();
		}
		//bool operator==(self &it) { return edge==it.edge; }
		//bool operator!=(self &it) { return !(edge==it.edge); }

		const Iterator& operator=(Iterator const &it) {
			edge = ((self*) &it)->edge;
			return *((Iterator*) this);
		}

	private:
		void getEdgeBuffer(EdgeBuffer& t) {
			t.end = edge.end();
			if (!t.end) {
				t.row = edge.row();
				t.col = edge.col();
				t.rate = edge.rate();
				//t.label = edge.label();
			}
		}

	private:
		SubMatrixAuxEdge edge;
	};

	class SubMatrixAux: public EdgeBaseContainer {
	public:
		SubMatrixAux(BaseStateLevel* _model = 0, SizeType _start_col = 0, SizeType _end_col = 0, SizeType _start_row = 0,
				SizeType _end_row = 0) {
			start_col = _start_col;
			end_col = _end_col;
			start_row = _start_row;
			end_row = _end_row;
			model = (FlatStateLevel*) _model;
		}
		~SubMatrixAux() {
		}
		;

		void init(BaseStateLevel* _model, SizeType _start_col, SizeType _end_col, SizeType _start_row, SizeType _end_row) {
			model = (FlatStateLevel*) _model;
			start_col = _start_col;
			end_col = _end_col;
			start_row = _start_row;
			end_row = _end_row;
		}

		void find_offsets(SizeType col, int& start_offset, int& end_offset) {
			if (model->major_size[col] > BINSEARCH_THRESHOLD)
				find_binary_search(col, start_offset, end_offset);
			else
				find_linear_search(col, start_offset, end_offset);
		}

		void begin(Iterator *it) {
			// We pass 'this' so that the iterator can access the submatrix
			// boundaries.
			((SubMatrixAuxIterator*) it)->init(model, this);
		}

		void begin(Iterator *it, EdgeBuffer& t) {
			((SubMatrixAuxIterator*) it)->init(model, this, t);
		}

	public:
		SizeType start_col;
		SizeType end_col;
		SizeType start_row;
		SizeType end_row;

	private:
		inline void find_binary_search(SizeType col, int& start_offset, int& end_offset) {
			//Finding the smallest i such that minor_index[i] >= start_row
			find_one_bin_search(col, 0, model->major_size[col] - 1, start_row, start_offset);
			find_one_bin_search(col, start_offset, model->major_size[col] - 1, end_row, end_offset);
		}
		inline void find_one_bin_search(SizeType col, int l, int h, SizeType row, int& offset) {
			int m;
			while (l <= h) {
				m = (l + h) / 2;
				SizeType content_m = model->minor_index[col][m];
				if (content_m == row)
					break;
				else if (content_m > row)
					h = m - 1;
				else
					l = m + 1;
			}
			if (l > h)
				offset = l;
			else
				offset = m;
		}

		inline void find_linear_search(SizeType col, int& start_offset, int& end_offset) {
			int i = 0;
			while ((i < model->major_size[col]) && (model->minor_index[col][i] < start_row)) {
				i++;
			}
			start_offset = i;
			while ((i < model->major_size[col]) && (model->minor_index[col][i] < end_row)) {
				i++;
			}
			end_offset = i;
		}

	private:
		FlatStateLevel *model;
	};

	/************* PVIterator and PV classes ***************/
	// Iterator and container classes for PVs of the LTS
	class PVIterator: public PVBaseIterator {
	public:
		typedef PVIterator self;

		PVIterator() {
			PVArray = EndPointer = NULL;
		}

		PVIterator(const self& it) {
			//avoid doing this cause its expensive
			PVArray = it.PVArray;
		}
		void init(PVValue *_PVArray, int length) {
			PVArray = _PVArray;
			EndPointer = PVArray + length;
		}

		void init(PVValue* _PVArray, int length, PVBuffer& t) {
			init(_PVArray, length);
			getPVBuffer(t);
		}

		~PVIterator() {
		}

		PVValue& operator*() {
			return *PVArray;
		}
		PVValue* operator->() {
			return PVArray;
		}

		void next(PVBuffer& t) {
			PVArray++;
			getPVBuffer(t);
		}

		void getPVBuffer(PVBuffer& t) {
			t.end = (PVArray == EndPointer);
			if (!t.end)
				t.reward = *PVArray;
		}

		void operator++() {
			PVArray++;
		}

		bool end() {
			return PVArray == EndPointer;
		}

		const Iterator& operator=(Iterator const &it) {
			//avoid doing this cause its expensive
			PVArray = ((self *) &it)->PVArray;
			EndPointer = ((self *) &it)->EndPointer;
			return *((Iterator *) this);
		}

	private:
		PVValue *PVArray;
		PVValue *EndPointer;
	};

	class PV: public PVBaseContainer {
	public:
		PV(BaseStateLevel* _model = 0) {
			model = (FlatStateLevel*) _model;
		}
		;
		void init(BaseStateLevel* _model, int _PVIndex) {
			model = (FlatStateLevel*) _model;
			PVIndex = _PVIndex;
			PVArray = model->PVValues[PVIndex];
			//EndIterator = &model->PVValues[PVIndex][model->getNumberOfStates()];
		}
		~PV() {
		}
		;

		void begin(Iterator* it) {
			((PVIterator*) it)->init(PVArray, model->getNumberOfStates());
		}

		void begin(Iterator* it, PVBuffer& t) {
			((PVIterator*) it)->init(model->PVValues[PVIndex], model->getNumberOfStates(), t);
		}

		void operator()(SizeType state_index, PVBuffer& t) {
			//For the sake of performance, this function assumes that
			//state_index is less than model->getNumberOfStates()
			t.reward = PVArray[state_index];
		}

	private:
		FlatStateLevel *model;
		PVValue* PVArray;
		//PVIterator EndIterator;
		int PVIndex;
	};

private:
	void switch_orientation(TOrientation new_orientation);
};

