/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#include "StateLevelObjects/BaseStateLevel.h"
#include <vector>
#include "BaseClasses/symbolic/mdnode.h"
#include "PVMTMDDReader.h"
#include "BaseClasses/symbolic/mdd.h"
#include "BaseClasses/symbolic/md.h"
#include "BaseClasses/MobiusThread.h"
#include "BaseClasses/symbolic/MappedMDD.h"

typedef PATH_TYPE PathType;
typedef INDEX_TYPE IndexType;

class MxDStateLevel: public BaseStateLevel {
public:
	MxDStateLevel(const char* _filename);
	~MxDStateLevel();

	SizeType getNumberOfStates();
	SizeType getInitialState();
	int getNumberOfPVs();
	char* getPVName(int i);

	void getAllEdges(EdgeBaseContainer* edges);
	void allocateAllEdgesContainer(EdgeBaseContainer*& edges);
	void deallocateAllEdgesContainer(EdgeBaseContainer* edges);
	void allocateAllEdgesIterator(EdgeBaseIterator*& it);
	void deallocateAllEdgesIterator(EdgeBaseIterator* it);

	void getRow(SizeType row_index, EdgeBaseContainer* edges);
	void allocateRowContainer(EdgeBaseContainer*& edges);
	void deallocateRowContainer(EdgeBaseContainer* edges);
	void allocateRowIterator(EdgeBaseIterator*& it);
	void deallocateRowIterator(EdgeBaseIterator* it);

	void getColumn(SizeType column_index, EdgeBaseContainer* edges);
	void allocateColumnContainer(EdgeBaseContainer*& edges);
	void deallocateColumnContainer(EdgeBaseContainer* edges);
	void allocateColumnIterator(EdgeBaseIterator*& it);
	void deallocateColumnIterator(EdgeBaseIterator* it);

	void getRowPartition(Partition& row_part);
	void getColPartition(Partition& col_part);
	void getSubMatrixPartition(Partition& _col_part, Partition& _row_part);
	void getSubMatrix(SizeType _start_col, SizeType _end_col, SizeType _start_row, SizeType _end_row, EdgeBaseContainer* s);
	void allocateSubMatrixContainer(EdgeBaseContainer*& edges);
	void deallocateSubMatrixContainer(EdgeBaseContainer* edges);
	void allocateSubMatrixIterator(EdgeBaseIterator*& it);
	void deallocateSubMatrixIterator(EdgeBaseIterator* it);

	void getPV(PVBaseContainer* pv, int PVIndex);
	void allocatePVContainer(PVBaseContainer*& pv);
	void deallocatePVContainer(PVBaseContainer* pv);
	void allocatePVIterator(PVBaseIterator*& it);
	void deallocatePVIterator(PVBaseIterator* it);

	typedef MDNode::DIM_TYPE DIM_TYPE;

private:
	//SymbolicSSG* ssg;
	MDNode* mdroot;
	short no_of_levels;
	SSNode* LumpedS;
	MappingNode* MappingMDD;
	MDD<SSNode> *SS;
	ComputeMappingMDDEnv* CMNE;
	MD* md;
	SSNode *Value0, *Value1;
	PVMTMDDReader* mtmdd_reader;

	bool ReadSSFile(const char* base_filename);
	bool ReadMappingMDDFile(const char* base_filename);
	bool ReadMxDFile(const char* base_filename);
	bool ReadPVFile(const char* base_filename);

#define BUFSIZE 10000
	class allEdges;

	class EnumerationParams {
	public:
		~EnumerationParams();
		void init();
		EdgeBuffer buffer[BUFSIZE];
		int size;
		bool finish;
		bool terminate;
		MobiusMutex mutex;
		MobiusCond consumer;
		MobiusCond producer;
		MobiusThread* thread;
	};

	/*************** allEdgesIterator and allEdges classes *************/
	// Iterator and container classes for all edges of the LTS
	class allEdgesIterator: public EdgeBaseIterator {
	public:
		typedef allEdgesIterator self;

		allEdgesIterator() {
		}
		allEdgesIterator(BaseStateLevel *Model) { /*edge.init(Model);*/
		}
		allEdgesIterator(const self& it) { /*edge = it.edge; */
		}
		~allEdgesIterator() {
		}
		;
		void init(EnumerationParams *_enum_params) {
			//edge.init(Model, _enum_params);
			enum_params = _enum_params;
			index = 0;
		}
		/*
		 void init(MxDStateLevel *Model, EdgeBuffer& t) {
		 edge.init(Model, _enum_params);
		 getEdgeBuffer(t);
		 }
		 */

		void getEdgeBuffer(EdgeBuffer& t) {
			t = enum_params->buffer[index];
		}

		//Edge& operator*() { return edge;}
		Edge* operator->() {
			return NULL;/*return &edge;*/
		}

		void next(EdgeBuffer& t) {
			index++;
			if (index == enum_params->size) {
				enum_params->size = 0;
				if (enum_params->finish) {
					t.end = true;
					enum_params->finish = false;
					enum_params->mutex.Release();
					return;
				} else {
					enum_params->producer.Signal();
					enum_params->consumer.Wait(enum_params->mutex);
					index = 0;
				}
			}
			getEdgeBuffer(t);
		}

		void operator++() { /*edge.nextEdge();*/
		}
		bool end() {
			return false;/*return edge.end();*/
		}
		//bool operator==(self &it) { return edge==it.edge;}
		//bool operator!=(self &it) { return !(edge==it.edge);}

		const Iterator& operator=(Iterator const &it) {
			//avoid doing this cause its expensive
			//edge = ((self *)&it)->edge;
			return *((Iterator *) this);
		}

	private:
		int index;
		EnumerationParams *enum_params;
	};

public:
	typedef allEdges *ThreadParameters;

private:
	class allEdges: public EdgeBaseContainer {
	public:
		allEdges(BaseStateLevel* _model = 0) {
			model = (MxDStateLevel*) _model;
		}
		;
		void init(BaseStateLevel* _model);
		~allEdges();

		void begin(Iterator* it) {
		}
		;
		/*
		 {
		 ((allEdgesIterator*)it)->init(model);
		 }
		 */
		void begin(Iterator* it, EdgeBuffer& t);
		void* StartEnumeration();
		void DFS_MxD(MDNode* mdnode);

	private:
		MxDStateLevel *model;
		EnumerationParams enum_params;
		ThreadParameters thread_params;

		PATH_TYPE* af_maj_indices;
		PATH_TYPE* af_min_indices;
		REAL* af_coefs;
		INDEX_TYPE* af_maj_offsets;
		INDEX_TYPE* af_min_offsets;
		SSNode** LumpedS_path;
		MappingNode** MappingMDD_path;
	};

	/********* Definition for PV container and iterator classes ******/

	class PVIterator: public PVBaseIterator {
	public:
		typedef PVIterator self;

		PVIterator() {
			maj_indices = NULL;
			path = NULL;
		}

		~PVIterator() {
			delete[] maj_indices;
			delete[] path;
		}

		void next(PVBuffer& t) {
			if (ComputeNextActIndices() == model->no_of_levels)
				t.end = true;
			else
				getPVBuffer(t);
		}

		const Iterator& operator=(Iterator const &it) {
			//avoid doing this cause its expensive
			model = ((self *) &it)->model;
			/*
			 PVNode = ((self *)&it)->PVNode;
			 System = ((self *)&it)->System;
			 state_space = ((self *)&it)->state_space;
			 ssg = ((self *)&it)->ssg;
			 */
			maj_indices = ((self *) &it)->maj_indices;
			path = ((self *) &it)->path;
			return *((Iterator *) this);
		}

		//Old-fashioned functions are not implemented at this point
		PVValue& operator*() {
		}
		;
		PVValue* operator->() {
		}
		;
		void operator++() {
		}
		;
		bool end() {
		}
		;

		void init(MxDStateLevel* _model, int PVIndex, PVBuffer& t);

	private:
		void getPVBuffer(PVBuffer& t);
		short ComputeNextActIndices();

	private:
		MxDStateLevel* model;
		PATH_TYPE* maj_indices;
		MTMDDNode** path;
	};

	class PV: public PVBaseContainer {
	public:
		void init(BaseStateLevel* _model, int _PVIndex) {
			model = (MxDStateLevel*) _model;
			PVIndex = _PVIndex;
		}

		void begin(Iterator* it, PVBuffer& t) {
			((PVIterator*) it)->init(model, PVIndex, t);
		}

		//Old-fashioned function are not implemented at this point.
		void begin(Iterator* it) {
		}
		void operator()(SizeType state_index, PVBuffer& t) {
			//Random access to PV values are not supported at this point
			throw PVRandomAccessNAException();
		}
	private:
		MxDStateLevel* model;
		int PVIndex;
		int no_of_states;
	};
};

THREADRETURN MxDAllEdgesEnumeratorThread(void* _thread_params);
