/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: BaseStudyClass.hpp
 *
 *      Authors: Patrick G. Webster, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2014 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef BASE_STUDY_CLASS_H_
#define BASE_STUDY_CLASS_H_

#include "../BaseClasses/GlobalVariables.h"
#include "../Performance_Variables/PVModel.hpp"
#include <string>
#include <vector>

/**
 *  BaseStudyClass -- This class defines a base study class from which all
 *      other study classes should be derived.
 *
 *  @author Patrick Webster
 *  @version Mobius Version 1.0
 */
class BaseStudyClass {

public:

	/**
	 *   Base Constructor for Studies
	 */
	BaseStudyClass();

	/**
	 *   Base Destructor for Studies
	 */
	virtual ~BaseStudyClass();

	/**
	 *   Returns Pointer to the PVModel which the study is defined on.
	 *   @return ThePVModel a pointer to the PVModel
	 */
	virtual PVModel* GetPVModel(bool expandtimearrays = false);

	/**
	 *   Returns the number of global variables defined in the study.
	 *
	 *   @return numGVs the number of global variables
	 */
	unsigned int GetNumGVs();

	/**
	 *   Returns the number of experiments in the study.
	 *
	 *   @return NumExps the number of experiments
	 */
	unsigned int GetNumExps();

	/**
	 *   Returns the experiment name for the current experiment
	 *
	 *   @return experiments name (i.e. 'Experiment X')
	 */
	const char * GetCurrentExpName() const;

	/**
	 *   Returns the experiment name for the given experiment
	 *
	 *   @return experiments name (i.e. 'Experiment X')
	 */
	char * GetExpName(int expNum) const;

	/**
	 *   Sets the global variable value arrays for each variable
	 *       of each type.
	 *
	 *   @param expNum The experiment number of the experiment
	 *                   for which the global variable values are to be set.
	 */
	virtual void SetGVs(int expNum);

	/**
	 *   Prints to standard out the global variable values for the given
	 *       experiment.  This method is called by the State Space
	 *       Generator and the Simulator.
	 *
	 *   @param  expNum  The experiment number corresponding to the
	 *           experiment for which global variable values will be printed.
	 */
	virtual void PrintGlobalValues(int expNum);

	/**
	 *   Returns pointer to the global variable value corresponding to the
	 *       global variable with name <code>TheGVName</code>
	 *
	 *   @param  *TheGVName  The global variable name for which the value
	 *           will be returned.
	 */
	virtual void* GetGVValue(char* TheGVName);

	/**
	 *   Returns gv name
	 *
	 *   @param  *num  the gv index
	 */
	char* GetGVName(int index);

	/**
	 *   Returns gv type string
	 *
	 *   @param  *num  the gv index
	 */
	char* GetGVType(int index);

	/**
	 *   Returns pointer to the global variable value corresponding to the
	 *       global variable with name <code>TheGVName</code>
	 *
	 *   @param  *TheGVName  The global variable name for which the value
	 *           will be returned.
	 *   @param  *TheGVValue  The global variable value to which the variable
	 *           will be set.
	 */
	virtual void OverrideGVValue(char *TheGVName, void *TheGVValue);
	void overrideGVs(const char* TheOverrideString);

	char * GetDefaultMobiusRoot();
	void SetDefaultMobiusRoot(char *root);

	/**
	 *   Overloaded method to set the value of a "Global Variable" class.
	 *   These classes act as global variables but restrict setting the
	 *   value to this base class.  These overloads provide the friendship
	 *   needed to access the Set method.
	 */
	void SetGvValue(Short& S,  short s)  { S.Set(s); }
	void SetGvValue(Int& I,    int i)	 { I.Set(i); }
	void SetGvValue(Long& L,   long l)	 { L.Set(l); }
	void SetGvValue(Float& F,  float f)  { F.Set(f); }
	void SetGvValue(Double& D, double d) { D.Set(d); }
	void SetGvValue(Bool& B,   bool b)	 { B.Set(b); }
	void SetGvValue(Char& C,   char c)	 { C.Set(c); }

protected:

	/**
	 *  Pointer to the array of global variable names
	 */
	char** GVNames;

	/**
	 *  Pointer to the array of global variable types
	 */
	char** GVTypes;

	/**
	 *  Pointer to the performance variable model
	 */
	PVModel* ThePVModel;

	/**
	 *  The number of global variables in this study
	 */
	unsigned int NumGVs;

	/**
	 *  The number of experiments (active + inactive)
	 */
	unsigned int NumExps;

	/**
	 *  Default path for MOBIUSROOT (helpful for some solvers
	 **/
	char* DefaultMobiusRoot;

	/**
	 *  The current experiment number
	 */
	unsigned int mCurrentExpNum;

	std::vector<std::string> mExperimentNames;

private:
	void init();

};

#endif
