/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File:  InstantOfTime.hpp
 *
 *      Authors:  Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1999 The Board of Trustees of the University of Illinois
 *                         All rights reserved.
 *
 \*****************************************************************************/

#ifndef DETERMINISTIC_LIST_HPP_
#define DETERMINISTIC_LIST_HPP_

#include <iostream>
#include <fstream>
#include "RateInfoType.hpp"
#include "../BaseClasses/StandardErrors.h"
#include "../BaseClasses/debug.h"
#include "../BaseClasses/MobiusEnums.h"

/**
 *  This defines the initial capacity of
 *  the RateInfo array if the default
 *  constructor is used.  Since there
 *  tends not be too many deterministic
 *  actions in a model this is set to
 *  be a small number initially.
 */
#define DEFAULT_RATE_INFO_CAPACITY 2

/**
 *  This class holds all the enabling
 *  states and value of the deterministic
 *  distribution for a single deterministic
 *  action in Mobius.  This data structure
 *  is used during state space exploration to
 *  remember all the enabling states and values
 *  such that a .det file can be created after
 *  the exploration is complete.
 *
 *  @author Jay M. Doyle
 *  @version 1.0
 */
class DeterministicList {

public:

	/**
	 *  Base constructor
	 */
	DeterministicList();

	/**
	 *  This constructor creates the RateInfo
	 *  data structure to the specified size
	 *
	 *  @param InitialSize The initial size of the
	 *         RateInfo data structure.
	 */
	DeterministicList(int InitialSize);

	/**
	 *  Destructor
	 */
	virtual ~DeterministicList();

	/**
	 *  This method stores the fact that the
	 *  action is enabled in the specified
	 *  state with the specified value for
	 *  the deterministic time distribution.
	 *
	 *  @param NewState The enabling state number
	 *  @param Value The value of the deterministic
	 *         action in the enabling state
	 */
	void insertEnablingState(int NewState, double Value);

	/**
	 *  This method writes the enabling state
	 *  information to a .det file using the
	 *  UltraSAN/Mobius format. <BR>
	 *
	 *  The .det file format is:
	 *  <OL>
	 *  <LI> A 1 to denote that the model had deterministic actions
	 *       otherwise and 0 to donote no deterministic actions
	 *  <LI> The time for the deterministic action
	 *  <LI> All states in which the deterministic activity can
	 *       complete with the time value in 1.
	 *  <LI> A 0 to mark the end of the possible states with the time value in item 1.
	 *  <LI> If more than one deterministic activity exists, (or the same deterministic
	 *       action with different time values) repeat items 2 through 4.
	 *  </OL>
	 *
	 *  @param OutputFile An open output file stream
	 *         used to write the enabling state
	 *         information to a file
	 *  @param TheType The type of the output file
	 *         create while writing the enabling
	 *         states (ASCII or Binary).
	 */
	void writeEnablingStates(fstream& OutputFile, OutputFileType TheType);

	/**
	 *  This method prints out all the values
	 *  and associated enabling states for
	 *  a determinitic action.  This method
	 *  also prints out the value of all internal
	 *  data members.  This method is used for
	 *  debugging only.
	 */
	void printEnablingStates();

	/**
	 *  This method is used to get
	 *  the total number of distinct
	 *  deterministic rates encapsulated
	 *  in RateInfo data structure.
	 */
	inline int getNumRates() {
		return RateInfoSize;
	} // end getNumRates()

protected:

	/**
	 *  This method increase the size of the
	 *  RateInfo data member to the specified
	 *  size.
	 *
	 *  @param NewSize The new size of the
	 *         RateInfo data member
	 */
	void increaseRateInfoSize(int NewSize);

	/**
	 *  The number of RateInfo elements
	 *  that are current being used.  This
	 *  number also equals to the total
	 *  number of unique values found
	 *  for the associated deterministic
	 *  action during state space exploration.
	 */
	int RateInfoSize;

	/**
	 *  This is the actual size of the RateInfo
	 *  data structure.
	 */
	int RateInfoCapacity;

	/**
	 *  This is an array that stores the
	 *  a list of elements, where each
	 *  elements conatains a list of enabling
	 *  states for a specfic determinisitic
	 *  value.
	 */
	RateInfoType* RateInfo;

private:

}; // end DeterministicList


#endif // end DETERMINISTIC_LIST_HPP_
