/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File:  InstantOfTime.hpp
 *
 *      Authors:  Jay M. Doyle, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1999 The Board of Trustees of the University of Illinois
 *                         All rights reserved.
 *
 \*****************************************************************************/

#ifndef RATE_INFO_TYPE_HPP_
#define RATE_INFO_TYPE_HPP_

#include <stdlib.h>
#include <iostream>
#include <fstream>
#include "../BaseClasses/debug.h"
#include "../BaseClasses/StandardErrors.h"
#include "../BaseClasses/MobiusEnums.h"

/**
 *  This is the default size of the StateList
 *  array
 */
#define DEFAULT_RATE_ARRAY_SIZE 10

/**
 *   This class encapsules a set of enabling states
 *   for a determinisitic action.  All states in this
 *   data structure have the same determinisitic firing
 *   time value.  An array of RateInfoType is used in
 *   the DeterministicList class to save all the enabling
 *   information for a deterministic action.
 *
 *   @see DeterministicList
 *
 *   @author Jay M. Doyle
 *   @version 1.0
 */
using namespace std;

class RateInfoType {

public:

	/**
	 *  Base constructor
	 */
	RateInfoType();

	/**
	 *  This constructor initialized the
	 *  StateList array to a specified size and
	 *  the value of the deterministic distribution.
	 *
	 *  @param InitialStateSize This is the initial
	 *         capacity of the StateList data
	 *         structure.
	 *  @param TheValue The value of the associated
	 *         deterministic action in all the enabling
	 *         states in this data structure.
	 */
	RateInfoType(int InitialStateSize, double TheValue);

	/**
	 *  Destructor
	 */
	virtual ~RateInfoType();

	/**
	 *  This method returns the value of the
	 *  deterministic action in all the states
	 *  saved in the StateList data structure.
	 *
	 *  @return The value of the deterministic
	 *          action in all the associated
	 *          enabling states
	 */
	inline double getValue() {
		return Value;
	} // end getValue()

	/**
	 *  This method sets the value of the deterministic
	 *  distribution for all the associated enabling
	 *  states.
	 */
	inline void setValue(double TheValue) {
		Value = TheValue;
	} // end setValue(double)


	/**
	 *   This method returns the number of
	 *   enabling states in StateList data
	 *   structure.
	 *
	 *   @return The number of enabling states
	 */
	inline unsigned int getNumStates() {
		return NumStates;
	} // end getNumState()

	/**
	 *  This method inserts the state into
	 *  the StateList data structure.  This
	 *  method does not prevent duplication in
	 *  the StateList.
	 *
	 *  @param NewState A state in which the
	 *         associated deterministic action
	 *         is enabled and has a value == Value.
	 */
	void insertState(int NewState);

	/**
	 *  This method forces the object into a
	 *  a state where the destructor will not
	 *  delete the StateList array upon object
	 *  destructor.  See MarkedForDeletion data
	 *  member for more details.
	 */
	inline void saveStateList() {
		MarkedForDeletion = false;
	} // end RateInfoType::saveStateList

	/**
	 *  This method writes all the enabling states
	 *  to an output file.  Each enabling state
	 *  is listed on a separate line.
	 *
	 *  @param OutputFile The an opend output file
	 *         stream used to write the enabling
	 *         states to file.
	 *  @param TheType The output file type (ASCII, Binary)
	 */
	void writeStates(fstream& OutputFile, OutputFileType TheType);

	/**
	 *  This is a method used to print the
	 *  enabling state information to standard
	 *  out.  This is only used for debugging.
	 */
	void printStates();

protected:

	/**
	 *  This method increases the size of
	 *  the StateSize array to the
	 *  specified size.
	 *
	 *  @param NewSize The desired new size
	 *   of the StateList data structure
	 */
	void increaseStateList(int NewSize);

	/**
	 *  The number of enabling states
	 *  current stored in StateList
	 */
	unsigned int NumStates;

	/**
	 *  The size of the StateList array. Not
	 *  all the elements may be used.
	 */
	unsigned int Capacity;

	/**
	 *  This is the deterministic firing time
	 *  for the associated action
	 */
	double Value;

	/**
	 *  A list of enabling states with the
	 *  same deterministic firing time value
	 */
	int* StateList;

	/**
	 *  This flag denotes if the StateList array
	 *  should be deleted when the destructor
	 *  is called.  An instance of RateInfoType may
	 *  be copied, in which case we have two objects
	 *  that point to a single StateList array.  One
	 *  may be deleted, but we want the other one to
	 *  maintain the StateList data structure.  See
	 *  DeterministicList::increaseRateInfoSize(int)
	 *  to see how this used in practice.
	 */
	bool MarkedForDeletion;

private:

};

#endif // RATE_INFO_TYPE_HPP__
