/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: StateSpaceGen.h
 *
 *      Authors: John Sowder, Daniel D. Deavours, William H. Sanders
 *      Creation Date: 1999
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

// This is the header file for the State Space Generator class

#ifndef STATESPACEGEN_H
#define STATESPACEGEN_H
#include "../BaseClasses/debug.h"
#include "../BaseClasses/TraceFileContainer.h"
#include "../BaseClasses/TraceFileMobiusSSG.h"
#include "../TRex/ResultsDB.h"
#include "gen.h"
#include "queue.h"
#include "HybridHash.h"
#include "FileType.h"
#include "Reward.h"
#include "DeterministicList.hpp"
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <math.h>
#include <map>
#ifndef MINGW
#include <sys/times.h>
#endif

/** StateSpaceGen
 * Version 1.0
 * 8-6-98
 * This class implements the state space generator.  The state space generator
 * supports exponential, deterministic, and instantaneous actions.  The
 * important methods are the findnext and fire instant method which are
 * action specific methods that find all next tangible states
 */

class StateSpaceGen {

private:

	/*
	 * An explanation will not be given for every, just the key ones
	 */
	unsigned long current_state_num;

	/*
	 * The sequence of next states and the their corresponding rates from
	 * the current state
	 */
	map<unsigned int, double> next_states;
	typedef map<unsigned int, double>::iterator map_iterator;

	/**
	 *  global_current_state is the current state number in the model
	 */
	unsigned long num_timed, global_current_state, State_Size;

	/*change */
	unsigned long num_erlang;
	/**
	 *
	 */
	ResultsDB *DB; // the database instrumentation methods

	/**
	 *
	 */
	unsigned long ActionTypeFlag;

	/**
	 *
	 */
	PVModel *The_Model;

	/**
	 *
	 */
	Reward *RewardModel;

	/**
	 *
	 */
	Queue *queue;

	/**
	 *
	 */
	HybridHash* StateTree;

	/**
	 *
	 */
	BaseActionClass **instant_list;

	/**
	 *
	 */
	BaseActionClass **timed_list;

	/* change */
	// Added by gaonkar_erlang_improvement
	BaseActionClass **erlang_list;

	/**
	 *  This is data structure used to collect
	 *  all the actions defined in the model.
	 *  It is passed as an argument to the
	 *  listActions method defined on a Mobius
	 *  model.
	 */
	List<BaseActionClass> ActionContainer;

	/**
	 *
	 */
	BaseActionClass **activity_list;

	/**
	 *  This is a data structure use to collect
	 *  the list of all the groups in the model.
	 *  It is passed to the listGroups method
	 *  defined on a Mobius model
	 */
	List<BaseGroupClass> GroupContainer;

	/**
	 *
	 */
	BaseGroupClass **Groups;

	/**
	 *
	 */
	state_struct *new_store;

	/**
	 *
	 */
	double global_rate;

	/**
	 *  This data structure maintains a list
	 *  of enabling states for all the deterministic
	 *  actions in the model.  For each state we
	 *  remember the value of the deterministic
	 *  action in that model state.  This information
	 *  is used to build the .det file used by
	 *  the adiss and diss solvers.
	 */
	DeterministicList detlist;

	/**
	 *
	 */
	char *new_state_holder;

	/**
	 *
	 */
	char *TheFilePath;

	/**
	 *
	 */
	char *TheExpName;

	/**
	 *
	 */
	short FlagAbsorbing;

	/** Finds all next states for a model containing only exponential
	 * actions
	 */
	void FindNextNormal(char *);
	/** Finds all next states for a model containing only instantaneous and
	 * exponential actions
	 */
	void FindNextInstant(char *);
	/** Finds all next states for a model containing cany combination of
	 * deterministic, exponential, or instantaneous
	 */
	void FindNextDeterministic(char *);
	/** Fire method for models containing instantaneous actions
	 */
	void FireInstant(double);
	/** Fire method for instantaneous actions for models with deterministic
	 * actions
	 */
	void FireInstantDet(double);

	/** Store the new state in the state tree
	 */
	void SaveStableState(double rate, int index);
	/** Fire an action, including erlang actions
	 */
	bool FireAction(BaseActionClass *action, char *state_holder);

	/** Prints the state to standard out for Java interface to use
	 */
	inline void OutputState();

	void SetMethodPointers();

public:
	/** constructor
	 */
	StateSpaceGen(PVModel *, const char *, const char *, short, double, const char*, const char*, char* trace_filename = nullptr);

	/** destructor
	 */
	~StateSpaceGen();

	/** begins initial setup and state generation
	 */
	void GenStateSpace();

	/** method for generation if only exponential and/or instantaneous action
	 * are in the model
	 */
	void GenStateSpaceNormal(char *, char *);

	/** method for generation if only deterministic,exponential and/or
	 * instantaneous actions are in the model
	 */
	void GenStateSpaceDeterministic(char *, char *, char*);

	/**
	 *  The selected output file format
	 */
	short filetype;

	/**
	 * trace file name
	 */
	char *TraceFilename;

	/** Mobius SSG trace */
	TraceFileContainer * trace;
	TraceFileMobiusSSG * traceMobSSG;
	int TraceLevel;
	void InitializeTrace();

	/**
	 *  The number of recursively fired instantaneous activities in the current
	 *  stable state.
	 */
	short num_ztrans;

	/**
	 *
	 */
	unsigned long num_instant;

	/**
	 *
	 */
	double hash_value;

	/**
	 *
	 */
	void setHashValue(double);

	/**
	 *  Set the output file type for state space files
	 */
	inline void SetFileType(int filetype) {
		this->filetype = filetype;
	}

	/**
	 * Sets the trace level.
	 */
	inline void SetTraceLevel(int Level) {
		cout << "SSG: SetTraceLevel: " << Level << endl;
		TraceLevel = Level;
	}
};

inline void StateSpaceGen::OutputState() {

	if (current_state_num % 1000 == 0)
		cout << "_$" << current_state_num << "$" << endl;
}

typedef void (StateSpaceGen::*PFMI)(char *);
#endif

