/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef GraphSymbolicSSG_H_
#define GraphSymbolicSSG_H_

#include "BaseClasses/symbolic/ssnode.h"
#include "BaseClasses/symbolic/mdd.h"
#include <ext/hash_map>
#include "BaseClasses/BaseModelClass.h"
#include "Performance_Variables/PVModel.hpp"
#include "BaseClasses/symbolic/DynamicObjects.h"
#include <map>
#include "BaseClasses/symbolic/mdnode.h"
#include "BaseClasses/symbolic/md.h"
#include "BaseClasses/symbolic/DoubleFunction.h"
#include "BaseClasses/symbolic/Visualize.h"
#include "BaseClasses/symbolic/MappedMDD.h"
#include <set>
#include <queue>
#include "PVMTMDDCreator.h"
#include "GraphSymDS.h"

#define MAX_IMM_TRANSITIONS 25
#define NUM_SYM_PARAMS 6

class GraphSymbolicSSG {

	/**************************************************/
	/** Public Methods                                */
	/**************************************************/
public:
	GraphSymbolicSSG(PVModel * _pvModel, int* SymbolicParams, bool CompositionalLumpingEnabled);
	~GraphSymbolicSSG();

	typedef BaseMDDNode<short> MDDNode;

	void InitializeMDD();
	void InitializeMD();
	int insertState(int level, State* state, BaseModelClass* model);

	void ConvertLocalR2MDNodes();
	void BuildLocalTree(int submodel_idx);
	void BuildGlobalTree(int submodel_idx);
	MDNode* BuildGlobalTree(TMDMatrix& mat, int submodel_idx, int ancestor_idx);
	void WriteFlatSSFile(MDDNode* node, const char* ss_filename);
	void WriteFlatSSFile(MDDNode* node, byte* buffer, FILE* f);
	void WriteSSFile(SSNode* root, char* ss_filename);
	void SkipCompositionalLumping();

	/** Writes all of the local state spaces
	 * for use in the solvers. */
	void WriteLocalSSFile(string &file_name);

	/** Writes the permutations for use in the solvers */
	void WritePermFile(string &file_name);

	void GenerateMCG();
	void GenerateSS();
	void LumpSSAndProjectMxD();
	void GenerateSprime();

	void ProjectMatrixDiagram();
	void ModelLevelLumpSS();
	MDDNode* ModelLevelLumpSS(MDDNode* R);
	void ModelLevelLumpSSWithTrim();
	MDDNode* ModelLevelLumpSSWithTrim(MDDNode* node);
	bool * af_partial_perm_true;
	int * af_lump_indices;
	MDDNode* CopyMDD(MDDNode* node);
	MDDNode* CreateLumpedMDD(MDDNode* node);
	void BuildMatrixDiagram();
	void BuildMappingMDD();
	void TestMappingMDD(SSNode* v);
	PATH_TYPE* path;
	void GeneratePVMTMDDs();
	int TotalStateSize;
	int GetStateSpaceIndex(int level, char* buff, int size);
	SSNode* BuildGraphLumpedNodes(int * StateIndices);
	void PrintMDDStates(MDDNode* node);
	void PrintMDDStates(MDDNode* node, byte* buffer);
	void PrintMD(MDNode* node, string prefix);
	void PrintIndices(int* indices);

	void ReportStateSpaceSizes(bool reset_offsets, SSNode* root, char* rootname);
	void ReportStatistics();

	PATH_TYPE getUnlumpedStateIndex(int level, PATH_TYPE lumped_state_index);
	INDEX_TYPE getNoOfUnlumpedStates() {
		return ((SSNode*) S)->getNStates();
	}

	INDEX_TYPE getNoOfLumpedStates() {
		return ((SSNode*) CMLumpedS)->getNStates();
	}

	vector<State*> *getStateSpaceVector() {
		return StateSpace;
	}

	GraphSymDS* getSystem() {
		return GraphSystem;
	}
	PVModel* getPVModel() {
		return pvModel;
	}

	int GetNumUnlumpedStates();
	// The MDDs used
	MDDNode *S, *U, *MLumpedS, *CLumpedS, *Sp, *CLumpedSp;
	MDDNode *SemiMLumpedS;
	MDDNode *CMLumpedS;
	MDDNode *R;
	PVMTMDDCreator* pvmtmdd_creator;

	MappingNode *MappingMDD;
	ComputeMappingMDDEnv* CMNE;

	MDNode* TransMat;

	/** Used in ModelLevelLump a lot */
	byte* input_state;
	byte* new_state;
	vector<int> temp_map;
	vector<int> StateSpaceSizes;
	vector<int> LumpedSSSizes;

	/** Debug Methods */

	/**************************************************/
	/** Private Attributes                            */
	/**************************************************/
private:

	/** === MD & MDD Attributes === */
	/** Size of the Unique Table Size of the MDD */
	int UniqueTableSize_MDD;

	/** Size of the Union Cache Size of the MDD */
	int UnionCacheSize;

	/** Size of the Local Exploration Cache Size */
	int SaturCacheSize;

	/** Size of the Unique Table Size of the MD */
	int UniqueTableSize_MD;

	/** Size of the Submatrix Cache Size of the MD */
	int SubmatCacheSize;

	/** Size of the Add Cache Size of the MD */
	int AddCacheSize;

	/** \todo */
	List<BaseGroupClass> GroupContainer;

	/** \todo Description */
	BaseGroupClass **Groups;

	/** Enable Compositional Lumping. Affects... */
	bool CompositionalLumpingEnabled;

	/** \todo */
	PVModel* pvModel;
	GraphSymDS * GraphSystem;

	/** Bool for each level marking whether the model
	 * is local or not.
	 *
	 * True for all atomic models containing state.
	 */
	bool * IsLocalLevel;

	// Terminal SSNodes
	SSNode* Value0;
	SSNode* Value1;

	/** MDD containing SSNode (MDD nodes) */
	MDD<SSNode> *SS;

	// The MDs used
	MD* md;

	//An array of MDNodes pointers to the trees corresponding to the representation of
	//local(nonlocal) behaviour (using the local(nonlocal) matrices) of the model.
	/** \todo */
	MDNode** LocalMDTree;
	MDNode** NonLocalMDTree;
	int * LastLocalIndex;
	TFunc NewStatesTree;

	/** Linked list of SSNodes.
	 *
	 * Contains an array of  linked lists of every node in the MDD
	 * at each level.  Uses a tagged list to ensure that only one
	 * copy of the node is inserted. \todo Is this correct?
	 */
	SSNode::node_slist_type* NodesList;

	/** Linked List of Useful SSNodes.
	 *
	 * \todo More here
	 */
	ssnode_list_t UsefulCopiesList;

	//row could be declared locally in the method it's used. However, there is
	//some cost associated with its allocation and grow that I decided to define
	//it global to the object.
	DynamicArray<REAL, true> *row;
	SparseMatrix<REAL> *localR;
	MDNode** localMDNode;

	/** \todo comments correct?
	 */
	//Balanced trees to store non-local matrices
	//The first pair corresponds to the indices of the shared states of the
	//destination and source states respectively. The second pair corresponds
	//to the indices of the local states of the destination and source states
	//respectively. 'double' corresponds to the rate of the transition
	TMDMatrix *nonlocalR;
	MDNode* BuildIdentityMatrixPath(int level_start, int level_end, MDNode* downPointer);

	/** \todo Descriptoin Bstar */
	typedef bool BstarType[500][500];
#ifdef FIXED_BSTAR
	BstarType *Bstar;
#else
	DynamicFullMatrix<bool> *Bstar;
#endif

	vector<pair<int, int> > *B;

	int NumLevels;

	int * MDDLevel2SubmodelIndex;
	int * SubmodelIndex2MDDLevel;
	int NumLocalFirings;
	int NumNonLocalFirings;

	State **NextState;

	/** List of variables that are passed across functions.
	 *
	 * AF = "Across Functions"
	 */
	int AF_NumImmTrans;
	BaseModelClass* AF_Submodel;

	/** Across Function Nodes.
	 *
	 * \todo Correct?
	 * I think this represents the nodes in a path along
	 * the MDD.
	 */
	SSNode** AF_Nodes;
	int AF_SubmodelIndex;

	/** Represents which state indices (downward pointers)
	 * to use in a MDD when firing global states
	 */
	int* AF_StateIndices;
	int AF_NumDependencies;
	int* AF_NextStateIndices;
	SSNode* AF_NewTail;
	int* AF_Levels;
	TMDMatrix AF_Transitions;

	/** Used in ModelLevelLump to search through the MDD */
	int* AF_LumpStateIndices;

	/** Used in PrintMDDStates to search through the MDD */
	int* AF_PrintStateIndices;
	int* AF_PrintMDMajorIndices;
	int* AF_PrintMDMinorIndices;
	int AF_Rate;

	/** \todo Is this correct here? Description? */
	map<SSNode*, SSNode*> ReplaceMap;

	/** Various Hash Tables \todo */
	/** \todo */
	//We want to have a cache for saturation operator. The operator takes
	//a node and returns the saturated node.
	class SaturationCacheKey {
	public:
		SaturationCacheKey() {
			R = NULL;
		}
		~SaturationCacheKey() {
			reset();
		}
		SaturationCacheKey(SSNode* _R) {
			R = _R;
		}

		//Computes the hash function on the key
		size_t operator()(SaturationCacheKey& key) {
			return (unsigned int) key.R->getID() * 97;
		}

		//Compares two keys
		bool operator==(SaturationCacheKey& key2) {
			return ((void*) R == (void*) key2.R);
		}
		void reset() {
			R = NULL;
		}

		//A cache entry is dead if the MDD node of the key is dead
		bool isDead() {
			return R->isDead();
		}
	private:
		MDDNode* R;
	};

	/** Saturation Cache.
	 *
	 * Caches local saturation procedures. This way, when the node is found again,
	 * we can just pull the result from the cache. */
	class CSaturationCache: public ComputedCache<SaturationCacheKey, MDDNode> {
	public:
		void GC();
	}*SaturationCache;

	struct hash_func_state {
		size_t operator()(State* s) const {
			return s->hash();
		}
	};

	struct eq_state {
		bool operator()(State* s1, State* s2) const {
			return (s1->compare(*s2) == 0);
		}
	};

	/** \todo Description */
	typedef __gnu_cxx::hash_map<State*, int, hash_func_state, eq_state> THashtable;
	THashtable *StateHashTable;

	/** A vector of State* for each model */
	vector<State*> * StateSpace;

	//void PrintState(void * state, int NumBytes);


	/**************************************************/
	/** Private Methods                               */
	/**************************************************/
	SSNode* LocalSSE(SSNode* root, int MDDlevel = -1);
	SSNode* Saturate_DFS(SSNode* v, int MDDlevel = -1);
	SSNode* SaturateLocallyMultiSteps(SSNode* v);
	SSNode* SaturateLocallyOneStep(SSNode* v);
	void ExploreOneStepLocally(int level, int x);

	SSNode* BuildNewGlobalStates(TFunc& tree, int k);
	MDDNode* GlobalSSE(MDDNode* root);
	MDDNode* FireAllGlobal(int k);
	void FireGlobal();

	void SetStateInLevels(int submodel_index, const int* const state_indices);

	void FireImmediateActionsSequence(int x, int level, REAL rate, bool local);

	/** \todo Inline or not? */
	MDDNode* Diff(MDDNode* S1, MDDNode *S2);
	void Union(MDDNode*& result, MDDNode* S1, MDDNode*S2);
	MDDNode* Union(MDDNode* S1, MDDNode *S2);
	void Intersect(MDDNode*& result, MDDNode* S1, MDDNode*S2);
	MDDNode* Intersect(MDDNode* S1, MDDNode *S2);
	SSNode* BuildLowerPart(int level, bool CompLumpEnabled);

	/** \todo Comments */
	void TagUsefulNodesAux(SSNode *head, int level, bool& path_to_tail_exists, ssnode_list_t* tagged_nodes_list);
	bool TagUsefulNodes(SSNode *head, SSNode* tail, ssnode_list_t* tagged_nodes_list);
	void UntagNodes(ssnode_list_t* tagged_nodes_list);

	SSNode* CopyUsefulNodes(SSNode *head);
	SSNode* CopyUsefulNodesAux(SSNode *head);
	SSNode* CopyUptoLevelk(SSNode *head, int k);
	SSNode* CopyUptoLevelkAux(SSNode *head, int k);
};

/** Sets state in each level.
 *
 * \todo Ensure correctness, inline?
 * - Function assumes that DepList and DepLevels are in the same order
 *   - Caused me to sort the lists in GraphSymDS
 */
inline void GraphSymbolicSSG::SetStateInLevels(int submodel_index, const int* const state_indices) {
	GraphAtomicModel* atomic_model = GraphSystem->AtomicModels[submodel_index];
	for (int i = 0; i < atomic_model->GetNumDependencies(); i++) {
		BaseModelClass* const model = atomic_model->DepList[i];
		const int SS_ancestor_index = GraphSystem->MDDLevel2SSIndex[atomic_model->DepLevels[i]];
		model->SetLocalState(StateSpace[SS_ancestor_index][state_indices[i]]);
	}
}

#endif
