/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef MappedMDD_H_
#define MappedMDD_H_

#include "ssnode.h"
#include <set>
#include <list.h>
#include <map>
#include "cache.h"

class TRepSpec {
public:
	TRepSpec() {
	}
	TRepSpec(short _level_index, short _no_of_replicas, short _replica_size, int _parent_rep_idx) {
		level_index = _level_index;
		no_of_replicas = _no_of_replicas;
		replica_size = _replica_size;
		parent_rep_idx = _parent_rep_idx;
	}
	short level_index;
	short no_of_replicas;
	short replica_size;
	short parent_rep_idx;
};

//In a path array, index 0 corresponds to the uppermost level.
//The order of reps in TOrder should be in a such a way that inner reps
//come before (i.e. lower indices in rep_spec array) outer ones

class TOrder {
public:
	TOrder(short _no_of_reps) {
		no_of_reps = _no_of_reps;
		cur_no_of_reps = 0;
		rep_spec = new TRepSpec[no_of_reps];
	}

	~TOrder() {
		delete[] rep_spec;
	}

	void AddRep(TRepSpec& RepSpec);

	void AddRep(short level_index, short no_of_replica, short replica_size, short parent_rep_idx);
	TRepSpec& operator[](short idx);

public:
	short no_of_reps;
private:
	TRepSpec* rep_spec;
	short cur_no_of_reps;
};

class MappingNode: public MDDNode {
public:
	MappingNode(short variable, short _size);
	MappingNode(short val);
	~MappingNode();

	INDEX_TYPE getOffset(short i) const {
		return offset[i];
	}
	void setOffset(short i, INDEX_TYPE off) {
		offset[i] = off;
	}
	INDEX_TYPE computeOffset(PATH_TYPE* indices);
	void merge(const MappingNode* v);
	void print(bool detailed = true) const;
	virtual int operator==(const MappingNode& node);
	void GetSpecification(int& no_of_nodes, int& size_in_memory);
	size_t hash();
	int SizeInMemory() {
		int result = 6 * 4 + 2;
		if (var >= 0)
			result += size * (8 + 4);
		return result;
	}
protected:
	virtual void writeNode(FILE* f);
private:
	void MarkNodes(set<MappingNode*>& l);
	INDEX_TYPE* offset;
};

struct MappingNodeCmp {
	bool operator()(MappingNode* v1, MappingNode* v2) const;
};

//typedef set<MappingNode*, MappingNodeCmp> TMNodeUniqueTable;
typedef list<MappingNode*> TMNodeUniqueTable;

class ComputeMappingMDDEnv {
private:
	typedef map<int, MappingNode*> mapping;
public:
	ComputeMappingMDDEnv(SSNode* _root, SSNode* _lumpedSS, TOrder* _order);
	ComputeMappingMDDEnv(FILE* f);
	~ComputeMappingMDDEnv();
	MappingNode* ComputeMappingMDD(SSNode* v, INDEX_TYPE& min_offset);
	MappingNode* ComputeMappingMDD();
	MappingNode* getMappingMDD() {
		return MappingMDD;
	}
	void DestructMappingMDD();
	MappingNode* readTree(FILE* f);
	MappingNode* dfsReadTree(FILE* f, mapping& nodetable);

#ifndef MERGE_NODES
	MappingNode* insertNodeUnique(MappingNode* node);
	pair<MappingNode*, bool> findOrAddNode(MappingNode* node);
	void EnableGC() {
		GCEnabled = true;
	}
	void DisableGC() {
		GCEnabled = false;
	}
#endif

private:

	void SetupUniqueTable(void);

	short no_of_levels;
	PATH_TYPE *path;
	PATH_TYPE *sorted_path;
	TOrder *order;
	SSNode* root;
	SSNode* lumpedSS;
	MappingNode* MappingMDD;
	INDEX_TYPE state_counter;

#ifdef MERGE_NODES
	TMNodeUniqueTable* UniqueTable;
#else
	NodeUniqueTable<MappingNode> UniqueTable;
	int MAXLOADFACTOR;
	float MAXDEAD;
#endif
	bool GCEnabled;
};

void ComputeMappingMDD(SSNode* root, SSNode* lumpedSS, TOrder* order);

//In a path array, index 0 corresponds to the uppermost level
void sort_path(PATH_TYPE* p, TOrder &order);
bool sort_any_path(PATH_TYPE* p, TOrder &order);
inline bool is_nongreater_sequence(PATH_TYPE *p1, PATH_TYPE *p2, short size);
inline bool sorted(PATH_TYPE *p, TOrder &order);
short rep_sorted(PATH_TYPE *p, TRepSpec& rep_spec);
inline bool is_replica_nongreater(PATH_TYPE *p, TRepSpec& rep_spec, short replica1, short replica2);
inline short start_of_replica(TRepSpec& rep_spec, short replica);
void move_replica_to_right(PATH_TYPE *p, TRepSpec& rep_spec, short i);
void move_replica_to_left(PATH_TYPE *p, TRepSpec& rep_spec, short i);
INDEX_TYPE Path2Offset(SSNode* node, PATH_TYPE* p);
void write_paths_rec(SSNode* node, PATH_TYPE* p, PATH_TYPE* sorted_p, short no_of_levels, TOrder& order);
void print_path(PATH_TYPE *p, short path_len);
void write_paths(SSNode* root, TOrder& order);

bool ValidUniqueTable(TMNodeUniqueTable& UT);

inline short start_of_replica(TRepSpec& rep_spec, short replica) {
	return rep_spec.level_index + 1 + replica * rep_spec.replica_size;
}

inline bool is_replica_nongreater(PATH_TYPE *p, TRepSpec& rep_spec, short replica1, short replica2) {
	return is_nongreater_sequence(p + start_of_replica(rep_spec, replica1), p + start_of_replica(rep_spec, replica2), rep_spec.replica_size);
}

inline bool sorted(PATH_TYPE *p, TOrder &order) {
	for (short rep_idx = 0; rep_idx < order.no_of_reps; rep_idx++) {
		if (rep_sorted(p, order[rep_idx]) != -1)
			return false;
	}
	return true;
}

inline bool is_nongreater_sequence(PATH_TYPE *p1, PATH_TYPE *p2, short size) {
	short i = 0;
	while (i < size) {
		if (p1[i] == p2[i])
			i++;
		else
			return (p1[i] < p2[i]);
	}
	return true;
}

#endif
