/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef SymbolicSSG_H_
#define SymbolicSSG_H_

#include <ext/hash_map>
#include "BaseClasses/symbolic/ssnode.h"
#include "BaseClasses/symbolic/mdd.h"
#include "BaseClasses/BaseModelClass.h"
#include "Performance_Variables/PVModel.hpp"
#include "BaseClasses/symbolic/DynamicObjects.h"
#include <map>
#include "BaseClasses/symbolic/mdnode.h"
#include "BaseClasses/symbolic/md.h"
#include "BaseClasses/symbolic/DoubleFunction.h"
#include "BaseClasses/symbolic/Visualize.h"
#include "BaseClasses/symbolic/MappedMDD.h"
#include <set>
#include <queue>
#include "SymDS.h"
#include "PVMTMDDCreator.h"
#include "LumpingMxD.h"

#define MAX_IMM_TRANSITIONS 25
#define NUM_SYM_PARAMS 6

void pois_pts(REAL lambda, REAL epsilon, unsigned long* L, unsigned long *R, REAL **points);

class SymbolicSSG {
public:
	SymbolicSSG(PVModel* _pvModel, int* SymbolicParams, bool CompositionalLumpingEnabled);
	~SymbolicSSG();
	void GenerateSS();
	void LumpSSAndProjectMxD();
	void PerformCompositionalLumping();
	void SkipCompositionalLumping();
	void GeneratePVMTMDDs();
	void PrepareBuildingMappingMDD();
	void BuildMappingMDD();
	void ReportStatistics();
	void ReportStateSpaceSizes(bool reset_offsets, SSNode* root, char* rootname);

	void BuildMatrixDiagram();

	void TestMappingMDD(SSNode* v);

	PVModel* getPVModel() {
		return pvModel;
	}
	SymDS* getSystem() {
		return System;
	}
	vector<State*> *getStateSpaceVector() {
		return state_space;
	}
	PATH_TYPE getUnlumpedStateIndex(int level, PATH_TYPE lumped_state_index);
	INDEX_TYPE getNoOfUnlumpedStates() {
		return ((SSNode*) S)->getNStates();
	}

	MDDNode *S, *U, *MLumpedS, *CMLumpedS, *CLumpedS, *Sp, *CLumpedSp;
	MDDNode *SemiMLumpedS;
	MappingNode *MappingMDD;
	MDD<SSNode> *SS;
	PVMTMDDCreator* pvmtmdd_creator;

	ComputeMappingMDDEnv* CMNE;
	MD* md;
	MDNode* TransMat;
private:
	typedef map<MDDNode*, MDDNode*> TLumpingMDDMap;
	SymDS* System;
	PVModel* pvModel;
	bool CompositionalLumpingEnabled;
	List<BaseGroupClass> GroupContainer;
	BaseGroupClass **Groups;
	static char* ErrorMessages[4];
	//System->no_of_levels() is used a lot so we cache it in the following
	//variable
	short no_of_levels;

	vector<int> SSSizes;
	vector<int> LumpedSSSizes;
	LumpingMxD* LumpMxDObject;
	PATH_TYPE** CLumped2UnlumpedStateIndex;
	void ModelLevelLumpSS();

	void LumpMatrixDiagram();
	void CompositionalLumpSS();
	MDDNode* CompositionalLumpMDD(MDDNode* node);
	MDDNode* CompositionalLumpMDDAux(MDDNode* node, TLumpingMDDMap& LumpedMDDNodes);
	MDDNode* CompositionalUnlumpMDD(MDDNode* lnode);
	MDDNode* CompositionalUnlumpMDDAux(MDDNode* lnode, TLumpingMDDMap& UnlumpedMDDNodes);
	MDDNode* ModelLevelLumpSS(MDDNode* node, bool CompLumpEnabled, bool SemiLumping);
	MDDNode* ModelLevelLumpSSDirectly(MDDNode* node, bool CompLumpEnabled, bool SemiLumping);
	INDEX_TYPE getNoOfLumpedStates() {
		return ((SSNode*) CMLumpedS)->getNStates();
	}
	INDEX_TYPE getNoOfSuccStates() {
		return ((SSNode*) CLumpedSp)->getNStates();
	}

	void GenerateSprime();

	SSNode* LocalSSE(SSNode* root, int MDDlevel = -1);
	SSNode* Saturate_DFS(SSNode* v, int MDDlevel = -1);
	void ExploreOneStepLocally(int level, int x);
	void FireImmediateActionsSequence(int x, int level, REAL rate, bool local);
	void ModifyBstar(int LM_index, int SS_index, int source, int dest);
	SSNode* SaturateLocallyMultiSteps(SSNode* v);
	SSNode* SaturateLocallyOneStep(SSNode* v);
	MDDNode* GlobalSSE(MDDNode* root);

	void ProjectMatrixDiagram();

	void SequenceLessOrEqual(MDDNode*& Set, int length, int start_level_1, int start_level_2, bool CompLumpEnabled);

	SSNode* BuildNewGlobalStates(TFunc& tree, int k);
	MDDNode* FireAllGlobal(int k);
	void FireGlobal();

	void SetStateInLevels(int submodel_index, const int* const state_indices);

	void TagUsefulNodesAux(SSNode *head, int level, bool& path_to_tail_exists, ssnode_list_t* tagged_nodes_list);
	bool TagUsefulNodes(SSNode *head, SSNode* tail, ssnode_list_t* tagged_nodes_list);
	void UntagNodes(ssnode_list_t* tagged_nodes_list);
	SSNode* CopyUsefulNodes(SSNode *head);
	SSNode* CopyUsefulNodesAux(SSNode *head);
	SSNode* CopyUptoLevelk(SSNode *head, int k);
	SSNode* CopyUptoLevelkAux(SSNode *head, int k);

	int insertState(int level, State* state, BaseModelClass* model);
	void InitializeMDD();
	void DeleteMDD();

	void ConvertLocalR2MDNodes();
	void InitializeMD();
	void DeleteMD();
	void BuildLocalTree(int submodel_idx);
	void BuildGlobalTree(int submodel_idx);
	MDNode* BuildGlobalTree(TMDMatrix& mat, int submodel_idx, int ancestor_idx);

	//Balanced trees to store non-local matrices
	//The first pair corresponds to the indices of the shared states of the
	//destination and source states respectively. The second pair corresponds
	//to the indices of the local states of the destination and source states
	//respectively. 'double' corresponds to the rate of the transition

	TMDMatrix *nonlocalR;
	void print_nonlocal_matrices();
	void print_local_matrixnodes();
	MDNode* BuildIdentityMatrixPath(int level_start, int level_end, MDNode* downPointer);

	void print_local_matrices();
	void print_local_state_spaces();
	void print_MDD_nodes(bool detailed);
	void print_all_nodes(bool detailed);
	void print_live_nodes();
	MDDNode* Union(MDDNode* S1, MDDNode *S2);
	void Union(MDDNode*& result, MDDNode* S1, MDDNode*S2);
	MDDNode* Diff(MDDNode* S1, MDDNode *S2);
	void Intersect(MDDNode*& result, MDDNode* S1, MDDNode*S2);
	MDDNode* Intersect(MDDNode* S1, MDDNode *S2);

	SSNode* BuildLowerPart(int level, bool CompLumpEnabled);
	SSNode* BuildMiddlePart(int level1, int level2, bool CompLumpEnabled, SSNode** node_ptrs);
	SSNode* BuildUpperPart(int level, bool CompLumpEnabled, SSNode* Node);
	SSNode* LessOrEqual(int level1, int level2, bool CompLumpEnabled);
	SSNode* MoreOrEqual(int level1, int level2, bool CompLumpEnabled);
	SSNode* Less(int level1, int level2, bool CompLumpEnabled);
	SSNode* More(int level1, int level2, bool CompLumpEnabled);
	SSNode* Equal(int level1, int level2, bool CompLumpEnabled);

	void print_dag_aux(SSNode* v, ssnode_tnodestatus_list_t& list);
	void print_dag(SSNode* v);

	void WriteVizFile(SSNode* root, char* viz_filename);
	void WriteSSFile(SSNode* root, char* ss_filename);

	//Declaration of the hash tables and required auxiliary functions
	struct hash_func_state {
		size_t operator()(State* s) const {
			return s->hash();
		}
	};

	struct eq_state {
		bool operator()(State* s1, State* s2) const {
			return (s1->compare(*s2) == 0);
		}
	};

	//An array of hashtables, one for each submodel
	//Used to map state object to state number
	typedef __gnu_cxx::hash_map<State*, int, hash_func_state, eq_state> THashtable;
	THashtable *state_hashtable;

	//An array of expandable vectors, one for each submodel
	//Used to map state number to state object
	vector<State*> *state_space;

	//Not sure whether Value0 will ever be used. Can be replaced by NULL pointer
	SSNode* Value0;
	SSNode* Value1;

public:
	typedef bool BstarType[500][500];

private:
#ifdef FIXED_BSTAR
	BstarType *Bstar;
#else
	DynamicFullMatrix<bool> *Bstar;
#endif

	vector<pair<int, int> > *B;

	//row could be declared locally in the method it's used. However, there is
	//some cost associated with its allocation and grow that I decided to define
	//it global to the object.
	DynamicArray<REAL, true> *row;
	SparseMatrix<REAL> *localR;
	MDNode** localMDNode;
	//An array of MDNodes pointers to the trees corresponding to the representation of
	//local(nonlocal) behaviour (using the local(nonlocal) matrices) of the model.
	MDNode** localMDTree;
	MDNode** nonlocalMDTree;
	int *last_local_index;

	SSNode::node_slist_type* nodes_lists;
	ssnode_list_t useful_copies_list;

	map<SSNode*, SSNode*> replace_map;
	TFunc NewStatesTree;

	//SubmodelIndex2MDDLevel[i] is the MDD level of the local state space of the ith submodel
	int *SubmodelIndex2MDDLevel;
	bool* is_local_level;
	int* MDDLevel2SubmodelIndex;
	int* SSIndex2MDDLevel;
	int* LMIndex2SubmodelIndex;
	int no_of_local_firings;
	int no_of_nonlocal_firings;

	State **next_state;

	//The list of variables that are passed without change across functions
	//So it's better to define them once instead of passing them as parameters
	//to functions. af stands for across functions.
	BaseModelClass* af_submodel;
	int af_submodel_index;
	SSNode** af_nodes;
	int* af_state_indices;
	int* af_levels;
	SSNode* af_new_tail;
	int af_no_of_ancestors;
	int* af_next_state_indices;
	TMDMatrix af_transitions;

	//Number of consecutive immediate transitions that have been performed
	int af_num_imm_trans;

	//We want to have a cache for saturation operator. The operator takes
	//a node and returns the saturated node.
	class SaturationCacheKey {
	public:
		SaturationCacheKey() {
			R = NULL;
		}
		~SaturationCacheKey() {
			reset();
		}
		SaturationCacheKey(SSNode* _R) {
			R = _R;
		}

		//Computes the hash function on the key
		size_t operator()(SaturationCacheKey& key) {
			return (unsigned int) key.R->getID() * 97;
		}

		//Compares two keys
		bool operator==(SaturationCacheKey& key2) {
			return ((void*) R == (void*) key2.R);
		}
		void reset() {
			R = NULL;
		}

		//A cache entry is dead if the MDD node of the key is dead
		bool isDead() {
			return R->isDead();
		}
	private:
		MDDNode* R;
	};

	class CSaturationCache: public ComputedCache<SaturationCacheKey, MDDNode> {
	public:
		void GC();
	}*SaturationCache;

	int unique_table_size_MDD;
	int union_cache_size;
	int satur_cache_size;
	int unique_table_size_MD;
	int submat_cache_size;
	int add_cache_size;

private:
	PATH_TYPE* path;
	PATH_TYPE* sorted_path;

};

inline MDDNode* SymbolicSSG::Union(MDDNode* S1, MDDNode *S2) {
	assert(S2 != NULL);
	assert(S1 != NULL);
	//NOTE:this is a minor optimization. If the two sets to be unioned are the
	//same set, return one of them. This will save some calls to Case and also
	//some nodes created during Case.
	if (S1 == S2)
		return S1;
	CaseParam Params(S1, S2, Value1);
	return SS->Case(Params);
}

inline void SymbolicSSG::Union(MDDNode*& result, MDDNode* S1, MDDNode*S2) {
	MDDNode* temp = Union(S1, S2);
	temp->incRef();
	S2->decRef();
	S1->decRef();
	result = temp;
}

inline MDDNode* SymbolicSSG::Diff(MDDNode* S1, MDDNode *S2) {
	/*
	 if (S2 == NULL) {
	 assert(false);
	 assert(S1 != NULL);
	 return S1;
	 }
	 else */{
		CaseParam Params(S2, S1, Value0);
		return SS->Case(Params);
	}
}

inline MDDNode* SymbolicSSG::Intersect(MDDNode* S1, MDDNode* S2) {
	CaseParam Params(S2, Value0, S1);
	return SS->Case(Params);
}

inline void SymbolicSSG::Intersect(MDDNode*& result, MDDNode* S1, MDDNode*S2) {
	MDDNode* temp = Intersect(S1, S2);
	temp->incRef();
	S2->decRef();
	S1->decRef();
	result = temp;
}

inline void SymbolicSSG::SetStateInLevels(int submodel_index, const int* const state_indices) {
	AtomicModel* atomic_model = System->atomic_models[submodel_index];
	for (int i = 0; i < atomic_model->no_of_ancestors; i++) {
		BaseModelClass* const model = atomic_model->ancestors[i];
		const int SS_ancestor_index = System->MDDLevel2SSIndex[atomic_model->ancestors_levels[i]];
		model->SetLocalState(state_space[SS_ancestor_index][state_indices[i]]);
	}
}

class ChangedSet {
public:
	ChangedSet(int _size) {
		changed = new DynamicArray<int, true> (_size, false);
	}
	~ChangedSet() {
		delete changed;
	}
	void add(int i);
	int pop();
	bool empty() {
		return changed_set.empty();
	}
	int size() {
		return changed->getSize();
	}
private:
	DynamicArray<int, true>* changed;
	queue<int> changed_set;
};

class HashTable {
public:
	HashTable(int _size) {
		int size = (1 << _size);
		mask = size - 1;
		table = new HashTableElem[size];
		for (int i = 0; i < size; i++)
			table[i].key = -1;
	}
	~HashTable() {
		delete[] table;
	}

	bool Lookup(INDEX_TYPE _key, INDEX_TYPE& _result) {
		HashTableElem& elem = table[_key & mask];
		if (elem.key == _key) {
			_result = elem.result;
			return true;
		} else
			return false;
	}

	void Insert(INDEX_TYPE _key, INDEX_TYPE _result) {
		HashTableElem& elem = table[_key & mask];
		elem.key = _key;
		elem.result = _result;
	}
	typedef struct {
		INDEX_TYPE key;
		INDEX_TYPE result;
	} HashTableElem;
private:
	HashTableElem* table;
	int mask;
};

#endif
