/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef __CACHE_H_
#define __CACHE_H_

#include "mddnode.h"
#include "BaseClasses/TraceFileMobius.h"
#include "BaseClasses/primes.h"
#include <typeinfo>
#include <vector>


// Determine if the C++ compiler support the C++11 forward list and unordered set
#if defined(__has_include)
#if __has_include(<forward_list>)
#define HAS_FORWARD_LIST 1
#endif

#if __has_include(<unordered_set>)
#define HAS_UNORDERED_SET 1
#endif

#elif (__cplusplus > 199711L) || (_MSC_VER >= 1800)
#define HAS_FORWARD_LIST 1
#define HAS_UNORDERED_SET 1
#endif

#if (HAS_FORWARD_LIST)
#include <forward_list>
#else
#include <ext/slist>
using namespace __gnu_cxx;
#endif

#if (HAS_UNORDERED_SET)
#include <unordered_set>
#else
#include <ext/hash_set>
using namespace __gnu_cxx;
#endif

unsigned long nextPrime(unsigned long hint);

class ComputedCacheBase {
public:
	virtual void GC() = 0;
};

// A general class that acts as cache for MDD operations.  You give an
// object of class KeyType and get the result of a specific operation
// on the object. This class is used to implement, among others,
// the Fire operation and Case operation cache. There are some requirement
// on the class KeyType. Some operations such as equality, etc. should
// be defined on it. The cache is implemented as a hash table with at most
// one element per bucket.
template<class KeyType, class ReturnType, class HashFcn = KeyType>
class ComputedCache: public ComputedCacheBase
{
public:
	ComputedCache();
	ComputedCache(int _bucket_count);
	~ComputedCache();

	virtual void GC() = 0;

	ReturnType* lookup(KeyType& key, bool& item_found);

	void insert(KeyType& key, ReturnType* value);

	void report();

	//resizes and rebuilds the hashtable
	void resize(int _bucket_count);

	void clear()
	{
		Cache.clear();
	}

protected:

	//The main data structure is a vector of pairs of KeyType (the argument
	//of the operation we are interested in) and ReturnType (the result of
	//the operation)
	typedef pair<KeyType, ReturnType*> BucketType;
	vector<BucketType> Cache;

private:
	// This class counts the number of all searches (search), the number
	// of hits (found) and the number of items inserted into the cache
	// (insertion).
	unsigned int found;
	unsigned int search;
	unsigned int insertion;
	int bucket_count;
	HashFcn hash_func;
};

// Hash table class for MDD nodes. Uses hash_set data structure in STL
template<class TNode>
class NodeUniqueTable
{
#if (HAS_FORWARD_LIST)
	typedef std::forward_list<ComputedCacheBase*> cache_list_t;
#else
    typedef slist<ComputedCacheBase*> cache_list_t;
#endif

public:
	NodeUniqueTable()
	{}

	NodeUniqueTable(int bucket_count)
	{
		HashTable.resize(bucket_count);
	}

	~NodeUniqueTable()
	{
		Traceprintf(2, "NodeUniqueTable::~NodeUniqueTable\n");
		//Write the code to free all the memory used by everything
		//I have to frist free the memory taken by the nodes themselves
		// and then free the memory taken by the hash function
		HashTable.clear();
	}

	//resizes and rebuilds the hash table
	void resize(int bucket_count)
	{
		Traceprintf(2, "Old UniqueTable bucket_count : %d\n", getNoOfBuckets());
#ifdef HAS_UNORDERED_SET
		HashTable.rehash(bucket_count);
#else
		HashTable.resize(bucket_count);
#endif
		Traceprintf(2, "New UniqueTable bucket_count : %d\n", getNoOfBuckets());
	}

	void GC()
	{
		//First we delete all cache entries that point to dead nodes
		//Then we have to delete all the nodes themselves and also
		//delete them from the UniqueTable.

		//First it does garbage collection for all the operation caches
		//that are registered with the uniqueness table
		for (cache_list_t::iterator it = CacheList.begin(); it != CacheList.end(); it++) {
			(*it)->GC();
		}

		//Going through all the elements in the hash table and delete all
		//those who are dead (have no references to them)
		//The weird pointer operations below is because we want to delete
		//from a linked list-like structure
		typename HashTableType::iterator NodeIter1, NodeIter2;
		for (NodeIter1 = HashTable.begin(); NodeIter1 != HashTable.end();) {
			TNode* NodePtr = *NodeIter1;
			//NodeIter2 corresponds to the node to which NodePtr points
			NodeIter2 = NodeIter1++;
			if (NodePtr->isDead()) {
				HashTable.erase(NodeIter2);
				delete NodePtr;
			}
		}
		Traceprintf(2, "%d nodes left ...", MDDNode::getNoOfNodes());
	}

	//Registers an operation cache with the uniqueness table
	void registerCache(ComputedCacheBase* Cache)
	{
		CacheList.push_front(Cache);
	}

	//pair.second is true if the element is already there
	pair<TNode*, bool> findOrAddNode(TNode* nodeptr)
	{
		//Finds the node if it is in the hashtable and inserts it into the
		//hashtable if it is NOT already there.
#if (HAS_UNORDERED_SET)
		pair<typename HashTableType::iterator, bool> result = HashTable.insert(nodeptr);
#else
		pair<typename HashTableType::iterator, bool> result = HashTable.insert_noresize(nodeptr);
#endif
		//result.second is true if the element is actually inserted.
		//What we want is to be true if the element is already there.
		return make_pair(*(result.first), !result.second);
	}

	void removeNode(TNode* nodeptr)
	{
		assert(HashTable.erase(nodeptr) == 1);
	}

	void print()
	{
		typename HashTableType::iterator NodeIter = HashTable.begin();
		while (NodeIter != HashTable.end()) {
			(*NodeIter)->print();
			NodeIter++;
		}
	}

	//Returns the number of nodes in the hash table
	int getNoOfNodes()
	{
		return HashTable.size();
	}

	int getNoOfBuckets()
	{
		return HashTable.bucket_count();
	}

private:
	struct NodePointerEqual
	{
		bool operator ()(TNode* node1ptr, TNode* node2ptr) const
		{
			return (*node1ptr) == (*node2ptr);
		}
	};

	struct NodeHashFunction
	{
		size_t operator ()(TNode* nodeptr) const
		{
			return nodeptr->hash();
		}
	};

public:
#if (HAS_UNORDERED_SET)
    typedef unordered_set<TNode*, NodeHashFunction, NodePointerEqual> HashTableType;
#else
	typedef hash_set<TNode*, NodeHashFunction, NodePointerEqual> HashTableType;
#endif

private:
	// Defines a hash table whose elements are pointers to MDD nodes the
	// hash function of the elements is declared as NodeHashFunction
	// (defined above) and element equality is tested by NodePointerEqual
	// function (defined above)
	HashTableType HashTable;
	//The list of operation caches that are associated with this uniqueness
	//table. This list is needed when doing garbage collection.
	cache_list_t CacheList;

public:
	typename HashTableType::iterator begin()
	{
		return HashTable.begin();
	}

	typename HashTableType::iterator end()
	{
		return HashTable.end();
	}

};

//Data structure initialization with default number of buckets
template<class KeyType, class ReturnType, class HashFcn>
ComputedCache<KeyType, ReturnType, HashFcn>::ComputedCache()
{
	bucket_count = nextPrime(10);
	Cache.resize(bucket_count);
	search = found = insertion = 0;
	for (int i = 0; i < bucket_count; i++) {
		Cache[i].second = NULL;
	}
}

//Data structure initialization with a specific number of buckets
template<class KeyType, class ReturnType, class HashFcn>
ComputedCache<KeyType, ReturnType, HashFcn>::ComputedCache(int _bucket_count)
{
	bucket_count = nextPrime(_bucket_count);
	Cache.resize(bucket_count);
	for (int i = 0; i < bucket_count; i++) {
		Cache[i].second = NULL;
	}
	search = found = insertion = 0;
}

//resizing the operation cache
template<class KeyType, class ReturnType, class HashFcn>
void ComputedCache<KeyType, ReturnType, HashFcn>::resize(int _bucket_count)
{
	int old_bucket_count = bucket_count;
	bucket_count = nextPrime(_bucket_count);
	Cache.resize(bucket_count);
	for (int i = old_bucket_count; i < bucket_count; i++) {
		Cache[i].second = NULL;
	}
}

template<class KeyType, class ReturnType, class HashFcn>
ComputedCache<KeyType, ReturnType, HashFcn>::~ComputedCache()
{
	Traceprintf(2, "CaseComputedCache::~CaseComputedCache\n");
	//Double check whether it needs other things
	Cache.clear();
}

/***********************************************************************
 * Looks up the result of the operation on the argument key. It returns *
 * the MDD node if the cache has the result of the operation and return *
 * NULL if it is not there.                                             *
 ***********************************************************************/
template<class KeyType, class ReturnType, class HashFcn>
inline ReturnType*
ComputedCache<KeyType, ReturnType, HashFcn>::lookup(KeyType& key, bool& item_found)
{
	search++;
	int hash_pos = (hash_func(key) % bucket_count);
	if (Cache[hash_pos].first == key) {
		found++;
		item_found = true;
		ReturnType* result = Cache[hash_pos].second;
		return Cache[hash_pos].second;
	} else {
		item_found = false;
		return NULL;
	}
}

/***********************************************************************
 * Inserts a key and the result of the operation on key (result) into   *
 * the operation cache.                                                 *
 ***********************************************************************/
template<class KeyType, class ReturnType, class HashFcn>
inline void ComputedCache<KeyType, ReturnType, HashFcn>::insert(KeyType& key, ReturnType* result)
{
	insertion++;
	//assert(!result->isDead());
	int hash_pos = (hash_func(key) % bucket_count);
	Cache[hash_pos].first = key;
	Cache[hash_pos].second = result;
}

template<class KeyType, class ReturnType, class HashFcn>
inline void ComputedCache<KeyType, ReturnType, HashFcn>::report()
{
	Traceprintf(2, "The number of look ups is %d/%d\n", found, search);
}

#endif
