/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      File: lumping-tarjan.h
 *
 *      Authors: Salem Derisavi, William H. Sanders
 *      Creation Date: 2002
 *
 *      Copyright (c) 1994-2003 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/
#ifndef __LUMPING_TARJAN_H__
#define __LUMPING_TARJAN_H__

#include <stdlib.h>
#include <iostream>
#include <string.h>
#include <fstream>
#include <list>
#include <ext/slist>
#include <queue>
#include <utility>
#include <stdio.h>
#include <map>
#include <assert.h>
#include <vector>

#include "mdnode.h"
#include <set>

#define EPSILON 1e-13

using namespace std;
using __gnu_cxx::slist;

template<class T>
class PVTuple {
public:
	PVTuple() {
		size = 0;
		values = NULL;
	}
	PVTuple(int _size) {
		size = _size;
		values = new T[size];
		memset(values, 0, size * sizeof(T));
	}
	PVTuple(const PVTuple& PVtuple) {
		size = PVtuple.size;
		values = new T[size];
		memcpy(values, PVtuple.values, size * sizeof(T));
	}
	~PVTuple() {
		delete[] values;
	}
	T& GetValue(int i) {
		return values[i];
	}
	void SetValue(int i, T value) {
		values[i] = value;
	}
	bool operator ()(const PVTuple& PVtuple1, const PVTuple& PVtuple2) const {
		assert(PVtuple1.size == PVtuple2.size);
		return memcmp(PVtuple1.values, PVtuple2.values, PVtuple1.size * sizeof(T)) < 0;
	}

	void print() const {
		printf("(");
		for (int i = 0; i < size; i++)
			printf("%f ", values[i]);
		printf(")");
	}
private:
	int size;
	T* values;
};

typedef enum {
	uninitalized, succlist, predlist
} TMode;

template<class PVType, class MatElement>
class Lumping {
public:
	struct successor {
		int dest;
		MatElement rate;
	};

	class predecessor {
	public:
		int source;
		MatElement rate;
	};

	struct TBlock;
	typedef list<TBlock> TPartition;
	typedef map<MatElement, typename TPartition::iterator> TSplitsRB;

	typedef pair<MatElement, typename TPartition::iterator> TSplitRBPair;
	typedef slist<pair<typename TPartition::iterator, MatElement> > TMapping;

	typedef struct {
		MatElement sum_rate;
		typename TPartition::iterator block;
	} TSplit;

	struct TBlock {
		int size;
		list<int> elements;
		TSplitsRB splits;
		bool marked;
		MatElement sum_rate;
		int index;
		PVTuple<PVType>* PVtuple;
	};

	typedef struct {
		MatElement sum_rate;
		typename TPartition::iterator block;
		list<int>::iterator elem_in_block;
		slist<successor> succ_list;
		slist<predecessor> pred_list;
		TMapping outgoing;
	} TState;

	typedef struct {
		vector<TState> states;
		int num_transitions;
		TMode mode;
	} TChain;

	static void print_chain(TChain&);
	static void print_partition(FILE* f, TPartition& Q, char* name);
	static void print_partition(TPartition& Q, char* name);
	static void print_block(FILE* f, TBlock& block, char* str);
	static void print_block(TBlock& block, char* str);
	static void print_splitter(list<int> &S);
	static bool is_partition_correct(TPartition& Q, TChain& chain);
	static bool is_partition_correct(TPartition& Q, vector<TState>& states);
	static void init_partition_with_one_block(TPartition& Q, int no_of_states);
	static void pred2succ(TChain &chain);
	static void ComputeSumsToSplitter(TChain& chain, list<int>& S, queue<int>& x_list);
	static void lump(TChain& chain, TPartition& Q);
	static void Split(vector<TState>& states, queue<int>& x_list, slist<list<int> >& L, TPartition& Q);
	static bool equal(vector<TState>& states, int x1, int x2);
	static void set_chain_initial_partition(TChain& chain, TPartition &Q);
	static void set_states_initial_partition(vector<TState>& states, TPartition& Q);
	static void add_split_blocks_to_L(slist<typename TPartition::iterator> &split_list, slist<list<int> > &L, TPartition &Q);
	static void copyQ2L(TPartition &Q, slist<list<int> > &L);
	static void verify(TChain& chain, TPartition& Q);
};

//TODO: Unify InitialPartition class for compositional lumping with the one
//for state-level lumping
template<class PVType, class MatElement>
class InitialPartition {
typedef	typename Lumping<PVType, MatElement>::TPartition TPartition;
public:
	InitialPartition (TPartition& _Partition) {
		Partition = &_Partition;
	}

	void InsertPVtuple(int state_num, PVTuple<PVType>& PVtuple) {
		typedef typename TTree::iterator TTreeIterator;
		pair<TTreeIterator, bool> result =
		tree.insert(make_pair(PVtuple, (typename TPartition::iterator)0));
		if (result.second) {
			typename Lumping<PVType, MatElement>::TBlock blk;
			blk.size = 1;
			blk.marked = false;
			blk.PVtuple = (PVTuple<PVType>*) &result.first->first; //Casting from (const PVTuple*) to (PVTuple*)
			Partition->push_front(blk);
			Partition->front().elements.push_front(state_num);
			result.first->second = Partition->begin();
		}
		else {
			result.first->second->size++;
			result.first->second->elements.push_front(state_num);
		}
	}

	void SetOneBlockPartition(int num_states) {
		Lumping<PVType, MatElement>::init_partition_with_one_block(*Partition, num_states);
	}

protected:
	TPartition* Partition;
private:
	typedef map<PVTuple<PVType>, typename TPartition::iterator, PVTuple<PVType> > TTree;
	TTree tree;
};

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::lump(TChain& chain, TPartition& Q) {
	set_chain_initial_partition(chain, Q);
	slist<list<int> > L; // List of potential splitters
	//Initialize the list of potential splitters to the partition incurred
	//by the rate reward of the states
	copyQ2L(Q, L);
	queue<int> x_list; // will eventually equal \union_{x \in S} e^{-1}(x)
	// this is actually L' in the pseudo code given in the paper

	while (!L.empty()) { //while there is a potential splitter
		//get the first splitter from L
		list<int> &S = L.front();
		// Compute x.sum_rate from every xEs such that s \in S
		ComputeSumsToSplitter(chain, S, x_list);
		// Since S is a reference to the first element of L (and not a
		// complete copy) we need to delete when we are done using it.
		L.pop_front();
		Split(chain.states, x_list, L, Q);
	}
}

// x_list (output) will finally contain all elements that have transitions
// to any element in S
template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::ComputeSumsToSplitter(TChain& chain, list<int>& S, queue<int>& x_list) {
	for (list<int>::iterator s = S.begin(); s != S.end(); s++) {
		for (typename slist<predecessor>::iterator x = chain.states[*s].pred_list.begin(); x != chain.states[*s].pred_list.end(); x++) {
			//membership in x_list is checked by checking whether sum_rate equals zero
			if (chain.states[x->source].sum_rate.isZero())
				x_list.push(x->source);
			chain.states[x->source].sum_rate += x->rate;
		}
	}
}

//x_list(input): list of all states that have any transition to the splitter
//L(input, output): list of potential splitters
template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::Split(vector<TState>& states, queue<int>& x_list, slist<list<int> >& L, TPartition& Q) {
	// The list of blocks which are split
	//This is the same as L'' in the pseudocode given in the paper
	slist<typename TPartition::iterator> split_list;
	typename TPartition::iterator B;

	while (!x_list.empty()) {
		int x = x_list.front();
		x_list.pop();
		TState &state = states[x];
		B = state.block;
		if (!B->marked) {
			split_list.push_front(B);
			B->marked = true; //add it to some B list later and unmark the elements of the list
		}

		//look up x.sum_rate in B->splits
		typename TSplitsRB::iterator result;
		result = B->splits.find(state.sum_rate);
		if (result == B->splits.end()) // This is the first sum_rate of this value
		{
			// So we have to create a new sub-block. This sub-block is instantly added
			// to the partition we have so far. The element x is added to the new block (blk)
			TBlock blk;
			blk.size = 1;
			blk.marked = false;
			blk.PVtuple = B->PVtuple;
			Q.push_front(blk);
			Q.front().elements.push_front(x);
			result = B->splits.insert(TSplitRBPair(state.sum_rate, Q.begin())).first;
		} else { // There has been a sum_rate with the same value
			result->second->size++;
			result->second->elements.push_front(x);
		}
		// The element x is removed from its previous block (B)
		B->elements.erase(state.elem_in_block);
		B->size--;
		// x has moved to a new block in Q so we have to set state.block
		// and state.elem_in_block. x's block is the last block found.
		// The position of x in its new block is the first element in the block
		// because x has just been added to the block.
		state.block = result->second;
		state.elem_in_block = result->second->elements.begin();
		state.sum_rate.Zero(); // This line practically deletes state from list x_list
	}

	//Adding all the split blocks to L except the biggest
	add_split_blocks_to_L(split_list, L, Q);
	split_list.clear();
	//print_partition(Q, "test");
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::set_chain_initial_partition(TChain& chain, TPartition &Q) {
	set_states_initial_partition(chain.states, Q);
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::set_states_initial_partition(vector<TState>& states, TPartition& Q) {
	for (typename TPartition::iterator block = Q.begin(); block != Q.end(); block++) {
		for (list<int>::iterator it = block->elements.begin(); it != block->elements.end(); it++) {
			TState &state = states[*it];
			state.block = block;
			state.elem_in_block = it;
			state.sum_rate.Zero();
		}
	}
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::copyQ2L(TPartition &Q, slist<list<int> > &L) {
	for (typename TPartition::iterator block = Q.begin(); block != Q.end(); block++) {
		// The copy constructor of list<int> takes care of the copying
		L.push_front(block->elements);
	}
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::add_split_blocks_to_L(slist<typename TPartition::iterator> &split_list, slist<list<int> > &L, TPartition &Q) {
	typename TPartition::iterator B;
	int biggest_split_size, i;
	typename TSplitsRB::iterator biggest_split_index;
	while (!split_list.empty()) {
		B = split_list.front();

		/*
		 TSplitsRB::iterator it1 = B->splits.begin(), it2;
		 if (it1 != B->splits.end()) {
		 it2 = it1;
		 it2++;
		 for (;it2 != B->splits.end(); it1++, it2++)
		 if (fabs(it1->first/it2->first - 1) < EPSILON)
		 printf("Two total outgoing rates are too close. The algorithm may give incorrect answers.\n");
		 }
		 */
		split_list.pop_front();

		//Find the largest among the V sets and what is remained in block B
		biggest_split_index = B->splits.end();
		biggest_split_size = B->size;
		for (typename TSplitsRB::iterator it = B->splits.begin(); it != B->splits.end(); it++) {
			if (it->second->size > biggest_split_size) {
				biggest_split_size = it->second->size;
				biggest_split_index = it;
			}
		}
		//Add all but the biggest split to L
		for (typename TSplitsRB::iterator it = B->splits.begin(); it != B->splits.end(); it++) {
			if (biggest_split_index != it) {
				L.push_front(it->second->elements);
			}
		}
		if ((biggest_split_index != B->splits.end()) && (B->size > 0))
			L.push_front(B->elements);

		B->splits.clear();
		assert(B->marked);
		B->marked = false;

		//Erase the initial block if it's empty
		if (B->size == 0)
			Q.erase(B);
	}
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::init_partition_with_one_block(TPartition& Q, int no_of_states) {
	// No PV to consider for partition refinement.
	// So the initial partition has only one block.
	Q.resize(1);
	typename TPartition::iterator B = Q.begin();
	B->size = no_of_states;
	B->marked = false;
	B->PVtuple = NULL;
	for (int i = 0; i < no_of_states; i++) {
		B->elements.push_front(i);
	}
}

/**************************************************************************
 ** Convert the predecessor lists in the chain to successor lists         **
 ** in order to check the lumping or to write the the lumped state space  **
 **************************************************************************/
template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::pred2succ(TChain &chain) {
	successor succ;
	predecessor pred;
	for (int i = 0; i < chain.states.size(); i++) {
		slist<predecessor> &pred_list = chain.states[i].pred_list;
		for (typename slist<predecessor>::iterator pred_it = pred_list.begin(); pred_it != pred_list.end(); pred_it++) {
			succ.rate = pred_it->rate;
			succ.dest = i;
			chain.states[pred_it->source].succ_list.push_front(succ);
		}
		pred_list.clear();
	}
	chain.mode = succlist;
}

/*************************************************************
 ** Checks whether the lumping satisfies the strong lumping  **
 ** condition. The code may not be time optimal but it does  **
 ** fairly well even for large chains. Returns true if the   **
 ** partition satisfies the conditions and false otherwise.  **
 ** Uses : equal(TChain &chain, int x1, int x2)              **
 *************************************************************/
template<class PVType, class MatElement>
bool Lumping<PVType, MatElement>::is_partition_correct(TPartition &Q, TChain &chain) {
	if (chain.mode == predlist)
		pred2succ(chain);

	return is_partition_correct(Q, chain);
}

template<class PVType, class MatElement>
bool Lumping<PVType, MatElement>::is_partition_correct(TPartition &Q, vector<TState>& states) {

	for (int i = 0; i < states.size(); i++) {
		TState &state = states[i];
		for (typename slist<successor>::iterator it = state.succ_list.begin(); it != state.succ_list.end(); it++) {
			typename TPartition::iterator& dest_block = states[it->dest].block;

			typename TMapping::iterator map_elem;
			for (map_elem = state.outgoing.begin(); map_elem != state.outgoing.end(); map_elem++) {
				if (map_elem->first == dest_block)
					break;
			}

			if (map_elem == state.outgoing.end()) {
				pair<typename TPartition::iterator, MatElement> p;
				state.outgoing.push_front(p);
				state.outgoing.front().second = it->rate;
				state.outgoing.front().first = dest_block;
			} else {
				map_elem->second += it->rate;
			}
		}
	}
	bool correct = true;
	for (typename TPartition::iterator block = Q.begin(); block != Q.end(); block++) {
		int x1 = block->elements.front();
		list<int>::iterator x2 = block->elements.begin();
		for (x2++; x2 != block->elements.end(); x2++) {
			if (!equal(states, x1, *x2)) {
				correct = false;
				break;
			}
		}
		if (!correct)
			break;
	}

	return correct;
}

/***************************************************************
 ** Check whether states x1 and x2 (in the same lumped state)  **
 ** satisfy the strong lumping theorem, that is the sum of the **
 ** rates of transitions from x1 and x2 to any other partition **
 ** is equal.                                                  **
 ** Used by : check_partition                                  **
 ***************************************************************/
template<class PVType, class MatElement>
bool Lumping<PVType, MatElement>::equal(vector<TState>& states, int x1, int x2) {
	TMapping &outgoing1 = states[x1].outgoing;
	TMapping &outgoing2 = states[x2].outgoing;
	if (outgoing1.size() != outgoing1.size())
		return false;
	typename TMapping::iterator map_elem1, map_elem2;
	for (map_elem1 = outgoing1.begin(); map_elem1 != outgoing1.end(); map_elem1++) {
		for (map_elem2 = outgoing2.begin(); map_elem2 != outgoing2.end(); map_elem2++) {
			if ((*map_elem1) == (*map_elem2))
				break;
		}
		if (map_elem2 == outgoing2.end())
			return false;
	}
	return true;
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::print_block(FILE* f, TBlock& block, char* str) {
	assert(block.size == block.elements.size());
	fprintf(f, "%s size(%d) ", str, block.size);
	for (list<int>::iterator elem = block.elements.begin(); elem != block.elements.end(); elem++) {
		fprintf(f, "%d ", *elem);
	}
	fprintf(f, "\n");
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::print_block(TBlock& block, char* str) {
	print_block(stdout, block, str);
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::print_splitter(list<int> &S) {
	cout << "Splitter : ";
	for (list<int>::iterator elem = S.begin(); elem != S.end(); elem++) {
		cout << *elem << " ";
	}
	cout << endl;
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::print_partition(FILE* f, TPartition& Q, char* name) {
	fprintf(f, "*********** %s partition ************\n", name);
	int counter = 0;
	for (typename TPartition::iterator block = Q.begin(); block != Q.end(); block++, counter++) {
		char title[50];
		sprintf(title, "block %d (%d): ", counter, block->index);
		print_block(f, *block, title);
	}
	fprintf(f, "------------ %s parition --------------\n", name);
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::print_partition(TPartition& Q, char* name) {
	print_partition(stdout, Q, name);
}

/***************************************************************
 **   Prints the entire chain to standard output               **
 ***************************************************************/
template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::print_chain(TChain& chain) {

	int num_states = chain.states.size();
	// Checking whether both sides work correctly
	for (int i = 0; i < num_states; i++) {
		TState &state = chain.states[i];
		for (typename slist<successor>::iterator it = state.succ_list.begin(); it != state.succ_list.end(); it++) {
			cout << "From " << i << " to " << it->dest << " with rate " << it->rate << endl;
		}
	}

	for (int i = 0; i < num_states; i++) {
		TState &state = chain.states[i];
		for (typename slist<predecessor>::iterator it = state.pred_list.begin(); it != state.pred_list.end(); it++) {
			cout << "From " << it->source << " to " << i << " with rate " << it->rate << endl;
		}
	}
}

template<class PVType, class MatElement>
void Lumping<PVType, MatElement>::verify(TChain& chain, TPartition& Q) {
	for (int i = 0; i < chain.states.size(); i++) {
		list<int>& elements = chain.states[i].block->elements;
		assert(find(elements.begin(), elements.end(), i) != elements.end());
		assert(*(chain.states[i].elem_in_block) == i);
	}
}

#endif
