/*****************************************************************************\
 *                            RESTRICTED RIGHTS LEGEND
 *
 *      Use, duplication, or disclosure by the Government is subject to
 *      restrictions as set forth in subparagraph (c) (1) (ii) of the Rights in
 *      Technical Data and Computer Software Clause at DFARS 252.227-7013.
 *
 *      The Board of Trustees of the University of Illinois
 *      109 Coble Hall
 *      801 South Wright Street
 *      Champaign, IL 61820-6242
 *
 *      Copyright (c) 1994-2008 William H. Sanders, and
 *      The Board of Trustees of the University of Illinois
 *      All rights reserved.
 *
 \*****************************************************************************/

#ifndef __MDNODE_H_
#define __MDNODE_H_
#include "DynamicObjects.h"
#include "typedefs.h"
#include "BaseClasses/TraceFileMobius.h"
#include <set>
#include <vector>

#if (HAS_FORWARD_LIST)
	#include <forward_list>
#else
	#include <ext/slist>
	using __gnu_cxx::slist;
#endif

#define MDNODE_OPT

class MDNode;
typedef double REAL;

typedef struct {
	INDEX_TYPE source;
	INDEX_TYPE dest;
	REAL rate;
} Transition;

typedef pair<REAL, MDNode*> TDownPair;

class TElement {
public:
#if (HAS_FORWARD_LIST)
	std::forward_list<TDownPair> d;
	typedef std::forward_list<TDownPair>::iterator iterator;
#else
	slist<TDownPair> d;
	typedef slist<TDownPair>::iterator iterator;
#endif
	virtual int operator==(TElement& e);
	void write(FILE* f);
};

typedef enum {
	row_oriented, col_oriented
} TOrientation;

class MDNode {
public:
	//Maximum dimensions of a matrix: 32767*32767
	typedef short DIM_TYPE;

	MDNode(short _var, TOrientation _orientation);
	MDNode() {
	}
	;
	void Construct(short _var, TOrientation _orientation);
	void initializeNode();
	~MDNode();
	void OK();
	DIM_TYPE getMajor(int i) {
		assert(i < major_minor.size()-1);
		return major_minor[i].major;
	}
	DIM_TYPE getMinor(int i) {
		return major_minor[i].minor;
	}
	void setMajor(int i, DIM_TYPE maj) {
		major_minor[i].major = maj;
	}
	void setMinor(int i, DIM_TYPE min) {
		major_minor[i].minor = min;
	}
	TElement* const getElement(int i) {
		return &elements[i];
	}
	int insertElement(DIM_TYPE _major, DIM_TYPE _minor);
	void insertDownPointer(int i, TDownPair p);
	bool ProceedToElement(int& index, DIM_TYPE major, DIM_TYPE minor);
	inline bool ProceedToElement(int& index, DIM_TYPE major);
	inline bool ProceedToElementOpt(int& index, DIM_TYPE major);
	void incRef();
	void decRef();
	virtual int operator==(MDNode& a);
	size_t hash(void);
	void writeTree(FILE* f);
	int readNode(FILE* f, vector<int>& indices, vector<double>& coefs, int& self_id, vector<int>& children_id);

	friend class TElement;

	typedef vector<MDNode*> NodeListType;

#if defined(ARRAY)
	typedef SparseVectorArray2<REAL> VectorType;
#elif defined(HASHVECTOR)
	typedef SparseVectorHash<REAL> VectorType;
#else
	typedef SparseVector2<REAL> VectorType;
#endif

	VectorType* getColumn(NodeListType* _md_nodes_lists, DIM_TYPE* col_indices);
	VectorType* getRow(NodeListType* _md_nodes_lists, DIM_TYPE* row_indices);

	void setOrientationShallow(TOrientation _orientation);
	void setOrientationDeep(TOrientation _orientation);
	TOrientation getOrientation() {
		return orientation;
	}
	int getSize() const {
		return major_minor.size() - 1;
	}
#ifdef MD_CACHE
	void ClearCache() {cache.Zero(); is_cache_valid = false;}
#endif
#ifdef MDNODE_OPT
	void OptimizeDS();
	void OptimizeDSTree();
	int SizeInMemory();
	void SizeInMemoryTree(int& no_of_nodes, int& memory_taken);
#endif
	int CheckIdentity();
	DIM_TYPE getNMajors() {
		return NMajors;
	}
	DIM_TYPE getNMinors() {
		return NMinors;
	}
	bool isDead() const {
		return (RefCount == 0);
	}
	virtual void print(bool detailed = true);
	void CheckValidity();
	void resetHash() {
		hash_value = UINT_MAX;
	}
	int getID() const {
		return id;
	}
	short topVariable() const {
		return var;
	}
	static int getNoOfDeads() {
		return NoOfDeads;
	}
	static int getNoOfNodes() {
		return NoOfNodes;
	}
	static int getCounter() {
		return counter;
	}
	static int getNoOfAlives() {
		return NoOfAlives;
	}
	static int getMaxNoOfAlives() {
		return MaxNoOfAlives;
	}
	static void AdjustNoOfDeads(int amount) {
		NoOfDeads += amount;
		NoOfAlives -= amount;
		MaxNoOfAlives = NoOfAlives > MaxNoOfAlives ? NoOfAlives : MaxNoOfAlives;
	}
	static void AdjustNoOfNodes(int amount) {
		NoOfNodes += amount;
		NoOfAlives += amount;
		MaxNoOfAlives = NoOfAlives > MaxNoOfAlives ? NoOfAlives : MaxNoOfAlives;
	}

	typedef enum {
		unmarked, marked
	} TMDNodeStatus;
	TMDNodeStatus status;
	DynamicArray<INDEX_TYPE, true> minor_offsets;

protected:
	int status_iden;

	static NodeListType* af_md_nodes_lists;
	static DIM_TYPE* af_major_indices;

	virtual void writeNode(FILE* f);
	void dfsWriteTree(FILE* f, slist<MDNode*>& tagged_list);
	void getMajor();

	void DFS(set<MDNode*>& marked_set);

	bool optimizedDS;
	typedef struct {
		DIM_TYPE major;
		DIM_TYPE minor;
#ifdef MDNODE_OPT
		int element_index;
#endif
	} Coord;
	typedef vector<Coord> major_minor_type;
	typedef vector<TElement> elements_type;
public:
	major_minor_type major_minor;
	//vector<DIM_TYPE> minor;
	elements_type elements;
#ifdef MDNODE_OPT
	//int* element_index;
	TDownPair* d;
#endif
protected:
	TOrientation orientation;
	DIM_TYPE NMajors;
	DIM_TYPE NMinors;
	/*
	 #if defined(ARRAY)
	 VectorType temp;
	 #endif
	 */
	VectorType cache;
#ifdef MD_CACHE
	bool is_cache_valid;
#endif

	//Number of nodes created so far (including those that has been deleted)
	//This is used for giving an ID to each node
	static int counter;
	//Number of all in the MD structure (including dead ones)
	static int NoOfNodes;
	//Number of dead nodes (those with RefCount == 0) in the MD structure.
	static int NoOfDeads;

	static int NoOfAlives;
	static int MaxNoOfAlives;
	//The index of the variable of the MD node.
	//Higher numbers mean higher in the tree
	short var;

	//ID of the node. IDs are assigned to nodes incrementally.
	int id;
	//The number of references to the node. Only external pointers and
	//nodes that have non-zero RefCount count.
	int RefCount;

	//Hash value of the node. Since the hash value depends on the downward
	//pointers, it should be computed after all downward pointers are assigned.
	unsigned int hash_value;

};

inline bool MDNode::ProceedToElement(int& index, DIM_TYPE major) {
	const int size = getSize();
	while (index < size && getMajor(index) < major)
		index++;
	return (index < size);
}

bool MDNode::ProceedToElementOpt(int& index, DIM_TYPE major) {
	const int size = getSize();
	while (index < size && getMajor(index) < major)
		index++;
	return (index < size);
}

#endif
